/*--

  \FileTitle{
  {\tt egctrl.c} --- Manage Electrocardiognosis recording control file
  }
  \CopyrightNotice{


    Copyright (c) 1991--1996 Alexandru Dan Corlan MD (http://dan.corlan.net)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


  }

  \section{The control file}

  \begin{code}
  --*/

#include <stdio.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/timeb.h>
#include <time.h>

#include "ilsix.h"

#include "gliks.h"
#include "ecgnosis.h"



/*--
\end{code}

\subsection{Assertion constituents}

\begin{code}

--*/

cell _recording_file;
cell _def_event;
cell _measurement;
cell _from;
cell _type;

init_assertion_constituents()

{
  _recording_file = cl_make_symbol("recording-file");
  _def_event = cl_make_symbol("def-event");
  _measurement = cl_make_symbol("measurement");
  _from = cl_make_symbol(":from");
  _type = cl_make_symbol(":type");
}

/*--
\end{code}

\subsection{Direct ctrl file access}

\begin{code}
--*/

#define MAX_NOF_ASSERTIONS 10000

cell ctrl_file = NIL;
cell ctrl_data[MAX_NOF_ASSERTIONS];
long ctrl_size  = 0;

void open_ctrl_file(char *ctrl_file_name)

{
  cell c;
  long i;

  ctrl_file = fileopen(ctrl_file_name, INPUT_OUTPUT, CHAR_FILE);
  /* check here non-existent file */

  for(i=0; i<MAX_NOF_ASSERTIONS; i++)
    ctrl_data[i] = NIL;
  ctrl_size = 0;
  
  while((c = lread(ctrl_file)) != eof_obj_)
    ctrl_data[ctrl_size ++] = c;
  

  /* file is left open, so that we can write in it */
}


/*--
\end{code}

\funparagraph{init\_ctrl\_file}



\PurposeReturn{
Creates a ctrl file, inserting at its beginning the
mandatory assertions. The recording file name and its
type are also given (it may be 0, meaning ECGNOSIS-1 file
or 1, meaning GLIKSEE-ANTICOMPRESSION file).
}
\Exceptions{
}
\GlobalEffects{

}
\Preconditions{

}
\Performance{

}

\begin{code}
--*/

/*--
\end{code}

\funparagraph{init\_ctrl\_file}

\begin{code}
--*/


void init_ctrl_file(char *ctrl_file_name, char *gwo_file_name, short filetype)

{
  struct stat sb;
  time_t tt;
  cell cc;

  /* RAISE an exception here if ctrl_file was already opened */

  ctrl_file = fileopen(ctrl_file_name, OUTPUT, CHAR_FILE);
  /* assumed to delete any similar existing file */

  init_file_with_site(ctrl_file);
  
  stat(gwo_file_name, &sb);
  tt = sb.st_mtime;

  switch(filetype) {
  case 0:
    cc = cons(cl_make_symbol(":format"),
	      cons(cl_make_symbol("ECGNOSIS-1"), NIL));
    break;
  case 1:
    cc = cons(cl_make_symbol(":format"),
	      cons(cl_make_symbol("GLIKSEE-ANTICOMPRESSION"), NIL));
    break;
  case 2:
    cc = cons(cl_make_symbol(":format"),
	      cons(cl_make_symbol("GLIKSEE-SHR"), NIL));
    break;
  default:
    /* EXCEPTION: no such file type */
    cc = NIL;
    break;
  }
  
  cc = cons(cl_make_symbol(":sampling-rate"),
	    cons(cl_make_integer(200), cc));
  cc = cons(_recording_file,
	    cons(cl_make_string(ctime(&tt)),
		 cons(NIL,
		      cons(cl_make_symbol(":name"),
					  cons(cl_make_string(gwo_file_name),
					       cc)))));
  
  pprint(cc, ctrl_file);

  fclose(disconnect(ctrl_file));
  ctrl_file = NIL;
  open_ctrl_file(ctrl_file_name);
  return ;   /* but some exception mechanism must be employed here */
}

void add_assertion(cell a)

{
  /* check here if the file is opened */

  ctrl_data[ctrl_size ++] = a;
  pprint(a, ctrl_file);
  fflush(file_of(ctrl_file));
}

void close_ctrl_file()

{
  /* DISSALLOCATE HERE ALL THE ctrl_data CONTENTS */

  ctrl_size = 0;
  fclose(disconnect(ctrl_file));
}

long n_of_assertions()

{
  return ctrl_size;
}

cell assertion_no(long k)

{
  return ctrl_data[k];
}


/*--
\end{code}

\subsection{Selectors}

\begin{code}
--*/

cell last_assertion_named(char *nm)

{
  cell c;
  long i;

  c = cl_make_symbol(nm);
  for(i=ctrl_size-1; i>=0; i--)
    if(cl_consp(ctrl_data[i]) && car(ctrl_data[i])==c)
      return ctrl_data[i];
  return NIL;
}

cell slot_of(cell assertion, cell keyword)

{
  if(!cl_consp(assertion))
    return NIL;
  assertion = cdr(assertion);
  while(assertion!=NIL && cdr(assertion)!=NIL) {
    if(car(assertion)==keyword)
      return cadr(assertion);
    assertion = cddr(assertion);
  }
  return NIL;
}

/*

THIS PART IS MORE SPECIFIC, TO EVENT MANAGEMENT

*/

cell all_measurements_between(long from, long to)

{
  cell r;
  cell a;
  cell b;
  long i;

  r = NIL;
  for(i=0; i<n_of_assertions(); i++) {
    a = assertion_no(i);
    if(is_cons(a) && car(a) == _def_event) {
      if(slot_of(a, _type) == _measurement) {
	b = slot_of(a, _from);
	if(is_binary(b) && binary(b) >= from && binary(b) <= to)
	  r = cons(a, r);
      }
    }
  }
  return r;
}


/*--
\end{code}

\subsection{Initialization}

\begin{code}
--*/


void egctrl()

{
  init_assertion_constituents();
}
