
/*--

  \FileTitle{
  EGDIAGRV --- Heart Rate Variability Diagrams Implementation
  }
  \CopyrightNotice{

    Copyright (c) 1991--1996 Alexandru Dan Corlan MD (http://dan.corlan.net)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  }

  \begin{code}
  --*/

#include <stdio.h>
#include <math.h>

#include "gliks.h"
#include "ecgnosis.h"

#include <gsl/gsl_fft_real.h>

/*--
  \end{code}

  \subsection{Heart Rate Variability Diagrams and Statistics}

  \subsubsection{Frequency Domain Variability}

  \funparagraph{rv\_compute\_window, rv\_apply\_window}

  \PurposeReturn{
  These functions compute, and respectively apply windows
  for the given number of elements, to the argument
  double vector---making it suitable for fourier transform.

  The only window that is computed for the moment is:

  $$ W(t) = {{e^{-(t-T)^2} - e^{-T^2}}\over{1 - e^{-T^2}}} $$

  Where T is half of the duration of the window.
  }
  \Exceptions{
  }
  \GlobalEffects{
  For performance reasons, in the case when the Fourier transform
  is repeatedly applied to many parts of the recording in sequence,
  the window function is computed once and the result stored
  in the \variabletag{four\_window} vector.
  }

  \begin{code}
  --*/

#define MAX_FOUR_SIZE 2048

double four_window[MAX_FOUR_SIZE];
short four_window_fill = 0;
double window_factor = 1.0;

void rv_compute_window(short win_size)

{
  double T, C;
  short t;

  T = win_size / 2.0;
  C = exp(- (T * T / ((double)win_size * win_size)));

  for(t=0; t<win_size; t++)
    four_window[t] = (exp(- ((t - T) * (t - T)/((double)win_size * win_size))) - C) 
                     / (1 - C);
  four_window_fill = win_size;
  for(t=0, window_factor=0; t<win_size; t++)
    window_factor += four_window[t];
  window_factor = t / window_factor;
}

void rv_apply_window(double *four, short win_size)

/* \variabletag{four} is the double vector, with real values alternating
   with the imaginary ones, to which the window is applied.
   */

{
  short i;

  if(win_size!=four_window_fill)
    rv_compute_window(win_size);
  
  for(i=0; i<win_size; i++) {
    four[2*i] *= four_window[i];
    four[2*i + 1] *= four_window[i];
  }
}


/*--
  \end{code}

  \funparagraph{fourier\_transform}  % The name of the function(s)

  \begin{code}
  --*/


void fourier_transform(double *data, int nn, int isign) {
  gsl_fft_real_radix2_transform (data, 2, nn);
}


/*--
  \end{code}

  \funparagraph{enneat\_TD, enneat\_VEB, enneat\_intercalates}
  \ImplementationNotes{
  These functions were used in the experimental {\bf GlikSee} program
  to filter any accidents which appeared on the heart rate
  signal. Original comments (almost close to the real code) are
  reproduced below:
  
  This section consists of three functions: |enneat_TD|() which
  makes the time domain neat, that is resolves isolated
  extrasystoles by means of interpolation of the sinus
  rhythm that should replace it there. It processes the 
  itab table by trying to detect the supraventricular
  and ventricular extrasystoles.
  
  Such a diagram is to be used also in other heart-variability
  assesments. An effort is done not to include in the
  display the shifts of heart rate which are
  not extrasystoles nor artefacts and thus could only result
  from sampling fast heart rate regulation processes.
  
  |enneat_TD| takes as arguments the number of complexes in
  the |itab| table and modifies the table by pattern-recognition 
  filtering.
  
  }
  \Portability{
  }
  \Changes{
  }

  \begin{code}
  --*/

#ifdef RESTORE_OLD_FUNCTIONS

#define itabfr(i) (12000.0/(itab[i+1].pl - itab[i].pl))

enneat_pauses(short nofcp)

  {
  double fr0, fr1, fr2, fr3;
  double frplus;
  short changed;
  int i, j;

  changed = 0;

  fr0 = itabfr(1);   /* i-4 */
  fr1 = itabfr(2);   /* i-3 */
  fr2 = itabfr(3);   /* i-2 */
  fr3 = itabfr(4);   /* i-1 */

  for(i=5; i<nofcp-5; i++)
    {

/* First class: presystole without compensatory pause */

/*    if(fr1 > fr0*1.3 && fr2 < fr0 * 1.3)
      {
      itab[i-3].pl = (itab[i-4].pl + itab[i-2].pl) / 2;
      goto goon;
      }*/

    /* Third class: pause of one beat */

    if((fr1 * 2) < (1.2 * fr2) && (fr1 * 2) > (0.8 * fr2)
       && (fr1 * 2) < (1.2 * fr0) && (fr1 * 2) > (0.8 * fr0))
      {
/*      printf("Detected %4d: %d %d %d %d %d %d\n", i-5, 
        itab[i-5].pl, itab[i-4].pl, itab[i-3].pl,
        itab[i-2].pl, itab[i-1].pl, itab[i].pl);*/
      for(j=nofcp; j>i-2; j--)
        itab[j].pl = itab[j-1].pl;
/*      printf("Detected>%4d: %d %d %d %d %d %d\n", i-5, 
        itab[i-5].pl, itab[i-4].pl, itab[i-3].pl,
        itab[i-2].pl, itab[i-1].pl, itab[i].pl);*/
      itab[i-2].pl = 
       ((long)(itab[i-3].pl) + ((itab[i].pl) - (itab[i-1].pl)));
/*      printf("Detected+%4d: %d %d %d %d %d %d\n", i-5, 
        itab[i-5].pl, itab[i-4].pl, itab[i-3].pl,
        itab[i-2].pl, itab[i-1].pl, itab[i].pl);*/
      changed++;
      goto goon;
      }

  goon:
    fr0 = fr1;
    fr1 = fr2;
    fr2 = fr3;
    fr3 = itabfr(i);
    }
/*  if(fr3 > 1.2*fr2 || fr3 < 0.8*fr2)
    itab[i-1].pl = itab[i-2].pl + (itab[i-2].pl - itab[i-1].pl);*/
  }


enneat_SVEBs(short nofcp)

  {
  double fr0, fr1, fr2, fr3;
  short mvq;
  int i, j;

  fr0 = itabfr(1);   /* i-4 */
  fr1 = itabfr(2);   /* i-3 */
  fr2 = itabfr(3);   /* i-2 */
  fr3 = itabfr(4);   /* i-1 */

  for(i=5; i<nofcp-5; i++)
    {

/* First class: presystole without compensatory pause */

    if(fr1 > (fr0 * 1.3) && fr2 < (fr0 * 1.2) && fr2 > (fr0 * 0.9))
      {
      mvq = itab[i-1].pl - itab[i-2].pl;
      mvq -= (itab[i-2].pl - itab[i-3].pl);
      for(j=i-2; j<nofcp; j++)
        itab[j].pl += mvq;
      
      }

  goon:
    fr0 = fr1;
    fr1 = fr2;
    fr2 = fr3;
    fr3 = itabfr(i);
    }
  }


enneat_VEBs(short nofcp)

  {
  double fr0, fr1, fr2, fr3;
  double frplus;
  short changed;
  int i, j;

  changed = 0;

  fr0 = itabfr(1);   /* i-4 */
  fr1 = itabfr(2);   /* i-3 */
  fr2 = itabfr(3);   /* i-2 */
  fr3 = itabfr(4);   /* i-1 */

  for(i=5; i<nofcp-5; i++)
    {
    /* Second class: presystole with compensatory pause */

    if(fr1 > fr0*1.15 && fr2 < 0.95*fr0 && fr3 < fr0*1.1)
      {
      itab[i-3].pl =(short)(((long)(itab[i-4].pl) * 2 + (long)(itab[i-1].pl)) / (long)3);
      itab[i-2].pl =(short)(((long)(itab[i-4].pl) + (long)(itab[i-1].pl) * 2) / (long)3);
      }

   goon:
     fr0 = fr1;
     fr1 = fr2;
     fr2 = fr3;
     fr3 = itabfr(i);
     }
  }

enneat_intercalates(short nofcp)

  {
  double fr0, fr1, fr2, fr3;
  double frplus;
  short changed;
  int i, j;

  changed = 0;

  fr0 = itabfr(1);   /* i-4 */
  fr1 = itabfr(2);   /* i-3 */
  fr2 = itabfr(3);   /* i-2 */
  fr3 = itabfr(4);   /* i-1 */

  for(i=5; i<nofcp-5; i++)
    {

    /* Fourth class: intercalated extrasystole */

    frplus = 60.0 / (60.0 / fr1 + 60.0 / fr2);
    if(frplus > 0.85 * fr3 && frplus < 1.15 * fr3)
      {
      for(j=i-2; j<nofcp-1; j++)
        itab[j].pl = itab[j+1].pl;
      changed --;
      }

  goon:
    fr0 = fr1;
    fr1 = fr2;
    fr2 = fr3;
    fr3 = itabfr(i);
    }
/*  if(fr3 > 1.2*fr2 || fr3 < 0.8*fr2)
    itab[i-1].pl = itab[i-2].pl + (itab[i-2].pl - itab[i-1].pl);*/
  }


enneat_TD(short n)

  {
  enneat_SVEBs(n);
  enneat_VEBs(n);
  enneat_pauses(n);
/*  enneat_intercalates(n);*/
  }

#define FFDSTEP 2

enneat_FD(short n) /* n is normally 1024, the number of samples */

  {
  short i, j, t, delta;
  short p1, p2;
  double asses;

  for(i=0; i<n; i++)
    {
    j = 2*i;  /* this is actually the pointer */
    if(j==0 && ((fftdata[j] > fftdata[j+1*FFDSTEP] * 1.3) ||
        (fftdata[j] < fftdata[j+1*FFDSTEP] * 0.8)))
      {
      fftdata[j] = fftdata[j+1*FFDSTEP];
      continue;
      }
    if(i==(n-1) && ((fftdata[j] > fftdata[j-FFDSTEP] * 1.3) ||
       (fftdata[j] < fftdata[j+FFDSTEP] * 0.8)))
      {
      fftdata[j] = fftdata[j-FFDSTEP];
      continue;
      }
    if(i==n-1 || i==0)
      continue;
    if(fftdata[j] < fftdata[j-FFDSTEP] * 0.8)
      {
      for(delta=0; i+delta<n && fftdata[j]==fftdata[j+delta*FFDSTEP];)
            delta++;
      if(fftdata[j-FFDSTEP] < 1.1 * fftdata[j+delta*FFDSTEP] &&
         fftdata[j-FFDSTEP] > 0.9 * fftdata[j+delta*FFDSTEP])
         {
         for(t=0; t<delta; t++)
           fftdata[j+t] = (fftdata[j-FFDSTEP] * (delta-t) +
             fftdata[j+delta*FFDSTEP] * t) / delta;
         }
      continue;
      }
/*    if(fftdata[j] > fftdata[j-FFDSTEP] * 1.3 && 
      fftdata[j-FFDSTEP] < fftdata[j+FFDSTEP] * 1.2 &&
      fftdata[j-FFDSTEP] > fftdata[j+FFDSTEP] * 0.8)
      {
      fftdata[j] = (fftdata[j-FFDSTEP] + fftdata[j+FFDSTEP])/2;
      continue;
      }*/
    if(i>n-3)
      continue;
    for(p1=0; fftdata[j+p1*FFDSTEP]==fftdata[j]; p1++);
    for(p2=p1; fftdata[j+p1*FFDSTEP]==fftdata[j+p2*FFDSTEP]; p2++);
    if(i+p2>=n) continue;

    if(fftdata[j]==0) fftdata[j] = 1;
    if(fftdata[j+p1*FFDSTEP]==0) fftdata[j+p1*FFDSTEP] = 1;

    asses = 60.0 / ( 60.0 / fftdata[j] + 60.0 / fftdata[j+p1*FFDSTEP]);
/*    if(p2>2)
       printf("p2=%d p1=%d f(p2)=%5.1f f(p1)=%5.1f\n",
         p2, p1, fftdata[j+p2*FFDSTEP], fftdata[j+p1*FFDSTEP]);*/

    if(fftdata[j-FFDSTEP] < fftdata[j+p2*FFDSTEP] * 1.2 &&
      fftdata[j-FFDSTEP] > fftdata[j+p2*FFDSTEP] * 0.8 &&
      asses < 1.2 * fftdata[j-FFDSTEP] && asses > 0.8 * fftdata[j-FFDSTEP])
      {
/*      printf("asses asses=%5.1f\n", asses);*/
      for(t=p2; t>=1; t--)
        {
        fftdata[j+(t-1)*FFDSTEP] = ((p2-t)*fftdata[j-FFDSTEP] + 
          t*fftdata[j+FFDSTEP*p2]) / p2;
        }
      i += p2-1;
      continue;
      }
    asses *= 2;
    if(fftdata[j] > 1.2 * fftdata[j-FFDSTEP] &&
      fftdata[j-FFDSTEP] < fftdata[j+p2*FFDSTEP] * 1.2 &&
      fftdata[j-FFDSTEP] > fftdata[j+p2*FFDSTEP] * 0.8  &&
      asses < 1.2 * fftdata[j-FFDSTEP] && asses > 0.8 * fftdata[j-FFDSTEP])
      {
/*      printf("asses=%5.1f p2=%d f(j-1)=%5.1f f(j+p2)=%5.1f\n", asses, p2,
         fftdata[j-FFDSTEP], fftdata[j+p2*FFDSTEP]);*/
      for(t=p2; t>=1; t--)
        {
        fftdata[j+(t-1)*FFDSTEP] = ((p2-t)*fftdata[j-FFDSTEP] + 
          t*fftdata[j+FFDSTEP*p2]) / p2;
        }
      i += p2-1;
      continue;
      }
    }
  }


/*--
  \end{code}

  \funparagraph{td\_hrv, build\_TD}  % The name of the function(s)
  \ImplementationNotes{
  % implementation decisions, discarded implementation alternatives

  The following functions are taken from the ancient {\bf GlikSee}
  experimental program, where they constructed the data for the application
  of the fourier transform, eventually using the ``enneat'' functions
  above---to filter the extrasystoles. The original comments follow:
  
  This function uses the |enneat_FD| function to create
  a filtered time-domain diagram and prepares
  data for Cooley-Tukey transformation by
  the |four| function.

  This function also calls the time-domain building function
  |td_hrv| which computes the time-domain indexes of heart
  rate variability.


  }
  \Portability{
  }
  \Changes{
  }

  \begin{code}
  --*/


long tdv[5000];
short hrvoffs = 0;

double pNN50f, pNN50t, pNN50;
double rMSSD;
double clv;

init_td_hrv()

{
  hrvoffs = 0;
}

td_hrv(short nofts)

{
  short i;
  
  for(i=3; i<nofts-3; i++)
    tdv[i-1+hrvoffs] = 5 * abs(itab[i].pl - itab[i-1].pl);
  hrvoffs += nofts-7;
}

compute_td_hrv()

{
  short i;
  double lmed;

  clv = 0.0;
  pNN50t = pNN50f = 0.0;
  rMSSD = 0.0;
  lmed = 0.0;
  for(i=0; i<hrvoffs-1; i++) 
    lmed += tdv[i];
  lmed /= (i+1);
/*  printf("lmed=%10.2f, hrvoffs=%d\n", lmed, hrvoffs);*/
  for(i=0; i<hrvoffs-1; i++) {
    clv += ((abs(tdv[i])-lmed) * (abs(tdv[i])-lmed));
/*    printf("%10.2f[%d] ", (abs(tdv[i])-lmed), i);*/
    if(i>0 && (abs(tdv[i])-abs(tdv[i-1]))>=50)
      pNN50t ++;
    else
      pNN50f ++;
    if(i>0)
      rMSSD += (abs(tdv[i])-abs(tdv[i-1])) * (abs(tdv[i])-abs(tdv[i-1]));
    }
  pNN50 = 100.0 * pNN50t / (pNN50t + pNN50f);
  rMSSD = sqrt(rMSSD/i);
  clv = sqrt(clv/hrvoffs);
    
}

build_FD(int tdneatf, int fdneatf)

  {
  int i,j,k;
  int nofts;
  int crtpt;
  double oldfr, crtfr, newcrtfr;

  for(i=0; i<2100; i++)
    fftdata[i] = 0.0;
  init_td_hrv();
  for(j=crtblock, crtpt = 0; j<nblocks && j<crtblock+8 && crtpt<1024; j++)
    {
    getblock(j);
    ungarble();
    nofts = makediag(extra, 16000, 0);
    if(tdneatf) 
      enneat_TD(nofts);

    td_hrv(nofts);

    k = 1;
    for(i=0; i<160 && (i+crtpt)<1024; i++)
      {
      while(itab[k+1].pl < i*100)
        k++;
/*      printf("%5d", k);*/
      newcrtfr = 12000.0/(itab[k+1].pl - itab[k].pl);
/*      if(newcrtfr > 150 && crtfr<120)
        newcrtfr = crtfr;*/
      if(i>157)
        newcrtfr = crtfr;
      crtfr = newcrtfr;
      fftdata[2 * (i+crtpt)] = crtfr;
      fftdata[2 * (i+crtpt) + 1] = 0.0;   /* the complex part */
      }
    crtpt += i;
    }
  compute_td_hrv();
  getblock(crtblock);
  ungarble();
/*  if(fdneatf)
    enneat_FD(1024);*/
  }
#endif

/*--
  \end{code}

  \funparagraph{hrv\_spectrum}  % The name of the function(s)
  \ImplementationNotes{
  % implementation decisions, discarded implementation alternatives
  }
  \Portability{
  }
  \Changes{
  }

  \begin{code}
  --*/

double hrv_spectrum[MAX_FOUR_SIZE];
double hrv_windowed[MAX_FOUR_SIZE];
double hrv_sampled[MAX_FOUR_SIZE];
double hrv_original[MAX_FOUR_SIZE]; /* before applying ectopy correction */

short ectopy_filter_error = 0;
short ectopy_ratio = 0;

double hrv_data[MAX_FOUR_SIZE * 2];
long  r_flag[MAX_FOUR_SIZE];    /* 0 if r_data steady, 1 if aberant */
long  r_data[MAX_FOUR_SIZE];    /* given or corrected RR interval */
long  qrs_data[MAX_FOUR_SIZE];  /* assumed not systematically more */
short data_fill = 0;            /* fill in qrs_data, which has one
				   more element than r_data */
long  restep = 1250;            /* 125 ms, 8/sec */


/* ADJUSTABLE PARAMETERS */

short ectopy_filter_f = 1;           /* use filter for ectopies */
short presystole_coefficient = 15;   /* % short, to be considered presystole */
short late_systole_coefficient = 10; /* % long, to be considered late */
short sustained_rhythm_length = 4;   /* RR's in a row to make
					a sustained rhythm */
short aberancy_len_tolerance = 3;    /* nr of aberant succesive
					RR's tolerated */

/* CUMULATIVE TIME DOMAIN MEASURE VARIABLES & FUNCTIONS */

long pNN50_total;
long pNN50_actual;
double CLV_Sx;
double CLV_Sx2;
long CLV_N;
double r_MSSD_Sx;
long r_MSSD_N;

double pNN50;
double r_MSSD;
double CLV;
double CLM;


init_time_domain_parameters()

{
  pNN50_total = 0;
  pNN50_actual = 0;
  CLV_Sx = 0.0;
  CLV_Sx2 = 0.0;
  CLV_N = 0;
  r_MSSD_Sx = 0.0;
  r_MSSD_N = 0;
}

add_to_time_domain_parameters()

{
  long i;

  for(i=0; i<data_fill; i++) {
    if(!r_flag[i]) {
      CLV_Sx += r_data[i] / 10; /* (in ms) */
      CLV_Sx2 += r_data[i] * (double)r_data[i] / 100.0; /* in ms*ms */
      CLV_N ++;
    }
  }

  for(i=1; i<data_fill; i++) {
    if(!r_flag[i] && !r_flag[i-1]) {
      pNN50_total ++;
      if(r_data[i] - r_data[i-1] > 500)
	pNN50_actual ++;
      r_MSSD_Sx += (r_data[i] - r_data[i-1]) *
	(double)(r_data[i] - r_data[i-1]) / 100.0;
      r_MSSD_N ++;
    }
  }
}


compute_time_domain_parameters()

{
  if(pNN50_total > 0)
    pNN50 = (double)pNN50_actual * 100.0 / pNN50_total;
  else
    pNN50 = -1.0;
  if(CLV_N > 0) {
    CLM = CLV_Sx / CLV_N;
    CLV = sqrt(CLV_Sx2 / CLV_N - CLV_Sx * CLV_Sx / (CLV_N * CLV_N));
  }
  else {
    CLM = -1.0;
    CLV = -1.0;
  }
  if(r_MSSD_N > 0)
    r_MSSD = sqrt(r_MSSD_Sx / r_MSSD_N);
  else
    r_MSSD = -1.0;
}

/* END OF ADJUSTABLE PARAMETERS */


short set_ectopy_filter(short f)

{
  ectopy_filter_f = f;
}

short ectopy_filter()

     /* return 0 if ectopies could be successfully filtered */
     /* return 1 if, for example, the signal was atrial fibrillation */
     /* return 2 if rhythm aberancy is intolerable */

{
  long srv;
  long first_srv;
  short i, j, k, n_in_row;

  ectopy_ratio = 0;
  if(ectopy_filter_f) {

    /* compute the first ``current sustained rhythm'' value */
    
    for(i=0; i<data_fill-sustained_rhythm_length-1; i++) {
      srv = r_data[i];
      for(j=1; j<sustained_rhythm_length; j++) {
	if((srv - r_data[i+j]) >
	   ((presystole_coefficient * (double)srv) / 100.0) )
	  goto presys_1;
	if((r_data[i+j] - srv) >
	   ((late_systole_coefficient * (double)srv) / 100.0) )
	  goto latesys_1;
	srv = (srv * j + r_data[i+j]) / (j+1);
      }
      goto srv_is_ok;
    presys_1:
    latesys_1:
      i = i+j+1;
      continue;
    }
    return 1;
    
  srv_is_ok:

    /* rhythm abberancy marking */

    first_srv = srv;
    n_in_row = 0;
    for(i=0; i<data_fill-1; i++) {
      r_flag[i] = 0;
      if((srv - r_data[i]) > (presystole_coefficient * (double)srv / 100.0))
	r_flag[i] = 1;
      if((r_data[i] - srv) > (late_systole_coefficient * (double)srv / 100.0))
	r_flag[i] = 2;
      
      if(!r_flag[i])
	n_in_row = 0;
      else
	n_in_row ++;

      if(n_in_row > aberancy_len_tolerance)
	return 2;
    }

    /* linear interpolation of the RR interval lengths */

    for(i=0; i<data_fill-1; i++) {
      if(r_flag[i]) {
	for(j=1; i+j < data_fill-1; j++) {
	  if(!r_flag[i+j])
	    break;
	}
	if(i==0) {
	  for(k=0; k<j; k++)  /* j assumed smaller than data_fill */
	    r_data[k] = r_data[j];
	  i += j+1;
	  ectopy_ratio += j;
	  continue;
	}

	if(i+j >= data_fill-1) {
	  for(k=0; i+k<data_fill-1; k++)  /* j assumed
					     smaller than data_fill */
	    r_data[i+k] = r_data[i-1];
	  i += j+1;
	  ectopy_ratio += j;
	  continue;  /* will break */
	}
	
	/* the general case, with proper linear interpolation */
	for(k=0; k<j; k++) 
	  r_data[i+k] = r_data[i-1] + ((r_data[i+j] - r_data[i-1]) *
				       (double)(k+1) /
				       (double)(j+1));
	i += j+1;
	ectopy_ratio += j;
	continue;
      }
    }
  }
  ectopy_ratio = (short)((double)ectopy_ratio * 100.0 / data_fill);
  return 0;
}


short sample_qrs_for_spectrum(PERFILE *pf, long m0, long n)

/* m0 is the starting moment in ccs */
/* n is the number of restep samples to compute HRV on, e.g. 1024 */
/* Returns -1 if error, 0..MAX_FOUR_SIZE-1 if some complexes were found */

{
  long i;
  short j;
  short p;
  short cbt;
  long r0, r1;
  long m1;

  m1 = m0 + n * restep;
  cbt = complexes_between(pf, m0, m1) + 10;
  if(cbt<=0)
    return -1;  /* no complexes found */
  if(cbt>=MAX_FOUR_SIZE)
    return -1;  /* too many found, something is wrong */
  fprintf(ecgnosislog, "Complexes_between succeded in hrv_sample\n");
  reqitab(pf, m0, m1);
  cpars_at(pf, qrs_data, CPQRSPOS, m0, cbt);
  fprintf(ecgnosislog, "cpars_at succeded in hrv_sample\n");

  data_fill = cbt;
  for(j=0; j < cbt-1; j++)
    r_data[j] = qrs_data[j+1] - qrs_data[j];
  ectopy_filter_error = ectopy_filter();

  j = 0; /* cursor in qrs_data */
         /* p is cursor in hrv_data */
  for(i=m0, p=0; i<m1 & j < cbt-2; i+=restep) {
    r0 = r_data[j];   /* qrs_data[j+1] - qrs_data[j]; */
    r1 = r_data[j+1]; /* qrs_data[j+2] - qrs_data[j+1];*/

    /* case 1: marginal interpolation before j=0 */
    if(i<=qrs_data[0]) {
      hrv_data[p++] = r0 - ((r1 - r0) * (qrs_data[0] - i)) / (double)r0;
      hrv_data[p++] = 0.0;
      continue;
    }

    /* increment j if necessary */
    while(i>qrs_data[j+1] && j < cbt-2) {
/*      fprintf(ecgnosislog, "i=%d j=%d cbt=%d qrs_data[j+1]=%d\n", i, j, cbt, qrs_data[j+1]);*/
      j++;
    }
    if(j>=cbt-2)
      return -1;  /* abnormal situation, too few data read */
        
    /* case 2: value between two known moments */
    /*         the most common case */
    /*         BUT also the case when i is past the last qrs_data[j] */
    
    r0 = r_data[j];   /* qrs_data[j+1] - qrs_data[j]; */
    r1 = r_data[j+1]; /* qrs_data[j+2] - qrs_data[j+1];*/

    hrv_data[p++] = r0 + ((r1 - r0) * (i - qrs_data[j])) / (double)r0;
    hrv_data[p++] = 0.0;
  }
  fprintf(ecgnosislog, "for terminated in hrv_sample, p=%d.\n", p);
  while(p<MAX_FOUR_SIZE * 2)
    hrv_data[p++] = 0;
  return 1;
}

double high_hf_boundary = 0.40;  /* Hz */
double low_hf_boundary  = 0.15;  /* Hz */
double high_lf_boundary = 0.15;  /* Hz */
double low_lf_boundary  = 0.04;  /* Hz */

double lf_area = 0.0;  /* sq s */
double hf_area = 0.0;  /* sq s */

double spec_resolution = 1.0;  /* Hz per component */

short compute_hf_lf_spectrum(PERFILE *p, long m0)

{
  short i;
  double pp, h0, hn;
  short n;

  n = 1024;

  if(ectopy_filter_f) {
    ectopy_filter_f = 0;
    if(sample_qrs_for_spectrum(p, m0, n) < 0) {
      ectopy_filter_f = 1;
      return -1;
    }
    for(i=0; i<n; i++)
      hrv_original[i] = hrv_data[2*i];
    ectopy_filter_f = 1;
  }
  
  if(sample_qrs_for_spectrum(p, m0, n) < 0)
    return -1;

  if(!ectopy_filter_f) 
    for(i=0; i<n; i++)
      hrv_original[i] = hrv_data[2*i];

  spec_resolution = (double)(10000 / restep) / n;

  init_time_domain_parameters();
  add_to_time_domain_parameters();
  compute_time_domain_parameters();

/*
  h0 = hrv_data[0];
  hn = hrv_data[2*n - 2];
  for(i=0; i<n; i++) {
    pp = h0 + (i * (hn - h0)) / n;
    hrv_data[2*i] -= pp;
  } */

  for(i=0, pp=0.0; i<n; i++)
    pp += hrv_data[2*i];
  pp /= n;
  for(i=0; i<n; i++) {
    hrv_data[2*i] -= pp;
    hrv_original[i] -= pp;
  }

  fprintf(ecgnosislog, "spec_resolution=%10.5f\n", spec_resolution);

  for(i=0; i<n; i++) {
    hrv_sampled[i] = hrv_data[2*i];
    /*    fprintf(ecgnosislog, "hrv_data[%d] = %10.2f\n", i, hrv_data[2*i]);*/
  }
  
  rv_compute_window(n);
  rv_apply_window(hrv_data, n);
  for(i=0; i<n; i++) {
    hrv_windowed[i] = hrv_data[2*i];
    /*    fprintf(ecgnosislog, "hrv_windowed[%d] = %10.2f\n", i, hrv_windowed[i]);*/
  }
  fourier_transform(hrv_data, n, 1);
  for(i=0; i<n; i++)
    hrv_spectrum[i] = (hrv_data[2*i] * hrv_data[2*i] + 
                           hrv_data[2*i+1] * hrv_data[2*i+1]) *
			     window_factor * window_factor
			     * 2.0 / n; /* 2 / N by the book */
  for(i=(short)(low_lf_boundary / spec_resolution), lf_area = 0.0; 
      i<(short)(high_lf_boundary / spec_resolution);
      i++)
    lf_area += hrv_spectrum[i] * spec_resolution; /* not / 100 */

  for(i=(short)(low_hf_boundary / spec_resolution), hf_area = 0.0; 
      i<(short)(high_hf_boundary / spec_resolution);
      i++)
    hf_area += hrv_spectrum[i] * spec_resolution; /* not / 100 */

  return 0;
}

/* The following subpackage will compute
   a HRV spectrum centered around the 0.02Hz (0.0033--0.04 Hz),
   in order to permit description of the VLF band
   */

double vlf_buffer[MAX_FOUR_SIZE];

double vlf_area;

double low_vlf_boundary = 0.0033;
double high_vlf_boundary = 0.04;

short vlf_filter_and_resample(double *where, double org_val)

{
  short i;

  /* filter the contents of the hrv_data */

  for(i=0; i<1024; i++) {
    hrv_data[2*i] = 0.2 * hrv_data[2*i] + 0.8 * org_val;
    org_val = hrv_data[2*i];
  }

  /* resample at every 4'th sample */
  for(i=0; i<256; i++)
    where[i] = hrv_data[i*8];

  return 1;
}

short compute_vlf_spectrum(PERFILE *p, long m0)

{
  short i;
  double pp, h0, hn, olv;
  short n;

  n = 1024;


  init_time_domain_parameters();
  
  for(i=0; i<4; i++) {
    if(sample_qrs_for_spectrum(p, m0 + 1024 * 1250 * i, 1024) < 0)
      return -1;  /* must somehow mark this error in the caller */
    
    add_to_time_domain_parameters();

    if(i==0)
      olv = hrv_data[0];
    else
      olv = vlf_buffer[256*i - 1];
    
    vlf_filter_and_resample(vlf_buffer + 256 * i, olv);
  }

  compute_time_domain_parameters();

  for(i=0; i<1024; i++) {
    hrv_data[2*i] = vlf_buffer[i];
    hrv_data[2*i + 1] = 0.0;
  }
  
  spec_resolution = (double)(10000 / restep) / (n) / 4;

  for(i=0, pp=0.0; i<n; i++)
    pp += hrv_data[2*i];
  pp /= n;
  for(i=0; i<n; i++) {
    hrv_data[2*i] -= pp;
    hrv_original[i] -= pp;
  }

  for(i=0; i<n; i++)
    hrv_sampled[i] = hrv_data[2*i];
  
  rv_compute_window(n);
  rv_apply_window(hrv_data, n);
  for(i=0; i<n; i++)
    hrv_windowed[i] = hrv_data[2*i];
  fourier_transform(hrv_data, n, 1);
  for(i=0; i<n; i++)
    hrv_spectrum[i] = (hrv_data[2*i] * hrv_data[2*i] + 
                           hrv_data[2*i+1] * hrv_data[2*i+1]) * window_factor
			     * window_factor * 2.0 / n; /* 2 / N by the book */

  for(i=(short)(low_vlf_boundary / spec_resolution), vlf_area = 0.0; 
      i<(short)(high_vlf_boundary / spec_resolution);
      i++)
    vlf_area += hrv_spectrum[i] * spec_resolution / 100;

  return 0;
}



/*--
  \end{code}

  \endoffile

  --*/

