
/*--

  \FileTitle{
  {\tt egdiagsc.c} --- Electrocardiognosis diagrams: the scatter diagram
  }
  \CopyrightNotice{

    Copyright (c) 1991--1996 Alexandru Dan Corlan MD (http://dan.corlan.net)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  }

  \subsection{Scatter Diagrams}

  \begin{code}
  --*/

#include <stdio.h>
#include <malloc.h>
#include <math.h>

#include "gliks.h"
#include "ecgnosis.h"



/*--
  \end{code}
  
  \dataparagraph{Scatter Diagram Description}
  
  \Definition{
  This data paragraph contains variables that are exported,
  or whose values are exported, with the results of computing
  scattergrams. 
  }
  \SemanticStructure{
  Scattergrams described in this section are squares of counts of
  points. They are stored in the dynamically allocated arrays
  \variabletag{scat} and \variabletag{dscat}. The size of the arrays
  is in \variabletag{scat\_dim}.
  }
  \ConsistencyRules{
  The \variabletag{scat} and the \variabletag{scat\_dim} arrays must contain
  either a vector of \variabletag{scat\_dim} pointers to vectors
  of \variabletag{scat\_dim} shorts, or be both NULL---in which case the
  value of \variabletag{scat\_dim is not used by anybody.}
  }

  \begin{code}
  --*/

long scat_dim = 200;

short **scat = NULL;
short **dscat = NULL;  /* i.e. the scat to display */

/*--
  \end{code}

  \funparagraph{alloc\_scats}
  \ImplementationNotes{

  \begin{code}
  --*/

short alloc_scats(short d)

{
  short i, j;

  /* Dissallocate the structures if they already exist */  
  if(scat!=NULL && scat_dim!=d) {
    for(i=0; i<scat_dim; i++)
      free(scat[i]);
    free(scat);
    scat = NULL;
  }
  if(dscat!=NULL && scat_dim!=d) {
    for(i=0; i<scat_dim; i++)
      free(dscat[i]);
    free(dscat);
    dscat = NULL;
  }
  if(scat!=NULL && dscat!=NULL) {
    clear_scat();
    return 1;  /* consider allocation unnecessary */
  }

  /* Allocate new structures now */
  scat_dim = d;
  scat = (short **)malloc(scat_dim * sizeof(short *));
  if(scat==NULL)
    return 0;  /* failure */
  for(i=0; i<scat_dim; i++) {
    scat[i] = (short *)malloc(scat_dim * sizeof(short));
    if(scat[i]==NULL) {
      i--;
      while(i>=0) {
	free(scat[i]);
	i--;
      }
      free(scat);
      return 0;
    }
    for(j=0; j<scat_dim; j++)
      scat[i][j] = 0;
  }
  dscat = (short **)malloc(scat_dim * sizeof(short *));
  if(dscat==NULL) {
    for(i=0; i<scat_dim; i++)
      free(scat[i]);
    free(scat);
    return 0;
  }
  for(i=0; i<scat_dim; i++) {
    dscat[i] = (short *)malloc(scat_dim * sizeof(short));
    if(dscat[i]==NULL) {
      i--;
      while(i>=0) {
	free(dscat[i]);
	i--;
      }
      free(dscat);
      for(i=0; i<scat_dim; i++)
	free(scat[i]);
      free(scat);
      return 0;
    }
    for(j=0; j<scat_dim; j++)
      dscat[i][j] = 0;
  }
  return 1;
}
  

/* SCATTER-GRAM CONSTRUCTION                                 */
/* This function will grow a lot when expressions involving  */
/* computation of each dimension will have to be interpreted */
/* For the moment it just computes the frequency diagram     */
/* It returns 0 if error and 1 if success                    */

short jadd_to_scat(long *cbuf, long cblen) {

  long i, p1, p2;

  for(i=1; i<cblen-1; i++) {
    p1 = 600000 / (cbuf[i]-cbuf[i-1]+1);
    p2 = 600000 / (cbuf[i+1]-cbuf[i]+1);
#ifdef MAINTRACE
    fprintf(ecgnosislog, "B");
    fflush(ecgnosislog);
#endif
    if(p1<scat_dim && p1>=0 && p2<scat_dim && p2>=0) {
#ifdef MAINTRACE
      fprintf(ecgnosislog, "I");
      fflush(ecgnosislog);
#endif
      scat[p1][p2] ++;
    }
#ifdef MAINTRACE
    fprintf(ecgnosislog, "A");
    fflush(ecgnosislog);
#endif
  }
}

short add_to_scat(PERFILE *p, long m0, long m1) {

  long cbt;
  long *cbuf;
  short c;

  cbt = complexes_between(p, m0, m1);
#ifdef MAINTRACE
  fprintf(ecgnosislog, "CBT in SCAT_ADD is %ld [%ld, %ld]\n", cbt, m0, m1);
  fflush(ecgnosislog);
#endif
  cbuf = NULL;
  cbuf = (long *)malloc(cbt * sizeof(long));
  if(cbuf==NULL)
    return 0;
  reqitab(p, m0, m1);
  cpars_at(p, cbuf, CPQRSPOS, m0, cbt);
#ifdef MAINTRACE
  fprintf(ecgnosislog, "cpars_at returned.\n");
  for(c=0; c<cbt; c++)
    fprintf(ecgnosislog, "CPARS RESULT IN SCAT_ADD #%d = %ld\n", c, cbuf[c]);
  fflush(ecgnosislog);
#endif MAINTRACE

  jadd_to_scat(cbuf, cbt);
  fprintf(ecgnosislog, "AFTER JADD\n");
  free(cbuf);
  return 1;
}

void clear_scat()

{
  short i, j;
  
  for(i=0; i<scat_dim; i++)
    for(j=0; j<scat_dim; j++)
      scat[i][j] = dscat[i][j] = 0;
}


/*--
  \end{code}

  \subsubsection{Scattergram processing}

  \begin{code}
  --*/


void scat_mediate(short d) {

  short i, j;
  short k, l;
  long s;

  for(i=0; i<scat_dim; i+=d)
    for(j=0; j<scat_dim; j+=d) {
      s = 0;
      for(k=0; k<d; k++) {
	if(i+k>=scat_dim)
	  continue;
	for(l=0; l<d; l++) {
	  if(l+j>=scat_dim)
	    continue;
	  s += scat[i+k][j+l];
	}
      }
      for(k=0; k<d; k++) {
	if(i+k>=scat_dim)
	  continue;
	for(l=0; l<d; l++) {
	  if(j+l>=scat_dim)
	    continue;
	  dscat[i+k][j+l] = s;
	}
      }
    }
}  /* done */



/*--
  \end{code}

  \subsubsection{Scattergram measurement}

  \ImplementationNotes{
  The computation
  primitives of this subpackage take data from the scattergrams
  and produce various results in the {\sl SCATMES} data objects below.

  The scattergrams may be divided into {\sl SECTORS} by the use
  of {\sl SEPARATION BOUNDARIES}. For the moment, the boundaries
  used are vertical and horizontal separation lines. The
  computation primitives
  of the package compute various measurements (such as the number of
  points in the sector) for every, or for a part of the sectors.

  Every sector contains is lower and its leftmost separation boundaries, but
  does not contain the rightmost and the upper boundaries.
  }

  \dataparagraph{Boundary descriptions}

  \Definition{
  The following data structure describes the location of boundaries
  for the scattergram sectors. For the moment, there are only
  vertical and horizontal boundaries.
  }
  \SemanticStructure{
  The \variabletag{hor\_bounds} and the \variabletag{ver\_bounds}
  vectors contain the locations of the boundaries. These vectors
  have fill cursors \variabletag{hor\_bounds\_fill} --- noted below $F_H$
  and \variabletag{ver\_bounds\_fill}, noted below $F_V$.
  At the beginning, each of the vectors 
  contains one value which is 0. 
  The number of sectors is $F_H \times F_V$. The sectors are numbered
  as in figure~\ref{fig:vectnum}, and so are placed the values in the vectors.

  \begin{figure}

% The following is in FIGURE/egdiagsc-vectnum.fig or
% FIGURE/egdiagsc-vectnum.latex, both produced by {\bf xfig.} Some
% hand editing is mandatory if the figure files are changed, because
% {\bf xfig} does not handle math in figure (or I don't know
% how to tell it to). A.D.C.

  \setlength{\unitlength}{0.012500in}%
%
\begingroup\makeatletter\ifx\SetFigFont\undefined
% extract first six characters in \fmtname
\def\x#1#2#3#4#5#6#7\relax{\def\x{#1#2#3#4#5#6}}%
\expandafter\x\fmtname xxxxxx\relax \def\y{splain}%
\ifx\x\y   % LaTeX or SliTeX?
\gdef\SetFigFont#1#2#3{%
  \ifnum #1<17\tiny\else \ifnum #1<20\small\else
  \ifnum #1<24\normalsize\else \ifnum #1<29\large\else
  \ifnum #1<34\Large\else \ifnum #1<41\LARGE\else
     \huge\fi\fi\fi\fi\fi\fi
  \csname #3\endcsname}%
\else
\gdef\SetFigFont#1#2#3{\begingroup
  \count@#1\relax \ifnum 25<\count@\count@25\fi
  \def\x{\endgroup\@setsize\SetFigFont{#2pt}}%
  \expandafter\x
    \csname \romannumeral\the\count@ pt\expandafter\endcsname
    \csname @\romannumeral\the\count@ pt\endcsname
  \csname #3\endcsname}%
\fi
\fi\endgroup
\begin{picture}(360,270)(40,540)
\thinlines
\put( 95,570){\framebox(240,240){}}
\thicklines
\put( 95,810){\line( 0,-1){240}}
\put( 95,570){\line( 1, 0){240}}
\put(165,810){\line( 0,-1){240}}
\put(205,810){\line( 0,-1){240}}
\put( 95,695){\line( 1, 0){240}}
\put(120,740){\makebox(0,0)[lb]{\smash{\SetFigFont{17}{20.4}{bf}0}}}
\put(180,740){\makebox(0,0)[lb]{\smash{\SetFigFont{17}{20.4}{bf}1}}}
\put(260,740){\makebox(0,0)[lb]{\smash{\SetFigFont{17}{20.4}{bf}2}}}
\put(120,620){\makebox(0,0)[lb]{\smash{\SetFigFont{17}{20.4}{bf}3}}}
\put(180,620){\makebox(0,0)[lb]{\smash{\SetFigFont{17}{20.4}{bf}4}}}
\put(260,620){\makebox(0,0)[lb]{\smash{\SetFigFont{17}{20.4}{bf}5}}}
\put( 80,540){\makebox(0,0)[lb]{\smash{\SetFigFont{12}{14.4}{rm}$V_0$}}}
\put(140,540){\makebox(0,0)[lb]{\smash{\SetFigFont{12}{14.4}{rm}$V_1$}}}
\put(200,540){\makebox(0,0)[lb]{\smash{\SetFigFont{12}{14.4}{rm}$V_2$}}}
\put( 40,560){\makebox(0,0)[lb]{\smash{\SetFigFont{12}{14.4}{rm}$H_0$}}}
\put( 40,700){\makebox(0,0)[lb]{\smash{\SetFigFont{12}{14.4}{rm}$H_1$}}}
\put(400,760){\makebox(0,0)[lb]{\smash{\SetFigFont{12}{14.4}{rm}$F_H=2$}}}
\put(400,740){\makebox(0,0)[lb]{\smash{\SetFigFont{12}{14.4}{rm}$F_V=3$}}}
\put(400,700){\makebox(0,0)[lb]{\smash{\SetFigFont{12}{14.4}{rm}$H_{0..F_H}$ 
is in {\tt hor\_bounds\_fill}}}}
\put(400,682){\makebox(0,0)[lb]{\smash{\SetFigFont{12}{14.4}{rm}$V_{0..F_V}$ 
is in {\tt ver\_bounds\_fill}}}}
\end{picture}


  \caption{Use of boundary vectors to segment the scattergram, and
           the numbering of the sectors}\label{fig:vectnum}
  \end{figure}

  }
  \ConsistencyRules{
  Every vector must have mutually different values, which are sorted
  in ascending order. The first element of each vector
  must be 0. The FILL-th element of the vector must be equal to
  \variabletag{scat\_dim}.
  }

  \begin{code}
  --*/

short ver_bounds[MAX_BOUNDS] = {0 };
short ver_bounds_fill = 1;

short hor_bounds[MAX_BOUNDS] = {0 };
short hor_bounds_fill = 1;


/*--
  \end{code}

  \funparagraph{clear\_scat\_bounds}

  \begin{code}
  --*/

void clear_scat_bounds()

{
  ver_bounds[0] = 0;
  ver_bounds[1] = scat_dim;
  ver_bounds_fill = 1;
  hor_bounds[0] = 0;
  hor_bounds[1] = scat_dim;
  hor_bounds_fill = 1;
}

/*--
  \end{code}

  \funparagraph{\_add\_scat\_bound}

  \PurposeReturn{  % be short
  Inserts a boundary value in one of the scat bound vectors. Returns
  1 if success (to be added to the fill value) and 0 if error---as
  the number of values added is 0. It uses \variabletag{scat\_dim}
  to check that the operation is not out of range.
  }
  \Exceptions{
  \begin{description}
  \item[Scat\_Bound\_Out\_Of\_Range] if the argument happens
  to be out of range (which is a calling program error).
  \item[Scat\_Bound\_Already\_Exists] if the argument happens
  to be already defined as a bound.
  \item[Too\_Many\_Scat\_Bounds] if there is no place
  for the new bound.
  \end{description}
  }
  \GlobalEffects{
  Changes the argument scat bounds vector.
  }
  \Preconditions{
  The argument scat bounds vector must contain that number
  of bounds specifyied by the argument fill, and these must
  be sorted in strictly ascending numerical order.
  }
  \Performance{
  Not of any concern.
  }

  \ImplementationNotes{
  An insertion in the list is done.
  }

  \begin{code}
  --*/

static short _add_scat_bound(short val, short *vector, short fill)

{
  short i, j;

  if(val<0 || val>scat_dim) {
    raise("Scat_Bound_Out_Of_Range");
    return 0;
  }
  if(fill >= MAX_BOUNDS) {
    raise("Too_Many_Scat_Bounds");
    return 0;
  }
  for(i=0; i<fill; i++) {
    if(vector[i]==val) {
      raise("Scat_Bound_Already_Exists");
      return 0;
    }
    if(val < vector[i])
      break;
  }
  for(j=fill; j>i; j--)
    vector[j] = vector[j-1];
  vector[i] = val;
  vector[fill+1] = scat_dim;
  return 1;  /* to be added to fill */
}


/*--
  \end{code}

  \funparagraph{add\_hor\_boundary, add\_ver\_boundary}

  \begin{code}
  --*/

short add_hor_boundary(short x)

{
  short r;

  r = _add_scat_bound(x, hor_bounds, hor_bounds_fill);
  hor_bounds_fill += r;
  return r;
}


short add_ver_boundary(short x)

{
  short r;

  r = _add_scat_bound(x, ver_bounds, ver_bounds_fill);
  ver_bounds_fill += r;
  return r;
}


/*--
  \end{code}

  \funparagraph{scat\_sector\_no} 

  \begin{code}
  --*/

short scat_sector_no(short x, short y)

{
  return (x + ver_bounds_fill * (hor_bounds_fill - 1 - y));
}


/*--
  \end{code}

  \funparagraph{ver\_bounds\_ix}

  \begin{code}
  --*/

short ver_bounds_ix(short sect)

{
  return (sect % ver_bounds_fill);
}


/*--
  \end{code}

  \funparagraph{hor\_bounds\_ix}

  \begin{code}
  --*/

short hor_bounds_ix(short sect)

{
  sect = sect / ver_bounds_fill;
  return (hor_bounds_fill - sect - 1);
}


/*--
  \end{code}

  \dataparagraph{}  % a name for the data structure

  \begin{code}
  --*/

struct scatmes sm[MAX_BOUNDS * MAX_BOUNDS];


/*--
  \end{code}

  \funparagraph{init\_sm} 

  \begin{code}
  --*/

void init_sm()

{
  short i;

  for(i=0; i<MAX_BOUNDS * MAX_BOUNDS; i++) {
    
    sm[i].volume = 0;
    sm[i].volg = SM_NOTCO;

    sm[i].surface = 0;
    sm[i].surg = SM_NOTCO;
    
    sm[i].xmed = 0.0;
    sm[i].xmeg = SM_NOTCO;

    sm[i].ymed = 0.0;
    sm[i].ymeg = SM_NOTCO;

    sm[i].lsup = 0;
    sm[i].lsg = SM_NOTCO;

    sm[i].linf = 0;
    sm[i].lig = SM_NOTCO;

    sm[i].lleft = 0;
    sm[i].llg = SM_NOTCO;

    sm[i].lright = 0;
    sm[i].lrg = SM_NOTCO;

    sm[i].xdisp = 0.0;
    sm[i].xdig = SM_NOTCO;

    sm[i].ydisp = 0.0;
    sm[i].ydig = SM_NOTCO;

    sm[i].corr = 0.0;
    sm[i].corg = SM_NOTCO;

    sm[i].oimed = 0.0;
    sm[i].oimg = SM_NOTCO;

    sm[i].oiimed = 0.0;
    sm[i].oiimg = SM_NOTCO;

    sm[i].dispi = 0.0;
    sm[i].dig = SM_NOTCO;

    sm[i].dispii = 0.0;
    sm[i].diig = SM_NOTCO;
  }
}

/*--
  \end{code}

  \funparagraph{_gen_scat_measures}

  \PurposeReturn{  % be short
  This function, with many arguments, performs the computation of
  the \variabletag{sm} data structure in the most general way.
  Particular forms of calling of this function should be
  used for practical purposes.
  }
  \GlobalEffects{
  Results in \variabletag{sm}.
  }
  \Preconditions{
  Bounds' vectors consistent with arguments.
  }
  \Performance{
  % How it is, not why and what to do about it
  }

  \ImplementationNotes{
  % How implementation of the function is done: algorithm---if special,
  % implementation decisions, discarded implementation alternatives

  Running it with float calculations would have taken very much
  time. Actually, about 1.000.000 FP computations would have
  had to be done for every double loop below. For this reason, we
  rewrote the formulae a little, but a limitation to a total
  of 30.000 points or so in the whole graphic might occur --- though
  we don't know what will actually happen in practice.
  }

  \begin{code}
  --*/

short _gen_scat_measures(short **sc, short scdim, short first, short last)

     /* compute all scat measurements from the first
	to the last of the sectors (including the first AND the last)
	*/

{
  short h_ix, v_ix;
  long i, x, y;
  long vol, sur;
  long xme, yme;
  short lsu, lin, lle, lri;
  long oim, oiim;
  long xdi, ydi, cor, di, dii;
  long sxy, sx2, sy2;

  /* CHECK THE ARGS HERE, ESP first AND last */
  for(i=first; i<=last; i++) {
    h_ix = hor_bounds_ix(i);
    v_ix = ver_bounds_ix(i);

    sxy = 0;
    sx2 = 0;
    sy2 = 0;
    
    sm[i].volume = 0;
    vol = 0;
    sm[i].volg = SM_NOTCO;

    sm[i].surface = 0;
    sur = 0;
    sm[i].surg = SM_NOTCO;
    
    sm[i].xmed = 0;
    xme = 0;
    sm[i].xmeg = SM_NOTCO;

    sm[i].ymed = 0;
    yme = 0;
    sm[i].ymeg = SM_NOTCO;

    sm[i].lsup = 0;
    lsu = 0;
    sm[i].lsg = SM_NOTCO;

    sm[i].linf = 0;
    lin = 0;
    sm[i].lig = SM_NOTCO;

    sm[i].lleft = 0;
    lle = 0;
    sm[i].llg = SM_NOTCO;

    sm[i].lright = 0;
    lri = 0;
    sm[i].lrg = SM_NOTCO;

    sm[i].xdisp = 0.0;
    xdi = 0.0;
    sm[i].xdig = SM_NOTCO;

    sm[i].ydisp = 0.0;
    ydi = 0.0;
    sm[i].ydig = SM_NOTCO;

    sm[i].corr= 0.0;
    cor = 0.0;
    sm[i].corg = SM_NOTCO;

    sm[i].oimed = 0.0;
    oim = 0.0;
    sm[i].oimg = SM_NOTCO;

    sm[i].oiimed = 0.0;
    oiim = 0.0;
    sm[i].oiimg = SM_NOTCO;

    sm[i].dispi = 0.0;
    di = 0.0;
    sm[i].dig = SM_NOTCO;

    sm[i].dispii = 0.0;
    dii = 0.0;
    sm[i].diig = SM_NOTCO;

    for(x=ver_bounds[v_ix]; x<ver_bounds[v_ix+1]; x++) {
      for(y=hor_bounds[h_ix]; y<hor_bounds[h_ix+1]; y++) {
	vol += sc[x][y];
	if(sc[x][y] > 0)
	  sur ++;
	xme  += sc[x][y] * x;
	yme  += sc[x][y] * y;
	
	oim  += (sc[x][y]) * (x + y);
	oiim += (sc[x][y]) * (x - y);
	
	sxy  += sc[x][y] * x * y;
	sx2  += sc[x][y] * x * x;
	sy2  += sc[x][y] * y * y;
      } /* for(y...) */
    } /* for(x...) */
    
    sm[i].volume = vol;
    sm[i].volg = SM_DONE;
    
    sm[i].surface = sur;
    sm[i].surg = SM_DONE;

    if(vol>0) {
      sm[i].xmed = ((float)xme) / vol;
      sm[i].xmeg = SM_DONE;
    }
    else {
      sm[i].xmeg = SM_ERROR;
    }

    if(vol>0) {
      sm[i].ymed = ((float)yme) / vol;
      sm[i].ymeg = SM_DONE;
    }
    else {
      sm[i].ymeg = SM_ERROR;
    }

    if(vol>0) {
      sm[i].oimed = oim / 1.4142 / vol;
      sm[i].oimg = SM_DONE;
    }
    else {
      sm[i].oimg = SM_ERROR;
    }

    if(vol>0) {
      sm[i].oiimed = oiim / 1.4142 / vol;
      sm[i].oiimg = SM_DONE;
    }
    else {
      sm[i].oiimg = SM_ERROR;
    }

    if(sx2>0 && sy2>0) {
      sm[i].corr = ((double)(sxy) * (double)(sxy)) /
	           ((double)(sx2) * (double)(sy2));
      sm[i].corg = SM_DONE;
    }
    else {
      sm[i].corg = SM_ERROR;
    }

    oim = (long)(sm[i].oimed * 1.4142);
    oiim = (long)(sm[i].oiimed * 1.4142);
    xme = sm[i].xmed;
    yme = sm[i].ymed;
    
    for(x=ver_bounds[v_ix]; x<ver_bounds[v_ix+1]; x++) {
      for(y=hor_bounds[h_ix]; y<hor_bounds[h_ix+1]; y++) {
	if(vol>0) {
	  xdi += sc[x][y] * (x - xme) * (x - xme);
	  ydi += sc[x][y] * (y - yme) * (y - yme);
	  di  += sc[x][y] * ((x + y) - oim) *
	                    ((x + y) - oim);
	  dii += sc[x][y] * ((x - y) - oiim) *
	                    ((x - y) - oiim);
	} /* if */
      } /* for(y...) */
    } /* for(x...) */

    di = di / 2;
    dii = dii / 2;

    if(vol>0) {
      sm[i].xdisp = sqrt((double)xdi / vol);
      sm[i].xdig  = SM_DONE;
    }
    else {
      sm[i].xdig  = SM_ERROR;
    }

    if(vol>0) {
      sm[i].ydisp = sqrt((double)ydi / vol);
      sm[i].ydig  = SM_DONE;
    }
    else {
      sm[i].ydig  = SM_ERROR;
    }

    if(vol>0) {
      sm[i].dispi = sqrt((double)di / vol);
      sm[i].dig   = SM_DONE;
    }
    else {
      sm[i].dig   = SM_ERROR;
    }

    if(vol>0) {
      sm[i].dispii = sqrt((double)dii / vol);
      sm[i].diig   = SM_DONE;
    }
    else {
      sm[i].diig   = SM_ERROR;
    }

#ifdef NOT_YET
    lsu = hor_bounds[h_ix+1] - 1;
    sm[i].lsg = SM_DONE;
    for(y=hor_bounds[h_ix+1] - 1; y>=hor_bounds[h_ix]; y--) {
      for(x = ver_bounds[v_ix]; x < ver_bounds[v_ix+1]; x++) {
	if(sc[x][y] > 0)
	  goto LSU_DONE;
      }
      lsu --;
    }
    sm[i].lsg = SM_ERROR;
  LSU_DONE:
    sm[i].lsup = lsu;

    lin = hor_bounds[h_ix];
    sm[i].lig = SM_DONE;
    for(y=hor_bounds[h_ix]; y<hor_bounds[h_ix+1]; y++) {
      for(x=ver_bounds[v_ix]; x<ver_bounds[v_ix+1]; x++) {
	if(sc[x][y] > 0)
	  goto LIN_DONE;
      }
      lin ++;
    }
    sm[i].lig = SM_ERROR;
  LIN_DONE:
    sm[i].linf = lin;

    lle = ver_bounds[v_ix];
    sm[i].llg = SM_DONE;
    for(x=ver_bounds[v_ix]; x<ver_bounds[v_ix+1]; x++) {
      for(y=hor_bounds[h_ix]; y<hor_bounds[h_ix+1]; y++) {
	if(sc[x][y] > 0)
	  goto LLE_DONE;
      }
      lle ++;
    }
    sm[i].llg = SM_ERROR;
  LLE_DONE:
    sm[i].lleft = lle;

    lri = ver_bounds[v_ix+1] - 2;
    sm[i].lrg = SM_DONE;
    for(x=ver_bounds[v_ix+1]-1; x>=ver_bounds[v_ix]; x--) {
      for(y = hor_bounds[h_ix]; y<hor_bounds[h_ix+1]; y++) {
	if(sc[x][y] > 0)
	  goto LRI_DONE;
      }
      lri --;
    }
    sm[i].lrg = SM_ERROR;
  LRI_DONE:
    sm[i].lright = lri;
#endif
  } /* for(i...) */    
}


/*--
  \end{code}

  \funparagraph{put\_scat\_to\_eps} 


  \begin{code}
  --*/


extern short n_scat_colors;
extern short n_scat_c[];
extern short n_scat_l[];

extern make_scat_limits();


short print_color_given_val(short v) { 			/* [FUN] */ 

  short i;

  for(i=0; i<4; i++) {
    if(v>=n_scat_l[i] && v<n_scat_l[i+1])
      return i;
  }
  return -1;
}

short put_scat_to_eps(FILE *f)

{
  short i, j, k, cc;

  /*--
    \end{code}

    \subparagraph{The bounding box and the preamble}

    \begin{code}
    --*/
  
  fprintf(f, "%%%%BoundingBox: 0 0 200 200\n");

  fprintf(f, "\n/grid-pen { 0.5 setgray } def\n");
  fprintf(f, "\n/spot0-pen { 0.1 setgray } def\n");
  fprintf(f, "\n/spot1-pen { 0.35 setgray } def\n");
  fprintf(f, "\n/spot2-pen { 0.7 setgray } def\n");
  fprintf(f, "\n/spot3-pen { 0.9 setgray } def\n");
 
  fprintf(f, "\n1.0 setlinewidth\n");

  fprintf(f, "\n/spot { moveto 1 1 rlineto stroke } def\n");

  fprintf(f, "\n/s0 { spot0-pen spot } def\n");
  fprintf(f, "\n/s1 { spot1-pen spot } def\n");
  fprintf(f, "\n/s2 { spot2-pen spot } def\n");
  fprintf(f, "\n/s3 { spot3-pen spot } def\n");

  /*--
    \end{code}

    \subparagraph{The grid of the scatter diagram}

    \begin{code}
    --*/

  fprintf(f, "\ngrid-pen 2.0 setlinewidth 0 0 moveto 200 0 rlineto\n");
  fprintf(f, "0 200 rlineto -200 0 rlineto 0 -200 rlineto closepath stroke\n");
  fprintf(f, "1.0 setlinewidth\n");
  fprintf(f, "25 0 moveto 0 200 rlineto stroke\n");
  fprintf(f, "50 0 moveto 0 200 rlineto stroke\n");
  fprintf(f, "75 0 moveto 0 200 rlineto stroke\n");
  fprintf(f, "2.0 setlinewidth 100 0 moveto 0 200 rlineto stroke\n");
  fprintf(f, "125 0 moveto 0 200 rlineto stroke\n");
  fprintf(f, "150 0 moveto 0 200 rlineto stroke\n");
  fprintf(f, "175 0 moveto 0 200 rlineto stroke\n");
  fprintf(f, "1.0 setlinewidth\n");
  fprintf(f, "0 25 moveto 200 0 rlineto stroke\n");
  fprintf(f, "0 50 moveto 200 0 rlineto stroke\n");
  fprintf(f, "0 75 moveto 200 0 rlineto stroke\n");
  fprintf(f, "2.0 setlinewidth 0 100 moveto 200 0 rlineto stroke\n");
  fprintf(f, "1.0 setlinewidth\n");
  fprintf(f, "0 125 moveto 200 0 rlineto stroke\n");
  fprintf(f, "0 150 moveto 200 0 rlineto stroke\n");
  fprintf(f, "0 175 moveto 200 0 rlineto stroke\n");

  make_scat_limits();
  for(i=0; i<4; i++) 
    fprintf(f, "%% scatl[i] %d\n", n_scat_l[i]);

  for(i=0; i<scat_dim; i++) {
    for(j=0; j<scat_dim; j++) {
      switch(cc = print_color_given_val(dscat[i][j])) {
      case -1: break;
      case 0:
	fprintf(f, "%d %d s0 ", i, j);
	break;
      case 1:
	fprintf(f, "%d %d s1 ", i, j);
	break;
      case 2:
	fprintf(f, "%d %d s2 ", i, j);
	break;
      case 3:
	fprintf(f, "%d %d s3 ", i, j);
	break;
      default:
	break;
      }
      fprintf(f, "%%%d\n", dscat[i][j]);
      if ((j % 5) == 0)
	fprintf(f, "\n");
    }
  }
  fprintf(f, "\n\n%%%%End\n\n");
}



void put_scat_to_latex_list()

{
  char locn[64];
  char scat_no_label[64];

  open_eps_file();
  sprintf(scat_no_label, "%d", current_eps_count());
  put_scat_to_eps(the_eps_file());
    /* not yet closed in order to make label name available */
  
  fprintf(the_latex_list(), "\\epsffile{%s}\n", the_eps_file_name());

  close_eps_file();  /* eps counter incremented */
}


/*--
  \end{code}

  \endoffile

  --*/

