/*--

  \FileTitle{
  {\tt egdiagsr.c} --- Electrocardiognosis, strips as diagrams
  }
  \CopyrightNotice{

    Copyright (c) 1991--1996 Alexandru Dan Corlan MD (http://dan.corlan.net)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  }

  \begin{code}
  --*/

#include <stdio.h>
#include <malloc.h>

#ifdef _MSDOS_
#include <ilfive/shpell5.h>
#else
#include <shpell5.h>
#endif

#include "gliks.h"
#include "ecgnosis.h"
#include "wipres.h"


/*--
\end{code}

\subsubsection{Leads}

\ImplementationNotes{ 
There are three kinds of leads: recorded leads (RL) which contain the data
that is actually recorded, normally 0 and 1; 
basic leads (BL), which contain a universal
translation of the recorded leads, normally A, B and C; 
classical leads (CL) which are widely used ECG leads that can
be computed from the basic leads. All the BL and CL are computed
by means of linear transformations of the
values they rely upon. Thus, their means of computation
can be established by vectors of factors to be multiplied
with the value of each data lead.
}

\begin{code}
--*/

#define MAX_NOF_RL   2
#define MAX_NOF_BL   3
#define MAX_NOF_CL   6
#define MAX_LEAD_LEN 15100

short bl_fact[MAX_NOF_BL][MAX_NOF_RL] = 
  { { 1,  0 },          /* A */
    { 0,  1 },          /* B */
    { 1, -1 } };        /* C */

short cl_fact[MAX_NOF_CL][MAX_NOF_BL] = 
  { { 2,  0,  0 },      /* DI */
    { 0,  0,  2 },      /* DII */
    { 0,  2,  0 },      /* DIII */
    {-1,  2,  0 },      /* aVF */
    { 2, -1,  0 },      /* aVL */
    {-1, -1,  0 } };    /* aVR */

short nof_rl = MAX_NOF_RL;
short nof_bl = MAX_NOF_BL;
short nof_cl = MAX_NOF_CL;

char *l_labels[MAX_NOF_RL+MAX_NOF_BL+MAX_NOF_CL] = 
  { "L0", "L1", "A", "B", "C", "DI", "DII", "DIII", "aVF", "aVL", "aVR" };


long crt_rl_offs = 0;
long crt_rl_len = 0;

short abuf[MAX_LEAD_LEN];
short rl[MAX_NOF_RL][MAX_LEAD_LEN];
short bl[MAX_NOF_BL][MAX_LEAD_LEN];

long get_lead_data(PERFILE *p, short *buf, short lead, long offs, 
                    long end, short resample_rate)

{
  long l, i, j;
  long len;

  len = (end-offs) / resample_rate;

  if(offs!=crt_rl_offs || len!=crt_rl_len) {
    for(l=0; l<nof_rl; l++) {
      j = get_samples(p, abuf, l, offs, end, resample_rate);
/*      fprintf(ecgnosislog, "got %d samples\n", j);*/
      fflush(ecgnosislog);
      for(i=0; i<len; i++)
        rl[l][i] = abuf[i];
    }
    for(l=0; l<nof_bl; l++) {
      for(i=0; i<len; i++) {
        bl[l][i] = 0;
        for(j=0; j<nof_rl; j++)
          bl[l][i] += bl_fact[l][j] * rl[j][i];
      }
    }
    crt_rl_len = len;
    crt_rl_offs = offs;
  }
  if(lead<nof_rl) {
/*    fprintf(ecgnosislog, "Well inside, lead=%d, len=%d\n", lead, len);
    fflush(ecgnosislog);*/
    for(i=0; i<len; i++)
      buf[i] = rl[lead][i];
    return len;
  }
  if(lead>=nof_rl && lead<nof_rl + nof_bl) {
    for(i=0; i<len; i++)
      buf[i] = bl[lead-nof_rl][i];
    return len;
  }
  if(lead>=nof_rl+nof_bl && lead<=nof_rl + nof_bl + nof_cl) {
    for(i=0; i<len; i++) {
      buf[i] = 0;
      for(j=0; j<nof_bl; j++)
        buf[i] += bl[j][i] * cl_fact[lead-nof_rl-nof_bl][j];
      buf[i] /= 2;
    }
    return len;
  }
  return 0;  /* exception: no such lead */
}

/*--
  \end{code}

  \dataparagraph{Strip diagram description}  % a name for the data structure

  \Definition{ % What the data structure is designed to represent
  The following variables make up, together, a complete description of
  a strip diagram.
  }
  \SemanticStructure{
  \begin{description}
  \item[strip\_buffer] contains the data to print;
  \item[strip\_fill] contains fill pointers in the strip buffer;
  \item[nof\_leads] is either 1 or 2;
  \item[first\_lead] is the code of the single, or first (if they are 2) lead
  to print;
  \item[strip\_ampl] 1000 means 10 mm/mV; 2000 means 5 mm/mV;
  if it is set to 0 when the strip output functions are called
  then it is automatically computed by the program to be the
  largest recommended amplification that assures complete
  view of both of the leads.

  \item[strip\_title] is the title of the picture containing the strip;
  \item[strip\_comment] is the comment, that appears in the text, on the
  strip; the strip comment should contain {\tt \%s} to indicate
  the place for a label reference.

  \item[paper\_speed] number of mm per second, divided by 10;
  \item[strip\_height] in points (72 in an inch);
  \item[strip\_width] in points;
  \item[strip\_grid\_type] 0=no grid, 1=dotted grid, 2=lined grid;
  \item[strip\_mvmark\_f] is larger than 0 if the mark is to be displayed;
  \item[strip\_frame\_f] is larger than 0 if a frame will be put
  around the strip;
  \item[strip\_exframe\_f] is larger than 0 if external labels are to be
  surrounded by a strip;
  \item[strip\_offs] contains the number of points where the middle of each
  strip is to be placed;
  \item[strip\_tech] is 0 if the strip diagram will be displayed in black and
  white, 1 if using grays and 2 if using color;
  \item[strip\_nof\_lines] is the number of succesive strips stacked in the
  diagram.
  \end{description}

  }
  \ConsistencyRules{
  % Rules that the data in the structure is expected to obey in order
  % to be legal, even if not enforced by the program itself
  }

  \subparagraph{Construction data}

  \begin{code}
  --*/

short strip_buffer[2][12002]; 					/* [VAR] */ 
short strip_fill[2]; 						/* [VAR] */ 
short nof_leads = 1; 						/* [VAR] */ 
short strip_ampl = 1000;					/* [VAR] */ 
short first_lead = 0; 						/* [VAR] */ 
PERFILE *strip_file = NULL; 					/* [VAR] */ 
long strip_start = 0; 						/* [VAR] */ 

char strip_title[256] = "";					/* [VAR] */ 
char strip_comment[4096] = ""; 					/* [VAR] */ 

/*--
  \end{code}

  \subparagraph{Typographic features}

  \begin{code}
  --*/


short paper_speed = 250;					/* [VAR] */
short strip_height = 144; 					/* [VAR] */ 
short strip_width = 432; 					/* [VAR] */
short mark_place_width = 30; 					/* [VAR] */ 
short strip_grid_type = 2; 					/* [VAR] */ 
short strip_mvmark_f = 1; 					/* [VAR] */ 
short strip_frame_f = 1; 					/* [VAR] */ 
short strip_exframe_f = 1; 					/* [VAR] */ 
short strip_tech = 0; 						/* [VAR] */ 
short nof_lines = 1; 						/* [VAR] */ 

/*--
  \end{code}

  \subparagraph{Accesory information}

  \begin{code}
  --*/

short strip_speed_label_f = 1; 					/* [VAR] */ 
short strip_ampl_label_f = 1; 					/* [VAR] */ 
short file_name_label_f = 0; 					/* [VAR] */ 
short strip_no_label_f = 0; 					/* [VAR] */ 
short date_label_f = 0; 					/* [VAR] */ 
short abs_time_on_f = 1; 					/* [VAR] */ 
short abs_time_off_f = 0; 					/* [VAR] */ 
short rel_time_on_f = 0; 					/* [VAR] */ 
short rel_time_off_f = 0; 					/* [VAR] */ 
short lead_label_f = 1; 					/* [VAR] */ 


/*--
  \end{code}

  \subparagraph{Typographic details}

  \begin{code}
  --*/

char *strip_signal_color[2] =
          { "Black", "Black" }; 			        /* [VAR] */ 
char *strip_grid_color = "WildStrawbery"; 			/* [VAR] */ 
char *strip_background_color = "White";				/* [VAR] */ 

float strip_signal_gray[2] = { 0.0, 0.0 };			/* [VAR] */ 
float strip_grid_gray = 0.0; 					/* [VAR] */ 
float strip_background_gray = 1.0; 				/* [VAR] */ 

float signal_line_size[2] = { 0.8, 0.8 }; 			/* [VAR] */ 
float large_grid_size = 0.4; 					/* [VAR] */ 
float small_grid_size = 0.1; 					/* [VAR] */ 
float frame_size = 0.8; 					/* [VAR] */ 
float exframe_size = 0.8; 					/* [VAR] */ 

float top_height = 14; 						/* [VAR] */ 
float top_baseline = 2; 					/* [VAR] */ 
float bottom_height = 14; 					/* [VAR] */ 
float bottom_baseline = 2; 					/* [VAR] */ 

short inter_strip_skip = 0; 					/* [VAR] */ 

/*--
  \end{code}

  \dataparagraph{Actual printing information}

  \Definition{ % What the data structure is designed to represent
  }
  \SemanticStructure{
  % Description of structure and associated semantics
  % Full definition of the intented data types by means of ADA declarations
  % may be appropiate
  }
  \ConsistencyRules{
  % Rules that the data in the structure is expected to obey in order
  % to be legal, even if not enforced by the program itself
  }

  \subparagraph{Accesory information contents}

  \begin{code}
  --*/

char strip_speed_label[32] = "";				/* [VAR] */ 
char strip_ampl_label[32] = "";					/* [VAR] */ 
char file_name_label[64] = "";					/* [VAR] */ 
char strip_no_label[16] = ""; 					/* [VAR] */ 
char date_label[16] = "";    					/* [VAR] */ 
char abs_time_on[16] = "";					/* [VAR] */ 
char abs_time_off[16] = "";					/* [VAR] */ 
char rel_time_on[16] = "";					/* [VAR] */ 
char rel_time_off[16] = "";  					/* [VAR] */ 
char lead_label[16] = "";					/* [VAR] */ 


/*--
  \end{code}

  \subparagraph{Label resources}

  \begin{code}
  --*/

char label_font[64] = "Helvetica"; 				/* [VAR] */ 
float label_font_size = 10; 					/* [VAR] */ 

short speed_label_x = 0; 					/* [VAR] */ 
short ampl_label_x = 70; 					/* [VAR] */ 
short lead_label_x = 200; 					/* [VAR] */ 
short file_name_label_x = 0; 					/* [VAR] */ 
short strip_no_label_x = 100; 					/* [VAR] */ 
short date_label_x = 200; 					/* [VAR] */ 
short time_on_label_x = 300; 					/* [VAR] */ 
short time_off_label_x = 380; 					/* [VAR] */ 



/*--
  \end{code}

  \subparagraph{Structural information}

  These variables describe features of the generated report:

  \begin{itemize}

  \item Figure
  environments are to be used if the report is to be transformed later
  into an article or another document (in which consistent text
  is added by the user, the generated diagrams reprezenting only
  an ilustration of ideas).
  \end{itemize}

  \begin{code}
  --*/

short use_figure_environments = 1;   				/* [VAR] */ 

/*--
  \end{code}

  \dataparagraph{Recommended amplitudes}  % a name for the data structure

  \begin{code}
  --*/

#define LAST_AMPL_DIVIDER      5
#define NO_AMPL_CHANGE         6
#define FIRST_AMPL_AMPLIFIER   7
#define LARGEST_AMPLIFIER     11

short recommended_ampls[12] = {
 10000,   /* 10.0 mV/cm */
  8000,   /*  8.0 mV/cm */
  5000,   /*  5.0 mV/cm */
  4000,   /*  4.0 mV/cm */
  2500,   /*  2.5 mV/cm */
  2000,   /*  2.0 mV/cm (LAST_AMPL_DIVIDER) */
  1000,   /*  1.0 mV/cm (NO_AMPL_CHANGE) */
   800,   /*  0.8 mV/cm (FIRST_AMPL_AMPLIFIER) */
   500,   /*  0.5 mV/cm */
   400,   /*  0.4 mV/cm */
   200,   /*  0.2 mV/cm */
   100,   /*  0.1 mV/cm */
};


/*--
  \end{code}

  \dataparagraph{Specifier symbols}

  \begin{code}
  --*/

static cell _strip;
static cell _data_file;
static cell _eruid;
static cell _start_moment;
static cell _diagram_no;

static cell _nof_leads;
static cell _amplification;
static cell _upper_lead;
static cell _nof_lines;

static cell _title;
static cell _comment;

static cell _paper_speed;
static cell _height;
static cell _width;
static cell _grid_type;
static cell _mark_width;
static cell _mark_f;
static cell _frame_f;
static cell _exframe_f;
static cell _tech;

static cell _speed_label_f;
static cell _ampl_label_f;
static cell _strip_no_label_f;
static cell _file_name_label_f;
static cell _date_label_f;
static cell _abs_time_on_f;
static cell _abs_time_off_f;
static cell _rel_time_on_f;
static cell _rel_time_off_f;
static cell _lead_label_f;

static cell _signal_color;
static cell _grid_color;
static cell _background_color;

static cell _signal_gray;
static cell _grid_gray;
static cell _background_gray;

static cell _signal_line_size;
static cell _large_grid_line_size;
static cell _small_grid_line_size;
static cell _frame_size;
static cell _exframe_size;
static cell _top_height;
static cell _top_baseline;
static cell _bottom_height;
static cell _bottom_baseline;

static cell _inter_strip_skip;

/*--
  \end{code}

  \subsubsection{Parameter specifiers}

  \begin{code}
  --*/

void set_strip_paper_speed(short speed)

{
  float ps;
  
  paper_speed = speed;
  ps = speed / 10.0;
  sprintf(strip_speed_label, "%5.1f mm/s", ps);
}

void set_strip_amplification(short ampl)

{
  float av;

  strip_ampl = ampl;
  if(ampl==0)
    sprintf(strip_ampl_label, "auto");
  else
    sprintf(strip_ampl_label, "1mV = %dmm", 10000/ampl);
}

void set_strip_grid_type(short gt)

{
  if(gt<0 || gt>2)
    return ; /* error */
  strip_grid_type = gt;
}

void set_strip_height(short ht)

{
  strip_height = ht;
}

void set_strip_width(short wd)

{
  strip_width = wd;
}

void set_strip_mv_mark_on() { strip_mvmark_f = 1; }
void set_strip_mv_mark_off() { strip_mvmark_f = 0; }

void set_strip_title(char *t)

{
  strncpy(strip_title, t, 256);
  strip_title[255] = 0;
}

void set_strip_comment(char *c)

{
  strncpy(strip_comment, c, 1024);
  strip_comment[1023] = 0;
}

void set_strip_leads_to_print(short leada, short leadb)

{
  if(leada==0 && leadb==0)
    return;
  if(leada!=0 && leadb!=0) {
    nof_leads = 2;
    first_lead = 0;
    return;
  }
  if(leada!=0) {
    nof_leads = 1;
    first_lead = 0;
    return;
  }
  if(leadb!=0) {
    nof_leads = 1;
    first_lead = 1;
    return;
  }
}

void set_number_of_strips(short ns)

{
  nof_lines = ns;
}

void set_strip_rendering_technology(short tech)

{
  if(tech<0 || tech>2)
    return ;
  strip_tech = tech;
}

/*--
  \end{code}

  \subsubsection{Output Functions}

  \begin{code}
  --*/


/*--
  \end{code}

  \funparagraph{strip\_extract}

  \PurposeReturn{  % be short
  Reads one of the strips to print in the \variabletag{strip\_buffer} variable,
  from the given \typetag{PERFILE}. This function may be considered
  a parameter specification function, just it must be the last one
  called before the put function.
  }
  \ImplementationNotes{
  This function is the---rather dummy---compute function for the strip
  diagrams.
  }

  \begin{code}
  --*/


static short strip_extract(PERFILE *p, long m0, short nth) 	/* [FUN] */ 

{
  long nof_samples;

  strip_file = p;
  strip_start = m0;
  
  nof_samples = (((strip_width * 250 * 200) / 72) / paper_speed);
  if (strip_mvmark_f)
    nof_samples -= (((mark_place_width * 250 * 200) / 72) / paper_speed);
  get_samples(p, strip_buffer[0], first_lead, m0 + nth * nof_samples * 50,
	      m0 + (nth+1) * nof_samples * 50, 50);
  if(nof_leads==2)
    get_samples(p, strip_buffer[1], !first_lead, m0 + nth * nof_samples * 50,
		m0 + (nth+1) * nof_samples * 50, 50);
  strip_fill[0] = strip_fill[1] = nof_samples;
  return 1;  /* i.e. success;
		exceptions from get_samples might lead to error */
}


/*--
  \end{code}

  \funparagraph{put\_strip\_to\_eps}  % The name of the function(s)

  \begin{code}
  --*/


/* The following variables will contain the results of the
 * analyses to be displayed as markings
 */

#define MAX_DISP_COMPLEXES 2000

long mp_qrspos[MAX_DISP_COMPLEXES];
long mp_bl0[MAX_DISP_COMPLEXES];
long mp_bl1[MAX_DISP_COMPLEXES];
long mp_st0[MAX_DISP_COMPLEXES];
long mp_st1[MAX_DISP_COMPLEXES];
long mp_j[MAX_DISP_COMPLEXES];
long mp_morph[MAX_DISP_COMPLEXES];
long mp_qrslen[MAX_DISP_COMPLEXES];

short act_pdisp_complexes = 0;

extern short bl_marking;
extern short j_marking; 	
extern short rr_marking; 	
extern short st_marking; 	
extern short qrsdet_marking;
extern short qrsref_marking;
extern short morph_marking; 
extern short qrslen_marking;

short put_strip_to_eps(FILE *f, PERFILE *p, long m0)

{
  float xl, yl, yfact, xss;
  float x, y, y0;
  short xsize, ysize;
  short maxheight;
  short choosen_scale;
  float mark_height;
  float height0, height1;
  short i, gn;
  short max0, max1, min0, min1;
  long absm;
  long xsize_in_ccs;
  short sp;
  long cbt;
  long xp;
  static char markbuf[64];

  absm = rel2abs_time(m0);
  sprintf(abs_time_on, "%02ld:%02ld:%02ld", absm / 36000000,
	  (absm % 36000000) / 600000, (absm % 600000) / 10000);


  /*--
    \end{code}

    \subparagraph{Compute and output the bounding box}

    \begin{code}
    --*/

  yl = top_height + bottom_height + strip_height * nof_lines +
    inter_strip_skip * (nof_lines - 1) + 2 * frame_size +
      2 * exframe_size;
  xl = strip_width + 2 * frame_size;
  fprintf(f, "%%%%BoundingBox: 0 0 %6.2f %6.2f\n", xl, yl);


  /*--
    \end{code}

    \subparagraph{Clear the background area}

    \begin{code}
    --*/

  if(strip_tech<2) /* not color */
    fprintf(f, "1.0 setgray\n");
  else
    fprintf(f, "%s\n", strip_background_color);
  fprintf(f, "newpath 0 0 moveto %d 0 rlineto 0 %d rlineto\n",
	  strip_width, strip_height);
  fprintf(f, "  %d 0 rlineto 0 %d rlineto closepath fill\n",
	  - strip_width, - strip_height);
  

  /*--
    \end{code}

    \subparagraph{Draw the frame and the exframes}

    \begin{code}
    --*/

  /* Not sure they are needed ... */

  /*--
    \end{code}

    \subparagraph{Compute the mark size}

    When done automatically, the summed sizes of the variation
    of both strips is taken to fill the entire available space.
    
    Only the first line of ECG is used for this determination, and
    considered to be good for the rest of the displayed strips
    (which is OK for the great majority of cases).

    \begin{code}
    --*/
  
  if(strip_ampl==0) { 
    strip_extract(p, m0, 0);
    
    min0 = 32000;
    max0 = -32000;
    min1 = 32000;
    max1 = -32000;
  
    for(i=0; i<strip_fill[0]; i++) {
      if(strip_buffer[0][i]<min0)
	min0 = strip_buffer[0][i];
      if(strip_buffer[0][i]>max0)
	max0 = strip_buffer[0][i];
      if(nof_leads==2) {
	if(strip_buffer[1][i]<min1)
	  min1 = strip_buffer[1][i];
	if(strip_buffer[1][i]>max1)
	  max1 = strip_buffer[1][i];
      }
    }

    maxheight = (max0 - min0);

    if(nof_leads==2)
      maxheight = maxheight + (max1 - min1);
    
    /* compute size in pixels at 1.0 amplification */
    
    y = 72.0 / 2.5          /* points per centimeter=mV */
      * maxheight / 20;     /* as there are 20 points per mV */
    
    /* the size of the result must be more than half the total
       height and less than total height itself */
    
    choosen_scale = NO_AMPL_CHANGE;
    if(y > strip_height) {
      for(choosen_scale = LAST_AMPL_DIVIDER;
	  choosen_scale >= 0;
	  choosen_scale --)
	if (y * 1000.0 / recommended_ampls[choosen_scale] <= strip_height)
	  break;
      /* left at 10.0 mV/cm */
    }
    else if(y < strip_height / 2) {
      for(choosen_scale = FIRST_AMPL_AMPLIFIER;
	  choosen_scale <= LARGEST_AMPLIFIER;
	  choosen_scale ++)
	if (y * 1000.0 / recommended_ampls[choosen_scale] >=
	    strip_height / 2)
	  break;
    }
    set_strip_amplification(recommended_ampls[choosen_scale]);
  }
  
  mark_height = 72.0 / 2.5 * 1000.0 / strip_ampl;
  
  /*--
    \end{code}

    \subparagraph{Output the grids}

    \begin{code}
    --*/

  xsize = strip_width;
  ysize = strip_height;

  if(strip_tech<2) /* not color */
    fprintf(f, "0.0 setgray \n");
  else
    fprintf(f, "%s\n", strip_grid_color);


  if(strip_grid_type > 0)
    for(gn=0; gn<nof_lines; gn++) {
      y0 = bottom_height + exframe_size +
	(nof_lines - gn - 1) * strip_height +
	  (nof_lines - gn) * inter_strip_skip;
      
      /* The dotted grid is ignored for the moment */
      /* The small lines --- always the same, at 1mm=0.04in=2.88pt */
      
      fprintf(f, "%5.2f setlinewidth [0.44 1.0] 0 setdash\n", small_grid_size);
      /* ((mvsize>0) ? (x = (float)mark_place_width + 2.88) : (x = 2.88)); */
      for(x = 2.88; x < xsize; x += 2.88) {
	  fprintf(f, "newpath %5.2f %5.2f moveto %5.2f %5.2f lineto stroke\n",
		  x, y0, x, y0 + ysize);
	}
      for(y=y0; y<y0 + ysize; y+=2.88) {
	fprintf(f, "newpath %d %5.2f moveto %d %5.2f lineto stroke\n",
		0, y, xsize, y);
      }

      /* The large lines --- always the same, at 5mm=0.2in=14.4pt */

      fprintf(f, "%5.2f setlinewidth [1.0 0.44] 0 setdash\n", large_grid_size);
      for(x = 0.0; x < xsize; x += 14.4) {
	fprintf(f, "newpath %5.2f %5.2f moveto %5.2f %5.2f lineto stroke\n",
		x, y0, x, y0 + ysize);
      }
      for(y=y0; y<y0 + ysize; y+=14.4) {
	fprintf(f, "newpath %d %5.2f moveto %d %5.2f lineto stroke\n",
		0, y, xsize, y);
      }
    }


  /*--
    \end{code}

    \subparagraph{Put the mV marks}

    First it goes to signal color, greys, etc, which will be used
    in the next section also.

    \begin{code}
    --*/
  

  if(strip_tech<2) /* not color */
    fprintf(f, "0.0 setgray\n");
  else
    fprintf(f, "%s\n", strip_signal_color[0]);
  fprintf(f, "%5.2f setlinewidth [] 0 setdash\n", signal_line_size[0]);

  for(gn=0; gn<nof_lines; gn++) {
    y0 = bottom_height + exframe_size +
      (nof_lines - gn - 1) * strip_height +
	(nof_lines - gn) * inter_strip_skip;
    
    /* The 5's below assume a mark space of 30 points */
    
    if(strip_mvmark_f) {
      fprintf(f, "newpath 5 %5.2f moveto 5 0 rlineto 0 %5.2f rlineto\n",
	      (strip_height - mark_height) / 2 + y0,
	      mark_height);
      fprintf(f, "  5 0 rlineto 0 %5.2f rlineto\n", -mark_height);
      fprintf(f, "5 0 rlineto stroke\n");
    }
  }

  /*--
    \end{code}

    \subparagraph{Vertical displacements of strips}

    \begin{code}
    --*/

  if(nof_leads==1)
    height0 = (strip_height -
	       72.0 / 2.5 * (max0 - min0) / 20.0 * 1000.0 / strip_ampl) /2;
  else {
    height1 = (strip_height - 72.0 / 2.5 * (max0 - min0 + max1 - min1) / 20.0
      * 1000.0 / strip_ampl) /3;
    height0 = 2 * height1 +
      72.0 / 2.5 * (max1 - min1) / 20.0 * 1000.0 / strip_ampl;
  }

  /*--
    \end{code}

    \subparagraph{Draw the strips}

    \begin{code}
    --*/

  fprintf(f, "\n\n/TOL { %6.2f exch rlineto } def\n\n",
	  xss = 72.0 / 25.4 * paper_speed / 10.0 / 200 /* = sample rate */);

  yfact = 72.0 * 10.0 / 25.4 * 1000.0 / strip_ampl / 20.0 /* aquis fact */;

  /* The first strip */

  for(gn=0; gn<nof_lines; gn++) {
    y0 = bottom_height + exframe_size +
      (nof_lines - gn - 1) * strip_height + (nof_lines - gn -1)
	* inter_strip_skip;
    strip_extract(p, m0, gn);
    fprintf(f, "newpath %d %6.2f moveto\n",
	    strip_mvmark_f ? mark_place_width : 0, y0 + height0 +
	    (strip_buffer[0][0] - min0) * yfact);
    for(i=1; i<strip_fill[0]; i++) {
      fprintf(f, "%6.2f TOL\n",
	      (strip_buffer[0][i] - strip_buffer[0][i-1]) * yfact);
      if(!(i%115)) {
	fprintf(f, "stroke\n");
	fprintf(f, "newpath %6.2f %6.2f moveto\n",
		strip_mvmark_f ? mark_place_width + i * xss : i * xss,
		y0 + height0 + (strip_buffer[0][i] - min0) * yfact);
      }
    }
    fprintf(f, "stroke\n\n");
  }

  if(nof_leads==2)
    for(gn=0; gn<nof_lines; gn++) {
      y0 = bottom_height + exframe_size +
	(nof_lines - gn - 1) * strip_height + (nof_lines - gn - 1)
	  * inter_strip_skip;
      strip_extract(p, m0, gn);
      fprintf(f, "newpath %d %6.2f moveto\n",
	      strip_mvmark_f ? mark_place_width : 0, y0 + height1 +
	      (strip_buffer[1][0] - min1) * yfact);
      for(i=1; i<strip_fill[1]; i++) {
	fprintf(f, "%6.2f TOL\n",
		(strip_buffer[1][i] - strip_buffer[1][i-1]) * yfact);
	if(!(i%115)) {
	  fprintf(f, "stroke\n");
	  fprintf(f, "newpath %6.2f %6.2f moveto\n",
		  strip_mvmark_f ? mark_place_width + i * xss : i * xss,
		  y0 + height1 + (strip_buffer[1][i] - min1) * yfact);
	}
      }
      fprintf(f, "stroke\n\n");
    }


  /*--
    \end{code}

    \subparagraph{Labels display}

    \begin{code}
    --*/

  fprintf(f, "/labelfont /%s findfont 10 scalefont def\n",
	  label_font);
  fprintf(f, "labelfont setfont\n");
  
  if(strip_speed_label_f) {
    fprintf(f, "%d %5.2f moveto (%s) show\n",
	    speed_label_x, top_baseline + 2 * exframe_size +
	    nof_lines * (strip_height + inter_strip_skip) + bottom_height,
	    strip_speed_label);
  }
  if(strip_ampl_label_f) {
    fprintf(f, "%d %5.2f moveto (%s) show\n",
	    ampl_label_x, top_baseline + 2 * exframe_size +
	    nof_lines * (strip_height + inter_strip_skip) + bottom_height,
	    strip_ampl_label);
  }
  if(file_name_label_f) {
    fprintf(f, "%d %5.2f moveto (%s) show\n",
	    file_name_label_x, bottom_baseline,
	    file_name_label);
  }
  if(date_label_f) {
    fprintf(f, "%d %5.2f moveto (%s) show\n",
	    date_label_x, bottom_baseline,
	    date_label);
  }
  if(strip_no_label_f) {
    fprintf(f, "%d %5.2f moveto ([%s]) show\n",
	    strip_no_label_x, bottom_baseline,
	    strip_no_label);
  }
  if(abs_time_on_f) {
    fprintf(f, "%d %5.2f moveto (%s) show\n",
	    time_on_label_x,
	    top_baseline + 2 * exframe_size +
	    nof_lines * (strip_height + inter_strip_skip) + bottom_height,
	    abs_time_on);
  }
  if(abs_time_off_f) {
    fprintf(f, "%d %5.2f moveto (%s) show\n",
	    time_off_label_x, bottom_baseline,
	    abs_time_off);
  }
  if(rel_time_on_f) {
    fprintf(f, "%d %5.2f moveto (%s) show\n",
	    time_on_label_x, bottom_baseline,
	    rel_time_on);
  }
  if(rel_time_off_f) {
    fprintf(f, "%d %5.2f moveto (%s) show\n",
	    time_off_label_x, bottom_baseline,
	    rel_time_off);
  }


  /* this code taken from the egdiagst code, to display the
     the markings on the ECG (such as QRS detection, distances
     between complexes, j point and the like */
  
/* Displaying the line markings and other markings */

  if(m0>50000 &&
     (bl_marking || j_marking || rr_marking || st_marking ||
      qrsdet_marking || qrsref_marking || qrslen_marking
      || morph_marking)) {

/* The sizes of the marking information */

    for(gn=0; gn<nof_lines; gn++) {
      y0 = bottom_height + exframe_size +
	(nof_lines - gn - 1) * strip_height + (nof_lines - gn - 1)
	* inter_strip_skip + strip_height;
      
      xsize_in_ccs = (((strip_width * 250 * 200) / 72) / paper_speed) * 50;
      if (strip_mvmark_f)
	xsize_in_ccs -= (((mark_place_width * 250 * 200) / 72) / paper_speed)
	  * 50;
      
      sp = reqitab(p, m0 + gn*xsize_in_ccs,
		   m0+(gn+1)*xsize_in_ccs+10000); /* exception!? */
      cbt = complexes_between(p, m0 + gn*xsize_in_ccs,
			    m0+(gn+1)*xsize_in_ccs+10000);
      /* exception! */
      if(cbt>MAX_DISP_COMPLEXES-2) /* exception! */
	cbt = MAX_DISP_COMPLEXES-2;
      
      /* obtaining the marking information */
      
      cpars_at(p, mp_qrspos, CPQRSPOS, m0 + gn*xsize_in_ccs, cbt);
      if(j_marking)
	cpars_at(p, mp_j, CPJPOS, m0 + gn*xsize_in_ccs, cbt);
      if(qrslen_marking)
	cpars_at(p, mp_qrslen, CPQRSLEN, m0 + gn*xsize_in_ccs, cbt);
      if(morph_marking)
	cpars_at(p, mp_morph, CPQRSMORPH, m0 + gn*xsize_in_ccs, cbt);
      if(bl_marking) {
	cpars_at(p, mp_bl0, CPBL1, m0 + gn*xsize_in_ccs, cbt);
      }
      
      if(rr_marking) {
	for(i=0; i<cbt-1; i++) {
	  sprintf(markbuf, "%4ldms", (mp_qrspos[i+1] - mp_qrspos[i])/10);
	  xp = (long)((double)(mp_qrspos[i] - (m0 + gn*xsize_in_ccs))
		      /10000.0 * paper_speed * 72.0 / 250.0);
	  fprintf(f, "%d %6.2f moveto ",
		  (strip_mvmark_f ? mark_place_width : 0) + xp, y0 - 40);
	  fprintf(f, "(%s) show ", markbuf);
	  sprintf(markbuf, "%4ld/'", 600000 / (mp_qrspos[i+1] - mp_qrspos[i]));
	  fprintf(f, "%d %6.2f moveto ",
		  (strip_mvmark_f ? mark_place_width : 0) + xp, y0 - 55);
	  fprintf(f, "(%s) show\n", markbuf);
	}
      }
    }
  }
}

extern cell last_assertion_named(char *);

void put_strip_to_report(PERFILE *p, long m0)

{
  FILE *rf;
  cell diagn;

  open_report_file();
  rf = the_report_file();
  if(rf==NULL)
    return ;  /* silently, you see, could not be done */

  diagn = last_assertion_named("new-diagram");
  if(diagn!=NIL)
    set_eps_counter((short)(binary(car(cdr(cdr(diagn))))));
  else
    set_eps_counter(1);
  
  fprintf(rf, "\\begin{gnode}{ecgnosis}{ECG%04d}\n\n", current_eps_count());
  fprintf(rf, "%s\n\n", strip_title);
  fprintf(rf, "\\slot{diaclass} ecg\n\n");
  fprintf(rf, "\\slot{params} \n\n");
  fprintf(rf, "     (  strip-amplification %d\n", strip_ampl);
  fprintf(rf, "        nof-leads %d\n", nof_leads);
  fprintf(rf, "        paper-speed %d\n", paper_speed);
  fprintf(rf, "        )\n\n");
  fprintf(rf, "\\slot{gparams}\n");
  fprintf(rf, "      ( strip-height %d\n", strip_height);
  fprintf(rf, "        strip-width %d\n", strip_width);
  fprintf(rf, "        strip-total-height %d\n",
                       strip_height * nof_lines);
  fprintf(rf, "        strip-frame-p %s\n", strip_frame_f ? "t" : "nil");
  fprintf(rf, "        strip-exframe-p %s\n", strip_exframe_f ? "t" : "nil");
  fprintf(rf, "        strip-mV-mark-p %s\n", strip_mvmark_f ? "t" : "nil");
  fprintf(rf, "        )\n\n");
  fprintf(rf, "\\slot{eps}\n\n");
  put_strip_to_eps(rf, p, m0);
  fprintf(rf, "\\slot{comment}\n\n");
  fprintf(rf, strip_comment);
  fprintf(rf, "\n\n\\end{gnode}\n\n");
  inc_eps_counter();
  close_report_file();
}


/*--
  \end{code}

  \funparagraph{put\_strip\_to\_latex\_list}
  \ImplementationNotes{
  % How implementation of the function is done: algorithm---if special,
  % implementation decisions, discarded implementation alternatives
  }
  \Portability{ % What is assumed about the operating environment
  }
  \Changes{ % that are anticipated, if special provision was made for them
  }

  \begin{code}
  --*/

void put_strip_to_latex_list(PERFILE *p, long m0)

{
  char locn[64];

  open_eps_file();
  sprintf(strip_no_label, "%d", current_eps_count());
  put_strip_to_eps(the_eps_file(), p, m0);
    /* not yet closed in order to make label name available */
  
  fprintf(the_latex_list(), "\n\n\n");

  if(use_figure_environments){
    fprintf(the_latex_list(), "\n\n\n\\begin{figure}[hbt]\n");
/*    fprintf(the_latex_list(), "\\begin{center}\n\n");*/
  }
  fprintf(the_latex_list(), "\\epsffile{%s}\n", the_eps_file_name());

  if(use_figure_environments)
    fprintf(the_latex_list(), "\\caption{%s}\\label{%s}\n",
	    strip_title, the_label_name());
  else
    fprintf(the_latex_list(), "\\centerline{%s}\n", strip_title);
  if(use_figure_environments) {
/*    fprintf(the_latex_list(), "\\end{center}\n");*/
    fprintf(the_latex_list(), "\\end{figure}\n\n");
  }
  sprintf(locn, "\\ref{%s}", the_label_name());
  fprintf(the_latex_list(), strip_comment, locn);
  close_eps_file();  /* eps counter incremented */
}


/*--
  \end{code}

  \funparagraph{clear\_strip\_description}

  \begin{code}
  --*/

void clear_strip_description()

{
nof_leads = 1; 						 
strip_ampl = 1000;					 
first_lead = 0; 						 
strip_file = NULL; 					 
strip_start = 0; 						 

strip_title[0] = 0;
strip_comment[0] = 0;

paper_speed = 250;					
strip_height = 144; 					 
strip_width = 432; 					
mark_place_width = 30; 					 
strip_grid_type = 2; 					 
strip_mvmark_f = 1; 					 
strip_frame_f = 1; 					 
strip_exframe_f = 1; 					 
strip_tech = 0; 						 
nof_lines = 1; 						 

strip_speed_label_f = 1; 					 
strip_ampl_label_f = 1; 					 
file_name_label_f = 0; 					 
strip_no_label_f = 0;
date_label_f = 0; 					 
abs_time_on_f = 1; 					 
abs_time_off_f = 0; 					 
rel_time_on_f = 0; 					 
rel_time_off_f = 0; 					 
lead_label_f = 0; 					 

strip_signal_color[0] = "Black";
strip_signal_color[1] = "Black";
strip_grid_color = "WildStrawbery"; 			 
strip_background_color = "White";				 

strip_signal_gray[0] = 0.0;
strip_signal_gray[1] = 0.0;
strip_grid_gray = 0.0;
strip_background_gray = 0.0;

signal_line_size[0] = 0.8;
signal_line_size[1] = 0.8;
large_grid_size = 0.0; 					 
small_grid_size = 0.0; 					 
frame_size = 0.8; 					 
exframe_size = 0.8; 					 

top_height = 14; 						 
top_baseline = 2; 					 
bottom_height = 14; 					 
bottom_baseline = 2; 					 

inter_strip_skip = 0; 					 

strip_speed_label[0] = 0;				 
strip_ampl_label[0] = 0;					 
file_name_label[0] = 0;					 
strip_no_label[0] = 0; 					 
date_label[0] = 0;    					 
abs_time_on[0] = 0;					 
abs_time_off[0] = 0;					 
rel_time_on[0] = 0;					 
rel_time_off[0] = 0;  					 
lead_label[0] = 0;					 

speed_label_x = 100; 					 
ampl_label_x = 170; 					 
lead_label_x = 100; 					 
file_name_label_x = 0; 					 
strip_no_label_x = 100; 					 
date_label_x = 200; 					 
time_on_label_x = 0; 					 
time_off_label_x = 380; 					 
}

/*--
  \end{code}

  \funparagraph{create\_strip\_specifier}

  \begin{code}
  --*/

extern cell makefloat(double);

cell create_strip_specifier(PERFILE *p, long m0)

{
  cell r;

  r = NIL;

  /* Construction will proceed from the end to the start of the list */
  
  r = cons(_bottom_baseline, cons(make(BINARY, bottom_baseline), r));
  r = cons(_bottom_height, cons(make(BINARY, bottom_height), r));
  r = cons(_top_baseline, cons(make(BINARY, top_baseline), r));
  r = cons(_top_height, cons(make(BINARY, top_height), r));


  r = cons(_exframe_size, cons(makefloat(exframe_size), r));
  r = cons(_frame_size, cons(makefloat(frame_size), r));
  r = cons(_large_grid_line_size, cons(makefloat(large_grid_size), r));
  r = cons(_small_grid_line_size, cons(makefloat(small_grid_size), r));
  r = cons(_signal_line_size,
	   cons(cons(makefloat(signal_line_size[0]),
		     cons(makefloat(signal_line_size[1]), NIL)), r));

  r = cons(_background_gray, cons(makefloat(strip_background_gray), r));
  r = cons(_grid_gray, cons(makefloat(strip_grid_gray), r));
  r = cons(_signal_gray,
	   cons(cons(makefloat(strip_signal_gray[0]),
		     cons(makefloat(strip_signal_gray[1]), NIL)), r));

  r = cons(_background_color, cons(strbuild(strip_background_color), r));
  r = cons(_grid_color, cons(strbuild(strip_grid_color), r));
  r = cons(_signal_color,
	   cons(cons(strbuild(strip_signal_color[0]),
		     cons(strbuild(strip_signal_color[1]), NIL)), r));
  
  r = cons(_speed_label_f, cons(strip_speed_label_f ? T : NIL, r));
  r = cons(_ampl_label_f, cons(strip_ampl_label_f ? T : NIL, r));
  r = cons(_strip_no_label_f, cons(strip_no_label_f ? T : NIL, r));
  r = cons(_date_label_f, cons(date_label_f ? T : NIL, r));
  r = cons(_file_name_label_f, cons(file_name_label_f ? T : NIL, r));
  r = cons(_abs_time_on_f, cons(abs_time_on_f ? T : NIL, r));
  r = cons(_abs_time_off_f, cons(abs_time_off_f ? T : NIL, r));
  r = cons(_rel_time_off_f, cons(rel_time_off_f ? T : NIL, r));
  r = cons(_rel_time_on_f, cons(rel_time_on_f ? T : NIL, r));
  r = cons(_lead_label_f, cons(lead_label_f ? T : NIL, r));

  r = cons(_tech, cons(make(BINARY, strip_tech), r));
  r = cons(_exframe_f, cons(strip_exframe_f ? T : NIL, r));
  r = cons(_frame_f, cons(strip_frame_f ? T : NIL, r));
  r = cons(_mark_f, cons(strip_mvmark_f ? T : NIL, r));
  r = cons(_mark_width, cons(make(BINARY, mark_place_width), r));
  r = cons(_grid_type, cons(make(BINARY, strip_grid_type), r));
  
  r = cons(_width, cons(make(BINARY, strip_width), r));
  r = cons(_height, cons(make(BINARY, strip_height), r));
  r = cons(_paper_speed, cons(makefloat(paper_speed / 10.0), r));

  r = cons(_comment, cons(strbuild(strip_comment), r));
  r = cons(_title, cons(strbuild(strip_title), r));

  r = cons(_nof_leads, cons(make(BINARY, nof_leads), r));
  r = cons(_amplification, cons(make(BINARY, strip_ampl), r));
  r = cons(_upper_lead, cons(make(BINARY, first_lead), r));
  r = cons(_nof_lines, cons(make(BINARY, nof_lines), r));
  
  r = cons(_start_moment, cons(make(BINARY, m0), r));
  r = cons(_eruid, cons(strbuild("No-ERUID"), r));
  r = cons(_data_file, cons(strbuild(the_ecg_file_name()), r));
  r = cons(_diagram_no, cons(make(BINARY, current_eps_count()), r));
  r = cons(_strip, r);

  return r;
}


/*--
  \end{code}

  \funparagraph{Initialisation function}

  \begin{code}
  --*/


egdiagsr()

{
  _strip = esym("strip-diagram");
  _diagram_no = esym(":diagram-number");
  _data_file = esym(":data-file");
  _eruid = esym(":eruid");
  _start_moment = esym(":starting-from");

  _nof_leads = esym(":number-of-leads");
  _amplification = esym(":amplification");
  _upper_lead = esym(":upper-lead");
  _nof_lines = esym(":number-of-strips");
  
  _title = esym(":title");
  _comment = esym(":comment");
  
  _paper_speed = esym(":paper-speed");
  _height = esym(":height");
  _width = esym(":width");
  _grid_type = esym(":grid-type");
  _mark_width = esym(":width-of-mark");
  _mark_f = esym(":markf");
  _frame_f = esym("framef");
  _exframe_f = esym(":exframef");
  _inter_strip_skip = esym(":inter-strip-skip");
  _tech = esym(":technology");
  
  _speed_label_f = esym(":speed-label-f");
  _ampl_label_f = esym(":amplification-label-f");
  _strip_no_label_f = esym(":diagram-number-label-f");
  _file_name_label_f = esym(":file-name-label-f");
  _date_label_f = esym(":date-label-f");
  _abs_time_on_f = esym(":absolute-on-time");
  _abs_time_off_f = esym(":absolute-off-time");
  _rel_time_on_f = esym(":relative-on-time");
  _rel_time_off_f = esym(":relative-off-time");
  _lead_label_f = esym(":lead-label-f");
  
  _signal_color = esym(":signal-color");
  _grid_color = esym(":grid-color");
  _background_color = esym(":background-color");
  
  _signal_gray = esym(":signal-gray-level");
  _grid_gray = esym(":grid-gray-level");
  _background_gray = esym(":background-gray-level");
  
  _signal_line_size = esym(":signal-line-size");
  _large_grid_line_size = esym(":large-grid-line-size");
  _small_grid_line_size = esym(":small-grid-line-size");
  _frame_size = esym(":frame-size");
  _exframe_size = esym(":exframe-size");
  _top_height = esym(":top-height");
  _top_baseline = esym(":top-baseline");
  _bottom_height = esym(":bottom-height");
  _bottom_baseline = esym(":bottom-baseline");
}

/*--
  \end{code}

  \endoffile

  --*/
