
/*--

  \FileTitle{
  {\tt egviewdt.c} --- Patient information interface
  }
  \CopyrightNotice{

    Copyright (c) 1991--1996 Alexandru Dan Corlan MD (http://dan.corlan.net)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  }
  \vskip 1in

  \subsection{Patient information}

  \begin{code}
  --*/


#include <stdio.h>   
#include <malloc.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/timeb.h>
#include <time.h>

#ifdef WIPORT2X
#include "wiport2x.h"
#include <wiptime.h>
#else
#include <wiport2.h>
#include <wiptime.h>
#endif

#include "ilsix.h"

#include "gliks.h"
#include "ecgnosis.h"



/*--
  \end{code}

  \dataparagraph{Site information}

  \Definition{
  This is a set of data on the site where the recording was made,
  similar to the ERUID.
  }
  \SemanticStructure{
  Data is stored in strings, associated with ``dirty flags''.

  The site strings are initialized from resources, and all
  the changes to them are persistent (in the .RES file).

  When an data item is changed, anyway, by pressing the \fbox{OK}
  button all the information must be changed in the resources. 
  The dirty flags are used for this pupose.
  }

  \begin{code}
  --*/

#define SHORT_S_SIZE   64   /* short string */
#define LONG_S_SIZE   256   /* longer string */
#define HUGE_S_SIZE  4095   /* huge string */


static short country_df = 0;
static char country[SHORT_S_SIZE] = "unknown";   
static cell _country;
  /* The name of the country, e.g. "Romania" */

static short language_df = 0;
short language = 5; /* ! THIS IS EXPORTED */
static cell _language;
   /* 5=^E=English, 18=^R=Romanian, 6=^F=French, 
      7=^G=German,   9=^I=Italian, 19=^S=Spanish */
static cell _eng; 
static cell _rom;
static cell _fra;
static cell _ger;
static cell _ita;
static cell _spa;

static short city_df = 0;
static char city[SHORT_S_SIZE] = "unknown";
static cell _city;
  /* The name of the city, e.g. "Bucuresti" */

static short inst_df = 0;
static char inst[LONG_S_SIZE] = "unknown";
static cell _inst;
  /* The name of the institution, e.g. "Spitalul Universitar Bucuresti" */

static short dept_df = 0;
static char dept[LONG_S_SIZE] = "unknown";
static cell _dept;
  /* The name of the department, e.g. "Cardiologie" */

static short instaddr_df = 0;
static char instaddr[LONG_S_SIZE] = "unknown";
static cell _instaddr;
  /* The address of the institution, e.g. "169 Spl. Independentei, 75980, 5" */

static short instemail_df = 0;
static char instemail[LONG_S_SIZE] = "unknown";
static cell _instemail;
  /* The email address, e.g.:  */
  /* "impersonal@cardio.univermed-uhosp.ro" */
  /* The email address need not be (and usually is not) the */
  /* name of the host where the ECG processing takes place */

static short phone_df = 0;
static char phone[SHORT_S_SIZE] = "unknown";
static cell _phone;
  /* The phone number of the department, e.g. "401-638.4380" */

static short fax_df = 0;
static char fax[SHORT_S_SIZE] = "unknown";
static cell _fax;
  /* The fax number of the department */

static short host_df = 0;
static char host[SHORT_S_SIZE] = "unknown";
static cell _host;
  /* The name of the host where processing takes place, for example: */
  /* tempest.cardio.univermed-uhosp.ro */

static cell _site;
static cell the_site = NIL;

static init_site_info()

{
  /* PUT INITIALIZATION FROM RESOURCES HERE */


  _site = esym("site");
  _country = esym(":country");
  _city = esym(":city");
  _inst = esym(":institution");
  _dept = esym(":department");
  _instaddr = esym(":surface-mail");
  _instemail = esym(":electronic-mail");
  _phone = esym(":phone");
  _fax = esym(":fax");
  _host = esym(":host");
  _language = esym(":language");

  _eng = esym(":English");
  _rom = esym(":Romanian");
  _fra = esym(":French");
  _ger = esym(":German");
  _ita = esym(":Italian");
  _spa = esym(":Spanish");
}


/*--
  \end{code}

  \dataparagraph{Users information}

  \Definition{
  This section will store the users of the installed holter system,
  i.e. the people who analyse the recordings, and whose comments
  and other actions must be documented as such. This system is
  designed to ease the introduction of the name of the user.

  Users are kept in the resource file as well.
  }
  \SemanticStructure{
  There is a limit to the number of users that can be entered in the
  system (MAX\_NOF\_USERS). A fill pointer establishes how many
  users actually are in the system.

  A number of paralel strings identify the various fields
  of each item: the name of each user, his/her full titulature and
  his/her email address.
  }

  \begin{code}
  --*/

#define MAX_NOF_USERS 100

static short users_fill = 0;
static short crt_user = 0;

static short user_df[MAX_NOF_USERS];
  /* If user contents was changed */

static char sign_name[MAX_NOF_USERS][LONG_S_SIZE];
static cell _sign_name;
  /* the full name and title, e.g. "Dr. Paul Georgescu, PhD" */

static char user_name[MAX_NOF_USERS][SHORT_S_SIZE];
static cell _user_name;
  /* user name of the user, which must be a symbol, e.g. pgescu */

static char user_email[MAX_NOF_USERS][LONG_S_SIZE];
static cell _user_email;
  /* user email address, e.g. pgescu@cardio.univermed-uhost.ro */

static cell _user;

extern char *bylanguage(char *, char *, char *, char *, char *, char *);

static init_user_info()

{
  user_df[0] = 0;
  crt_user = 0;
  users_fill = 1;
  strcpy(sign_name[0], "Dr. ______________________");
  strcpy(user_name[0], bylanguage("unspecified", "nespecificat",
				  "", "", "", ""));
  strcpy(user_email[0], "no.email");
  _sign_name = esym(":sign-name");
  _user_email = esym(":user-email");
  _user = esym("user");
}


/*--
  \end{code}

  \dataparagraph{Patient information}

  \Definition{
  All the clinical and paraclinical information concerning a patient
  will be stored here, in a set of string, logical and numeric variables.
  }
  \SemanticStructure{
  The description of a patient has the following sections:
  \begin{description}
  \item[General] contains the name, date of birth and other
    information concerning the patient and the way in which his/her
    medical records can be identified.
  \item[History] contains a description of the characteristics
    of the case before undertaking the recording. Mainly, this
    consists of his general diagnostic, of the diagnostic of his
    current condition, the chronic treatments he followed until the
    recording started.
  \item[Actual moment] contains a description of laboratory
    --- including ecocardiographic ---
    and drug administration parameters at the moment of the
    recording.
  \item[Technical] contains a description of the technical
    conditions in which the recording was done (such as the
    leads used, the recording device used, special filters that
    were applied, etc).
  \item[Events] contains a description of the events that took place
    during the recording, both symptoms and drug taking.
  \end{description}
  }

  \begin{code}
  --*/



/*--
  \end{code}

  \subparagraph{General}

  \begin{code}
  --*/


static short p_name_df = 0;
static char p_name[SHORT_S_SIZE];
static cell _p_name;
  /* The full name of the patient */

static char p_birth_df = 0;
static char p_birth[SHORT_S_SIZE];
static cell _p_birth;
  /* The birth date of the patient */

static char p_address_df = 0;
static char p_address[SHORT_S_SIZE];

static char p_phone_df = 0;
static char p_phone[SHORT_S_SIZE];

static char p_id_df = 0;
static char p_id[SHORT_S_SIZE];

static char p_med_ref_df = 0;
static char p_med_ref[SHORT_S_SIZE];

static short p_sex_df = 0;
static short p_sex = 'u';    /* 'u'=undefined, 'm'=male, 'f'=female */
static cell _p_sex;
  /* The sex of the patient */

static short p_profession_df = 0;
static char p_profession[SHORT_S_SIZE];
static cell _p_profession;
  /* The profession of the patient */

static short rec_requested_df = 0;
static char rec_requested[SHORT_S_SIZE];
static cell _rec_requested;
  /* The person who requested the recording */

static cell _rec_start;
static cell _rec_duration;

static char rec_start[SHORT_S_SIZE];
static short rec_start_df = 0;

static short rec_by_df = 0;
static char rec_by[SHORT_S_SIZE];
static cell _rec_by;
  /* Who performed the recording */

static short rec_interpreted_df = 0;
static char rec_interpreted[SHORT_S_SIZE];
static cell _rec_interpreted;
  /* Who interpreted the recording */


/*--
  \end{code}

  \subparagraph{History}

  \begin{code}
  --*/

#define MAX_NOF_DIA_CODES 30

static cell _dia_codes[MAX_NOF_DIA_CODES];

#define MAX_NOF_DIAGNOSTICS 20

static short diagnostics_df[MAX_NOF_DIAGNOSTICS];
static char diagnostics[MAX_NOF_DIAGNOSTICS][LONG_S_SIZE];
static cell _diagnostics;
  /* All the diagnostics of the patient */

  /* DETAILED DIAGNOSTICS */

  /* Sorry for the following magics; they are to be found in
   * the keys of the corresponding grels and in the initialiser
   * that follows
   */

static short diabetes_df = 0;  
static short diabetes = 9;
  /* diabetes mell: 9=unkn, 6=absent, 11=type 1 (Ins), 12=type 2 (non-Ins) */
static cell _diabetes;
  /* Presence and/or type of diabetes */

static short nyha_df = 0;
static short nyha = 8;
  /* NYHA class: 8=unkn, 0=no CCF, 1=NYHA_I,
                 2=NYHA_II, 3=NYHA_III, 4=NYHA_IV */
static cell _nyha;
static cell _no_nyha;
static cell _nyha_I;
static cell _nyha_II;
static cell _nyha_III;
static cell _nyha_IV;

static short hta_df = 0;
static short hta = 7;
  /* arterial hypertension: 7=unkn, 5=absent, 21--24=stage I--IV */
static cell _hta;
static cell _no_hta;
static cell _hta_I;
static cell _hta_II;
static cell _hta_III;
static cell _hta_IV;


  /* OTHER DIAGNOSTICS */

static short record_reason_df = 0;
static char record_reason[HUGE_S_SIZE];
static cell _record_reason;
  /* reason for recording the patient */

static short episode_history_df = 0;
static char episode_history[HUGE_S_SIZE];
static cell _episode_history;
  /* history of the patient's arrhythmia/pain/other episode */

/*--
  \end{code}

  \subparagraph{Actual moment}

  \subsubparagraph{Treatment}

  The current treatment of the patient refers to
  the drugs the patient took up to the moment of the
  recording, and which can also be active during the
  recording. For every such drug, the following
  general cases are possible:

  \begin{description}
  \item[unknown] it is not known whether the drug is taken [anymore];
  \item[absent] it is known that the drug is not taken;
  \item[charging] for some drugs---like Digoxin and Amiodarone---there
     is a period of time in which the effective drug concentration
     was not yet reached;
  \item[chronic-1] the first (minimal drug level) dose in chronic
     administration; most drugs only have this stage;
  \item[chronic-2] the next chronical dose;
  \item[chronic-3] the next chronical dose;
  \item[chronic-4] the next chronical dose;
  \item[discharging] drug interrupted, but may be still active;
  \item[potentially overdosed] overdose is suspected;
  \item[surelly overdosed] overdose is sure;
  \item[toxic] toxic levels of the drug.
  \item[specific] specific situation related to that drug (for example,
  we may consider recently interrupted Digoxin treatment).
  \end{description}

For every drug that may be considered, there exist the following
slots to be filled:

  \begin{description}
  \item[Common International Name] which is a generic name for
    the administered substance;
  \item[Brand Name] or Comercial Name, which is the name given
    to a specific preparation by the manufacturer;
  \item[Possible Administration Routes] which specifies how the
    drug can possibly be administered---for drugs with 
    conceivable action on the heart; they are actually the following:
    \begin{description}
    \item[p.o.] by mouth;
    \item[i.v.] intravenous;
    \item[i.m.] intramuscular;
    \item[s.c.] subcutaneous;
    \item[p.c.] percutaneous;
    \item[i.c.] intraconjunctival;
    \item[i.r.] intrarachidian;
    \item[p.r.] intrarectal;
    \end{description}
  \item[Dose Options] Which of the dozage options above are possible
    for the given drug;
  \item[Dose Labels] How should the dozage options above
    be named for each of the drugs;
  \item[Dose Descriptions] How should the dozage options
    above be described to the physician so that he makes a
    reliable selection.
  \end{description}

  Apart from these, drug specification implies classification of
  drugs. This is done by means of a Lisp tree which has the following
  structure:

\begin{verbatim}
<drug-tree>         ::= (<drug-class> <drug-element> *)
<drug-class>        ::= drugs | <drug-class-symbol>
<drug-class-symbol> ::= <symbol>
<drug-element>      ::= <drug-tree> | <DCI-drug-name>
<DCI-drug-name>     ::= <symbol> ;; common international denomination
\end{verbatim}

\begin{code}
--*/

/* FOLLOWS A DESCRIPTION OF THE POSSIBLE DRUGS */

#define MAX_NOF_DRUGS 100

#define _R_PO 0
#define _R_IV 1
#define _R_IM 2
#define _R_SC 3
#define _R_PC 4
#define _R_IC 5
#define _R_IR 6
#define _R_PR 7

#define N_ROUTES 8

#define _D_UNKNOWN     0
#define _D_ABSENT      1
#define _D_CHARGING    2
#define _D_CHRONIC_1   3
#define _D_CHRONIC_2   4
#define _D_CHRONIC_3   5
#define _D_CHRONIC_4   6
#define _D_DISCHARGING 7
#define _D_POT_OVER    8
#define _D_SURE_OVER   9
#define _D_TOXIC      10
#define _D_SPECIFIC   11

#define N_DOSES       12

struct drug_info {
  cell dci;                  /* symbol */
  cell brand;                /* symbol */
  short routef[N_ROUTES];    /* flags  */
  short dosef[N_DOSES];      /* flags  */
  cell dose_labels[N_DOSES]; /* strings */
  cell dose_docs[N_DOSES];   /* strings */
};

/* The symbol names of the dose classes */

cell _unknown_d;
cell _absent_d;
cell _charging_d;
cell _chronic_1_d;
cell _chronic_2_d;
cell _chronic_3_d;
cell _chronic_4_d;
cell _discharging_d;
cell _potential_over_d;
cell _sure_over_d;
cell _toxic_d;
cell _specific_d;

/* Lisp to C symbol mapping */

cell _dose_map[N_DOSES];

struct drug_info the_drugs[MAX_NOF_DRUGS];
short n_drugs = 0;        /* number of already defined drugs */
cell drug_tree = NULL;    /* drug classification */

/* Drug information for the actual patient 
 * The  C symbols _D_UNKNOWN --- _D_SPECIFIC are used 
 * The above-mentiones n_drugs is a fill pointer in these
 * vectors too.
 */

short drug_df[MAX_NOF_DRUGS];
short drug[MAX_NOF_DRUGS];
cell _chronic_treatment;


/*--
  \end{code}

  \subsubparagraph{Echocardiography}

  A whole echocardiography result package might
  be inserted here in the foreseeable future. This should include
  trans\oe sophageal echography and pictographic information taken
  from the echocardiographers.

  For the moment, only basic elements such as basic echocardiographic
  diagnostics (like thrombi and tumors) and very simple measurements
  such as the cavity dimensions are stored.

  \begin{code}
  --*/

static short any_echo_f = 0;               /* activates the whole section */

static char last_echo_date[SHORT_S_SIZE];
static short last_echo_date_df = 0;
static cell _last_echo_date;

/* AORTA */

static short disec_ao;
static short disec_ao_df;
static cell _disec_ao;

/* LEFT ATRIUM */

static short thromb_la;
static short thromb_la_df;
static cell _thromb_la;

static short tumor_la;
static short tumor_la_df;
static cell _tumor_la;

static short long_la;
static short long_la_df;
static cell _long_la;

static short transv_la;
static short transv_la_df;
static cell _transv_la;


/* RIGHT ATRIUM */

static short thromb_ra;
static short thromb_ra_df;
static cell _thromb_ra;

static short tumor_ra;
static short tumor_ra_df;
static cell _tumor_ra;

static short long_ra;
static short long_ra_df;
static cell _long_ra;

static short transv_ra;
static short transv_ra_df;
static cell _transv_ra;


/* LEFT VENTRICLE */

static short thromb_lv;
static short thromb_lv_df;
static cell _thromb_lv;

static short tumor_lv;
static short tumor_lv_df;
static cell _tumor_lv;

static short sys_lv;
static short sys_lv_df;
static cell _sys_lv;

static short diast_lv;
static short diast_lv_df;
static cell _diast_lv;

static short post_wall_lv_df;
static short post_wall_lv;
static cell  _post_wall_lv;


/* RIGHT VENTRICLE */

static short thromb_rv;
static short thromb_rv_df;
static cell _thromb_rv;

static short tumor_rv;
static short tumor_rv_df;
static cell _tumor_rv;

static short sys_rv;
static short sys_rv_df;
static cell _sys_rv;

static short diast_rv;
static short diast_rv_df;
static cell _diast_rv;


/* AORTIC VALVE */

static short calc_sig_ao;
static short calc_sig_ao_df;
static cell _calc_sig_ao;

static short veg_si_ao;
static short veg_si_ao_df;
static cell _veg_si_ao;

static short bicusp_ao;
static short bicusp_ao_df;
static cell _bicusp_ao;


/* MITRAL VALVE */

static short fib_mv;
static short fib_mv_df;
static cell _fib_mv;

static short calc_mv;
static short calc_mv_df;
static cell _calc_mv;

static short veg_mv;
static short veg_mv_df;
static cell _veg_mv;

static short prol_mv;
static short prol_mv_df;
static cell _prol_mv;

static short bal_mv;
static short bal_mv_df;
static cell _bal_mv;

static short mix_mv;
static short mix_mv_df;
static cell _mix_mv;

static short rupt_cdj_mv;
static short rupt_cdj_mv_df;
static cell _rupt_cdj_mv;

static short fib_cdj_mv;
static short fib_cdj_mv_df;
static cell _fib_cdj_mv;

static short calc_cdj_mv;
static short calc_cdj_mv_df;
static cell _calc_cdj_mv;

static short lax_cdj_mv;
static short lax_cdj_mv_df;
static cell _lax_cdj_mv;

static short pil_imob_mv;
static short pil_imob_mv_df;
static cell _pil_imob_mv;

static short pil_rupt_mv;
static short pil_rupt_mv_df;
static cell _pil_rupt_mv;

/* INTER-ATRIAL SEPTUM */

static short ia_sept_cont_df;
static short ia_sept_cont;
static cell  _ia_sept_cont;

/* INTER-VENTRICULAR SEPTUM */

static short iv_sept_cont_df;
static short iv_sept_cont;
static cell  _iv_sept_cont;

static short iv_sept_base_df;
static short iv_sept_base;
static cell  _iv_sept_base;

static short iv_sept_med_df;
static short iv_sept_med;
static cell  _iv_sept_med;

/* DOPPLER */

static short e_on_a_df;
static short e_on_a;
static cell _e_on_a;

/* PERICARDIUM */

static short peric_width_df;
static short peric_width;
static cell _peric_width;

static short peric_lich_df;
static short peric_lich;
static cell _peric_lich;


/*--
  \end{code}

  \subsubparagraph{Clinical}

  \begin{code}
  --*/

static short at_df;
static short at_max, at_min;
static cell _at_max;
static cell _at_min;

static short ausc_h_r_df;
static short ausc_h_r;
static cell _ausc_h_r;

static short pulm_load_df;
static short pulm_load;
static cell _pulm_load;

static short syst_stasis_df;
static short syst_stasis;
static cell _syst_stasis;

static short temp_df;
static short temp;
static cell _temp;

static short goitre_df;
static short goitre;
static cell _goitre;

static short na_df;
static short na;
static cell _na;

static short k_df;
static short k;
static cell _k;

static short ca_df;
static short ca;
static cell _ca;

static short mg_df;
static short mg;
static cell _mg;

static short urea_df;
static short urea;
static cell _urea;

static short creat_df;
static short creat;
static cell _creat;

static short hemoglobin_df;
static short hemoglobin;
static cell _hemoglobin;

static short co2_df;
static short co2;
static cell _co2;

static short o2_df;
static short o2;
static cell _o2;

static short rx_cardiomeg_df;
static short rx_cardiomeg;
static cell _rx_cardiomeg;


/*--
  \end{code}

  \subparagraph{Technical}

  This section contains information on the recording conditions,
  including the placement of the electrodes, on the file and
  the volume where it is stored---in the long term, and
  on various other issues such as the compression type and
  the size of the file. Some of this information is
  completed by the Electrocardiognosis program.

  \subsubparagraph{Electrode placement}

  Electrode placement will be represented by a ``bitwise-or'' combination 
  of two ``chest coordinates'', one ``craniocaudal''---designating
  the intercostal space where the electrode is placed, and the
  other---horizontal.

  A single such code is used for each of the derivations. Nevertheless,
  in the Lisp file, a rather complete description is given to the place
  of each of the leads---which combines the two coordinates.

  On the screen, the placement is described in plain words, but
  sometimes a shorter name such as ``jugular incisure'' which
  is much better understood is to be given.

  \begin{code}
  --*/

/* The Y coordinate */

#define SUPRACLAV   0x0001
#define SUBCLAV     0x0002
#define IC_SPACE_1  0x0003
#define IC_SPACE_2  0x0004
#define IC_SPACE_3  0x0005
#define IC_SPACE_4  0x0006
#define IC_SPACE_5  0x0007
#define IC_SPACE_6  0x0008
#define IC_SPACE_7  0x0009
#define IC_SPACE_8  0x000a
#define SUBCOST     0x000b

static char *y_coord_names[0x000c] = 
  {   "", /* for 0 */
      "supraclavicular-space",
      "subclavicular-space",
      "first-intercostal-space",
      "second-intercostal-space",
      "third-intercostal-space",
      "fourth-intercostal-space",
      "fifth-intercostal-space",
      "sixth-intercostal-space",
      "seventh-intercostal-space",
      "eighth-intercostal-space",
      "subcostal"
  };

/* The X coordinate */

#define CENTRAL            0x0100
#define RIGHT_LAT_ST       0x0200
#define LEFT_LAT_ST        0x0300
#define RIGHT_MEDC_LINE    0x0400
#define LEFT_MEDC_LINE     0x0500
#define RIGHT_ANT_AX_LINE  0x0600
#define LEFT_ANT_AX_LINE   0x0700
#define RIGHT_MED_AX_LINE  0x0800
#define LEFT_MED_AX_LINE   0x0900
#define RIGHT_POST_AX_LINE 0x0a00
#define LEFT_POST_AX_LINE  0x0b00

static char *x_coord_names[0x000c] = 
    { "",
      "median-line",
      "right-laterosternal",
      "left-laterosternal",
      "right-medioclavicular",
      "left-medioclavicular",
      "right-anterior-axilar-line",
      "left-anterior-axilar-line",
      "right-median-axilar-line",
      "left-median-axilar-line",
      "right-posterior-axilar-line",
      "left-posterior-axilar-line"
    };

/* THE ACTUAL PLACEMENTS FOR THE ELECTRODES */

static short a_plus_df = 0;
static short a_plus = IC_SPACE_5 | LEFT_MED_AX_LINE;
static short a_minus_df = 0;
static short a_minus = IC_SPACE_2 | RIGHT_LAT_ST;
static short b_plus_df = 0;
static short b_plus = CENTRAL | SUBCOST;
static short b_minus_df = 0;
static short b_minus = IC_SPACE_2 | RIGHT_LAT_ST;


/*--
  \end{code}

  \subsubparagraph{File and storage}

  The file name is not represented here, as it cannot be changed.
  Nor is the file size, the date of its creation, etc.

  Storage on various volumes and similar
  information is not related to the patient data entry package.

  \begin{code}
  --*/


cell _unkn;    

static init_patient_info()

{
  short i;

  _p_name = esym("patient-name");
  _p_birth = esym("patient-birth-date");
  _p_sex = esym("patient-sex");
  _p_profession = esym("patient-profession");  /* not used anymore */
  _rec_requested = esym("recording-requested-by");
  _rec_start = esym("recording-start");
  _rec_duration = esym("recording-duration");
  _rec_by = esym("recording-done-by");
  _diagnostics = esym("diagnostics");

  /* Initialisation of the special diagnostics follows */

  for(i=0; i<MAX_NOF_DIA_CODES; i++)
    _dia_codes[i] = NULL;
  
  _unkn = esym("unknown");
  _diabetes = esym(":diabetes");
/*    _dia_codes[9] = _unkn;
    _type_I_diabetes = esym("type-I-diabetes");
    _dia_codes[11] = _type_I_diabetes;
    _type_II_diabetes = esym("type-II-diabetes");
    _dia_codes[12] = _type_II_diabetes;
    _no_diabetes = esym("no-diabetes");
    _dia_codes[6] = _no_diabetes; */
  
  _nyha = esym(":NYHA-class-of-cardiac-failure");
    _dia_codes[8] = _unkn;
    _no_nyha = esym("no-congestive-cardiac-failure");
    _dia_codes[0] = _no_nyha;
    _nyha_I = esym("NYHA-class-I");
    _dia_codes[1] = _nyha_I;
    _nyha_II = esym("NYHA-class-II");
    _dia_codes[2] = _nyha_II;
    _nyha_III = esym("NYHA-class-III");
    _dia_codes[3] = _nyha_III;
    _nyha_IV = esym("NYHA-class-IV");
    _dia_codes[4] = _nyha_IV;

  _hta = esym(":systemic-hypertension-stage");
    _dia_codes[7] = _unkn;
    _no_hta = esym("no-hypertension");
    _dia_codes[5] = _no_hta;
    _hta_I = esym("stage-I-hypertension");
    _dia_codes[21] = _hta_I;
    _hta_II = esym("stage-II-hypertension");
    _dia_codes[22] = _hta_II;
    _hta_III = esym("stage-III-hypertension");
    _dia_codes[23] = _hta_III;
    _hta_IV = esym("stage-IV-hypertension");
    _dia_codes[24] = _hta_IV;

  _record_reason = esym(":reason-for-recording");
  _episode_history = esym(":episode-history");

  _dose_map[_D_UNKNOWN] = _unknown_d = esym("unknown-dose");
  _dose_map[_D_ABSENT] = _absent_d = esym("absent-dose");
  _dose_map[_D_CHARGING] = _charging_d = esym("charging-dose");
  _dose_map[_D_CHRONIC_1] = _chronic_1_d = esym("chronic-1-dose");
  _dose_map[_D_CHRONIC_2] = _chronic_2_d = esym("chronic-2-dose");
  _dose_map[_D_CHRONIC_3] = _chronic_3_d = esym("chronic-3-dose");
  _dose_map[_D_CHRONIC_4] = _chronic_4_d = esym("chronic-4-dose");
  _dose_map[_D_DISCHARGING] = _discharging_d = esym("discharging-dose");
  _dose_map[_D_POT_OVER] = _potential_over_d = esym("potential-overdose");
  _dose_map[_D_SURE_OVER] = _sure_over_d = esym("sure-overdose");
  _dose_map[_D_TOXIC] = _toxic_d = esym("toxic-dose");
  _dose_map[_D_SPECIFIC] = _specific_d = esym("specific-dose");

  _chronic_treatment = esym(":chronic-treatment");  

  _last_echo_date = esym(":date-of-the-last-echo");

  _disec_ao = esym("aortic-disection");
  _thromb_la = esym("left-atrial-thrombus");
  _tumor_la = esym("left-atrial-tumor");
  _long_la = esym("left-atrial-longitudinal-diameter");
  _transv_la = esym("left-atrial-transversal-diameter");
  _thromb_ra = esym("right-atrial-thrombus");
  _tumor_ra = esym("right-atrial-tumor");
  _long_ra = esym("right-atrial-longitudinal-diameter");
  _transv_ra = esym("right-atrial-transversal-diameter");
  _thromb_lv = esym("left-ventricular-thrombus");
  _tumor_lv = esym("left-ventricular-tumor");
  _sys_lv = esym("left-ventricular-systolic-diameter");
  _diast_lv = esym("left-ventricular-diastolic-diameter");
  _post_wall_lv = esym("left-ventricular-posterior-wall");
  _thromb_rv = esym("right-ventricular-thrombus");
  _tumor_rv = esym("right-ventricular-tumor");
  _sys_rv = esym("right-ventricular-systolic-diameter");
  _diast_rv = esym("right-ventricular-diastolic-diameter");
  _calc_sig_ao = esym("aortic-sigmoid-calcification");
  _veg_si_ao = esym("aortic-sigmoid-vegetation");
  _bicusp_ao = esym("aortic-sigmoid-bicuspidy");
  _fib_mv = esym("mitral-valve-fibrosis");
  _calc_mv = esym("mitral-valve-calcification");
  _veg_mv = esym("mitral-valve-vegetations");
  _prol_mv = esym("mitral-valve-prolaps");
  _bal_mv = esym("mitral-valve-balonization");
  _mix_mv = esym("mitral-valve-mixoid-degenerescence");
  _rupt_cdj_mv = esym("mitral-cordage-rupture");
  _fib_cdj_mv = esym("mitral-cordage-fibrosis");
  _calc_cdj_mv = esym("mitral-cordage-calcification");
  _lax_cdj_mv = esym("mitral-cordage-laxity");
  _pil_imob_mv = esym("mitral-papilary-imobilisation");
  _pil_rupt_mv = esym("mitral-papilary-rupture");
  _ia_sept_cont = esym("interatrial-foramen");
  _iv_sept_cont = esym("interventricular-foramen");
  _iv_sept_base = esym("interventricular-septum-size-basal");
  _iv_sept_med = esym("interventricular-septum-size-medium");
  _e_on_a = esym("E-on-A");

  _peric_width = esym("width-of-pericardium");
  _peric_lich = esym("pericarditis");

  _at_max = esym("max-arterial-pressure");
  _at_min = esym("min-arterial-pressure");
  _ausc_h_r = esym("auscultatory-heart-rate");
  _pulm_load = esym("pulmonary-load");
  _syst_stasis = esym("systemic-stasis");
  _temp = esym("temperature");
  _goitre = esym("goitre");
  _na = esym("natremia");
  _k = esym("kaliemia");
  _ca = esym("calcemia");
  _mg = esym("magnesemia");
  _urea = esym("blood-urea");
  _creat = esym("blood-creatinine");
  _hemoglobin = esym("blood-haemoglobin");
  _co2 = esym("blood-carbon-dioxide");
  _o2 = esym("blood-oxigen");
  _rx_cardiomeg = esym("radiologic-cardiomegaly");

}


/*--
\end{code}

\subsubsection{Users information}

\begin{code}
--*/

cell extract_user_spec(short u)  /* user count; time ignored for the moment */

{
  cell l;

  l = cons(_user_email, cons(strbuild(user_email[u]), NIL));
  l = cons(_sign_name, cons(strbuild(sign_name[u]), l));
  l = cons(strbuild(""), cons(esym(user_name[u]), l));
  return cons(_user, l);
}

cell the_current_user()

{
  return esym(user_name[crt_user]);
}

/*--
  \end{code}

  \subsubsection{Site information}

  \ImplementationNotes{
  The site information is stored in {\tt \$ECGROOT/ecgnosis.sit} which 
  is a lisp file. The snapshot of this, at a given moment, is copied
  in the {\tt .ctl} file which controls that case, thus characterising the
  case. Nevertheless, this subpackage does not deal with the filename, but
  only with editing the site information and with loading and
  saving it.

  NOTE: $\uparrow$ this is not a good coment.
  }

  \begin{code}
  --*/


make_site_info_screen(GWIN *gg)

{
  cleardwin(gg, 0, PLANE_DWIN);
  cleardwin(gg, upd, PLANE_DWIN);
  cleardwin(gg, lod, PLANE_DWIN);
  cleardwin(gg, umctrld, PLANE_DWIN);
  cleardwin(gg, lctrld, PLANE_DWIN);
  cleardwin(gg, fulld, PLANE_DWIN);
  cleardwin(gg, rjumps, PLANE_DWIN);

  defbutton(gg, rjumps, 18,  23, 90, 20, "Users", 'U');
  defbutton(gg, rjumps, 18,  53, 90, 20, "Done", '\r');
  defbutton(gg, rjumps, 18,  83, 90, 20, "Back", 0x001b);
  defhbutton(gg, rjumps, 24);

  gsetcolor(gg, fulld, 12);
  gtext(gg, fulld, 10, 20, "Information on the local workstation.");

  gsetcolor(gg, fulld, 14);
  gtext(gg, fulld, 10, 50, "Country [c]");
  deftextpad(gg, fulld, 150, 50, 30, 1, 30, country, 'c');

  gtext(gg, fulld, 10, 80, "Language [l]");
  defaltswitch(gg, fulld,  10, 100, 130, 16, "^E: English", &language, 5);
  defaltswitch(gg, fulld, 150, 100, 130, 16, "^R: Romana", &language, 18);
  defaltswitch(gg, fulld, 300, 100, 130, 16, "^F: Francaise", &language, 6);
  defaltswitch(gg, fulld,  10, 120, 130, 16, "^G: Deutsche", &language, 7);
  defaltswitch(gg, fulld, 150, 120, 130, 16, "^I: Italiano", &language, 9);
  defaltswitch(gg, fulld, 300, 120, 130, 16, "^S: Espanol", &language, 19);

  gtext(gg, fulld, 10, 150, "City [y]");
  deftextpad(gg, fulld, 150, 150, 30, 1, 30, city, 'y');

  gtext(gg, fulld, 10, 180, "Institution [i]");
  deftextpad(gg, fulld, 150, 180, 30, 1, 30, inst, 'i');

  gtext(gg, fulld, 10, 210, "Department [d]");
  deftextpad(gg, fulld, 150, 210, 30, 1, 30, dept, 'd');

  gtext(gg, fulld, 10, 240, "Surface mail [s]");
  deftextpad(gg, fulld, 150, 240, 30, 1, 30, instaddr, 's');

  gtext(gg, fulld, 10, 270, "Host [h]");
  deftextpad(gg, fulld, 150, 270, 30, 1, 30, host, 'h');

  gtext(gg, fulld, 10, 300, "Email [e]");
  deftextpad(gg, fulld, 150, 300, 30, 1, 30, instemail, 'e');

  gtext(gg, fulld, 10, 330, "Telephone [t]");
  deftextpad(gg, fulld, 150, 330, 30, 1, 30, phone, 't');

  gtext(gg, fulld, 10, 360, "FAX [F]");
  deftextpad(gg, fulld, 150, 360, 30, 1, 30, fax, 'F');
}

/*--
\end{code}

\funparagraph{read\_site\_info}

\begin{code}
--*/

extern char *the_site_file_name();

void read_site_info()

{
  cell f;
  cell l;
  short u;

  if(!access(the_site_file_name(), R_OK))
    f = fileopen(the_site_file_name(), INPUT, CHAR_FILE);
  else
    f = NIL;
  if(f!=NIL) {
    l = lread(f);
    if(car(l)!=_site) {
      /* EXCEPTION: invalid site file */
      return ;
    }
    the_site = l;
    l = cdr(l);
    while(l!=NIL) {
      if(cdr(l)==NIL) {
	/* EXCEPTION: invalid site specification */
	return ;
      }
      if(car(l)==_country) {
        if(is_string(cadr(l))) {
          strncpy(country, string(cadr(l)), SHORT_S_SIZE);
        } else {
          /* EXCEPTION: invalid country spec */
        }
        l = cddr(l);
        continue;
      }
      if(car(l)==_language) {
        if(is_symbol(cadr(l))) {
	  language = 5; /* i.e. English, the default */
	  if(cadr(l)==_rom) {
	    language = 18;
	  }
	  if(cadr(l)==_fra) {
	    language = 6;
	  }
	  if(cadr(l)==_ger) {
	    language = 7;
	  }
	  if(cadr(l)==_ita) {
	    language = 9;
	  }
	  if(cadr(l)==_spa) {
	    language = 19;
	  }
	  /* The ``others'' exception must be raised here also,
	   * but the C language does not have an ELSIF instruction
	   * and the complexity of the code would grow too much 
	   * by using else, making it unreadable and hard to debug
	   */
	} else {
	  /* EXCEPTION: invalid language spec */
	}
	l = cddr(l);
	continue;
      }
      if(car(l)==_city) {
	if(is_string(cadr(l))) {
	  strncpy(city, string(cadr(l)), SHORT_S_SIZE);
	}
	else {
	  /* EXCEPTION: invalid city spec */
	}
	l = cddr(l);
	continue;
      }
      if(car(l)==_inst) {
	if(is_string(cadr(l))) {
	  strncpy(inst, string(cadr(l)), LONG_S_SIZE);
	}
	else {
	  /* EXCEPTION: invalid institution spec */
	}
	l = cddr(l);
	continue;
      }
      if(car(l)==_dept) {
	if(is_string(cadr(l))) {
	  strncpy(dept, string(cadr(l)), LONG_S_SIZE);
	}
	else {
	  /* EXCEPTION: invalid department spec */
	}
	l = cddr(l);
	continue;
      }
      if(car(l)==_instaddr) {
	if(is_string(cadr(l))) {
	  strncpy(instaddr, string(cadr(l)), LONG_S_SIZE);
	}
	else {
	  /* EXCEPTION: invalid institution address spec */
	}
	l = cddr(l);
	continue;
      }
      if(car(l)==_instemail) {
	if(is_string(cadr(l))) {
	  strncpy(instemail, string(cadr(l)), LONG_S_SIZE);
	}
	else {
	  /* EXCEPTION: invalid institution email spec */
	}
	l = cddr(l);
	continue;
      }
      if(car(l)==_phone) {
	if(is_string(cadr(l))) {
	  strncpy(phone, string(cadr(l)), SHORT_S_SIZE);
	}
	else {
	  /* EXCEPTION: invalid phone spec */
	}
	l = cddr(l);
	continue;
      }
      if(car(l)==_fax) {
	if(is_string(cadr(l))) {
	  strncpy(fax, string(cadr(l)), SHORT_S_SIZE);
	}
	else {
	  /* EXCEPTION: invalid fax spec */
	}
	l = cddr(l);
	continue;
      }
      if(car(l)==_host) {
	if(is_string(cadr(l))) {
	  strncpy(host, string(cadr(l)), SHORT_S_SIZE);
	}
	else {
	  /* EXCEPTION: invalid host spec */
	}
	l = cddr(l);
	continue;
      }
      /* EXCEPTION: invalid slot/site specification */
      l = cddr(l);
      continue;
    }
    u = 0;
    while((l = lread(f))!=eof_obj_) {
      if(car(l)!=_user)
	continue; /* EXCEPTION: user specification expected in the site file */
      l = cddr(l);  /* time stamp ignored */
      strncpy(user_name[u], symname(car(l)), SHORT_S_SIZE);
      l = cddr(l);  /* :sign-name ignored */
      strncpy(sign_name[u], string(car(l)), LONG_S_SIZE);
      l = cddr(l);  /* :user-email ignored */
      strncpy(user_email[u], string(car(l)), LONG_S_SIZE);
      u ++;
    }
    if(u>0)
      users_fill = u;
    else
      users_fill = 1;  /* the initialised default is preserved */
    if(crt_user>=users_fill)
      crt_user = 0;
    fclose(disconnect(f));
  }
}


/*--
\end{code}

\funparagraph{extract\_site\_spec}

\begin{code}
--*/

cell the_language()

{
  switch(language) {
  case 5:
    return _eng;
  case 18:
    return _rom;
  case 6:
    return _fra;
  case 7:
    return _ger;
  case 9:
    return _ita;
  case 19:
    return _spa;
  }
  return _eng;
}

cell extract_site_spec()

{
  cell l;

  l = NIL;
  l = cons(_host, cons(strbuild(host), NIL));
  l = cons(_fax, cons(strbuild(fax), l));
  l = cons(_phone, cons(strbuild(phone), l));
  l = cons(_instemail, cons(strbuild(instemail), l));
  l = cons(_instaddr, cons(strbuild(instaddr), l));
  l = cons(_dept, cons(strbuild(dept), l));
  l = cons(_inst, cons(strbuild(inst), l));
  l = cons(_city, cons(strbuild(city), l));
  l = cons(_language, cons(the_language(), l));
  l = cons(_country, cons(strbuild(country), l));
  return cons(_site, l);
}

/*--
\end{code}

\funparagraph{write\_site\_info}

\begin{code}
--*/

void write_site_info()

{
  cell f;
  short u;

  f = fileopen(the_site_file_name(), OUTPUT, CHAR_FILE);
  pprint(extract_site_spec(), f);
  for(u=0; u<users_fill; u++)
    pprint(extract_user_spec(u), f);
  fclose(disconnect(f));
}


/*--
\end{code}

\funparagraph{init\_file\_with\_site}

\begin{code}
--*/


void init_file_with_site(cell f)

{
  short u;

  pprint(extract_site_spec(), f);
  for(u=0; u<users_fill; u++)
    pprint(extract_user_spec(u), f);
}


/*--
  \end{code}

  \subsection{Language support}

  \begin{code}
  --*/

char *bylanguage(char *eng, char *rom, char *fra,
  char *ger, char *ita, char *spa)

{
  switch(language) {
  case 5: /* ^E */
    return eng;
  case 18: /* ^R */
    if (rom!=NULL)
      return rom;
    else
      return eng;
  case 6: /* ^F */
    if(fra!=NULL)
      return fra;
    else
      return eng;
  case 7: /* ^G */
    if(ger!=NULL)
      return ger;
    else
      return eng;
  case 9: /* ^I */
    if(ita!=NULL)
      return ita;
    else
      return eng;
  case 19: /* ^S */
    if(spa!=NULL)
      return spa;
    else
      return eng;
  default:
    return eng;
  }
}

/*--
\end{code}

\funparagraph{egviewdt_init}

\begin{code}
--*/

egviewdt_init() {
  init_site_info();
  init_user_info();
  init_patient_info();
}


/*--
\end{code}

\funparagraph{users\_display}

\begin{code}
--*/

short selected_user = 0;

void users_display(GWIN *gg, short key0)  /* always on fulld */

{
  short i, j, u;

  for(i=0; i<3; i++) { /* x */
    for(j=0; j<18; j++) {
      u = i * 18 + j;
      if(u>=users_fill)
	break;
      defaltswitch(gg, fulld, i * 160 + 10, j * 20 + 20, 150, 16,
		   user_name[u], &selected_user, u + key0);
    }
  }
}


/*--
\end{code}

\funparagraph{user\_edit}

\begin{code}
--*/

void user_edit(GWIN *gg, short u)

{
  short c;

  cleardwin(gg, fulld, PLANE_DWIN);
  cleardwin(gg, rjumps, PLANE_DWIN);

  gsetcolor(gg, fulld, 14);
  gtext(gg, fulld, 10, 100, "User [u] ");
  deftextpad(gg, fulld, 150, 100, 30, 1, 30, user_name[u], 'u');

  gtext(gg, fulld, 10, 130, "Sign name [s] ");
  deftextpad(gg, fulld, 150, 130, 30, 1, 30, sign_name[u], 's');

  gtext(gg, fulld, 10, 160, "Email address [e] ");
  deftextpad(gg, fulld, 150, 160, 30, 1, 30, user_email[u], 'e');

  defbutton(gg, rjumps, 18, 23, 90, 20, "Done", '\r');
  defbutton(gg, rjumps, 18, 53, 90, 20, "Cancel", 0x001b);
  defhbutton(gg, rjumps, 24);

  while((c=gwinteract(gg))!='\r' && c!=0x001b && c!=24);
  switch(c) {
  case '\r':
    gwaccept(gg);
    break;
  case 0x001b:
    break;
  case 24:
    break;
  default:
    break; /* impossible */
  }
  cleardwin(gg, fulld, PLANE_DWIN);
  cleardwin(gg, rjumps, PLANE_DWIN);
}

/*--
\end{code}

\funparagraph{users_interaction}

\begin{code}
--*/

void users_interaction(GWIN *gg)

{
  short c, i;
  short redraw;

  redraw = 1;
  selected_user = crt_user + 'a';
  do {
    if(redraw) {
      cleardwin(gg, fulld, PLANE_DWIN);
      cleardwin(gg, rjumps, PLANE_DWIN);
      users_display(gg, 'a');
      defbutton(gg, rjumps, 18,  23, 90, 20, "Edit", 'E');
      defbutton(gg, rjumps, 18,  53, 90, 20, "Set", 'S');
      defbutton(gg, rjumps, 18,  83, 90, 20, "Delete", 'D');
      defbutton(gg, rjumps, 18, 113, 90, 20, "New", 'N');
      defbutton(gg, rjumps, 18, 143, 90, 20, "Back", '\r');
      defhbutton(gg, rjumps, 24);
      redraw = 0;
    }
    while((c=gwinteract(gg))>='a')
      gwaccept(gg);
    switch(c) {
    case 'E':
      user_edit(gg, selected_user-'a');
      redraw = 1;
      break;
    case 'S':
      crt_user = selected_user - 'a';
      break;
    case 'N':
      user_name[users_fill][0] = 0;
      sign_name[users_fill][0] = 0;
      user_email[users_fill][0] = 0;
      users_fill ++;
      user_edit(gg, users_fill-1);
      redraw = 1;
      break;
    case 'D':
      if(users_fill<=1)
	break;
      for(i=selected_user-'a'+1; i<users_fill; i++) {
	strncpy(user_name[i-1], user_name[i], SHORT_S_SIZE);
	strncpy(sign_name[i-1], sign_name[i], SHORT_S_SIZE);
	strncpy(user_email[i-1], user_email[i], SHORT_S_SIZE);
      }
      users_fill --;
      if(selected_user-'a'>=users_fill)
	selected_user = 'a';
      if(crt_user>=users_fill)
	crt_user = 0;
      redraw = 1;
      break;
    default:
      break;
    }
  } while(c!='\r');
  cleardwin(gg, fulld, PLANE_DWIN);
  cleardwin(gg, rjumps, PLANE_DWIN);
    
}

/*--
\end{code}

\funparagraph{file\_interaction}

\begin{code}
--*/


file_interaction(GWIN *gg)

{
  short c;

  read_site_info();
  make_site_info_screen(gg);
  while((c = gwinteract(gg))!='\r' && c!=0x001b)
    switch(c) {
    case 'U':
      users_interaction(gg);
      make_site_info_screen(gg);
      continue;
    default:
      continue;
    }
  if(c=='\r') {
    gwaccept(gg);
    write_site_info();
  }
  cleardwin(gg, fulld, PLANE_DWIN);
}


/*--
\end{code}

\funparagraph{rec\_info\_edit}

\begin{code}
--*/

extern char *the_ecg_file_name();
extern char *bylanguage(char *, char *, char *, char *, char *, char *);
extern cell the_current_user();
extern cell slot_of(cell, cell);
extern cell last_assertion_named(char *);


#define EL_UNKNOWN   18
#define EL_ECGNOSIS3 19
#define EL_STANDARD3 20
#define EL_ECGNOTHR3 21
#define EL_ECGNOSIS5 22
#define EL_ECGNOTHR5 23
#define EL_GLIKSEEV2 24
#define EL_GLIKSOTHR 25

#define EL_DIRA      16
#define EL_DIRB      17

short el_place = EL_UNKNOWN; /* unknown */

short electrodes_df = 0;

char dira[6] = "0";  /* for unknown */
char dirb[6] = "90";

void rec_info_edit(GWIN *gg)

{
  struct stat sb;
  time_t tt;
  short c, r;
  short redraw_f;
  cell as;

  redraw_f = 1;

  if((as = slot_of(last_assertion_named("recording-start"),
		   cl_make_symbol(":is"))) != NIL && is_string(as))
    strcpy(rec_start, string(as));
  else {
    stat(the_ecg_file_name(), &sb);
    tt = sb.st_mtime;
    strcpy(rec_start, ctime(&tt));
  }

  if((as = slot_of(last_assertion_named("recording-requested-by"),
		   cl_make_symbol(":name"))) != NIL && is_string(as))
    strcpy(rec_requested, string(as));
  else
    rec_requested[0] = 0;
  rec_requested_df = 0;

  if((as = slot_of(last_assertion_named("recording-done-by"),
		   cl_make_symbol(":name"))) != NIL && is_string(as))
    strcpy(rec_by, string(as));
  else
    rec_by[0] = 0;
  rec_by_df = 0;

  el_place = EL_UNKNOWN;

  if((as = slot_of(last_assertion_named("electrodes"),
		   cl_make_symbol(":generic-electrodes-placement"))) != NIL) {
    if(as==cl_make_symbol("single-V21"))
      el_place = EL_GLIKSEEV2;
    if(as==cl_make_symbol("frontal-3"))
      el_place = EL_ECGNOSIS3;
    if(as==cl_make_symbol("frontal-5"))
      el_place = EL_ECGNOSIS5;
    if(as==cl_make_symbol("Einthoven"))
      el_place = EL_STANDARD3;
  } else { /* i.e. one of the ``UNKNOWNs or OTHERs'' */
    if((as = slot_of(last_assertion_named("electrodes"),
		     cl_make_symbol(":cable-type"))) != NIL)  {
      if(as==cl_make_symbol("GLIKSEE"))
	el_place = EL_GLIKSOTHR;
      if(as==cl_make_symbol("ECGNOSIS-3"))
	el_place = EL_ECGNOTHR3;
      if(as==cl_make_symbol("ECGNOSIS-5"))
	el_place = EL_ECGNOTHR5;
    }
  }

  if((as = slot_of(last_assertion_named("electrodes"),
		   cl_make_symbol(":lead-orientation"))) != NIL) {
    sprintf(dira, "%d", binary(car(as)));
    sprintf(dirb, "%d", binary(cadr(as)));
  }
    

  do {
    if(redraw_f) {
      cleardwin(gg, fulld, PLANE_DWIN);
      cleardwin(gg, rjumps, PLANE_DWIN);
      gsetcolor(gg, fulld, 12);
      
      gtext(gg, fulld, 20, 0, bylanguage("Recording start",
				      "Inceputul inregistrarii",
				      "", "", "", ""));
      
      wi_time_panel(fulld, 20, 50, rec_start, YEAR2, SECOND, 'k', 'A', 'N');
      
      gtext(gg, fulld, 20, 120, bylanguage("Requested by:",
				       "Indicat de:", "", "", "", ""));
      deftextpad(gg, fulld, 160, 120, 30, 1, 30, rec_requested, 'r');
      defbutton(gg, fulld, 430, 120, 60, 20,
		bylanguage("List", "Lista", "", "", "", ""), 'q');
  
      gtext(gg, fulld, 20, 150, bylanguage("Done by:",
				       "Executat de:", "", "", "", ""));
      deftextpad(gg, fulld, 160, 150, 30, 1, 30, rec_by, 'd');
      defbutton(gg, fulld, 430, 150, 60, 20,
		bylanguage("List", "Lista", "", "", "", ""), 'x');

      gtext(gg, fulld, 20, 200, bylanguage("Electrode placement",
					   "Plasarea electrozilor",
					   "", "", "", ""));
      defaltswitch(gg, fulld, 20, 220, 200, 20,
		   bylanguage("Unknown", "Necunoscut", "", "", "", ""),
		   &el_place, EL_UNKNOWN);
      defaltswitch(gg, fulld, 20, 240, 200, 20,
		   bylanguage("Electrocardiognosis 3 el, V1/V5/Xif",
			      "Electrocardiognosis 3 el, V1/V5/Xif",
			      "", "", "", ""),
		   &el_place, EL_ECGNOSIS3);
      defaltswitch(gg, fulld, 20, 260, 200, 20,
		   bylanguage("Electrocardiognosis 3 el, Einthoven",
			      "Electrocardiognosis 3 el, Einthoven",
			      "", "", "", ""),
		   &el_place, EL_STANDARD3);
      defaltswitch(gg, fulld, 20, 280, 200, 20,
		   bylanguage("Electrocardiognosis 3 el, other",
			      "Electrocardiognosis 3 el, other",
			      "", "", "", ""),
		   &el_place, EL_ECGNOTHR3);
      defaltswitch(gg, fulld, 20, 300, 200, 20,
		   bylanguage("Electrocardiognosis 5 el, SV5/SP5",
			      "Electrocardiognosis 5 el, SV5/SP5",
			      "", "", "", ""),
		   &el_place, EL_ECGNOSIS5);
      defaltswitch(gg, fulld, 20, 320, 200, 20,
		   bylanguage("Electrocardiognosis 5 el, other",
			      "Electrocardiognosis 5 el, other",
			      "", "", "", ""),
		   &el_place, EL_ECGNOTHR5);
      defaltswitch(gg, fulld, 20, 340, 200, 20,
		   bylanguage("Gliksee 2 el, V2'",
			      "Gliksee 2 el, V2'",
			      "", "", "", ""),
		   &el_place, EL_GLIKSEEV2);
      defaltswitch(gg, fulld, 20, 360, 200, 20,
		   bylanguage("Gliksee 2 el, other",
			      "Gliksee 2 el, other",
			      "", "", "", ""),
		   &el_place, EL_GLIKSOTHR);

      gtext(gg, fulld,  20, 390,
	    bylanguage("Direction of A", "Directia lui A", "", "", "", ""));
      deftextpad(gg, fulld, 170, 390, 4, 1, 4, dira, EL_DIRA);

      gtext(gg, fulld, 270, 390,
	    bylanguage("Direction of B", "Directia lui B", "", "", "", ""));
      deftextpad(gg, fulld, 420, 390, 4, 1, 4, dirb, EL_DIRB);

      defbutton(gg, rjumps, 18,  23, 90, 20,
		bylanguage("OK", "Accepta", "", "", "", ""), 1);
      defbutton(gg, rjumps, 18,  53, 90, 20,
		bylanguage("Cancel", "Anuleaza", "", "", "", ""), 2);

      redraw_f = 0;
    }
      
    c = gwinteract(gg);
    r = wi_time_input(c);
    if (r) {
      strncpy(rec_start, wi_time_val('k'), SHORT_S_SIZE);
      rec_start_df = 1;
    }
    else
      switch(c) {
      case 'r':
	gwaccept(gg);
	rec_requested_df = 1;
	break;
      case 'd':
	gwaccept(gg);
	rec_by_df = 1;
	break;
      case 'q':
	users_interaction(gg);
	strncpy(rec_requested, sign_name[selected_user-'a'], SHORT_S_SIZE);
	rec_requested_df = 1;
	redraw_f = 1;
	break;
      case 'x':
	users_interaction(gg);
	strncpy(rec_by, sign_name[selected_user-'a'], SHORT_S_SIZE);
	rec_by_df = 1;
	redraw_f = 1;
	break;
      case EL_UNKNOWN:
	strcpy(dira, "0");
	strcpy(dirb, "90");
	gset_tp_contents(gg, fulld, EL_DIRA, dira);
	gset_tp_contents(gg, fulld, EL_DIRB, dirb);
	electrodes_df = 1;
	break;
      case EL_ECGNOSIS3:
	strcpy(dira, "30");
	strcpy(dirb, "90");
	gset_tp_contents(gg, fulld, EL_DIRA, dira);
	gset_tp_contents(gg, fulld, EL_DIRB, dirb);
	electrodes_df = 1;
	break;
      case EL_STANDARD3:
	strcpy(dira, "0");
	strcpy(dirb, "60");
	gset_tp_contents(gg, fulld, EL_DIRA, dira);
	gset_tp_contents(gg, fulld, EL_DIRB, dirb);
	electrodes_df = 1;
	break;
      case EL_ECGNOTHR3:
	strcpy(dira, "0");
	strcpy(dirb, "90");
	gset_tp_contents(gg, fulld, EL_DIRA, dira);
	gset_tp_contents(gg, fulld, EL_DIRB, dirb);
	electrodes_df = 1;
	break;
      case EL_ECGNOSIS5:
	strcpy(dira, "30");
	strcpy(dirb, "90");
	gset_tp_contents(gg, fulld, EL_DIRA, dira);
	gset_tp_contents(gg, fulld, EL_DIRB, dirb);
	electrodes_df = 1;
	break;
      case EL_ECGNOTHR5:
	strcpy(dira, "0");
	strcpy(dirb, "90");
	gset_tp_contents(gg, fulld, EL_DIRA, dira);
	gset_tp_contents(gg, fulld, EL_DIRB, dirb);
	electrodes_df = 1;
	break;
      case EL_GLIKSEEV2:
	strcpy(dira, "90");
	strcpy(dirb, "0");
	gset_tp_contents(gg, fulld, EL_DIRA, dira);
	gset_tp_contents(gg, fulld, EL_DIRB, dirb);
	electrodes_df = 1;
	break;
      case EL_GLIKSOTHR:
	strcpy(dira, "90");
	strcpy(dirb, "0");
	gset_tp_contents(gg, fulld, EL_DIRA, dira);
	gset_tp_contents(gg, fulld, EL_DIRB, dirb);
	electrodes_df = 1;
	break;
      case EL_DIRA:
      case EL_DIRB:
	electrodes_df = 1;
	break;
      default:
	break;
      }
    
  } while((c & 0x00ff) != 1 && (c & 0x00ff) != 2);

  c = c & 0x00ff;
  switch(c) {
  case 1:
    tt = time(NULL);
    if (rec_requested_df) {
      as = cons(cl_make_symbol(":name"),
		cons(cl_make_string(rec_requested), NIL));
      as = cons(_rec_requested, as);

/* cons(cl_make_string(ctime(&tt)), cons(the_current_user(), as)));*/

      add_assertion(as);
      rec_requested_df = 0;
    }
    if (rec_by_df) {
      as = cons(cl_make_symbol(":name"),
		cons(cl_make_string(rec_by), NIL));
      as = cons(_rec_by, as);
 
  /*cons(cl_make_string(ctime(&tt)), cons(the_current_user(), as)));*/
      add_assertion(as);
      rec_by_df = 0;
    }
    if (rec_start_df) {
      /* accepted, generate the information in the control file */
      strncpy(rec_start, wi_time_val('k'), SHORT_S_SIZE);
      as = cons(cl_make_symbol(":is"),
		cons(cl_make_string(rec_start), NIL));
      as = cons(_rec_start, as);
/* cons(cl_make_string(ctime(&tt)), cons(the_current_user(), as))); */
      add_assertion(as);
      rec_start_df = 0;
    }
    if(electrodes_df) { /* DEVICE TYPE SPECIFIC */
      gwaccept(gg);
      as = cons(cl_make_symbol(":lead-orientation"),
		cons(cons(cl_make_integer(atoi(dira)),
			  cons(cl_make_integer(atoi(dirb)), NIL)), NIL));
      switch(el_place) {
      case EL_ECGNOSIS3:
      case EL_STANDARD3:
      case EL_ECGNOTHR3:
      case EL_ECGNOSIS5:
      case EL_ECGNOTHR5:
	as = cons(cons(cl_make_symbol("bipolar"),
		       cons(cl_make_symbol("bipolar"), NIL)), as);
	as = cons(cl_make_symbol(":lead-types"), as);

	as = cons(cl_make_symbol(":nr-of-leads"),
		  cons(cl_make_integer(2), as));

	as = cons(cl_make_symbol(":lead-labels"),
		  cons(cons(cl_make_symbol("A"),
			    cons(cl_make_symbol("B"), NIL)), as));
	break;
      case EL_GLIKSEEV2:
      case EL_GLIKSOTHR:
	as = cons(cons(cl_make_symbol("unipolar"), NIL), as);
	as = cons(cl_make_symbol(":lead-types"), as);

	as = cons(cl_make_symbol(":nr-of-leads"),
		  cons(cl_make_integer(1), as));

	as = cons(cl_make_symbol(":lead-labels"),
		  cons(cons(cl_make_symbol("GL"), NIL), as));
	break;
      default:
	break;
      }

      switch(el_place) { /* CABLE SPECIFIC */
      case EL_ECGNOSIS5:
      case EL_ECGNOTHR5:
	as = cons(cl_make_symbol(":electrode-labels"),
		  cons(cons(cl_make_symbol("Ref"),
			    cons(cl_make_symbol("A-"),
				 cons(cl_make_symbol("A+"),
				      cons(cl_make_symbol("B-"),
					   cons(cl_make_symbol("B+"),
						NIL))))),
		       as));

	as = cons(cl_make_symbol(":lead-composition"),
		  cons(cons(cons(cl_make_symbol("A-"),
				 cons(cl_make_symbol("A+"), NIL)),
			    cons(cons(cl_make_symbol("B-"),
				      cons(cl_make_symbol("B+"), NIL)), NIL)),
		       as));

	as = cons(cl_make_symbol(":nr-of-electrodes"),
		  cons(cl_make_integer(5), as));

	as = cons(cl_make_symbol(":cable-type"),
		  cons(cl_make_symbol("ECGNOSIS-5"), as));
	break;
      case EL_ECGNOSIS3:
      case EL_ECGNOTHR3:
      case EL_STANDARD3:
	as = cons(cl_make_symbol(":electrode-labels"),
		  cons(cons(cl_make_symbol("common"),
			    cons(cl_make_symbol("A+"),
				 cons(cl_make_symbol("B+"), NIL))),
		       as));

	as = cons(cl_make_symbol(":lead-composition"),
		  cons(cons(cons(cl_make_symbol("common"),
				 cons(cl_make_symbol("A+"), NIL)),
			    cons(cons(cl_make_symbol("common"),
				      cons(cl_make_symbol("B+"), NIL)), NIL)),
		       as));

	as = cons(cl_make_symbol(":nr-of-electrodes"),
		  cons(cl_make_integer(3), as));

	as = cons(cl_make_symbol(":cable-type"),
		  cons(cl_make_symbol("ECGNOSIS-3"), as));
	break;
      case EL_GLIKSEEV2:
      case EL_GLIKSOTHR:
	as = cons(cl_make_symbol(":electrode-labels"),
		  cons(cons(cl_make_symbol("Ref"),
			    cons(cl_make_symbol("GL+"), NIL)),
		       as));

	as = cons(cl_make_symbol(":lead-composition"),
		  cons(cons(cons(cl_make_symbol("Ref"),
				 cons(cl_make_symbol("GL+"), NIL)), NIL),
		       as));

	as = cons(cl_make_symbol(":nr-of-electrodes"),
		  cons(cl_make_integer(2), as));

	as = cons(cl_make_symbol(":cable-type"),
		  cons(cl_make_symbol("GLIKSEE"), as));
	break;
      default:
	break;
      }

      switch(el_place) {  /* PLACEMENT SPECIFIC */
      case EL_GLIKSEEV2:
	as = cons(cl_make_symbol(":electrode-placement"),
		  cons(cons(cl_make_symbol("left-laterosternal"),
			    cons(cl_make_symbol("second-intercostal-space"),
				 NIL)), as));
	as = cons(cl_make_symbol(":generic-electrodes-placement"),
		  cons(cl_make_symbol("single-V21"), as));
	break;
      case EL_ECGNOSIS3:
	as = cons(cl_make_symbol(":electrode-placement"),
		  cons(cons(cons(cl_make_symbol("right-laterosternal"),
			    cons(cl_make_symbol("second-intercostal-space"),
				 NIL)),
		       cons(cons(cl_make_symbol("median-line"),
				 cons(cl_make_symbol("subcostal"),
				      NIL)),
			    cons(cons(cl_make_symbol("left-anterior-axilar-line"),
				      cons(cl_make_symbol("fifth-intercostal-space"),
					   NIL)),
				 NIL))),
		  as));
	as = cons(cl_make_symbol(":generic-electrodes-placement"),
		  cons(cl_make_symbol("frontal-3"), as));
	break;
      case EL_STANDARD3:
	as = cons(cl_make_symbol(":electrode-placement"),
		  cons(cons(cons(cl_make_symbol("right-anterior-axilar-line"),
			    cons(cl_make_symbol("supraclavicular"),
				 NIL)),
		       cons(cons(cl_make_symbol("median-line"),
				 cons(cl_make_symbol("subcostal"),
				      NIL)),
			    cons(cons(cl_make_symbol("left-anterior-axilar-line"),
				      cons(cl_make_symbol("supraclavicular"),
					   NIL)),
				 NIL))),
		  as));
	as = cons(cl_make_symbol(":generic-electrodes-placement"),
		  cons(cl_make_symbol("Einthoven"), as));
	
    break;
      case EL_ECGNOSIS5:
	as = cons(cl_make_symbol(":electrode-placement"),
/* Ref */	  cons(cons(cons(cl_make_symbol("right-anterior-axilar-line"),
			    cons(cl_make_symbol("fifth-intercostal-space"),
				 NIL)),
/* A- */	       cons(cons(cl_make_symbol("right-laterosternal"),
				 cons(cl_make_symbol("second-intercostal-space"),
				      NIL)),
/* A+ */		    cons(cons(cl_make_symbol("left-anterior-axilar-line"),
				      cons(cl_make_symbol("fifth-intercostal-space"),
					   NIL)),
/* B- */			 cons(cons(cl_make_symbol("right-laterosternal"),
					   cons(cl_make_symbol("second-intercostal-space"),
						NIL)),
/* B+ */			      cons(cons(cl_make_symbol("median-line"),
						cons(cl_make_symbol("subcostal"),
						     NIL)),
				 NIL))))),
		  as));

	as = cons(cl_make_symbol(":generic-electrodes-placement"),
		  cons(cl_make_symbol("frontal-5"), as));
	
	break;
      default:
	break;
      }
      as = cons(cl_make_symbol("electrodes"), as);

/* cons(cl_make_string(ctime(&tt)), cons(the_current_user(), as)));*/

      add_assertion(as);
      electrodes_df = 0;
    }
    break;
  case 2:
    /* canceled, ignore all changes */
    rec_start_df = 0;
    electrodes_df = 0;
    rec_requested_df = 0;
    rec_by_df = 0;
    break;
  default:
    break;
  }
  wi_time_del_dwin(fulld);

}

/*--
\end{code}

\funparagraph{pat\_info\_edit}

\begin{code}
--*/

void pat_info_edit(GWIN *gg)

{
  struct stat sb;
  time_t tt;
  short c, r;
  short redraw_f;
  cell as;

  /* restore the information already known on the patient */

  if((as = slot_of(last_assertion_named("subject-identification"),
		  cl_make_symbol(":name"))) != NIL && is_string(as))
    strncpy(p_name, string(as), SHORT_S_SIZE);
  else
    p_name[0] = 0;

  if((as = slot_of(last_assertion_named("subject-identification"),
		   cl_make_symbol(":birth-date")))!=NIL && is_string(as))
    strncpy(p_birth, string(as), SHORT_S_SIZE);
  else {
    if((as = slot_of(last_assertion_named("recording-start"),
		     cl_make_symbol(":is"))) != NIL && is_string(as))
      strcpy(rec_start, string(as));
    else {
      stat(the_ecg_file_name(), &sb);
      tt = sb.st_mtime;
      strcpy(rec_start, ctime(&tt));
    }
  }

  p_sex = 'u';
  if((as = slot_of(last_assertion_named("subject-identification"),
		   cl_make_symbol(":sex")))!=NIL && is_symbol(as)) {
    if(as==cl_make_symbol("male"))
      p_sex = 'm';
    if(as==cl_make_symbol("female"))
      p_sex = 'f';
  }

  if((as = slot_of(last_assertion_named("subject-identification"),
		   cl_make_symbol(":address")))!=NIL && is_string(as)) 
    strncpy(p_address, string(as), SHORT_S_SIZE);
  else
    p_address[0] = 0;

  if((as = slot_of(last_assertion_named("subject-identification"),
		   cl_make_symbol(":phone")))!=NIL && is_string(as)) 
    strncpy(p_phone, string(as), SHORT_S_SIZE);
  else
    p_phone[0] = 0;


  if((as = slot_of(last_assertion_named("subject-identification"),
		   cl_make_symbol(":id")))!=NIL && is_string(as)) 
    strncpy(p_id, string(as), SHORT_S_SIZE);
  else
    p_id[0] = 0;

  if((as = slot_of(last_assertion_named("subject-identification"),
		   cl_make_symbol(":medical-reference")))!=NIL
     && is_string(as)) 
    strncpy(p_med_ref, string(as), SHORT_S_SIZE);
  else
    p_med_ref[0] = 0;

  redraw_f = 1;
  do {
    if(redraw_f) {
      cleardwin(gg, fulld, PLANE_DWIN);
      cleardwin(gg, rjumps, PLANE_DWIN);
      gsetcolor(gg, fulld, 12);

      gtext(gg, fulld, 20, 20, bylanguage("Patient name",
					  "Numele si prenumele pacientului",
					  "", "", "", ""));
      deftextpad(gg, fulld, 20, 40, 50, 1, 50, p_name, 'n');

      gtext(gg, fulld, 20, 70, bylanguage("Birth date", "Data nasterii",
					  "", "", "", ""));
      wi_time_panel(fulld, 20, 115, p_birth, YEAR4, DAY, 'b', 'A', 'N');

      gtext(gg, fulld, 270, 70, bylanguage("Sex", "Sexul", "", "", "", ""));
      defaltswitch(gg, fulld, 270,  90, 100, 20,
		   bylanguage("unknown", "necunoscut", "", "", "", ""),
		   &p_sex, 'u');
      defaltswitch(gg, fulld, 270, 110, 100, 20,
		   bylanguage("female", "femeiesc", "", "", "", ""),
		   &p_sex, 'f');
      defaltswitch(gg, fulld, 380, 110, 100, 20,
		   bylanguage("male", "barbatesc", "", "", "", ""),
		   &p_sex, 'm');

      gtext(gg, fulld, 20, 160, bylanguage("Patient address",
					  "Adresa pacientului",
					  "", "", "", ""));
      deftextpad(gg, fulld, 20, 180, 50, 1, 50, p_address, 'a');

      gtext(gg, fulld, 20, 210, bylanguage("ID Card No", "Seria si Nr BI",
					  "", "", "", ""));
      deftextpad(gg, fulld, 20, 230, 20, 1, 20, p_id, 'i');

      gtext(gg, fulld, 20, 260, bylanguage("Patient phone",
					  "Telefonul pacientului",
					  "", "", "", ""));
      deftextpad(gg, fulld, 20, 280, 20, 1, 20, p_phone, 'p');

      gtext(gg, fulld, 20, 310, bylanguage("Medical reference",
					  "FO sau Nr. Registru",
					  "", "", "", ""));
      deftextpad(gg, fulld, 20, 330, 20, 1, 20, p_med_ref, 'r');

      defbutton(gg, rjumps, 18,  23, 90, 20,
		bylanguage("OK", "Accepta", "", "", "", ""), 1);
      defbutton(gg, rjumps, 18,  53, 90, 20,
		bylanguage("Cancel", "Anuleaza", "", "", "", ""), 2);
      
      redraw_f = 0;
    }
    c = gwinteract(gg);
    r = wi_time_input(c);
    if (r) {
      strncpy(p_birth, wi_time_val('b'), SHORT_S_SIZE);
      p_birth_df = 1;
    }
    else
      switch(c) {
      case 'n':
	gwaccept(gg);
	p_name_df = 1;
	break;
      case 'u':
      case 'f':
      case 'm':
	gwaccept(gg);
	p_sex_df = 1;
	break;
      case 'i':
	gwaccept(gg);
	p_id_df = 1;
	break;
      case 'a':
	gwaccept(gg);
	p_address_df = 1;
	break;
      case 'p':
	gwaccept(gg);
	p_phone_df = 1;
	break;
      case 'r':
	gwaccept(gg);
	p_med_ref_df = 1;
	break;
      default:
	break;
      }
  } while((c & 0x00ff) != 1 && (c & 0x00ff) != 2);

  c = c & 0x00ff;
  switch(c) {
  case 1:
    if(p_name_df || p_birth_df || p_sex_df ||
       p_address_df || p_phone_df || p_id_df || p_med_ref_df) {
      as = cons(cl_make_symbol(":medical-reference"),
		cons(cl_make_string(p_med_ref), NIL));
      as = cons(cl_make_symbol(":id"),
		cons(cl_make_string(p_id), as));
      as = cons(cl_make_symbol(":phone"),
		cons(cl_make_string(p_phone), as));
      as = cons(cl_make_symbol(":address"),
		cons(cl_make_string(p_address), as));
      switch(p_sex) {
      case 'f':
	as = cons(cl_make_symbol(":sex"),
		  cons(cl_make_symbol("female"), as));
	break;
      case 'm':
	as = cons(cl_make_symbol(":sex"),
		  cons(cl_make_symbol("male"), as));
	break;
      case 'u':
      default:
	as = cons(cl_make_symbol(":sex"),
		  cons(cl_make_symbol("unknown"), as));
	break;
      }
      as = cons(cl_make_symbol(":birth-date"),
		cons(cl_make_string(p_birth), as));

      as = cons(cl_make_symbol(":name"),
		cons(cl_make_string(p_name), as));

      tt = time(NULL);
      as = cons(cl_make_symbol("subject-identification"), as);

/* cons(cl_make_string(ctime(&tt)), cons(the_current_user(), as)));*/

      add_assertion(as);

      p_name_df = 0;
      p_birth_df = 0;
      p_sex_df = 0;
      p_address_df = 0;
      p_phone_df = 0;
      p_id_df = 0;
      p_med_ref_df = 0;
    }
    break;
  case 2:
    p_name_df = 0;
    p_birth_df = 0;
    p_sex_df = 0;
    p_address_df = 0;
    p_phone_df = 0;
    p_id_df = 0;
    p_med_ref_df = 0;
    break;
    
  }
  wi_time_del_dwin(fulld);

}


/*--
  \end{code}

  \endoffile

  --*/

