
/*== ================================================================
 * \FileTitle      {glikintv.c---Interval superimposition and query package}

    Copyright (c) 1991--1996 Alexandru Dan Corlan MD (http://dan.corlan.net)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *
 *== ================================================================
 *
 * \FileContents
 */


/*-- ----------------------------------------------------------------
 * \end{code}
 *
 * \section                {glikintv.c} 
 *
 * \ImplementationNotes
 * \Portability
 * \Changes
 *      Generic functions and types would be much needed in the
 *      implementation, so that
 *      the interval-specification mechanism is available for any
 *      type of value of the parameter. The final library module
 *      for the \packagetag{glikintv} package thus needs to be
 *      written in Ada.
 *
 * \begin{code}
 */


#include "glikintv.h"
#include <malloc.h>


/*-- ----------------------------------------------------------------
 * \end{code}
 *
 * \funparagraph                {new\_fintvs, new\_iintvs}
 * \ImplementationNotes
 * \Portability
 *      \macrotag{sizeof} includes alignement in the result, so that
 *      a vector of n elements of a given type have n*sizeof(type) size.
 * \Changes
 *
 * \begin{code}
 */


FINTVS *new_fintvs(double defa)

{
  FINTVS *r;

  r = (FINTVS *)malloc(sizeof(FINTVS));
  if (r==NULL) {
    memory_error(1);
  }
  r->size = INIT_VS_SIZE;
  r->fill = 1;
  r->vv   = (FLENT *)malloc(INIT_VS_SIZE * sizeof(FLENT));
  if (r->vv==NULL) {
    memory_error(1);
  }
  r->vv[0].at = 0;
  r->vv[0].v  = defa;
  return r;
}



IINTVS *new_iintvs(long defa)

{
  IINTVS *r;

  r = (IINTVS *)malloc(sizeof(IINTVS));
  if (r==NULL) {
    memory_error(1);
  }
  r->size = INIT_VS_SIZE;
  r->fill = 1;
  r->vv   = (ILENT *)malloc(INIT_VS_SIZE * sizeof(ILENT));
  if (r->vv==NULL) {
    memory_error(1);
  }
  r->vv[0].at = 0;
  r->vv[0].v  = defa;
  return r;
}


/*-- ----------------------------------------------------------------
 * \end{code}
 *
 * \funparagraph                {ins\_i\_intv, ins\_f\_intv}
 * \ImplementationNotes
 * \Portability
 * \Changes
 *
 * \begin{code}
 */


void ins_i_intv(IINTVS *intvs, long from, long to, long val)

{
  short i, j, k;
  short targfill;
  ILENT *new_vv;
  long toval;

  /* find out how many data points will be left after the function call */
  
  targfill = intvs->fill + 2;  /* in the case that none is deleted */
  for(i=0; i<intvs->fill; i++)
    if ((intvs->vv[i].at >= from) && (intvs->vv[i].at <= to))
      targfill --;  /* this one will be deleted */

  /* Why <= to when to is actually exclusive? Because, if it is exactly
   * equal to to we don't need add an "end of interval" pair anymore,
   * so intvs->fill should not have been added 2
   */

  /* decide whether the vv should be enlarged, and do it if necessary */
  /* because of the modification algorithm below, in the worst case,
   * two pairs more may be (transiently) needed during the process */

  if (targfill+2>=intvs->size) {
    new_vv   = (ILENT *)malloc((intvs->size + VS_SIZE_INCREMENT)
			       * sizeof(ILENT));
    if (new_vv==NULL) {
      memory_error(1);
    }
    for(i=0; i<intvs->fill; i++) {
      new_vv[i].v = intvs->vv[i].v;
      new_vv[i].at = intvs->vv[i].at; /* let's hope no slots are added later */
    }
    free(intvs->vv);
    intvs->vv = new_vv;
  }

  /* now, we have a large enough pair vector */
  /* first, we insert |from| and |to|, by replacing already existent
   * values in the corresponding places if necessary */
  
  /* find the value to be given to the |to| point */

  toval = intvs->vv[0].v;  /* some initialisation */

  for(i=intvs->fill-1; i>=0; i--)
    if(intvs->vv[i].at <= to) {
      toval = intvs->vv[i].v;
      break;
    }

  for(i=0; i<intvs->fill; i++)       /* insert |from| */
    if(intvs->vv[i].at >= from) {
      if(intvs->vv[i].at == from)
	intvs->vv[i].v = val;
      else {
	for(j=intvs->fill; j>i; j--) {
	  intvs->vv[j].v = intvs->vv[j-1].v;
	  intvs->vv[j].at = intvs->vv[j-1].at;
	}
	intvs->vv[i].at = from;
	intvs->vv[i].v  = val;
	intvs->fill ++;
      }
      break;
    }
  if(i==intvs->fill) { /* i.e. all were smaller than from */
    intvs->vv[i].at = from;
    intvs->vv[i].v  = val;
    intvs->fill ++;
  }

  
  for(i=0; i<intvs->fill; i++)       /* insert |to| */
    if(intvs->vv[i].at >= to) {
      if(intvs->vv[i].at == to)
	/* do nothing, the previous continuation is preserved */ ;
      else {
	for(j=intvs->fill; j>i; j--) {
	  intvs->vv[j].v = intvs->vv[j-1].v;
	  intvs->vv[j].at = intvs->vv[j-1].at;
	}
	intvs->vv[i].at = to;
        intvs->vv[i].v  = toval;
	intvs->fill ++;
      }
      break;
    }
  if(i==intvs->fill) { /* i.e. all were smaller than |to| */
    intvs->vv[i].at = to;
    intvs->vv[i].v  = toval;
    intvs->fill ++;
  }

  /* and now delete everything between |from| and |to| excluding
   * the two */

  for(i=0; i<intvs->fill; i++)
    if(intvs->vv[i].at==from) {  /* it must be true at least once */
      for(j=i+1; j<intvs->fill; j++)
	if(intvs->vv[j].at==to && j>i+1) {
	  while(j<intvs->fill) {
	    intvs->vv[i+1].at = intvs->vv[j].at;
	    intvs->vv[i+1].v  = intvs->vv[j].v;
	    i++;
	    j++;
	  }
	  intvs->fill -= (j - i - 1);
	}
      break;
    }


  /* final adjustment: if there are two or more succesive
   * pairs with the same value, keep only one position-value pair
   * in place; apply repeatedly until no succesion if found any more
   */

  do {
    k = 0;
    for(i=0; i<intvs->fill-1; i++)
      if(intvs->vv[i].v == intvs->vv[i+1].v) {
	for(j=i+1; j<intvs->fill && (intvs->vv[j].v==intvs->vv[i].v); j++);
	k = j - i - 1;
	while(j<intvs->fill) {
	  intvs->vv[i+1].v = intvs->vv[j].v;
	  intvs->vv[i+1].at = intvs->vv[j].at;
	  i++;
	  j++;
	}
	intvs->fill -= k;
      }
  } while(k>0);
  return ;
  
}




void ins_f_intv(FINTVS *intvs, long from, long to, double val)

{
  short i, j, k;
  short targfill;
  FLENT *new_vv;
  double toval;

  /* find out how many data points will be left after the function call */
  
  targfill = intvs->fill + 2;  /* in the case that none is deleted */
  for(i=0; i<intvs->fill; i++)
    if ((intvs->vv[i].at >= from) && (intvs->vv[i].at <= to))
      targfill --;  /* this one will be deleted */

  /* Why <= to when to is actually exclusive? Because, if it is exactly
   * equal to to we don't need add an "end of interval" pair anymore,
   * so intvs->fill should not have been added 2
   */

  /* decide whether the vv should be enlarged, and do it if necessary */
  /* because of the modification algorithm below, in the worst case,
   * two pairs more may be (transiently) needed during the process */

  if (targfill+2>=intvs->size) {
    new_vv   = (ILENT *)malloc((intvs->size + VS_SIZE_INCREMENT)
			       * sizeof(ILENT));
    if (new_vv==NULL) {
      memory_error(1);
    }
    for(i=0; i<intvs->fill; i++) {
      new_vv[i].v = intvs->vv[i].v;
      new_vv[i].at = intvs->vv[i].at; /* let's hope no slots are added later */
    }
    free(intvs->vv);
    intvs->vv = new_vv;
  }

  /* now, we have a large enough pair vector */
  /* first, we insert |from| and |to|, by replacing already existent
   * values in the corresponding places if necessary */
  
  /* find the value to be given to the |to| point */

  toval = intvs->vv[0].v;  /* some initialisation */

  for(i=intvs->fill-1; i>=0; i--)
    if(intvs->vv[i].at <= to) {
      toval = intvs->vv[i].v;
      break;
    }

  for(i=0; i<intvs->fill; i++)       /* insert |from| */
    if(intvs->vv[i].at >= from) {
      if(intvs->vv[i].at == from)
	intvs->vv[i].v = val;
      else {
	for(j=intvs->fill; j>i; j--) {
	  intvs->vv[j].v = intvs->vv[j-1].v;
	  intvs->vv[j].at = intvs->vv[j-1].at;
	}
	intvs->vv[i].at = from;
	intvs->vv[i].v  = val;
	intvs->fill ++;
      }
      break;
    }
  if(i==intvs->fill) { /* i.e. all were smaller than from */
    intvs->vv[i].at = from;
    intvs->vv[i].v  = val;
    intvs->fill ++;
  }

  
  for(i=0; i<intvs->fill; i++)       /* insert |to| */
    if(intvs->vv[i].at >= to) {
      if(intvs->vv[i].at == to)
	/* do nothing, the previous continuation is preserved */ ;
      else {
	for(j=intvs->fill; j>i; j--) {
	  intvs->vv[j].v = intvs->vv[j-1].v;
	  intvs->vv[j].at = intvs->vv[j-1].at;
	}
	intvs->vv[i].at = to;
        intvs->vv[i].v  = toval;
	intvs->fill ++;
      }
      break;
    }
  if(i==intvs->fill) { /* i.e. all were smaller than |to| */
    intvs->vv[i].at = to;
    intvs->vv[i].v  = toval;
    intvs->fill ++;
  }

  /* and now delete everything between |from| and |to| excluding
   * the two */

  for(i=0; i<intvs->fill; i++)
    if(intvs->vv[i].at==from) {  /* it must be true at least once */
      for(j=i+1; j<intvs->fill; j++)
	if(intvs->vv[j].at==to && j>i+1) {
	  while(j<intvs->fill) {
	    intvs->vv[i+1].at = intvs->vv[j].at;
	    intvs->vv[i+1].v  = intvs->vv[j].v;
	    i++;
	    j++;
	  }
	  intvs->fill -= (j - i - 1);
	}
      break;
    }


  /* final adjustment: if there are two or more succesive
   * pairs with the same value, keep only one position-value pair
   * in place; apply repeatedly until no succesion if found any more
   */

  do {
    k = 0;
    for(i=0; i<intvs->fill-1; i++)
      if(intvs->vv[i].v == intvs->vv[i+1].v) {
	for(j=i+1; j<intvs->fill && (intvs->vv[j].v==intvs->vv[i].v); j++);
	k = j - i - 1;
	while(j<intvs->fill) {
	  intvs->vv[i+1].v = intvs->vv[j].v;
	  intvs->vv[i+1].at = intvs->vv[j].at;
	  i++;
	  j++;
	}
	intvs->fill -= k;
      }
  } while(k>0);
  return ;
  
}


#define TESTING
#ifdef TESTING


/*-- ----------------------------------------------------------------
 * \end{code}
 *
 * \subsection                {Test package} 
 *
 * \begin{code}
 */


#include <stdio.h>


/*-- ----------------------------------------------------------------
 * \end{code}
 *
 * \funparagraph                {main}
 * 
 * \PurposeReturn
 * \Exceptions
 * \GlobalEffects
 * \Preconditions
 * \Performance
 *
 * \begin{code}
 */


main()

{
  IINTVS *iin;
  char bf[100];
  char *s;
  long from, to, val;
  short i;

  iin = new_iintvs(999);
  while(1) {
    gets(bf);
    switch(bf[0]) {
    case 'a':        /* add an interval */
    case 'A':
      s = bf+1;
      from = atoi(s);
      while(*s==' ')
	s++;
      while(*s<='9' && *s>='0')
	s++;
      to = atoi(s);
      while(*s==' ')
	s++;
      while(*s<='9' && *s>='0')
	s++;
      val = atoi(s);
      ins_i_intv(iin, from, to, val);
      printf("\n");
      for(i=0; i<iin->fill; i++)
	printf("  [%d] at %4d -> %6d\n", i, iin->vv[i].at, iin->vv[i].v);
      break;
    default:
      printf("Error\n");
      break;
    }
  }
}
      
    

#endif
