/*--

  \FileTitle{
  {\tt glikrec.c} --- patient file structure management
  }
  \CopyrightNotice{

    Copyright (c) 1991--1996 Alexandru Dan Corlan MD (http://dan.corlan.net)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
  }
  
  \section{Gliks Recording}
  
  
  \begin{code}
  --*/

#include <stdio.h>
#include <malloc.h>

#include "gliks.h"
extern FILE *ecgnosislog;

/*--
  \end{code}

  \dataparagraph{Exception mechanism}

  \begin{code}
  --*/


short glikrecerr;
char glikerrbuf[100] = "No error message.";
char *glikerrmsg = glikerrbuf;


/*--
  \end{code}

  \subsection{Integrating signal analysis}

  \ImplementationNotes{

  The {\tt itab} and the {\tt stab} buffers
  from the {\tt PERFILE} data structures 
  are computed in two kinds of conditions: when their dimension is previously
  known (they have been computed before, the size is known, but memory 
  was meanwhile disallocated) and when their
  dimension is not known in advance (but a kind of maximum can usually be 
  estimated). The function that computes {\tt itab} ``from scratch'' is called
  {\tt icompute\_itab}. Functions that
  recalculate them once already known are {\tt recompute\_itab} and
  {\tt recompute\_stab}.
  
  When it is desired to compute the itab of the segment that starts at
  sample number S$_0$ and ends at sample S$_1$ the first thing to do is to 
  detect the fast complexes that have their reference point in this 
  region. After this first step is done, we already know how many complexes
  are there, where are their positions, and we can allocated itab.
  
  In order to determine the position of a complex, a piece of recording must
  be known ``before'' the complex (so that the detection algorithm can
  see the raising in speed). This part of the recording is called
  ``the preamble'' and it's size (normally 1 second) is established 
  (in ccs) by the {\tt PREAMBLE\_SIZE} symbol below.
  
  Some algorithms require that the end of the
  fast complex is identified to say that a given piece of record is a complex,
  so after the last identified QRS we need some region of the recording again. 
  This part of the recording is called ``the postamble''
  and it's size (normally
  1 second also) is given (in ccs) by the {\tt POSTAMBLE\_SIZE} symbol.

  Thus, in order to determine the {\tt itab} of a segment we need all the
  data of that segment, plus the preamble and the postamble. 
  Actually, for the first and for the last of the segments there is no 
  preamble and respectivelly, no postamble. In records which have 
  internal holes (not treated for the moment by this program --- but considered
  as an extension in the close future) it is possible for many other
  segments not to have preamble or postamble or to have a partial one.
  The {\tt act\_pre\_size} and 
  the {\tt act\_post\_size} variables keep the number
  of samples actually in the preamble and postamble.
  
  After the positions of the fast complexes are determined, identification of
  the ``reference points'', class and various others features also may involve
  the preamble and the postamble.
  
  The buffer used for computing an {\tt itab} or an {\tt stab} is usually
  called a ``glead'', and the {\tt itab} computed before it's size is
  preciselly known is called a ``gitab''. Pleas note that {\tt gitab} only
  records the {\tt pl} slot
  of the {\tt STEPONE} struct, as after this one becomes 
  known the corresponding structures in the segment can be correctly filled.
  
  The most frequent manner in
  which the {\tt icompute} and the {\tt recompute}
  functions are called is ``in sequence''. Usually, at the beginning of the 
  analysis, the functions are called once for each segment in the recording 
  and, afterwards, are called for clusters of segments. In order for the
  overlaping regions not to be read twice in such processings, the 
  {\tt getgbuf} function, which reads data into the buffer, keeps track
  of what is in it between calls, so it can reuse part of the buffer if
  indicated.
  
  Please note that there can be one buffer for each lead in the data structure
  but, at least the first implementation, chooses to ignore all the
  leads other than lead[0]. The number of the actual lead that is kept
  there is stored in {\tt act\_lead}.
  }
  \Portability{ % What is assumed about the operating environment
  }
  \Changes{ % that are anticipated, if special provision was made for them
  }

  \dataparagraph{Analysis results}  % a name for the data structure

  \begin{code}
  --*/

PERFILE *lst_perfile = NULL;    /* [VAR] */ 

long  *gitab = NULL;           /* in samples from glead_origin ! */ /* [VAR] */
short *gst1tab = NULL;         /* in record units from the baseline */
short *gst2tab = NULL;         /* idem, for the ``other'' strip */
short *gjtab = NULL;           /* J~point position from the qrs detect place */
short *gbl1tab = NULL;         /* baseline position */
short *gbl2tab = NULL;         /* baseline position */
short *gqrsltab = NULL;        /* QRS width */
short *gmtab = NULL;           /* morphology code */

long gitab_als = 0;             /* [VAR] */ 
long gitab_fill = 0;  /* [VAR] */ 



/*--
  \end{code}

  \dataparagraph{Data for analysis}

  \begin{code}
  --*/


#define PREAMBLE_SIZE 10000    /* 1 second */ /* [DEF] */ 
#define POSTAMBLE_SIZE 10000   /* 1 second */ /* [DEF] */ 

short *oth_glead = NULL;                                        /* [VAR] */ 
short act_oth_glead = 0;                                        /* [VAR] */ 

short gleads_are_sync = 0;                                      /* [VAR] */ 

short *glead = NULL; /* [VAR] */ 
short act_lead = 0; /* [VAR] */ 
long glead_als = 0;            /* size in samples allocated to the glead */ /* [VAR] */ 
long glead_origin;             /* in samples (position of glead[act_pre_size] !) */ /* [VAR] */ 
long glead_fill;               /* size in samples of the actual data block */ /* [VAR] */ 
 /* not including the act_pre_size and the act_post_size */
long act_pre_size = 0; /* [VAR] */ 
long act_post_size = 0; /* [VAR] */ 



/*--
  \end{code}

  \funparagraph{ensure\_glead\_has}

  \PurposeReturn{
  The following function allocates glead to the size required for a given 
  task. It glead is already large enough, it's not allocated again. 

  This function
  returns 1 if job is done and 0 (also seting {\tt glikrecerr}) if it
  couldn't be done --- because of lack of memory.
  }

  \begin{code}
  --*/

short ensure_glead_has(long k) /* [FUN] */ 

{
  if(glead_als>=k)
    return 1;
  if(glead!=NULL) {
    free(glead);
    glead = NULL;
    if(oth_glead!=NULL)
      free(oth_glead);
    oth_glead = NULL;
  }
  glead = (short *)malloc(sizeof(short)*k);
  oth_glead = (short *)malloc(sizeof(short)*k);
  if(glead==NULL || oth_glead==NULL)
    {
      glikrecerr = NO_MEMORY;
      glead_als = 0;
      if(glead!=NULL) {
        free(glead);
        glead = NULL;
      }
      if(oth_glead!=NULL) {
        free(oth_glead);
        oth_glead = NULL;
      }
      return 0;
    }
  glead_als = k;
  return 1;
}


/*--
  \end{code}

  \funparagraph{glead\_require}

  \PurposeReturn{
  The following function is given a {\tt PERFILE} structure and a region to be
  from a given lead of the {\tt PERFILE} it reallocated (if necessary) {\tt glead}
  and reads the required information --- reusing what it can from the old
  contents of the buffer. 

  The org and len are the requirements to be read in the {\tt glead} buffer,
  and are given in samples. Samples are used everywhere so that ccs to sample 
  convertion is done at the higher level in a localized region (it might imply, 
  for some of the Holter recorders, special rounding issues).
  The {\tt nof\_leads} argument is 0 if only the required lead is to be
  read from the disk, and 1 or larger if all the leads are to be read.
  
  The following function returns 0 on failure and 1 on success. 
  }
  \Exceptions{
  On failure
  (including attempted read past end of file) the {\tt glikrecerr} is set
  to some value.
  }

  \begin{code}
  --*/

short glead_require(PERFILE *p, short ld,
                    long org, long len,
                    short nof_leads)                            /* [FUN] */ 

{
  long i, delta, rv;
  long shift, sz;
  short oth_ld;

#ifdef RECTRACE
  fprintf(ecgnosislog, "CALL glead_require(,%d,%ld,%ld)\n", ld, org, len);
  fflush(ecgnosislog);
#endif

  /* compute the other lead --- for more than two leads, a different
     approach must be employed */

  if(ld==0)
    oth_ld = 1;
  else
    oth_ld = 0;

  /* first make sure that glead actually exists */
  if(ensure_glead_has(len)==0)
    return 0;

  if(act_pre_size!=-1 && act_post_size !=-1
     && glead_fill!=-1 && glead_origin!=-1) {
    
/**** CASE 0:

The segment that is required is in the same position as the
one already in the buffer. Only the segment in {\tt glead} is actually
required, or all of the segments are required but they
are also already present in their buffers (i.e. {\tt gleads\_are_syn}
is set to 1).

*/

    if(ld==act_lead && glead_origin - act_pre_size == org &&
       glead_fill + act_pre_size + act_post_size == len) {
      if(nof_leads==0)
        return 1;        /* the segment was already there */
      if(nof_leads!=0) {  /* all the other segments must be read */
        if(gleads_are_sync)
          return 1;      /* they are already there */
        rv = ecg_read(p->e, oth_glead, oth_ld, len);
        /* validity of the arguments in ecg_read above is questionable */
        if(rv<delta) {
          glikrecerr = READ_PAST_EOF;
          return 0;
        }
        gleads_are_sync = 1;
        return 1;
      }
    }

    if(ld==oth_ld && glead_origin - act_pre_size == org &&
       glead_fill + act_pre_size + act_post_size == len && gleads_are_sync) {
      for(i=0; i<len; i++) {
        shift = glead[i];
        glead[i] = oth_glead[i];
        oth_glead[i] = shift;
      }
    }
  
    
    
    /* if the end of the old buffer will become the beginning
     * of the new buffer: case 2 --- the most frequent */
    if(p==lst_perfile && ld==act_lead && org > glead_origin - act_pre_size &&
       org+len > glead_origin + glead_fill + act_post_size) {
#ifdef RECTRACE
      fprintf(ecgnosislog, "     branch ``case 2'' --- read next.\n");
      fflush(ecgnosislog);
#endif
      delta = org + len - glead_origin - glead_fill - act_post_size;
      shift = org + act_pre_size - glead_origin;
      sz = act_pre_size + glead_fill + act_post_size - shift;
      for(i=0; i<sz; i++)
        glead[i] = glead[i+shift];
      if(nof_leads!=0 && gleads_are_sync)
        for(i=0; i<sz; i++)
          oth_glead[i] = oth_glead[i+shift];
      if(ecg_seek(p->e, org)==0) {
        glikrecerr = BAD_SEEK;
        p = NULL;   /* buffer is corrupted */
        return 0;
      }
      rv = ecg_read(p->e, glead+shift, ld, delta);
      if(rv<delta) {
        glikrecerr = READ_PAST_EOF;
        p = NULL;
        return 0;
      }
      if(nof_leads!=0 && gleads_are_sync) {
        if(ecg_seek(p->e, org)==0) {  /* shouldn't it be org+shift? */
          glikrecerr = BAD_SEEK;
          p = NULL;   /* buffer is corrupted */
          return 0;
        }
        rv = ecg_read(p->e, oth_glead+shift, oth_ld, delta);
        if(rv<delta) {
          glikrecerr = READ_PAST_EOF;
          p = NULL;
          return 0;
        }
      }
      if(nof_leads!=0 && !gleads_are_sync) {
        /* must read the other lead completely */
        if(ecg_seek(p->e, org)==0) {
          glikrecerr = BAD_SEEK;
          p = NULL;   /* ? */
          return 0;
        }
        rv = ecg_read(p->e, oth_glead, oth_ld, len);
        if(rv<delta) {
          glikrecerr = READ_PAST_EOF;
          p = NULL;
          return 0;
        }
      }
      glead_origin = -1;
      act_pre_size = -1;
      act_post_size = -1;   /* they aren't known anymore ? */
      return 1;  /* but data was successfuly read */
    }
      

    /* if the end of the new buffer overlaps with the beginning
     * of the old one --- the less likely situation: case 1 */
    if(p==lst_perfile && ld==act_lead && org < glead_origin - act_pre_size && 
         org+len > glead_origin - act_pre_size &&
         org+len < glead_origin + glead_fill + act_post_size) {

/***** MULTIPLE LEAD PROGRAM NOT INSERTED HERE BECAUSE THIS IS NEVER CALLED
 *****/


#ifdef RECTRACE
  fprintf(ecgnosislog, "     branch ``case 1'' --- read back.\n");
  fflush(ecgnosislog);
#endif
      delta = glead_origin - act_pre_size - org;
      for(i=len-1; i>=delta; i--)
        glead[i] = glead[i-delta];
      if(ecg_seek(p->e, org)==0) {
        glikrecerr = BAD_SEEK;
        p = NULL;   /* buffer is corrupted */
        return 0;
      }
        rv = ecg_read(p->e, glead, ld, delta);
/*      if(ld==0)
      else
        rv = delta;*/
      if(rv<delta) {
        glikrecerr = READ_PAST_EOF;
        p = NULL;
        return 0;
      }
      glead_origin = -1;
      act_pre_size = -1;
      act_post_size = -1;   /* they aren't known anymore */
      return 1;  /* but data was successfuly read */
    }
  }

/* otherwise, just read the data; glead is large enough */
#ifdef RECTRACE
  fprintf(ecgnosislog, "     branch ``case 3'' --- read random.\n");
  fflush(ecgnosislog);
#endif
  if(ecg_seek(p->e, org)==0) {
    glikrecerr = BAD_SEEK;
    p = NULL;   /* NO THIS HAS NO SENSE, REWRITE AND RESPECIFY */
    return 0;
  }
  if(nof_leads==0)
    rv = ecg_read(p->e, glead, ld, len);
  else
    rv = ecg_read2(p->e, glead, ld, oth_glead, oth_ld, len);
  if(rv<len) {
    glikrecerr = READ_PAST_EOF;
    p = NULL;
    return 0;
  }
  if(nof_leads!=0)
    gleads_are_sync = 1;
  glead_origin = -1;
  act_lead = ld;
  act_pre_size = -1;
  act_post_size = -1;   /* they aren't known anymore */
  return 1;  /* but data was successfuly read */
}


/*--
  \end{code}

  \funparagraph{seg\_require}
  % The name of the function is expected to be prezent
  % in the sectioning element above

  \PurposeReturn{
  The following function is given a segment start and a segment end and
  tries to read in {\tt glead} the contents of the segment, with a required
  (or possible) preamble and postamble, and also to set the variables
  accordingly. 

  All durations are in samples. 

  The following function returns 0 on failure and 1 on success. 
  }
  \Exceptions{
  }
  \Performance{
  While {\tt ecg\_size} is called twice the second call
  is rare enough (for large files) to make a local variable for
  the size of the ecg file unnecessary.
  }

  \begin{code}
  --*/

short seg_require(PERFILE *p,                                   /* [FUN] */ 
                  long segorg, long segend,                     /* [FUN] */ 
                  short ld, short nof_leads)                    /* [FUN] */ 

{
  long pre, post;

  pre = PREAMBLE_SIZE / ecg_sample_size(p->e);
  post = POSTAMBLE_SIZE / ecg_sample_size(p->e);
  if (pre > segorg)
    pre = segorg;
  if(post > ecg_size(p->e) - segend)
    post = ecg_size(p->e) - segend;
#ifdef RECTRACE
  fprintf(ecgnosislog, "CALL seg_require segorg=%ld segend=%ld pre=%ld post=%ld\n",
          segorg, segend, pre, post);
  fflush(ecgnosislog);
#endif
  if(glead_require(p, ld, segorg - pre, segend - segorg + pre + post,
     nof_leads)) {
    act_pre_size = pre;
    act_post_size = post;
    glead_origin = segorg;
    glead_fill = segend - segorg + pre + post;
    return 1;
  }
  else {
#ifdef RECTRACE
  fprintf(ecgnosislog, "     glead_require returned in error, with %d\n", glikrecerr);
  fflush(ecgnosislog);
#endif TRACE
    return 0;
}
}



/*--
  \end{code}

  \funparagraph{icompute\_itab}

  \PurposeReturn{
  The following is the local function for allocating and computing the 
  {\tt itab} of one segment from scratch. Please note that it's arguments are
  in samples, and everything during the process is in samples.

  The {\tt icompute\_itab} function returns -1 if error and
  a positive integer to hold the
  number of {\tt STEPONE} entries needed for the segment if the
  function is sucessfull.
  }
  \GlobalEffects{
  This function allocates itab if necessary. 
  }
  \Performance{
  As {\tt qrs\_detect\_init} is rather expensive to run a strategy might
  later be implemented to call it just a few times for each file.
  }

  \ImplementationNotes{
  This function does not know
  the method used for QRS detection. Instead, it uses two {\tt glikalys}
  functions: {\tt next\_qrs} and {\tt min\_complex\_size}. On the other hand,
  these functions may use the {\tt parameter\_value} from the {\tt glikrec}
  package and also the {\tt ecg\_sample\_size} function from the 
  {\tt glikcess} package. There can also be a function for ``automatic detection
  of parameters'' using both {\tt glikrec} and {\tt glikalys} functions,
  which will be probably located in the {\tt glikalys} package.
  
  The value returned by {\tt min\_complex\_size} is the length, in samples, in
  which no QRS is assumed to be found following a given QRS; thus, it
  can be used for computing the longest possible size of the {\tt itab}, which
  is the allocated size for {\tt gitab}.
  
  The specification of the {\tt next\_qrs} function, as assumed by {\tt icompute\_itab}
  is the following: the function takes as argument a {\tt PERFILE} pointer, 
  a pointer to a vector of 
  shorts (which are samples), the size of the vector, and a flag which is 
  -1 if the threshold to use is unknown, 0 if the one set by
  the previous call to {\tt next\_qrs} or {\tt qrs\_detect\_init}
  is to be used and a threshold
  value (a short integer) if the threshold is given by the caller.

  It returns -1 if no fast complex is found to the end of the vector of samples
  and the number of samples up to the ``detection point'' of the next
  qrs if one was found.
  
  The {\tt qrs\_detect\_init} function is called with the whole buffer
  before finding qrs's in the buffer, so that threshold and other
  parameters can be initialised by the qrs detection module.
  }

  \begin{code}
  --*/


short icompute_itab(PERFILE *p, long segstart, long segend, short ld) /* [FUN] 
*/ 

{
  long gitab_maxlen;
  long gli;   /* index in glead */
  long stril;
  long gloff;

  if(!seg_require(p, segstart / ecg_sample_size(p->e), 
      segend / ecg_sample_size(p->e), ld, 2))
    return -1;

#ifdef RECTRACE
  fprintf(ecgnosislog, "CALL icompute_itab(,%ld,%ld,%d)\n", segstart, segend, ld);
  fprintf(ecgnosislog, "     segment successfully read.\n");
  fflush(ecgnosislog);
#endif

  gitab_maxlen = 3 + (segend - segstart) / min_complex_size(p);

  /* ensure now that gitab exists and is of the desired size */
  /* change: in parralel with gitab we use other vectors for the storage */
  /*         of the st levels (on the two derivations), the baseline
  /*         the position of the J~point */

  if(gitab==NULL || gitab_als < gitab_maxlen) {
    if(gitab!=NULL)
      free(gitab);
    if(gst1tab!=NULL)
      free(gst1tab);
    if(gst2tab!=NULL)
      free(gst2tab);
    if(gjtab!=NULL)
      free(gjtab);
    if(gqrsltab!=NULL)
      free(gqrsltab);
    if(gbl1tab!=NULL)
      free(gbl1tab);
    if(gbl2tab!=NULL)
      free(gbl2tab);
    if(gmtab!=NULL)
      free(gmtab);

    gitab   = (long *) malloc(gitab_maxlen * sizeof(long));
    gst1tab = (short *)malloc(gitab_maxlen * sizeof(short));
    gst2tab = (short *)malloc(gitab_maxlen * sizeof(short));
    gjtab   = (short *)malloc(gitab_maxlen * sizeof(short));
    gqrsltab   = (short *)malloc(gitab_maxlen * sizeof(short));
    gbl1tab = (short *)malloc(gitab_maxlen * sizeof(short));
    gbl2tab = (short *)malloc(gitab_maxlen * sizeof(short));
    gmtab   = (short *)malloc(gitab_maxlen * sizeof(short));

    if(gitab==NULL || gst1tab==NULL || gst2tab==NULL || gjtab==NULL
       || gbl1tab==NULL || gbl2tab==NULL || gmtab==NULL || gqrsltab==NULL) {
      if(gitab!=NULL)
        free(gitab);
      gitab = NULL;

      if(gst1tab!=NULL)
        free(gst1tab);
      gst1tab = NULL;

      if(gst2tab!=NULL)
        free(gst2tab);
      gst2tab = NULL;

      if(gjtab!=NULL)
        free(gjtab);
      gjtab = NULL;
      
      if(gbl1tab!=NULL)
        free(gbl1tab);
      gbl1tab = NULL;
      
      if(gbl2tab!=NULL)
        free(gbl2tab);
      gbl2tab = NULL;

      if(gmtab!=NULL)
	free(gmtab);
      gmtab = NULL;

      if(gqrsltab!=NULL)
	free(gqrsltab);
      gqrsltab = NULL;
      
      glikrecerr = NO_MEMORY;
      gitab_als = 0;
      return -1;  /* Can't do the trick, not enough memory */
    }
    gitab_als = gitab_maxlen;
  }

  gitab_fill = 0;
  gli = 0;
  stril = glead_fill; /* see later for possible bugs */

#ifdef RECTRACE
  fprintf(ecgnosislog, "     gitab ensured, of %ld long items, stril=%ld.\n",
     gitab_als, stril);
  fflush(ecgnosislog);
#endif

  qrs_detect_init(p, glead, stril);

#ifdef RECTRACE
  fprintf(ecgnosislog, "     QRS detection initiated.\n");
  fflush(ecgnosislog);
#endif

  gloff = next_qrs(p, glead, stril, 0);

#ifdef RECTRACE
  fprintf(ecgnosislog, "     First QRS detected after %ld samples.\n", gloff);
  fflush(ecgnosislog);
#endif

  if(gloff==-1)  /* incredible: no fast complex in the segment! */
    return 0;    /* it is a success anyway */
  gli = gloff-8;
  /* this correction (8) is due to the detection algorithm */
  if(gli >= act_pre_size && gli < act_pre_size + glead_fill) {
    gitab[gitab_fill] = gli - act_pre_size;
    gbl1tab[gitab_fill] = preceding_baseline(glead, stril, gli,
                                            ecg_sample_size(p->e));
    gbl2tab[gitab_fill] = preceding_baseline(oth_glead, stril, gli,
                                            ecg_sample_size(p->e));
    gjtab[gitab_fill] = locate_j_point(glead, stril, gli,
                                       ecg_sample_size(p->e));
    gqrsltab[gitab_fill] = last_qrs_len();
    gst1tab[gitab_fill] = glead[gli+gjtab[gitab_fill]+p->st_detect_offs] -
                                       gbl1tab[gitab_fill];
    gst2tab[gitab_fill] = glead[gli+gjtab[gitab_fill]+p->st_detect_offs] -
                                       gbl2tab[gitab_fill];
    gmtab[gitab_fill] = morphology(p, glead, stril, gli, 150000);
    gitab_fill ++;
  }
/*  else
    return gitab_fill; */
  while(gli<stril) {
    gli += 40;
    gloff = next_qrs(p, glead+gli, stril-gli, 0);
#ifdef RECTRACE
  fprintf(ecgnosislog, "     Next QRS detected after %ld samples.\n", gli);
  fflush(ecgnosislog);
#endif

    if(gloff==-1)
      return gitab_fill;
    gli += gloff;
    gli -= 8;
    /* correction due to the detection algorithm */
    if(gli >= act_pre_size && gli < stril - act_post_size) {
      gbl1tab[gitab_fill] = preceding_baseline(glead, stril, gli,
                                               ecg_sample_size(p->e));
      gbl2tab[gitab_fill] = preceding_baseline(oth_glead, stril, gli,
                                               ecg_sample_size(p->e));
      gjtab[gitab_fill] = locate_j_point(glead, stril, gli,
                                         ecg_sample_size(p->e));
      gqrsltab[gitab_fill] = last_qrs_len();
      gst1tab[gitab_fill] = glead[gli+gjtab[gitab_fill]+p->st_detect_offs] -
        gbl1tab[gitab_fill];
      gst2tab[gitab_fill] = glead[gli+gjtab[gitab_fill]+p->st_detect_offs] -
        gbl2tab[gitab_fill];
      gitab[gitab_fill] = gli - act_pre_size;
      gmtab[gitab_fill] = morphology(p, glead, stril, gli, 150000);
      gitab_fill ++;
    }
    if(gli >= stril - act_post_size) {
#ifdef RECTRACE
  fprintf(ecgnosislog, "     Returning %ld.\n", gitab_fill);
  fflush(ecgnosislog);
#endif
      return gitab_fill;  /* we are already beyond end of the stripe */
    }
  }
/* ??? Where should this end ??? */
}


/*--
  \end{code}

  \subsection{Segment Operations}

  \funparagraph{ss\_inq}

  \begin{code}
  --*/

short ss_inq(PERFILE *p, short s) /* [FUN] */ 

{
  /* check for the ``null'' state */
  if(p->sgs[s]->itabn==-1) {
    /* verify here it is OK */
    return SS_NULL;
  }
  
  /* check for the ``full'' state */
  if(p->sgs[s]->stab!=NULL) {
    /* verify */
    return SS_FULL;
  }

  /* check for the ``morph'' state */
  if(p->sgs[s]->itab!=NULL && p->sgs[s]->itab[0].morph!=-1) {
    /* verify */
    return SS_MORPH;
  }

  /* check for the ``sketch'' state */
  if(p->sgs[s]->itab!=NULL && p->sgs[s]->itab[0].morph==-1) {
    /* don't forget to do some checks here */
    return SS_SKETCH;
  }

  /* take the risk to assume it is in the ``point'' state */
  return SS_POINT;
}



/*--
  \end{code}

  \funparagraph{ss\_chg}
  \ImplementationNotes{
  When the {\tt ss\_chg} function changes the state of a segment
  to a higher level (with more information) it does two things:
  
  \begin{enumerate}
  \item it verifies that all the previous segments are at least in the
  {\tt SS\_POINT} state --- usually by calling itself recursivelly;
  \item it allocates the memory needed to store the new information
  (QRS positions, complex classes, various measures) and calls the appropriate
  functions to fill these values.
  \end{enumerate}
  
  When it is called to change the state to a lower level, it just desallocates
  the buffers refered by the segment.
  }
  \Changes{
  Please consider reducing the number of calls to {\tt ecg\_sample\_size} 
  (maybe
  by adding a new field in the {\tt PERFILE} data structure).
  }

  \begin{code}
  --*/

short ss_chg(PERFILE *p, short s, short state) /* [FUN] */ 

{
  short ss, i, j;
  long morpos;

  ss = ss_inq(p, s);
  if(ss==state)
    return;

  if(state>ss) { /* we are going up --- must allocate new information */
    if(state==SS_POINT) { /* it can't be NULL when we go up */
      if(s==0) {
        p->sgs[s]->itabn = 0;  /* it starts the whole thing */
        return 1;
      }
      if(p->sgs[s-1]->itabn>=0) { /* the previous exists */
        if(p->sgs[s-1]->itab!=NULL) {
          p->sgs[s]->itabn = p->sgs[s-1]->itabn + p->sgs[s-1]->itabfill;
          return 1;
        }
        else {
          /* but must be changed to make analysis faster, for example
             to avoid continuing to do the terminal phase analysis */

          j = icompute_itab(p, p->sgs[s-1]->segbegin,
                            p->sgs[s-1]->segend, p->crt_qrs_lead);
          /* HERE: a different lead strategy */
          if(j==-1)
            return 0;
          p->sgs[s]->itabn = p->sgs[s-1]->itabn + j;
        }
      }
      else {  

/* now, take them in ascending order to prevent too many recursive calls */
/* this is a costly procedure; the calling functions (of ss_chg) must take */
/* care to avoid going through it frequently */

        for(i=0; i<=s; i++)
          if(ss_inq(p, i)<SS_POINT)
            if(!ss_chg(p, i, SS_POINT))
              return 0;
        return 1;
      }
    }
    if(state==SS_SKETCH) {
      if(ss==SS_NULL)
        if(!ss_chg(p, s, SS_POINT))
          return 0;
      if(p->assqrsf!=NULL) {
        read_qrs_data(p, s);
        return 1;
      }
      j = icompute_itab(p, p->sgs[s]->segbegin, p->sgs[s]->segend, 
           p->crt_qrs_lead);
      /* HERE: a different lead strategy */
      if(j==-1)
        return 0;

      /* allocate the itab segment */
      p->sgs[s]->itab = (STEPONE *)malloc(j * sizeof(STEPONE));
      if(p->sgs[s]->itab==NULL)
        return 0;

      /* initialise fields in the itab segment with gitab ones */
      for(i=0; i<j; i++) {
        p->sgs[s]->itab[i].pl = p->sgs[s]->segbegin + gitab[i] * 
          ecg_sample_size(p->e);
        p->sgs[s]->itab[i].st1 = gst1tab[i];
        p->sgs[s]->itab[i].st2 = gst2tab[i];
        p->sgs[s]->itab[i].j = gjtab[i] * ecg_sample_size(p->e);
	p->sgs[s]->itab[i].qrsl = gqrsltab[i] * ecg_sample_size(p->e);
        p->sgs[s]->itab[i].bl1 = gbl1tab[i];
        p->sgs[s]->itab[i].bl2 = gbl2tab[i];
        p->sgs[s]->itab[i].reff = -8;
        p->sgs[s]->itab[i].morph = gmtab[i];
      }

      /* adjust segment entry fields */
      p->sgs[s]->itabsz = p->sgs[s]->itabfill = j;

      /* adjust PERFILE memory counters */
      p->totitabsz += j;
      p->itabn ++;
      return 1;
    }
    if(state==SS_MORPH) {
      if(ss_chg(p, s, SS_SKETCH)==0)
        return 0;
      
      /* we have the {\tt itab} allocated and detection positions filled */
      /* make sure that the lead information is also in memory */
      
      if(!seg_require(p, p->sgs[s]->segbegin / ecg_sample_size(p->e),
                      p->sgs[s]->segend / ecg_sample_size(p->e), 0, 2));

      /* take care about the lead 0 --- must be replaced with the
       * ``morphology identification lead'' later */

/*      
      for(i=0; i<p->sgs[s]->itabfill; i++) {
        morpos = (p->sgs[s]->itab[i].pl - p->sgs[s]->segbegin) /
          ecg_sample_size(p->e) + act_pre_size;
        p->sgs[s]->itab[i].morph =
          gmclass(p, glead, glead_fill, morpos, morpos+40, morpos-40);
	  } */
      
      return 1;      
    }
    if(state==SS_FULL) {
      return 0;
    }  /* not implemented yet */
  }
  else { /* necessarily state < ss (old state) */
    switch(state) {
    case SS_FULL: /* can't be! */
      break;
    case SS_MORPH: /* can't be! */
      break;
    case SS_SKETCH: /* can't be yet, as long as stab not implemented */
      break;
    case SS_POINT:
    case SS_NULL:

/* At a second thought, lead information will not be disallocated by
   this function any more, as it does not have any connection with
   the segment status.

      for(j=0; j<MAX_LEADS; j++)
        if(p->sgs[s]->lead[j]!=NULL) {
          free(p->sgs[s]->lead[j]);
          p->sgs[s]->lead[j] = NULL;
          p->sgs[s]->leadsz[j] = -1;
        } */
      if(p->sgs[s]->itab!=NULL) {
        free(p->sgs[s]->itab);
        p->sgs[s]->itab = NULL;
        p->totitabsz -= p->sgs[s]->itabsz;
        p->itabn --;
        p->sgs[s]->itabsz = p->sgs[s]->itabfill = -1;
      }
 /* MUST CONSIDER FLUSHING TO DISK HERE! */
/*      if(p->sgs[s]->etab!=NULL) { 
        free(p->sgs[s]->etab);
        p->sgs[s]->etab = NULL;
        p->totetabsz -= p->sgs[s]->etabsz;
        p->etabn --;
        p->sgs[s]->etabsz = p->sgs[s]->etabfill = -1;
      } */
      if(p->sgs[s]->stab!=NULL) {
        free(p->sgs[s]->stab);
        p->sgs[s]->stab = NULL;
        p->totstabsz -= p->sgs[s]->stabsz;
        p->stabn --;
        p->sgs[s]->stabsz = p->sgs[s]->stabfill = -1;
      }
      if(state==SS_NULL)
        p->sgs[s]->itabn = 0;
      break;
    default:  /* this can't exist! */
      break;  /* should be internal error! */
    }
  }
}


/*--
  \end{code}

  \funparagraph{segat}

  \begin{code}
  --*/

long segat(PERFILE *p, long to) /* [FUN] */ 

{
  long segto;

  /* an euristic search for the segment in which {\tt to} occurs: */
  /* every segment is assumed to be of the size of the first segment. */

  if(p->sgs[0]->segend <= p->sgs[0]->segbegin)
    segto = 0;
  else
    segto = to / (p->sgs[0]->segend - p->sgs[0]->segbegin);
  if(p->sgs[segto]->segbegin > to)
    while(segto > 0 && p->sgs[segto]->segbegin > to)
      segto --;
  if(p->sgs[segto]->segend < to)
    while(segto < p->sgsfill && p->sgs[segto]->segend < to)
      segto ++;

  /* the very unlikely situation that the segment could not be found */
  if(segto==0 && p->sgs[segto]->segbegin > to ||
     segto >= p->sgsfill) {
    glikrecerr = BAD_ARGUMENT;
    return -1;
  }
  return segto;
}


/*--
  \end{code}

  \subsection{Memory Management}

  \ImplementationNotes{
  The implementation strokes a balance between two subgoals:

  \begin{enumerate}
  \item do not use too much memory (no more than a given upper limit);
  \item make the metrics required through the {\tt glikeval} primitives
  avaiable with as little processing as possible (considered in both disk
  accesses and proessing of the raw data involved).
  \end{enumerate}
  
  The following principles were followed in the design of the
  memory management system:
  
  \begin{enumerate}
  \item Memory management strategy is confined to a single file.
  \item The limit set by {\tt set\_total\_size} is orientative. This means
  that during processing the system might allocate a little more
  (usually not exceeding with more than 64K).
  \item If a requirement (such as {\tt itabreq}) is done, any 
  previous requirement of the same type becomes ineffective;
  \item The last of the memory requirements done in step~2 takes
  precedence (may result in freeing memory due to) previous requirements
  [of a different type].
  \item When a requirement for memory is done, it must be possible to 
  resolve it within the limit fixed by {\tt set\_total\_size}, otherwise
  it is IGNORED.
  \end{enumerate}
  
  When a memory requirement is done, the necessary segments are constructed
  in memory from the disk using the {\tt ss\_chg} call. Before each of the calls
  to {\tt ss\_chg} the implementation tries to see if it doesn't overflow
  the memory allocated to it. If it does, it frees the first segment outside
  the last of the requirements of each type. If this can't be done, it
  frees the first segment outside the current requirement. When a segment 
  is freed, all the information it contains is freed --- stab, itab, etab
  and leads --- except the {\tt itabn} variable which contains the count of the
  first complex in the segment. This is accomplished using the {\tt ss\_chg}
  function call.
  
  When a {\tt glikeval} call is done outside the current requirement an implicit
  requirement is considered to be made to the segment[s] affected by that
  call. The processing described in the previous paragraph takes place.
  
  More than one segment can be implicated in a {\tt glikeval}
  call in two conditions:  (1) when it involves a complex or a moment
  inside a complex near a segment boundary --- because calls usually go
  to and fro a few segments; (2) when it involves resampling a long
  piece of a lead, larger than a segment (which is rarely a case if ever).
  }
  \Portability{ % What is assumed about the operating environment
  }
  \Changes{ % that are anticipated, if special provision was made for them
  If processings involving {\tt glikeval} become very ``jumpy'' in future 
  applications, a different strategy of freing segments might be involved:
  from the freeable segments one at random is choosen and freed. It is
  proven that random freeing segments is as good as any other general
  algorithm (but may be better than one-end freeing).

  The pieces of table involving {\tt etab} will be allocated and freed solidary
  with the {\tt itab} tables.
  }

  \funparagraph{just\_free}

  \begin{code}
  --*/

void just_free(PERFILE *p, short s) /* [FUN] */ 

{
  short i;

  if(ss_inq(p, s)>SS_POINT) /* for otherwise it would do some allocation! */
    ss_chg(p, s, SS_POINT);
  for(i=0; i<MAX_LEADS; i++)
    if(p->sgs[s]->lead[i]!=NULL) {
      p->totleadsz -= p->sgs[s]->leadsz[i];
      p->leadn --;
      free(p->sgs[s]->lead[i]);
      p->sgs[s]->lead[i] = NULL;
      p->sgs[s]->leadsz[i] = 0;
    }
}


/*--
  \end{code}

  \funparagraph{free\_one\_segment}
  \ImplementationNotes{
  There are two levels of segment
  freeing: outside any of the types of requirement, outside the last required
  region of the last requirement call. The two levels are used in
  sequence.
  }

  \begin{code}
  --*/

short free_one_segment(PERFILE *p) /* [FUN] */ 

{  /* A very inefficient, strightforward procedure! 
    * One might consider adjusting this
    * if it proves to require significant time from the total program execution
    */
  
  short i, j;

/* level 1 */

  for(i=0; i<p->sgsfill; i++) {
    if(p->lintrs_on!=-1)    /* if some lead domain was allocated */
      if(i>=p->lintrs_on && i<p->lintrs_off) continue;
    if(p->iintrs_on!=-1)       /* if some itab domain was allocated */
      if(i>=p->iintrs_on && i<p->iintrs_off) continue;
    if(p->sintrs_on!=-1)    /* if some stab domain was allocated */
      if(i>=p->sintrs_on && i<p->sintrs_off) continue;
    if(ss_inq(p, i)>SS_POINT) { /* we can free it! */
      just_free(p, i);           /* it also adjusts the sizes! */
      return 1;
    }
    for(j=0; j<MAX_LEADS; j++)
      if(p->sgs[i]->leadsz[j]>0) {
        just_free(p, i);
        return 1;
      }
    /* continue; */
  }


/* level 2 --- same as level 1, but continues in more restricted cases */

  for(i=0; i<p->sgsfill; i++) {
    if(p->lintrs_on!=-1 && p->lasttabtype==LTT_L)
      /* if lead domain was last allocated */
      if(i>=p->lintrs_on && i<p->lintrs_off) continue;
    if(p->iintrs_on!=-1 && p->lasttabtype==LTT_I)
      /* if some itab domain was allocated */
      if(i>=p->iintrs_on && i<p->iintrs_off) continue;
    if(p->sintrs_on!=-1 && p->lasttabtype==LTT_S)
      /* if some stab domain was allocated */
      if(i>=p->sintrs_on && i<p->sintrs_off) continue;
    if(ss_inq(p, i)>SS_SKETCH) { /* we can free it! */
      just_free(p, i);           /* it also adjusts the sizes! */
      return 1;
    }
    for(j=0; j<MAX_LEADS; j++)
      if(p->sgs[i]->leadsz[j]>0) {
        just_free(p, i);
        return 1;
      }
    /* continue; */
  }
  return 0;  /* Nothing could be disallocated */  
}

/*--
  \end{code}

  \funparagraph{adjust\_memory}

  \begin{code}
  --*/


short adjust_memory(PERFILE *p) /* [FUN] */ 

{
  long sz;
  
  while((sz = p->totleadsz+p->totitabsz+p->totetabsz+p->totstabsz)
        > p->totlim) {
    if(free_one_segment(p)==0)
      return 0;
    if(p->totleadsz+p->totitabsz+p->totetabsz+p->totstabsz>=sz)
      return 0;
  }
  return 1;
}



/*--
  \end{code}

  \funparagraph{set\_total\_size}

  \begin{code}
  --*/

short set_total_size(PERFILE *p, long n) /* [FUN] */ 

{
  short t;
  long no;

  while(n < (no=p->totleadsz+p->totitabsz+p->totstabsz)) {
    t = free_one_segment(p);
    if(t==0)
      return 0;
    if(no < p->totleadsz+p->totitabsz+p->totstabsz)
      return 0;   /* This must never happen */
    }
  p->totlim = n;
  return 1;
}


/*--
  \end{code}

  \funparagraph{reqlead}

  \begin{code}
  --*/


short reqlead(PERFILE *p, short leadno, long segfrom, long segto) /* [FUN] */ 

{
  long i, j;
  long ss;
  long sgssz;  /* number of samples */
  short r;

#ifdef RECTRACE
  fprintf(ecgnosislog, "CALL reqlead(%8x, %d, %ld, %ld)\n", 
    p, leadno, segfrom, segto);
  fflush(ecgnosislog);
#endif
  ss = ecg_sample_size(p->e);

  p->lintrs_on = segfrom;
  p->lintrs_off = segto;
  p->lasttabtype = LTT_L;
  
  for(i=segfrom; i<=segto; i++) {
    if(p->sgs[i]->lead[leadno]!=NULL)
      continue;
    sgssz = (p->sgs[i]->segend - p->sgs[i]->segbegin)/ss;
    sgssz = sgssz * sizeof(short);  /* this being the means of representation */
#ifdef RECTRACE
    fprintf(ecgnosislog, "     iteration nr %ld, ss=%ld, sgssz=%ld.\n", i, ss, sgssz);
    fprintf(ecgnosislog, "          lim0=%ld, lim1=%ld\n", p->sgs[i]->segbegin/ss, p->sgs[i]->segend/ss);
    fflush(ecgnosislog);
#endif
/* WHEN DEBUG
    while(p->totitabsz+p->totstabsz+p->totleadsz+sgssz+p->totetabsz
          > p->totlim) {
      r = free_one_segment(p);
      if(r==0) {
        glikrecerr = MEMORY_IS_FULL;
        return 0;
      }
    } */
    r = seg_require(p, p->sgs[i]->segbegin/ss,
                    p->sgs[i]->segend/ss, leadno, 2);
    if(r==0) {
      glikrecerr = CANT_READ;
      return 0;
    }
    p->sgs[i]->lead[leadno] = (short *)malloc(sgssz);
    if(p->sgs[i]->lead[leadno] == NULL) {
      glikrecerr = NO_MEMORY;
      return 0;
    }
    for(j=0; j<sgssz/2; j++)
      p->sgs[i]->lead[leadno][j] = glead[act_pre_size+j];
    p->sgs[i]->leadsz[leadno] = sgssz/2;
    p->totleadsz += sgssz;
  }
  return 1;
}


/*--
  \end{code}

  \funparagraph{reqitab}

  \begin{code}
  --*/


short reqitab(PERFILE *p, long from, long to)

{
  long segto, segfrom;
  long i;
  short r;

  segfrom = segat(p, from);
  if(segfrom==-1)
    return 0;

  segto = segat(p, to);
  if(segto==-1)
    return 0;

#ifdef RECTRACE
  fprintf(ecgnosislog,
          "CALL reqitab(), segemnts from %ld to %ld.\n", segfrom, segto);
  fflush(ecgnosislog);
#endif

  p->iintrs_on = segfrom;
  p->iintrs_off = segto;
  p->lasttabtype = LTT_I;

  for(i=segfrom; i<segto; i++) {
    if(ss_inq(p, i)>SS_POINT)
      continue;
#ifdef RECTRACE
  fprintf(ecgnosislog, "     SSINQ returned something le SS_POINT\n");
  fflush(ecgnosislog);
#endif
    if((r=ss_chg(p, i, SS_SKETCH))==0) {
      glikrecerr = CANT_ADVANCE;
      return 0;
    }
/*    while(p->totitabsz+p->totstabsz+p->totleadsz+p->totetabsz
          > p->totlim) {
      r = free_one_segment(p);
      if(r==0) {
        glikrecerr = MEMORY_IS_FULL;
        return 0;
      }
    }*/
  }
  return 1;
}



/*--
  \end{code}

  \subsection{Recording\_File}

  \funparagraph{rec\_file\_open}

  \begin{code}
  --*/

PERFILE *rec_file_open(char *sif, char *rcf, short ftype) /* [FUN] */ 

  {
  PERFILE *r;
  long i, j;


  /*--
    \end{code}

    \subparagraph{Allocate PERFILE}

    \begin{code}
    --*/
  

  r = (PERFILE *)malloc(sizeof(PERFILE));
  if(r==NULL) {
    glikrecerr = NO_MEMORY;
    glikerrmsg = "Not enough memory in the rec_file_open() function.\n";
    return NULL;
  }


  /*--
    \end{code}

    \subparagraph{Try to opem ecg file}

    \begin{code}
    --*/
  
  r->e = ecg_open(rcf, ftype);
  if(r->e==NULL) {
    free(r);
    glikrecerr = NO_SUCH_RECORD_FILE;
    sprintf(glikerrbuf, "No record file named %s", rcf);
    glikerrmsg = glikerrbuf;
    return NULL;
  }


  /*--
    \end{code}

    \subparagraph{Fetch recording parameters}

    \begin{code}
    --*/
  

  r->nleads = ecg_noft(r->e);
  for(i=0; i<r->nleads; i++) {
    r->leadnm[i] = NULL;
  }

  /*--
    \end{code}
    
    \subparagraph{Establish and allocate segments}
    

  The input file is divided in segments that are contiguous pieces of
  recording (in all simultaneous derivations). The segments must not
  be larger than 32k (that is 16k samples) and yet not more than
  {\tt MAX\_SEGS}; the second requirement takes precedence over the first. 
  To start with, the recommended segment size is 1 minute at 200 samples
  per second (this should lead to 24kb lead buffers and 1440 segments in 
  24h --- the maximum number of samples, normally 2000, would thus be 
  enough for 33h and 20 minutes).

  \begin{code}
  --*/
  


  r->sgsfill = (ecg_size(r->e) / 16000) + 1;
  r->sgssz = 16000;

#ifdef TRACE
  fprintf(ecgnosislog, "Size %ld samples, %d segments created\n",
          ecg_size(r->e), r->sgsfill);
  fflush(ecgnosislog);
#endif

  /* if it didn't work, make segments as large as needed */
  if (r->sgsfill>MAX_SEGS) {
    r->sgssz = (ecg_size(r->e) / (MAX_SEGS+3));
    r->sgsfill = (ecg_size(r->e) / r->sgssz) + 1;
  }

  r->sgscssz = ecg_sample_size(r->e) * r->sgssz;
  for(i=0; i<r->sgsfill; i++) {  /* a better allocator may be written */
    r->sgs[i] = (SEGBODY *)malloc(sizeof(SEGBODY));
    if(r->sgs[i]==NULL) {
      i--;
      while(i>=0)
        free(r->sgs[i]);
      free(r);
      glikrecerr = NO_MEMORY;
      glikerrmsg = "Not enough memory for segment allocation.";
      return NULL;
    }
    r->sgs[i]->segbegin = i * r->sgscssz;  
       /* the count of the first sample, numbered beginning with 0 */
    r->sgs[i]->segend = (i+1) * r->sgscssz; 
       /* the count of the next sample after the last in this segment */
    for(j=0; j<r->nleads; j++) {
      r->sgs[i]->lead[j] = NULL;
      r->sgs[i]->leadsz[j] = r->sgssz;
      r->sgs[i]->leadfill[j] = 0;
    }
    r->sgs[i]->itabsz = -1;   /* UNKNOWN */
    r->sgs[i]->itabfill = -1; /* itab not allocated */
    r->sgs[i]->itab = NULL;   /* not yet allocated */
    r->sgs[i]->itabn = -1;    /* UNKNOWN */

    r->sgs[i]->stabsz = -1;   /* UNKNOWN */
    r->sgs[i]->stabfill = -1; /* stab not allocated */
    r->sgs[i]->stab = NULL;   /* not yet allocated */

#ifdef ETAB
    r->sgs[i]->etabsz = -1;   /* UNKNOWN, not allocated */
    r->sgs[i]->etabfill = -1; /* not allocated */
    r->sgs[i]->etab = NULL;   /* not yet allocated */
#endif

  }


  /*--
    \end{code}
    
    \subparagraph{set VM strategy counters}
    
    \begin{code}
    --*/

  r->totleadsz = 0;
  r->leadn = 0;
  r->totitabsz = 0;
  r->itabn = 0;
  r->totstabsz = 0;
  r->stabn = 0;

  r->lasttabtype = LTT_NONE;

  r->lintrs_on = r->lintrs_off = -1;
  r->iintrs_on = r->iintrs_off = -1;
  r->sintrs_on = r->sintrs_off = -1;



  /*--
    \end{code}

    \subparagraph{Inititalize morphology classification functions}

    \begin{code}
    --*/

  clear_morphologies(r);

  /*--
    \end{code}

    \subparagraph{Set other global variables}

    \begin{code}
    --*/
  
  /* initialize the associated QRS files to NULL */
  r->outqrsf = NULL;
  r->assqrsf = NULL;

  /* clear the gmo for any case */
  r->gmofill = 0;

  /* initialize the crt_qrs_lead */
  r->crt_qrs_lead = 0;
  r->st_detect_offs = 8;

  return r;  
  }


     /*--
       \end{code}

       \funparagraph{rec\_file\_close}

       \begin{code}
       --*/
     
void rec_file_close(PERFILE *p) 

{
  long i, j;
  
  if(p->e!=NULL) {
    ecg_close(p->e);
    p->e = NULL;
  }
  for(i=0; i<p->sgsfill; i++) {
    if(p->sgs[i]==NULL)
      continue;
    for(j=0; j<MAX_LEADS; j++)
      if(p->sgs[i]->lead[j]!=0)
        free(p->sgs[i]->lead[j]);
    if(p->sgs[i]->itab!=NULL) {
      free(p->sgs[i]->itab);
      p->sgs[i]->itab=NULL;
    }
    if(p->sgs[i]->stab!=NULL) {
      free(p->sgs[i]->stab);
      p->sgs[i]->stab = NULL;
    }

    /* INCLUDE TEXT FOR FREEING GEF STRUCTURES HERE */

    free(p->sgs[i]);
    p->sgs[i] = NULL;
  }
  free(p);
}


/*--
  \end{code}

  \funparagraph{rec\_file parameters}

  \begin{code}
  --*/



long rec_file_size(PERFILE *p)

{
  return ecg_size(p->e);
}

long rec_file_duration(PERFILE *p)

{
  long r;

  r = ecg_size(p->e);
  r *= p->e->ecgsamplesz;
  return r;
}

short rec_file_traces(PERFILE *p)

{
  return ecg_noft(p);
}


char *rec_file_name(PERFILE *p)

{
  return p->e->ecgfn;
}



/*--
  \end{code}

  \subsection{Results}

  \begin{code}
  --*/


/*--
  \end{code}

  \dataparagraph{Last complex location}

  \begin{code}
  --*/

static long last_complex_n = -1;  /* [VAR] */ 
static long last_complex_seg = -1; /* [VAR] */ 
static long last_complex_mom = -1; /* [VAR] */ 
static short skiped_complex_p; /* [VAR] */ 

/*--
  \end{code}

  \funparagraph{complex\_at}
  \Changes{
  Let {\tt itab[i]...itab[i+1]} be a complex; this function assumes that
  it starts with the sample given by {\tt itab[i]+1} and ends with
  the sample given by {\tt itab[i+1]} inclusively. Other functions
  in this package might assume a different situation (but I can't know
  which and how until the whole program is written). If a misbehaviour
  of the program could be thus explained, please follow that interaction.
  
  Boundaries between segments are not treated by this function.
  }

  \begin{code}
  --*/


long complex_at(PERFILE *p, long mom) /* [FUN] */ 


{
  long segmom;
  long deltac;

  segmom = segat(p, mom);
  if(segmom == -1) {
    last_complex_seg = -1;
    last_complex_n = -1;
    last_complex_mom = -1;
    fprintf(ecgnosislog, "segat failed.\n");
    fflush(ecgnosislog);
    return -1;
  }

  last_complex_seg = segmom;
  last_complex_mom = mom;

  if(ss_inq(p, segmom)<SS_SKETCH)
    ss_chg(p, segmom, SS_SKETCH);

  /* We know that the moment IS inside the segment; in order to determine
   * its exact position we must first euristically choose a complex and
   * move around until to right one is found.
   */

  deltac = p->sgs[segmom]->itabfill *
    ((float)(mom - p->sgs[segmom]->segbegin) / 
    (float)(p->sgs[segmom]->segend - p->sgs[segmom]->segbegin));
  if(p->sgs[segmom]->itab[deltac].pl > mom) {
    while(p->sgs[segmom]->itab[deltac].pl > mom)
      deltac --;
    return (last_complex_n = p->sgs[segmom]->itabn + deltac);
  }
  if(p->sgs[segmom]->itab[deltac].pl <= mom) {
    while(p->sgs[segmom]->itab[deltac].pl <= mom)
      deltac ++;
    return (last_complex_n = p->sgs[segmom]->itabn + deltac);
  }
  return (last_complex_n = p->sgs[segmom]->itabn + deltac);
}

/*--
  \end{code}

  \funparagraph{next\_cat}
  \ImplementationNotes{
  % How implementation of the function is done: algorithm---if special,
  % implementation decisions, discarded implementation alternatives
  The {\tt skiped\_complex\_p} is needed when a parameter is required
  at constant intervals of time (for example, the R--R distance for
  the complex found at each 0.5 seconds). If the intervals are too large
  it is possible for a sequence of complexes to be found inside
  an interval, and thus for a value to be omitted. To permit detecting
  this situation (in quite sofisticated applications),
  the {\tt next\_cat} function sets the {\tt skiped\_complex\_p} to 1 if it
  occurs and to 0 if it doesn't.

  This function is not used any more, the {\tt cpars\_at} making it
  unnecessary.
  }

  \begin{code}
  --*/

static short skiped_complex_p; /* [VAR] */ 

long next_cat(PERFILE *p, long mom) /* [FUN] */ 

{
  long rv;
  
  if(last_complex_n==-1 || last_complex_seg==-1 || last_complex_mom==-1)
    return complex_at(p, mom);

  skiped_complex_p = 0;

  rv = last_complex_n - p->sgs[last_complex_seg]->itabn;
  if(mom>last_complex_mom) {
    while(mom<p->sgs[last_complex_seg]->itab[rv].pl) {
      rv ++;
      skiped_complex_p ++;
      if(rv >= p->sgs[last_complex_seg]->itabfill) {
        last_complex_seg ++;
        rv = 0;
        if(last_complex_seg >= p->sgsfill) {
          last_complex_seg = -1;
          last_complex_n = -1;
          last_complex_mom = -1;
          return -1;  /* out of bounds */
        }
      }
    }
    skiped_complex_p --;
    if(skiped_complex_p==-1)
      skiped_complex_p = 0;
    return (last_complex_mom = mom,
            last_complex_n = rv+p->sgs[last_complex_seg]->itabn);
  }
  if(mom<last_complex_mom) {
    while(mom<p->sgs[last_complex_seg]->itab[rv].pl) {
      rv --;
      skiped_complex_p ++;
      if(rv < 0) {
        last_complex_seg --;
        if(last_complex_seg < 0) {
          last_complex_seg = -1;
          last_complex_n = -1;
          last_complex_mom = -1;
          return -1;  /* out of bounds */
        }
        rv = p->sgs[last_complex_seg]->itabfill - 1;
      }
    }
    skiped_complex_p --;
    if(skiped_complex_p==-1)
      skiped_complex_p = 0;
    return (last_complex_mom = mom,
            last_complex_n = rv+p->sgs[last_complex_seg]->itabn);
  }
  return last_complex_n;  /* i.e. if it is the same moment */
}


/*--
  \end{code}

  \funparagraph{complexes\_between}

  \begin{code}
  --*/

short complexes_between_eor = 0; /* [VAR] */ 

long complexes_between(PERFILE *p, long m0, long m1) /* [FUN] */ 

{
  long r0, r1;

  r0 = complex_at(p, m0);
  fprintf(ecgnosislog, "r0=%ld\n", r0);
  r1 = complex_at(p, m1);
  fprintf(ecgnosislog, "r1=%ld\n", r1);
  fflush(ecgnosislog);
  complexes_between_eor = 0;
  if(r0==-1) {
    return -1;
  }
  if(r1==-1) {  /* normally because it is past EOF */
    if(ss_inq(p, p->sgsfill-1)<SS_SKETCH)
      r1 = ss_chg(p, p->sgsfill-1, SS_SKETCH);
    if(r1==0)
      return -1;
    return p->sgs[p->sgsfill-1]->itabfill-1 - r0;
  }
  return r1-r0;
}


/*--
  \end{code}

  \funparagraph{cpar\_of}
  \ImplementationNotes{
  The following describes the \functiontag{\_cpar\_of} function, which
  is used by \functiontag{cpar\_of} and \functiontag{cpars\_of}

  This function computes the value of an analysis parameter
  when both the segment and the complex count within the
  segment are given. The various means of determining the segment
  and the count efficiently are implemented by the functions that
  follow.

  Apart from reducing the size of the code, this function
  also keeps the computation of a parameter in a single place.

  {\tt seg} is the segment and {\tt n} is the complex count inside
  the segment; {\tt segss} is the result of calling the {\tt ss\_inq}
  function on {\tt seg}; {\tt par} is the symbol that
  describes the parameter. 

  The {\tt segss} is included here because, in the frequent
  case in which this function is called repeatedly for each 
  complex the {\tt ss\_inq} function should only be called once.
  
  The {\tt \_cpar\_of} function returns the value of the parameter
  as a signed long quantity --- if everything worked ok.

  As we cannot predict what this value
  may be in the future extentions of the function, a sepparate 
  variable is provided for errors ({\tt \_cpar\_of\_err}). This
  must normally be 0, otherwise trouble occoured. 

  The trouble can be ``1'' if the level of the segment is not
  high enough to return the value, or ``2'' if there is an unknown
  parameter.

  The {\tt \_cpar\_err\_clue} variable contains a clue --- normally the
  level that must be touched by the segment in order to permit
  the computation of the required parameter --- indicating how the
  error can be corrected.
  }

  \begin{code}
  --*/

static _cpar_of_err = 0; /* [VAR] */ 
static _cpar_err_clue = 0; /* [VAR] */ 

long _cpar_of(PERFILE *p, long seg, short segss, short par, short n) /* [FUN] */ 

{
  /* ALWAYS ASSUME ARGUMENTS ARE RIGHT */

  _cpar_of_err = 0;
  _cpar_err_clue = 0;
  switch(par) {
  case CPST1:
    if(segss<SS_SKETCH) {
      _cpar_of_err = 1;
      return _cpar_err_clue = SS_SKETCH;
    }
    return p->sgs[seg]->itab[n].st1;
  case CPST2:
    if(segss<SS_SKETCH) {
      _cpar_of_err = 1;
      return _cpar_err_clue = SS_SKETCH;
    }
    return p->sgs[seg]->itab[n].st2;
  case CPJPOS:
    if(segss<SS_SKETCH) {
      _cpar_of_err = 1;
      return _cpar_err_clue = SS_SKETCH;
    }
    return p->sgs[seg]->itab[n].j;
  case CPBL1:
    if(segss<SS_SKETCH) {
      _cpar_of_err = 1;
      return _cpar_err_clue = SS_SKETCH;
    }
    return p->sgs[seg]->itab[n].bl1;
  case CPBL2:
    if(segss<SS_SKETCH) {
      _cpar_of_err = 1;
      return _cpar_err_clue = SS_SKETCH;
    }
    return p->sgs[seg]->itab[n].bl2;
  case CPQRSPOS:
    if(segss<SS_SKETCH) {
      _cpar_of_err = 1;
      return _cpar_err_clue = SS_SKETCH;
    }
    return p->sgs[seg]->itab[n].pl;
  case CPQRSLEN:
    if(segss<SS_SKETCH) {
      _cpar_of_err = 1;
      return _cpar_err_clue = SS_SKETCH;
    }
    return p->sgs[seg]->itab[n].qrsl;
  case CPQRSMORPH:
    if(segss<SS_SKETCH) {
      _cpar_of_err = 1;
      return _cpar_err_clue = SS_SKETCH;
    }
    return p->sgs[seg]->itab[n].morph;
  case CPQRSREF:
    if(segss<SS_MORPH) {
      _cpar_of_err = 1;
      return _cpar_err_clue = SS_MORPH;
    }
    return p->sgs[seg]->itab[n].reff;
  case CPQRSREFPOS:
    if(segss<SS_MORPH) {
      _cpar_of_err = 1;
      return _cpar_err_clue = SS_MORPH;
    }
    return p->sgs[seg]->itab[n].reff + p->sgs[seg]->itab[n].pl;
  default:
    _cpar_of_err = 2;
    return 0;
  }
  _cpar_of_err = 2;
  return 0;
}

long segwith(PERFILE *p, long n)

{
  
}


long cpar_of(PERFILE *p, short par, long n)  			/* [FUN] */ 

{
  long segn;
  long rv;

  segn = segwith(p, n);
  if(segn==-1) {
    glikrecerr = NO_SUCH_COMPLEX;
    return 0;
  }
  rv = _cpar_of(p, segn, ss_inq(p, segn), par, n - p->sgs[segn]->itabn);
  if(_cpar_of_err>0) {
    switch(_cpar_of_err) {
    case 1:
      ss_chg(p, segn, _cpar_err_clue);
      if(adjust_memory(p)==0) {
        glikrecerr = TOO_LITTLE_ALLOCATED_MEMORY;
        return 0;
      }
      rv = _cpar_of(p, segn, ss_inq(p, segn), par, n - p->sgs[segn]->itabn);
      if(_cpar_of_err>0) {
        glikrecerr = INTRACTABLE_SEG_TYPE;
        return 0;
      }
      return rv;
    case 2:
      glikrecerr = NO_SUCH_PARAMETER;
      return 0;
    default:
      glikrecerr = INTERNAL_ERROR;
      return 0;
    }
  }
  return rv;
}


/*--
  \end{code}
  
  \funparagraph{cpar\_at, cpars\_at, cpars\_between}
  \ImplementationNotes{
  This function should use the result of {\tt segat} as 
  set by {\tt complex\_at} (thus avoiding reading it twice).
  }
  \Changes{ 
  \functiontag{cpar\_at} should use the result of {\tt segat} as 
  set by {\tt complex\_at} (thus avoiding reading it twice).
  }
  
  \begin{code}
  --*/
  
long cpar_at(PERFILE *p, short par, long pos) /* [FUN] */ 

{
  long segn;
  long rv;
  long n;

  segn = segat(p, pos);
  n = complex_at(p, pos);
  if(segn==-1 || n==-1) {
    glikrecerr = NO_SUCH_COMPLEX;
    return 0;
  }
  rv = _cpar_of(p, segn, ss_inq(p, segn), par, n - p->sgs[segn]->itabn);
  if(_cpar_of_err>0) {
    switch(_cpar_of_err) {
    case 1:
      ss_chg(p, segn, _cpar_err_clue);
      if(adjust_memory(p)==0) {
        glikrecerr = TOO_LITTLE_ALLOCATED_MEMORY;
        return 0;
      }
      rv = _cpar_of(p, segn, ss_inq(p, segn), par, n - p->sgs[segn]->itabn);
      if(_cpar_of_err>0) {
        glikrecerr = INTRACTABLE_SEG_TYPE;
        return 0;
      }
      return rv;
    case 2:
      glikrecerr = NO_SUCH_PARAMETER;
      return 0;
    default:
      glikrecerr = INTERNAL_ERROR;
      return 0;
    }
  }
  return rv;

}

long cpars_at(PERFILE *p, long *buf, short par, long pos, long nsam) /* [FUN] */ 

{
  long segn;
  long rv;
  long i, n;

  segn = segat(p, pos);
  n = complex_at(p, pos);
  if(segn==-1 || n==-1) {
    glikrecerr = NO_SUCH_COMPLEX;
    return 0;
  }
  if(n < p->sgs[segn]->itabn)
    segn --;
  for(i=0; i<nsam; i++) {
    rv = _cpar_of(p, segn, ss_inq(p, segn), par, n - p->sgs[segn]->itabn);
    
#ifdef RECTRACE
    fprintf(ecgnosislog, "     Called _cpar_of(, %ld, , , %ld) -> %ld, err=%d\n", 
      segn, n - p->sgs[segn]->itabn, rv, _cpar_of_err);
    fflush(ecgnosislog);
#endif

    if(_cpar_of_err>0) {
      switch(_cpar_of_err) {
      case 1:
        ss_chg(p, segn, _cpar_err_clue);
        if(adjust_memory(p)==0) {
          glikrecerr = TOO_LITTLE_ALLOCATED_MEMORY;
          return 0;
        }
        rv = _cpar_of(p, segn, ss_inq(p, segn), par, n - p->sgs[segn]->itabn);
        if(_cpar_of_err>0) {
          glikrecerr = INTRACTABLE_SEG_TYPE;
          return 0;
        }
        return rv;
      case 2:
        glikrecerr = NO_SUCH_PARAMETER;
        return 0;
      default:
        glikrecerr = INTERNAL_ERROR;
        return 0;
      }
    }
    n++;
    if(n >= p->sgs[segn]->itabn + p->sgs[segn]->itabfill) {
      /* no more complexes in this segment */
      segn++;
      if(segn >= p->sgsfill) { /* no more complexes in this file */
        return i+1;
      }
    }
    buf[i] = rv;
  }
  return nsam;
}


/*--
  \end{code}

  \funparagraph{get\_samples}
  \Changes{ % that are anticipated, if special provision was made for them
  Only sampling at the same frequency as the buffer representation
  is provided. Resamplin may be desirable in the close future.
  }

  \begin{code}
  --*/

long get_samples(PERFILE *p, short *buf, short stripno, long t0, long t1, long newper) /* [FUN] */ 

{
  long segno0, segno1;
  short i, j;
  long pp;
  long lim;


#ifdef TRACE
  fprintf(ecgnosislog, "CALL get_samples(%ld, %ld).\n", t0, t1);
  fflush(ecgnosislog);
#endif
  if(t0>=rec_file_duration(p))
    return 0;
  if(t1>=rec_file_duration(p))
    t1 = rec_file_duration(p);
  segno0 = segat(p, t0);
  if(segno0<0)
    return -1;
#ifdef RECTRACE
  fprintf(ecgnosislog, "     Succesfully located segno0.\n"); fflush(ecgnosislog);
#endif
  segno1 = segat(p, t1);
  if(segno1<0)
    return -1;  /* please note that both must be inside the recording! */
#ifdef RECTRACE
  fprintf(ecgnosislog, "     Succesfully located segno1.\n"); fflush(ecgnosislog);
#endif

  if(newper!=ecg_sample_size(p)) {
    /* this general case is not treated yet */
    /* when it will be treated a sepparate branch for the */
    /* ``same sampling frequency'' will remain */
    return -1;
  }
#ifdef RECTRACE
  fprintf(ecgnosislog, "     newper is of the ecg_sample_size size, can proceed.\n"); fflush(ecgnosislog);
#endif

    if(reqlead(p, stripno, segno0, segno1)==0) {
#ifdef TRACE
    fprintf(ecgnosislog, 
      "     bad return from reqlead(), error=%d\n", glikrecerr);
    fflush(ecgnosislog);
#endif
    return -1;   /* could not get the lead into the memory */
  }
#ifdef RECTRACE
  fprintf(ecgnosislog, "     reqlead() worked.\n");
  fflush(ecgnosislog);
#endif
  
  /* Now, find the first sample in the first segment */
  i = (t0 - p->sgs[segno0]->segbegin) / newper /* == samplesize */ ;
  j = 0;  /* cursor in buf */
  for(pp = segno0; pp<=segno1 && j<(t1-t0)/newper; pp++) {
    if(pp<segno1)
      lim = p->sgs[pp]->leadsz[stripno];
    else
      lim = (t1 - p->sgs[segno1]->segbegin) / newper;
#ifdef RECTRACE
  fprintf(ecgnosislog, "  segno0=%ld, segno1=%ld, i=%d, j=%d\n", segno0, segno1, i, j);
  fprintf(ecgnosislog, "  lim=%d, stripno=%d, (t1-t0)/newper=%ld lead[%d]=%8x\n", 
          lim, stripno, (t1-t0)/newper, pp, p->sgs[pp]->lead[stripno]);
  fflush(ecgnosislog);
#endif
    for(; i<lim && j<(t1-t0)/newper; i++)
      buf[j++] = p->sgs[pp]->lead[stripno][i];
    i = 0;
  }
  return j;
}


/*--
  \end{code}

  \subsection{Results from morphology processing}

  \begin{code}
  --*/

short nof_morphologies(PERFILE *p) 				/* [FUN] */ 

{
  return p->gmofill;
}

short sample_morphology(PERFILE *p, short m, short *result) 	/* [FUN] */ 

{
  short i;

  for(i=0; i<60; i++) /* magic 60, 2*0.15s --- see glikalys */
    result[i] = p->gmo[m].gm[i];
}

short type_of_morph(PERFILE *p, short m) 			/* [FUN] */ 

{
  return p->gmo[m].gtype;
}

long nof_gliks_with_morph(PERFILE *p, short m) 			/* [FUN] */ 

{
  return p->gmo[m].ngliks;
}

short supclass_of(PERFILE *p, short m)   			/* [FUN] */ 

{
  return p->gmo[m].supclass;
}

short set_supclass_of(PERFILE *p, short m, short sc) 		/* [FUN] */ 

{
  p->gmo[m].supclass = sc;
}

short set_type_of_morph(PERFILE *p, short m, short t) 		/* [FUN] */ 

{
  p->gmo[m].gtype = t;
}

/*--
  \end{code}

  \subsection{Intermediary results file}
  
  \funparagraph{open\_qrs\_file}

  \Portability{ % What is assumed about the operating environment
  This function assumes that \filetag{fopen} works in a Unix-like manner,
  i.e. that "w" opens a text as well as a binary file (no translation
  is done). In a (for example) DOS environment "wb" is to be used.
  
  If a QRS file is created on a little-endian machine, then it must
  be read on a little-endian machine also; more precisely, the representation
  of 32-bit integers on the writing and on the reading machines
  must be the same.
  }
  
\begin{code}
--*/


short open_qrs_file(PERFILE *p, char *qrs_file_name) 		/* [FUN] */ 

{
  FILE *f;

  if(p->outqrsf!=NULL) {
    raise(QRS_FILE_NOT_CLOSED);
    return 0;
  }
  f = fopen(qrs_file_name, "wb");
  if(f==NULL) {
    raise(CANT_OPEN_QRS_FILE_FOR_OUT);
    return 0;
  }
  p->outqrsf = f;
  return 1;
}

/*--
  \funparagraph{write\_to\_qrs\_file}

  \ImplementationNotes{
  We did not write segments on disk---instead arbitrary periods of time---%
  because that would have violated the hiding of segment existence.
  
  We did not design the write function to only have the last of the time
  limits (the first always being the last one of the previous call---normally)
  because in order to identify what is written to the disk, the maintenance
  programmer would have had to simulate in mind all program execution
  to the point (but this decision of ours is still disputable).
  }

\begin{code}
--*/
  
static void write_4_bytes(long s, FILE *fo) {                   /* [FUN] */ 

  unsigned char *c;

  c = (unsigned char *)(&s);
  putc(*c++, fo);
  putc(*c++, fo);
  putc(*c++, fo);
  putc(*c++, fo);
}

short write_to_qrs_file(PERFILE *p, long t0, long t1) 		/* [FUN] */ 

{
  long segix, seg;

  complex_at(p, t0); /* in order to set the \variabletag{last_complex_n}
                        and the \variabletag{last_complex_tag} values
                        */
  if(last_complex_seg==-1 || last_complex_n==-1) {
    raise(BAD_ARGUMENTS_TO_QRS_WRITE);
    return 0;
  }
  segix = last_complex_n - p->sgs[seg=last_complex_seg]->itabn;
  segix++;
  if(segix >= p->sgs[seg]->itabfill) {
    seg ++;
    segix = 0;
    if(seg >= p->sgsfill) {
      raise(BAD_ARGUMENTS_TO_QRS_WRITE);
      return 0;
    }
  }
  while(p->sgs[seg]->itab[segix].pl < t1) {
    write_4_bytes(p->sgs[seg]->itab[segix].pl, p->outqrsf);
    putc(p->sgs[seg]->itab[segix].st1, p->outqrsf);
    putc(p->sgs[seg]->itab[segix].st2, p->outqrsf);
    segix++;
    if(segix >= p->sgs[seg]->itabfill) {
      seg ++;
      segix = 0;
      if(seg >= p->sgsfill) {
        raise(BAD_ARGUMENTS_TO_QRS_WRITE);
        return 0;
      }
    }
  }
  return 1;
}


/*--
  \end{code}                        


  \begin{code}
--*/
  
short close_qrs_file(PERFILE *p) 				/* [FUN] */ 

{
  if(p->outqrsf==NULL) {
    raise(CLOSE_NO_QRS_FILE);
    return 0;
  }
  fclose(p->outqrsf);
  p->outqrsf = NULL;
  return 1;
}

/*--
  \end{code}

  \funparagraph{assign\_qrs\_file}

  \ImplementationNotes{
  Appart from opening a previously existing QRS file, this function also
  deletes all \variabletag{itab}s existing in the \typetag{PERFILE} structure
  (as they are now obsolete and may be inconsistent with the ones
  read from the QRS file). All the \variabletag{itabn}s are recomputed and
  all the segments are forced into the \constanttag{SS\_POINT} status. 
  
  Please note that the \variabletag{itabn}s can be used as indexes in the
  file, as there are 4 bytes per itab value.
  
  QRS files that are too short are not taken into account here. Nor are
  the too long ones. We never consider that such a file is associated
  to a wrong PERFILE structure. The application is expected
  (and can) ensure consistence between these files.
  }
  \Portability{ % What is assumed about the operating environment
  The file system is assumed to be Unix-like, i.e. "r" specification
  in \functiontag{fopen} implies that text translation is not done.
  On MS-DOS for example, "rb" should be used.
  }

  \begin{code}
  --*/
  
static long thel;

static long read_4_bytes(FILE *f)  {                            /* [FUN] */ 

/*  unsigned char *c;

  c = (unsigned char *)(&l);
  *c++ = getc(f);
  *c++ = getc(f);
  *c++ = getc(f);
  *c++ = getc(f); */
  fread(&thel, 1, 4, f);
/*  if(feof(f))
    return -1;*/
  return thel;
}


short assign_qrs_file(PERFILE *p, char *qrsfn) 			/* [FUN] */ 

{
  FILE *f;
  long crtn;
  long seg;
  long nextv;

  if(p->assqrsf!=NULL) {
    raise(QRS_FILE_ALREADY_ASSIGNED);
    return 0;
  }
  f = fopen(qrsfn, "rb");
  if(f==NULL) {
    raise(NO_QRS_FILE_TO_ASSIGN);
    return 0;
  }
  p->assqrsf = f;
  crtn = 0;
  nextv = 0;
  for(seg=0; nextv>-1 && seg < p->sgsfill; seg++) {
    p->sgs[seg]->itabn = crtn;
    if(crtn==0)
      crtn++;
    if(p->sgs[seg]->itab==NULL)
      free(p->sgs[seg]->itab);
    p->sgs[seg]->itab = NULL;
    p->sgs[seg]->itabfill =
      p->sgs[seg]->itabsz = -1; /* as in \functiontag{rec_file_open} */
    do {
/*      nextv = read_4_bytes(f);*/
      fread(&nextv, 4, 1, f);
      getc(f);
      getc(f);
      if(feof(f))
        break;
/*      fprintf(ecgnosislog, "%ld\n", nextv);
      fflush(ecgnosislog);*/
      crtn++;
    } while(nextv < p->sgs[seg]->segend && nextv!=-1);
/*    fprintf(ecgnosislog, "\n");
    fflush(ecgnosislog);*/
  }  /* shortness of files is not taken into account */
  return 1;
}
  

/*--
  \end{code}

  \funparagraph{read\_qrs\_data}

\PurposeReturn{
  This is a function local to the \filetag{glikrec} module, which
reads in a \typetag{PERFILE}, from the temporary results file that
is assigned, all the qrs analysis information pertaining to 
a given segment.  Returns 0 if failure and 1 if success.
}
\Exceptions{
\begin{itemize}
\item{\exceptiontag{NO\_ASSIGNED\_QRS\_FILE}} when called for a \typetag{PERFILE}
that does not have any associated QRS file.
\item{\exceptiontag{NO\_MEMORY}} when memory allocation of an eventually
new {\tt itab} segment fails.
\end{itemize}
}
\GlobalEffects{
  % Global effects on exported or external variables
The \concepttag{segment status} of the argument segment is advanced to reflect
the new situation.
}
\Preconditions{
  % Conditions to be met when the function is called
The \typetag{PERFILE} argument must have an assigned qrs file.
}


\Portability{ % What is assumed about the operating environment
See observations of the way integers are represented at the beginning
of this package.
}
\Changes{ % that are anticipated, if special provision was made for them
It should later read more than just the detection position of the QRS.
}

\begin{code}
--*/

int read_qrs_data(PERFILE *p, short seg) {                    /* [FUN] */ 

  long i, lim;
  long po;

  fprintf(ecgnosislog, "read_qrs_data(, %d)\n", seg);
  fflush(ecgnosislog);  
  if(p->assqrsf==NULL) {
    raise(NO_ASSIGNED_QRS_FILE);
    return 0;
  }
  po = p->sgs[seg]->itabn;
  po *= 6;
  lim = p->sgs[seg+1]->itabn - p->sgs[seg]->itabn;
  fseek(p->assqrsf, po, 0); /* i.e. from the start of the file */
  if(p->sgs[seg]->itabsz < lim) {
    if(p->sgs[seg]->itab!=NULL)
      free(p->sgs[seg]->itab);
    p->sgs[seg]->itab = (STEPONE *)malloc(lim * sizeof(STEPONE));
    if(p->sgs[seg]->itab==NULL) {
      p->sgs[seg]->itabsz = -1;
      p->sgs[seg]->itabfill = -1;
      raise(NO_MEMORY);
      return 0;
    }
    p->sgs[seg]->itabsz = lim;
  }
  for(i=0; i<lim; i++) {
    if(i==0 && seg==0)
      p->sgs[seg]->itab[i].pl = 0;
    else {
      p->sgs[seg]->itab[i].pl = read_4_bytes(p->assqrsf);
      p->sgs[seg]->itab[i].st1 = getc(p->assqrsf);
      if(p->sgs[seg]->itab[i].st1>127)
        p->sgs[seg]->itab[i].st1 = p->sgs[seg]->itab[i].st1 - 256;
      p->sgs[seg]->itab[i].st2 = getc(p->assqrsf);
      if(p->sgs[seg]->itab[i].st2>127)
        p->sgs[seg]->itab[i].st2 = p->sgs[seg]->itab[i].st2 - 256;
    }
  }
  p->sgs[seg]->itabfill = lim;
  fprintf(ecgnosislog, "Done, lim=%ld\n", lim);
  fflush(ecgnosislog);
  return 1;  /* done */
}

/*--
  \end{code}

  \funparagraph{unassign\_qrs\_file}

\begin{code}
--*/

short unassign_qrs_file(PERFILE *p) 				/* [FUN] */ 

{
  if(p->assqrsf==NULL) {
    raise(NO_ASSIGNED_QRS_FILE);
    return 0;
  }
  fclose(p->assqrsf);
  p->assqrsf = 0;
  return 0;
}



/*--
  \end{code}

  \subsection{Functions that set-get analysis parameters}

  \funparagraph{Parameter set-get functions}  % The name of the function(s)

  \begin{code}
  --*/


void set_crt_qrs_lead(PERFILE *p, short v) 			/* [FUN] */ 

{
  p->crt_qrs_lead = v;
}

void set_st_detect_position(PERFILE *p, short stpos) 		/* [FUN] */ 

{
  p->st_detect_offs = stpos;
}

short get_st_detect_position(PERFILE *p) 			/* [FUN] */ 

{
  return p->st_detect_offs;
}


/*--
  \end{code}

  \endoffile

  --*/

