/*--

\FileTitle{ % format: <short title> --- <long title>
  e2100 --- E2100 communication package
  }
  \CopyrightNotice{ % format: Copyright \copyright <year(s)> <who>
  Copyright \copyright 1993--1994 Alexandru Dan CORLAN
  Copyright \copyright 1994--1995 Marius SERITAN

  \begin{verbatim}
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    \end{verbatim}

  }
  \RevisionHistory{
  This file is based upon the gwo series of communication programs
  developed by Alexandru CORLAN, Radu CORLAN and Marius SERITAN. In fact
  this is the implementation of one end of the communication protocol.

\begin{description}
\item{23'rd of May 1993} created by Alexandru Dan Corlan
\item{1994}   serlio version by Radu Corlan
\item{20 Jul 1994} djgpp version by Marius Seritan and Radu Corlan
\item{1 Dec 1994} ERUID implementation by Marius Seritan, design
by Alexandru Corlan. Two new subprotocols added to the Holter
interface in order to read holter specific info and internal ram.
\item{9 dec 1994} mseritan:
  gwo_read_packet improves transfer succes by retrying at the
  packet level (gwo_read_data is a special case because of
  the state built into the transmission protocol).
\item{10 dec 1994} mseritan:
             vertical parity added to the transmission packets
\item{22 dec 1994} mseritan: removed vertical parity, no need
\item{3 jan 1995} mseritan:
             modified send_gwo_a_byte, gwo echos each character send
               in order to implement flow control 
               (finally PC got faster than gwo)
\item{5 jan 1995} mseritan:
             cosmetical changes, making the file world readable
             using the `taw' package, introducing the e2100 package
\item{19 jan 1995} mseritan:
             introduced the function e2100\_refresh.
\item{1 mar 1995} mseritan:
             introduced the function e2100\_refresh\_address in order to
	     correct yet another transmision protocol error when
	     block transfers are not atomic.
\item{6 mar 1995} mseritan:
             introduced a modular device driver with major number 19,
	     minor number the serial port in order to implement
	     the atomic portion of the library. The transimision
	     program seems at last to work under Linux,
	     no more strange race conditions occur.
\item{21 mar 1995} mseritan:
             extended the device approach to gwo\_set\_address, there
             there is no more need to run the program as root (but
             root authority is needed to install it). Cool.
\item{23 may 1995} mseritan:
             in order to increase transfer speed under Linux, more than 
             a block can be read in a call to the holter device. The dimension
             of the buffer must be a multiple of 32 (the dimension of 
             the block).
\end{itemize}
}
  \vskip 1in

  \begin{code}
  --*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#ifdef unix
#include <unistd.h>
#else
#include <io.h>
#include <bios.h>
#include <dos.h>
#endif

#include <fcntl.h>


#include "serlio.h"
#include "e2100.h"

/*--
  \end{code}

  \section{e2100}
  \ImplementationNotes{This is a  multilevel package.
  What you see here is half of the history. The other half
  is the program located in the E2100 device.}
  \Portability{This package is based upon the serlio library
  that has been ported to DOS (Microsoft and Djgpp compiler) and
  to Linux.}
  \Changes{}

  \begin{code}
  --*/


#ifndef min
#define min(a,b) ((a) <= (b) ? (a) : (b))
#endif /* min */


/*--
  \end{code}

  \typeparagraph{uchar}

  \Definition{Industry quasi standard unsigned char type.}

  \begin{code}
  --*/

typedef unsigned char uchar;


/*--
  \end{code}

  \subsection{gwo\_byte --- level 1 gwo communication functions}
  \PackagePurpose{}
  \FunctionInterraction{
  \functiontag{gwo\_init} initializes the serial interface and the low
  level communication automata in the holter,
  \functiontag{gwo\_send\_byte} and \functiontag{gwo\_receive\_byte}
  are used to transmit bytes to the holter.
  }
  \Exceptions{}
  \GlobalEffects{}
  \Preconditions{}
  \Performance{}

  \begin{code}
  --*/

/*--
  \end{code}

  \funparagraph{gwo\_init}

  \ImplementationNotes{ 
  Initialize the serial port and try to clear E2100's output buffers
  left dirty
  by a previous command. For this bytes are read from the serial port
  until no more bytes are available (\function{receive\_byte} returns
  an error).
  }

  \Portability{ The function \function{receive_byte} must include
  a timeout feature.
  }
  \Changes{}

  \begin{code}
  --*/


/*--
  \end{code}

  \constantparagraph{GWO\_BYTE\_MAX\_JUNK}

  \Definition{The maximum number of junk bytes that could be received
  from the holter.}
  \SemanticStructure{  }
  \ConsistencyRules{  }

  \begin{code}
  --*/

#ifdef LINUX_DEVICE
static int holter_handle;
#endif

#define GWO_BYTE_MAX_JUNK 100


int gwo_init( int serial_port, long baud_rate ) {
  int received;

#ifndef LINUX_DEVICE
  /* initialization of the serial port */
  clear_serial_error(serial_port);
  set_timeout(serial_port, 1000L );
  serial_port_init( serial_port, baud_rate, 0x03);
  modem_control(serial_port, 1);

  /* empty serial line, maybe holter is replying to some command
   * if the response is too verbose something must be wrong  
   */

  for( received=0;  receive_byte(serial_port) > 0; received++ )
    if( received >= GWO_BYTE_MAX_JUNK ) return -1;

  clear_serial_error(serial_port); /* of course there is one serial error */
#else
  holter_handle = open( LINUX_DEVICE, O_RDWR );
  if( holter_handle == -1 ) return -1;
#endif /* LINUX_DEVICE */
  
  return 0;
}

   
/*--
  \end{code}
    
  \funparagraph{gwo_send_byte}
  \ImplementationNotes{
  This function takes care of the echo feature of the holter
  communication protocol. Holter echo was added in order to
  implement flow control as the host is faster than the holter device.
  }
  \Portability{}
  \Changes{}

  \begin{code}
  --*/


short gwo_send_byte(int c, int serial_port) {
  int r;
  r = transmit_byte(serial_port, c);
  if( r<0 ) return r;
  r = receive_byte( serial_port );
  if( (r<0) || (r!=c) ) return -1;
}


/*--
  \end{code}

  \funparagraph{gwo_get_byte}
  % The name of the function is expected to be prezent
  % in the sectioning element above

  \PurposeReturn{Reads a byte from the serial port and returns it. If some
  error occured a negative value is returned.}
  \Exceptions{}
  \GlobalEffects{}
  \Preconditions{}
  \Performance{}

  \begin{code}
  --*/

int gwo_get_byte( int serial_port ) {
  return( receive_byte( serial_port ));
}


/*--
  \end{code}

  \subsection{gwo\_packet --- level 2 gwo communication functions}
  \PackagePurpose{Implements reliable communication through transaction
  like requests to the holter.
  There are two kind of transactions:
  \begin{itemize}
  \item simple transactions used to transfer a small amount of information
  (less than one packet size);
  \item addresed transactions used to receive a chunk from a bigger information
  block;
  \end{itemize}}
  \FunctionInterraction{
  \functiontag{gwo\_read\_packet} is used to read an ordinary packet,
  \functiontag{gwo\_set\_address} and 
  \functiontag{gwo\_read\_addressed\_packet} are used to transfer an addressed
  packet.
  }
  \Exceptions{
  }
  \GlobalEffects{
  % Global effects on exported or external variables
  }
  \Preconditions{
  % Conditions to be met when the package starts to be used
  }
  \Performance{
  % How it is, not why and what to do
  }

  \begin{code}
  --*/


/*--
  \end{code}

  \constantparagraph{Gwo commands}

  \ConstantSemantic{
 The possible commands to give to the GWO 

  }

  \begin{code}
  --*/

#define GWO_READ_STATUS     0x00a0
#define GWO_READ_DATA       0x00a1
#define GWO_READ_INFO       0x00a2
#define GWO_READ_RAM        0X00a3
#define GWO_WRITE_ADDRESS   0x00a8

/*--
  \end{code}

  \constantparagraph{GWO\_PACKET\_RETRY}

  \ConstantSemantic{ The number of times a transaction with gwo should be
  retried.}

  \begin{code}
  --*/

#define  GWO_PACKET_RETRY   10


/*--
  \end{code}

  \funparagraph{gwo_set_address}
  Set next read address.

  \PurposeReturn{ 0 if ok else -1}
  \Exceptions{}
  \GlobalEffects{}
  \Preconditions{}
  \Performance{}

  \begin{code}
  --*/

int gwo_set_address(long adr, int serial_port) {
  /* adr --- the address to be set 
   * serial_port --- number of the serial port used
   */
  short r, c, i;

#ifdef LINUX_DEVICE
  if( holter_handle == -1 ) return -3;
   if( lseek( holter_handle, adr, 0) ) return -1;
  return 0;
#else
  r = gwo_send_byte(GWO_WRITE_ADDRESS, serial_port );
  if( r < 0 ) {
#ifdef GWO_SET_ADDRESS_VERBOSE
    printf( "Error while sending address command.\n" );
#endif
    return -1;
  }

  for(i=0; i<4; i++) {
    c = 0x00ff & (adr >> 24);
    r = gwo_send_byte(c, serial_port );
    if( r<0 ) {
      return r;
    }
    adr = adr << 8;
    }

  return 0;  /* OK */
#endif /* LINUX_DEVICE */
  }


/*--
  \end{code}

  \funparagraph{gwo_read_packet}

  \PurposeReturn{An integer, negative if some error occured. A packet is
  read from gwo and placed in the memory buffer given as argument. If some
  error occurs then a new simple transaction is initiated.}
  \Exceptions{}
  \GlobalEffects{}
  \Preconditions{}
  \Performance{}

  \begin{code}
  --*/


/*--
\end{code}

\funparagraph{gwo\_read\_packet}  % The name of the function(s)
\ImplementationNotes{ 
This is the most critical function of the package. The implementation
is rather convoluted as the function was designed to work under
both the Linux operationg system and under DOS. Under Linux the
critical section of the function was moved in a kernel space
device driver where it can run without being interrupted. A write
to the device sends the request code and the read returns a 32 byte
packet. While this approach breaks the semantic of the system calls,
it reuses most of the code of the DOS implementation and hopefully
this improves the mentenability of the software.
The handle used to access the device is passed in the
global variable \variabletag{holter\_handle}.
}
\Portability{}
\Changes{}

\begin{code}
--*/

 
#define GWO_GET_PACKET_VERBOSE

int gwo_read_packet( int request_code /* the command code to send to gwo */
		    , char * buffer /* storage for the packet */
		    , int len	/* dimension of the above buffer */
		    , int retries /* number of retries before renouncing */
		    , int serial_port /* ... */
		    ) {
  int i				/* local index in the buffer */
    , result			/* of serial operations */
    , retry;			/* # of the actual retry */
  
  
#ifndef LINUX_DEVICE
  for( retry = 0; receive_byte(serial_port ) > 0; retry++ );
  if(retry > 0) {
    printf( "Warning: %d pending bytes in holter's buffers.\n", retry );
    fflush( stdout );
  }
#endif
  for( retry=0; retry<retries; retry++ ) {
#ifdef LINUX_DEVICE
    if( holter_handle == -1 ) return -1;
    buffer[0] = (char)request_code;
    if( write( holter_handle, buffer, 1 ) != 1 ) return -1;
    result = read( holter_handle, buffer, len );
    if( result != len ) result = -1;
#else 
    _disable();
    result = gwo_send_byte(request_code, serial_port );
    if( result < -1 ) {
#ifdef GWO_GET_PACKET_VERBOSE
      printf( "Error while sending command.\n" );
#endif
      _enable();
      return -1;
    }
    
    for(i=0; i<len && result>=0; i++) {
      result  = gwo_get_byte(serial_port);
      buffer[i] = result;
      }
    _enable();
#endif /* LINUX_DEVICE */
    if( result>=0 ) {
#ifdef GWO_GET_PACKET_VERBOSE
      if( retry ) printf( "We *made* it, Yogi!\n" );
#endif
      return 0;	/* allright */
    }
#ifdef GWO_GET_PACKET_VERBOSE
    printf(
      "Warning: request %x at attempt %d got %d bytes instead of %d, error code %d.\n"
	   , request_code, retry, i, len, result );
#endif
  }
  return result;		/* return the error code */
}




/*--
       \end{code}

       \funparagraph{gwo_read_addressed_packet}
     
       Reads a packet from holter, if some error occurs the address
       is retransmitted.

       \PurposeReturn{0 ok, -1 error}
       \Exceptions{}
       \GlobalEffects{}
       \Preconditions{The transfer address must be properly set 
       for the first and (hopefully) only attempt.}
       \Performance{}

       \begin{code}
       --*/
     

int gwo_read_addressed_packet( long address, int code
			      , char * buffer, int len, int retries 
			      , int serial_port ) {
  /* address --- holter adress used if the first attempt failed
   * buffer  --- storage for the packet
   * len     --- dimension of the previous buffer
   * retries --- number of retries
   * serial_port --- number of the serial port 
   */
  int result, retry;

  
  result = gwo_set_address( address, serial_port );
  if(result<0) {
#ifdef GWO_READ_ADDRESSED_PACKET_VERBOSE
    printf( "Error while resending the start address, aborting.\n" );
#endif
    return result;
  }
  for( retry=0; retry<retries; retry++ ) {
    result = gwo_read_packet( code, buffer, len, 1, serial_port );

    if( result>=0 ) {
      return 0;			/* allright */
    }
#ifdef GWO_READ_ADDRESSED_PACKET_VERBOSE
    printf( "Data read error at attempt %d got %d bytes instead of %d.\n"
	   , retry, i, len );
#endif
    result = gwo_set_address( address, serial_port );
    if(result<0) {
#ifdef GWO_READ_ADDRESSED_PACKET_VERBOSE
      printf( "Error while resending the start address, aborting.\n" );
#endif
      return result;
    }
  }
  return result;		/* return the error code */
}



/*--
  \end{code}

  \subsection{gwo\_read --- level 3 gwo communication functions}

  \PackagePurpose{Read holter information in specific structures.}
  \FunctionInterraction{
  Holter identification data is read through the function
  \functiontag{gwo\_read\_info}, internal holter variables through
  the function{gwo\_read\_ram}, patient data through 
  \functiontag{gwo\_read\_data}
  and an obsoleted way to get holter status information is the function
  \functiontag{gwo\_read\_status}.
  }
  \Exceptions{}
  \GlobalEffects{}
  \Preconditions{The serial port must be properly initialized}
  \Performance{}

  \begin{code}
  --*/

/*--
  \end{code}

  \funparagraph{holter_short, holter_long}
  \ImplementationNotes{}
  \Portability{
  The holter is assumed to be big endian, the host is assumed to
  be little endian}
  \Changes{Both functions must be completely rewritten if the above
  assumption changes, but this is not difficult.}

  \begin{code}
  --*/



long holter_long( uchar * access ) {
  long temp;
  
  temp = 0;
  temp += access[0]; temp = temp << 8;
  temp += access[1]; temp = temp << 8;
  temp += access[2]; temp = temp << 8;
  temp += access[3];
  return temp;
}

long holter_short( uchar * access ) {
  long temp;
  
  temp = 0;
  temp += access[0]; temp = temp << 8;
  temp += access[1];
  return temp;
}




/*--
  \end{code}

  \dataparagraph{Size of communication packets}

  \Definition{}
  \SemanticStructure{}
  \ConsistencyRules{}

  \begin{code}
  --*/
#define GWO_SS 32
#define GWO_DS 32



/*--
  \end{code}

  \funparagraph{gwo_read_data}
  \ImplementationNotes{That's easy.}
  \Portability{}
  \Changes{}

  \begin{code}
  --*/

int gwo_read_data( char * buffer, int len, long address, int serial_port ) {

  return gwo_read_addressed_packet( address , GWO_READ_DATA
				   , buffer , len
				   , GWO_PACKET_RETRY
				   , serial_port );
}

/*--
  \end{code}

  \subsubsection{holter\_ram}
  \PackagePurpose{Access holter internal variables in order to
  obtain status information.}
  \FunctionInterraction{
  \functiontag{gwo\_read\_ram} is used to obtain a snapshot of holter's
  internal memory, access components through \functiontag{holter\_long}
  or \functiontag{holter\_short} --- they take into account differences
  in byte orderings between the holter and the host.
  }
  \Exceptions{}
  \GlobalEffects{}
  \Preconditions{The holter must be initialized.}
  \Performance{}

  \begin{code}
  --*/


/*--
  \end{code}

  \dataparagraph{holter_ram}  % a name for the data structure

  \Definition{This structure is an exact image of the
  internal variables of the holter.}
  \SemanticStructure{}
  \ConsistencyRules{
  Two and four bytes members must be accessed
  through the functions \function{holter_short} and \function{holter_long}.
  }

  \begin{code}
  --*/
#define HLTR_FIFOMK 0x3f
#define HLTR_DDATAMK 0x1f
#define HOLTER_RAM_SIZE 256

struct holter_ram {
  uchar fifo[HLTR_FIFOMK+1];
  uchar ddata[HLTR_DDATAMK+1];
  uchar fsp;
  uchar fep;
  uchar dfsp;
  uchar dfep;
  uchar dsaved;
  uchar addr1[4];
  uchar addr2[4];
  uchar ctimer[4];
  uchar btnad[4];
  uchar status;
  uchar savaddr[4];
  uchar blkst;
  uchar btncnt;
  uchar btntim;
  uchar btnst;
  uchar btnpr;
  uchar tdbytes;
  uchar tsbytes;
  uchar tabytes;
  uchar sbptr;
  uchar adptr;
  uchar lcntch[2];
  uchar ledst;
  uchar qrsdata[8];
  uchar cstate;
  uchar cmeta;
  uchar casval;
  uchar casnbb;
  uchar fval[2];
  uchar srerror;
  uchar fill[255];
};


/*--
  \end{code}

  \funparagraph{gwo\_read\_ram}
  \PurposeReturn{Returns 0 if ok, negative value if puzzled. Places
  a binary image of the holter internal ram in the given memory buffer.
  }
  \Exceptions{  }
  \GlobalEffects{}
  \Preconditions{The serial interface should be initialized.}
  \Performance{}

  \begin{code}
  --*/

int gwo_read_ram( struct holter_ram * ram, int serial_port ) {
  
  int address;			/* actual address read */
  int result;			/* of the serial operation */
  char buffer[ GWO_DS ];	/* buffer used to read a packet */

  for( address = 0; address < HOLTER_RAM_SIZE; address += GWO_DS) {
    result = gwo_set_address( address, serial_port );
    if( result<0 ) return result;
    result = gwo_read_addressed_packet( address , GWO_READ_RAM 
				       , buffer , GWO_DS
				       , GWO_PACKET_RETRY
				       , serial_port );
    if( result < 0 ) return result;
    bcopy( buffer, (char*)ram + address
	  , min(HOLTER_RAM_SIZE-address,GWO_DS) );
  }

  return 0;   /* i.e. OK */
}


/*--
  \end{code}

  \subsubsection{holter_info}
  \PackagePurpose{Provides an encapsulated data type to access
  EEPROM stored identification information about the holter.}
  \FunctionInterraction{
  Call \functiontag{gwo\_read\_info} in order to obtain a
  snapshot of the relevant holter EEPROM location. Access
  directly the information in the structure; use \functiontag{holter\_long}
  in order to access a four byte value and \functiontag{holter\_short}
  in order to access a two byte value.
  }
  \Exceptions{
  }
  \GlobalEffects{  }
  \Preconditions{The holter must be properly initialized.}
  \Performance{
  }

  \begin{code}
  --*/

/*--
  \end{code}

  \dataparagraph{holter_info}  % a name for the data structure

  \Definition{EPROM stored identification information.}
  \SemanticStructure{}
  \ConsistencyRules{
  The information must be read through the function \function{gwo_read_info}.
  Two bytes members must be accessed through \function{holter_short}.
  }

  \begin{code}
  --*/

struct holter_info {
  uchar sampling_period[2];	/* in 10^-6 sec */
  uchar resolution;		/* in bits */
  uchar file_format;
  uchar compression_format;
  uchar version;			/* version of the holter program */
  uchar patch_level;		/* of the holter program */
  uchar magic;			/* some magic number --- 77 */
  uchar fill[32];
};



/*--
  \end{code}

  \funparagraph{gwo_read_info}


  \PurposeReturn{0 if ok, negative if error}
  \Exceptions{}
  \GlobalEffects{}
  \Preconditions{}
  \Performance{}

  \begin{code}
  --*/

int gwo_read_info( struct holter_info * info, int serial_port ) {
  short r;

  r = gwo_read_packet( GWO_READ_INFO, (char*)info, 32
		      , GWO_PACKET_RETRY, serial_port );
   
  if( (r<0) || (info->magic != 77) ) {
#ifdef GWO_READ_INFO_VERBOSE
    fprintf( stderr, "Something wrong in info.\n" );
#endif
    return r;
  }

  return 0;   /* i.e. OK */
}



/*--
  \end{code}

  \subsubsection{holter\_status}
  \PackagePurpose{Get some status information from the holter. This is an
  obsoleted package.}
  \FunctionInterraction{
  Call \functiontag{gwo\_read\_status} and access directly the global
  variables provided by the package.
  }
  \Exceptions{  }
  \GlobalEffects{
  A call to \functiontag{gwo\_read\_status} will affect all the
  exported global variables.
  }
  \Preconditions{The holter should be properly initialized.}
  \Performance{  }

  \begin{code}
  --*/



/*--
  \end{code}

  \dataparagraph{Status variables}  % a name for the data structure

  \Definition{ Those variables are used to export holter status information.
  Only the first variable -- addr1 -- is correct and usefull.}
  \SemanticStructure{  }
  \ConsistencyRules{  }

  \begin{code}
  --*/


unsigned char gwo_status[GWO_SS]; /* Buffer to keep the read status */

long addr1;			/* fill pointer in the record ram */
long addr2;			/* transfer related variables */
short tdbytes;			/* number of data bytes to transfer */
short tsbytes;			/* number of status bytes to transfer */
short tabytes;			/* number of address bytes to transfer */
short sbptr;			/* pointer used in a status read operation */
short adptr;			/* pointer used in an address write operation */



/*--
  \end{code}

  \constantparagraph{Status buffer access}

  \Definition{Offsets in the status packet where various variables
  are located. Apart from addr1 those offsets might not be correct.
  }
  \SemanticStructure{  }
  \ConsistencyRules{  }

  \begin{code}
  --*/


#define ADDR1   0x00
#define ADDR2   0x04
#define TDBYTES 0x0c
#define TSBYTES 0x0d
#define TABYTES 0x0e
#define SBPTR   0x0f
#define ADPTR   0x10


/*--
  \end{code}

  \funparagraph{gwo_read_status}  % The name of the function(s)
  \ImplementationNotes{This is the first implementation of the
  status function. It is still kept here --- although it might not be used
  --- because it is a cleaner way to get status information from the holter.}
  \Portability{A little endian arhitecture is assumed.}
  \Changes{}

  \begin{code}
  --*/


int gwo_read_status( int serial_port )

  {
  short r;

  r = gwo_read_packet(  GWO_READ_STATUS, gwo_status
		      , GWO_SS, GWO_PACKET_RETRY, serial_port );
  if( r<0 ) {
    return r;
  }

  addr1 = *((unsigned char *)(gwo_status + ADDR1 + 1)) * 65536L +
    *((unsigned char *)(gwo_status + ADDR1 + 2)) * 256L +
    *((unsigned char *)(gwo_status + ADDR1 + 3));
  addr2 = *((unsigned char *)(gwo_status + ADDR2 + 1)) * 65536L +
    *((unsigned char *)(gwo_status + ADDR2 + 2)) * 256L +
    *((unsigned char *)(gwo_status + ADDR2 + 3));

  tdbytes = *((unsigned char *)(gwo_status + TDBYTES));
  tsbytes = *((unsigned char *)(gwo_status + TSBYTES));
  tabytes = *((unsigned char *)(gwo_status + TABYTES));
  sbptr = *((unsigned char *)(gwo_status + SBPTR));
  adptr = *((unsigned char *)(gwo_status + ADPTR));
  return 0;   /* i.e. OK */
  }


/*--
  \end{code}

  \subsection{ERUID package implementation}

  \ImplementationNotes{
  Time handling is quite clumsy (a printed string is re-parsed).
  }
  \Portability{Time related functions are available both under
  MSC6.0 and DJGPP.
  }
  \Changes{ % that are anticipated, if special provision was made for them
  }

  \begin{code}
  --*/



/*--
  \end{code}

  \dataparagraph{eruid_on_file}  % a name for the data structure

  \Definition{Binary structure for the representation of the
  eruid on file.
  }
  \SemanticStructure{
  }
  \ConsistencyRules{
  Most of the fields will be represented as ASCII strings in order to
  increase portability of the file.
  }

  \begin{code}
  --*/


struct eruid_on_file {		/* this is a structure imposed by A. Corlan */
  char site[16];		/* literal site identification */
  char host[5];			/* literal host identification */
  char year[2];			/* last two digits of the year */
  char month[2];		/* month in numerical format */
  char day[2];			/* day of the month */
  char hour[2];			/* hour in 00-24 format */
  char minute[2];		/* minute of the hour */
  char order[1];		/* used if all above match for two eruids */

  char eruid_version[2];
  char sampling_period[5];	/* in 10^-6 sec */
  char resolution[2];		/* in bits */
  char file_format[2];
  char file_dimension[9];	/* in octets */
  char checksum[4];
  char fill[9];
};



/*--
  \end{code}

  \dataparagraph{eruid_data}  % a name for the data structure

  \Definition{This is the representation of some of the ERUID fields
  in memory.}
  \SemanticStructure{
  % Description of structure and associated semantics
  % Full definition of the intented data types by means of ADA declarations
  % may be appropiate
  }
  \ConsistencyRules{
  At the end of the processing the relevant information is written
  in the structure \datastructure{eruid_on_file}.  }

  \begin{code}
  --*/

struct eruid_data {
  long lenght;			/* length of the file */
  char checksum[4];
  int checksum_ndx;		/* index of the character in the checksum */
  struct eruid_on_file on_file;
} the_eruid_data;


/* copy the value of an environment variable at the memory location */
void get_from_env( char * variable_name, char * location, int max_len ) {
  int i;			/* some index variable */
  char * value;			/* value of the environment variable */

  value = getenv( variable_name );
  if( value == NULL ) {
#ifdef GET_FROM_ENV_VERBOSE
    fprintf( stderr, "Warning, environment variable %s not defined\n"
	    , variable_name );
#endif
    for( i=0; i<max_len; i++ ) location[i]=0;
  } else {
    if( strlen( value ) >= max_len ) {
#ifdef GET_FROM_ENV_VERBOSE
      fprintf( stderr
	      , "Warning, value too long for %s, truncated at %d characters.\n"
	      , variable_name, max_len);
#endif
    }
    strncpy( location, value, max_len );
  }
}

/* computes the numerical representation of a month */
static char * months[] = {
  "Jan", "Feb", "Mar", "Apr", "May", "Jun"
    , "Jul", "Aug" , "Sep", "Oct", "Nov", "Dec"
};

void month_to_number( char * month ) {
  int i;
  for( i=0; i<12; i++ ) {
    if( !strcmp( month, months[i] ) ) {
      sprintf( month, "%2d", i+1 );
      return;
    }
  }
}



/*--
  \end{code}

  \funparagraph{eruid_set_time}
  Update ERUID information related to the start of the recording.
  \ImplementationNotes{Literal time representation is re-parsed.}
  \Portability{Time functions used are available both under MSC6.0 and
  Djgpp}
  \Changes{There should be a simpler way to obtain time information.}

  \begin{code}
  --*/


void eruid_set_time( struct eruid_data * eruidd	/* structure to modify */
		    , time_t nb_seconds /* length of the record in seconds */
		    ) {
  time_t  now;
  /* pointer used in the parsing process */
  char *full_date, *year, *month, *day, *hour, *minute;

  
  time( &now );		/* in now number of seconds since 1/1/70 */
  now -= nb_seconds;
  full_date = ctime( &now );
  /* decompose the string */
  month = full_date+4; month[3]=0; month_to_number( month );
  day = month + 4; day[2]=0;
  hour = day + 3; hour[2]=0;
  minute = hour + 3; minute[2]=0;
  year = minute + 6; year[4]=0;
  /* place time information in the eruid */
  memcpy( eruidd->on_file.year, year+2, 2); /* get the last two digits */
  memcpy( eruidd->on_file.month, month, 2);
  memcpy( eruidd->on_file.day, day, 2);
  memcpy( eruidd->on_file.hour, hour, 2);
  memcpy( eruidd->on_file.minute, minute, 2);
  eruidd->on_file.order[1]=0;	/* what is this used for ?*/
}


/*--
  \end{code}

  \funparagraph{eruid_init}
  Place information in the ERUID  data structure.
  \ImplementationNotes{There is no holter dependent functions
  in the implementation, the apropriate information is passed by the caller.
  The function accesses environment variables.}
  \Portability{}
  \Changes{}

  \begin{code}
  --*/


void eruid_init(  struct eruid_data * eruidd /* structure to modify */
		/* data to place in the structure */
		, int sampling_period
		, int resolution
		, short file_format
		, short compression_format
		, short nb_seconds /* since the start of the record */
		) {

  /* initialize the processing part */
  eruidd -> lenght = 0;
  eruidd -> checksum[0] = 0;
  eruidd -> checksum[1] = 0;
  eruidd -> checksum[2] = 0;
  eruidd -> checksum[3] = 0;
  eruidd -> checksum_ndx = 0;

  /* initialize the permanent part */
  /* from the environment */
  get_from_env( "ECGN_SITE", eruidd -> on_file . site,  16 );
  get_from_env( "ECGN_HOST", eruidd -> on_file . host, 5 );

  /* place data sent by the caller */
  sprintf( eruidd->on_file.sampling_period
	  , "%5d"
	  ,  sampling_period);
  sprintf( eruidd->on_file.resolution
	  , "%2d"
	  , resolution) ;
  eruidd->on_file.file_format[0]= file_format;
  eruidd->on_file.file_format[1]
    = compression_format;
  eruid_set_time( eruidd, nb_seconds );
}

#ifdef USE_ERUID_UPDATE
/*  call this function before writing a holter packet to disk
 * in order to update information in the eruid
 */
void eruid_update( struct eruid_data * data
		  , char * read_block, int dimension ) {
  data->lenght += dimension;
  for( ; dimension; dimension --, read_block ++ ) {
    data -> checksum[data -> checksum_ndx ] ^= read_block[0];
    data -> checksum_ndx ++;
  }
}
#endif /* USE_ERUID_UPDATE */

/*--
  \end{code}

  \funparagraph{eruid_write}
  Write to the file the information related to the eruid.
  It is assumed that the file contains valuable information
  and that the first 64 bytes were left free for the ERUID.
  \ImplementationNotes{The information gathered in written
  in the {\tt on_file} structure. All the structure is then
  written to the disk with a \function{write} call in order
  to preserve the contents of the file.}
  \Portability{The memory representation must match the file
  representation. Some simple sanity check is made as an error
  would render useless the first information bytes in the file.}
  \Changes{}

  \begin{code}
  --*/

void eruid_write( char * filename, struct eruid_data * data ) {
  int handle;
  struct eruid_on_file * eruid;

  /* sanity check */
  if( sizeof( struct eruid_on_file ) != 64 ) {
    fprintf( stderr, "Internal error, wrong dimension of eruid structure.\n" );
    return;
  }

  eruid = &(data->on_file);
  /* move temporary information to the proper location */
  memcpy( eruid->checksum, data->checksum, 4);
  sprintf( eruid->file_dimension, "%9ld", data->lenght );

  /* all done, write the eruid to file */
  handle = open( filename, O_WRONLY );
  if( handle == -1 ) {
    fprintf( stderr, "Could not open file to write eruid.\n" );
    exit( 1 );
  }
  if( write( handle, eruid, 64 ) == -1 ) {
    fprintf( stderr, "" );
  }
}


/*--
  \end{code}

  \section{e2100 implementation}

  \ImplementationNotes{  }
  \Portability{  }
  \Changes{   }

  \begin{code}
  --*/


/*--
  \end{code}

  \dataparagraph{HOLTER}  % a name for the data structure

  \Definition{Encapsulates all the information necessary to handle
  a holter connection and data transfer.}
  \SemanticStructure{
  \begin{description}
  \item {record_ram_address} pointer in the holter record ram,
  \item {serial_port} number of the serial port, 0 for {\tt COM1},
  2 for {\tt COM2}, 3 for {\sl Electrocardiognosis'} internal
  high speed interface. This structure member was added in the
  hope that sometimes more than one holter will be connected to
  the same host $\ldots$
  \item {info}  holter ROM stored information;
  \item {ram}   holter internal RAM --- used to get variables;
  \item {eruid} Dan's structure for file identification;
  incremented after each call to \function{e2100_read}
  \end{description}
  }
  \ConsistencyRules{}

  \begin{code}
  --*/

struct holter {
  long record_ram_address;   
  int serial_port;
  struct holter_info info;
  struct holter_ram ram;
  struct eruid_data eruid;
};




/*--
  \end{code}

  \dataparagraph{baud\_rates}  % a name for the data structure

  \Definition{Default baud rates to be used on the serial port. As the
  transfer speed is hardwired in the holter it is not particularly usefull
  to have this customization here.}
  \SemanticStructure{  }
  \ConsistencyRules{  }

  \begin{code}
  --*/
/* default baud rates for the each serial port */
static long baud_rates[] = { 19200L, 19200L, 76800L };


/*--
  \end{code}

  \dataparagraph{Exception management} 

  \Definition{   }
  \SemanticStructure{  }
  \ConsistencyRules{  }

  \begin{code}
  --*/

int e2100_exception;
#define E2100_no_holter           1
#define E2100_communication_error 2
#define E2100_no_memory           3

#define raise(exception,return_code) {e2100_exception=exception;\
					return return_code;}

/*--
  \end{code}

  \funparagraph{e2100_open}
  \PurposeReturn{A pointer to a E2100 structure, NULL if error occured}
  }
  \Exceptions{E2100_no_holter, E2100_no_memory}
  \GlobalEffects{
  }
  \Preconditions{
  }
  \Performance{
  }

  \begin{code}
  --*/

struct holter * e2100_open( int serial_port ) {
  struct holter * the_holter;

  /* alloc memory for the result */
  the_holter = malloc(sizeof(struct holter));
  if( the_holter == NULL ) raise(E2100_no_memory, NULL);
  
  the_holter -> serial_port = serial_port;

  /* initialize the serial port */
  gwo_init( serial_port, baud_rates[serial_port] );
  /* read holter identification */
  if( gwo_read_info( &(the_holter->info), serial_port ) < 0 ) {
    raise(E2100_no_holter,NULL);
  }
  /* read holter internal variables */
  if( gwo_read_ram( &(the_holter->ram), serial_port ) < 0 ) {
    raise(E2100_communication_error, NULL);
  }
  /* set the initial value of the  record_ram_address */
  the_holter -> record_ram_address = gwo_set_address( 0L, serial_port );
  if( the_holter -> record_ram_address < 0 ) {
    raise( E2100_communication_error, NULL);
  }
  /* build the eruid */
  eruid_init( &(the_holter->eruid)
	     , atol(the_holter -> info . sampling_period)
	     , the_holter -> info . resolution
	     , the_holter -> info . file_format
	     , the_holter -> info . compression_format
	     , e2100_nb_seconds( the_holter )
	     );
 
  return the_holter; /* OK */
}


/*--
  \end{code}

  \funparagraph{e2100\_refresh}
  \ImplementationNotes{  }
  \Portability{  }
  \Changes{  }

  \begin{code}
  --*/

int e2100_refresh( struct holter * the_holter ) {
  /* read holter internal variables */
  if( gwo_read_ram( &(the_holter->ram), the_holter -> serial_port ) < 0 ) {
    raise(E2100_communication_error, 0);
  }
  return 1;
}



/*--
  \end{code}

  \funparagraph{e2100_set_address}
  \ImplementationNotes{}
  \Portability{}
  \Changes{}

  \begin{code}
  --*/

int e2100_set_address( struct holter * this /* object to operate on */
		      , long address /* in the record RAM, in bytes */
		      )  {
  if( gwo_set_address( address, this->serial_port ) < 0 ) {
    this->record_ram_address = -1; /* mark undefined */
    raise( E2100_communication_error, -1);
  }
  this -> record_ram_address = address;
  return 0;
}

/*--
  \end{code}

  \funparagraph{e2100\_refresh\_address}
  \ImplementationNotes{}
  \Portability{}
  \Changes{}

  \begin{code}
  --*/

int e2100_refresh_address( struct holter * this /* object to operate on */
		      )  {
  if( gwo_set_address( this->record_ram_address, this->serial_port ) < 0 ) {
    raise( E2100_communication_error, -1);
  }
  return 0;
}


/*--
  \end{code}

  \funparagraph{e2100_read_block}
  \ImplementationNotes{
  The block buffer passed as an argument must point to a slice
  of memory at least {\tt E21OO_BLOCK_SIZE} long.
  }
  \Portability{{}
  \Changes{}

  \begin{code}
  --*/

int e2100_read_block( struct  holter * this, void * block ) {
  int result; /* of the last serial operation */
  result = gwo_read_data( block, E2100_BLOCK_SIZE
		  , this->record_ram_address, this->serial_port);
  if( result < 0 ) {
    raise( E2100_communication_error, -1);
  }
  this->record_ram_address += E2100_BLOCK_SIZE;
  return 0;
}

int e2100_read_blocks( struct holter * this, void * block, int nb) {
  int result; /* of the last serial operation */
  result = gwo_read_data( block, E2100_BLOCK_SIZE * nb
		  , this->record_ram_address, this->serial_port);
  if( result < 0 ) {
    raise( E2100_communication_error, -1);
  }
  this->record_ram_address += E2100_BLOCK_SIZE * nb;
  return 0;
}

/*--
  \end{code}

  \funparagraph{e2100\_sampling\_period}
  % The name of the function is expected to be prezent
  % in the sectioning element above

  \PurposeReturn{The sampling period in microseconds.}
  \Exceptions{  }
  \GlobalEffects{}
  \Preconditions{The holter must be initialized.}
  \Performance{}

  \begin{code}
  --*/

short e2100_sampling_period( struct holter * this ) {
  return holter_short( this->info.sampling_period );
}


/*--
  \end{code}

  \funparagraph{e2100\_resolution}
  % The name of the function is expected to be prezent
  % in the sectioning element above

  \PurposeReturn{The resolution of the holter in bits.}
  \Exceptions{  }
  \GlobalEffects{  }
  \Preconditions{ The holter must be initialized.  }
  \Performance{  }

  \begin{code}
  --*/


short e2100_resolution( struct holter * this ) {
  return ( this->info.resolution );
}


/*--
  \end{code}

  \funparagraph{e2100\_nb\_seconds}
  \PurposeReturn{ The number of seconds elapsed since the holter start.
  The returned value is based on data read by \functiontag{e2100\_init}
  or by \functiontag{e2100\_refresh}.}
  \Exceptions{  }
  \GlobalEffects{  }
  \Preconditions{ The  holter must be initialized.}
  \Performance{  }

  \begin{code}
  --*/


long e2100_nb_seconds( struct holter * holter ) {
  return( (float)holter_long( holter->ram . ctimer )
	 *  (float)e2100_sampling_period( holter )
	 / 1e6);
}

/*--
  \end{code}

  \funparagraph{e2100_read_fill}
  \ImplementationNotes{}
  \Portability{}
  \Changes{}

  \begin{code}
  --*/

long e2100_read_fill( struct holter * this ) {
  return holter_long(this -> ram . addr1);
}

/*--
  \end{code}

  \funparagraph{e2100_read_size}
  \ImplementationNotes{}
  \Portability{}
  \Changes{}

  \begin{code}
  --*/


long e2100_read_size( struct holter * this ) {
  return holter_long( this -> ram . ctimer);
}


/*--
  \end{code}

  \funparagraph{e2100_close}  % The name of the function(s)
  \ImplementationNotes{  }
  \Portability{ }
  \Changes{ }

  \begin{code}
  --*/

void e2100_close( struct holter *  this ) {
  free( this );
}

/*--
  \end{code}

  \funparagraph{e2100_transfer_file}  % The name of the function(s)
  \ImplementationNotes{  }
  \Portability{ }
  \Changes{ }

  \begin{code}
  --*/

void e2100_transfer_file( struct holter * the_holter, FILE * the_file ) {
  long nb, nbt;			/* number of blocks */
  int terr;			/* number of temporary errors */
  char buffer[ E2100_BLOCK_SIZE ];
  char test_buffer[ E2100_BLOCK_SIZE ];
  long address;

  nb = ((e2100_read_fill(the_holter) - 1) / E2100_BLOCK_SIZE) + 1;
  nbt = nb;
  terr = 0;
  address = 0L;
  e2100_set_address( the_holter, address );
  for( ; nb;  ) {
#ifdef PRINT_PROGRESS_MESSAGE
    if( (nb % 100 == 0 ) ) {
      char message[64];
      printf( message, " Transfer %2d%% completed %3d temporary errors."
	      , ( (nbt - nb) * 100 ) / nbt, terr);
    }
#endif
    if( e2100_read_block( the_holter, (void*)buffer ) == 0 ) {
      nb --;
      fwrite( buffer, E2100_BLOCK_SIZE, 1, the_file);
    } else {
#ifdef PRINT_PROGRESS_MESSAGE
      printf( "Temporary error in block %d.\n", nbt-nb );
#endif
      while( e2100_refresh_address( the_holter ) != 0);
      terr ++;
    }
  }
#ifdef PRINT_PROGRESS_MESSAGE
  printf( "Transfer done.");
#endif  
}

/*--
  \end{code}

  \endoffile

  --*/

     

