
/*--

  \FileTitle{e2100.h  --- e2100 communication package}
  \CopyrightNotice{Copyright \copyright 1995 Marius SERITAN

  \begin{verbatim}
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  \end{verbatim}
  }
  
  \AuthorAffiliation{
  Electrocardiognosis SRL, 28 Argentina str, 71206, Bucharest 1, ROMANIA
  }
  \RevisionHistory{
  \begin{description}
  \item{19 Jan 1994}
  Initial release under Linux SVGA, however this file should not be system
  dependent.
  \end{description}
  }
  \vskip 1in

  \begin{code}
  --*/




/*--
  \end{code}

  \section{e2100}
  \PackagePurpose{Connects to the holter in order to obtain usefull information
  and to transfer the recorded signal.}
  \FunctionInterraction{
  \begin{itemize}
  \item \functiontag{e2100\_open} and \functiontag{e2100\_close}
  are the constructor and the destructor of the \typetag{HOLTER} 
  encapsulated type. Repeated close/open sequences will not affect
  patient information (signal and events) stored in the holter.
  \item \functiontag{e2100\_set\_address} and \functiontag{e2100\_read\_block}
  are used to transfer information.
  \item \functiontag{e2100\_read\_fill} and the like are accessors.
  \item \functiontag{e2100\_refresh} is used to get up to date information
  for dinamic values like {\tt fill}, {\tt size} and {\tt nb\_seconds}.
  \end{itemize}
  }
  \Exceptions{No exceptions are raised. However a NULL return from 
  \functiontag{e2100\_open} indicates a severe error and a negative error
  from \functiontag{e2100\_read\_block} indicates a temporary read error.
  }
  \GlobalEffects{Each call to \functiontag{e2100\_open} resets the value
  of some holter internal variables. The variables affected are not
  directly accessible to the user (for example the address of the last record
  memory accessed).}
  \Preconditions{None, this package takes care of him self.}
  \Performance{Small transfer blocks are used in order to improve the
  transfer reliability and this might decrease the overall transfer rate.
  This rate is limited at the  moment by the holter hardware and
  software performance.}

  \begin{code}
  --*/


/*--
  \end{code}

  \dataparagraph{HOLTER}  % a name for the data structure

  \Definition{Encapsulated type holding all the information necessary
  for a connection to the holter.}
  \SemanticStructure{}
  \ConsistencyRules{}

  \begin{code}
  --*/

typedef struct holter * HOLTER;

/*--
\constantparagraph{E2100_BLOCK_SIZE}
\Definition{Dimension of a block read from the holter.}
--*/
#define E2100_BLOCK_SIZE 32


/*--
  \end{code}

  \funparagraph{e2100\_open}
  \PurposeReturn{A properly initialized variable of the type
  \typetag{HOLTER} or NULL if no holter was detected. Accessors
  can be applied on the returned variable. Dinamic values accessed
  in these way (number of seconds since the start of the record for
  example) are not updated afterwards from the holter. 
  }
  \Exceptions{}
  \GlobalEffects{Resets some holter internal variables.}
  \Preconditions{None}
  \Performance{}

  \begin{code}
  --*/
HOLTER e2100_open( int serial_port );


/*--
  \end{code}

  \funparagraph{e2100\_close}
  \PurposeReturn{None}
  \Exceptions{None}
  \GlobalEffects{Perform various housekeeping actions on the HOLTER variable.
  Of course it is not safe to use the variable afterwards.}
  \Preconditions{None}
  \Performance{}

  \begin{code}
  --*/

void e2100_close( HOLTER this );



/*--
  \end{code}

  \funparagraph{e2100\_refresh}
  \PurposeReturn{0 if OK, -1 if some communication error occured.}
  \Exceptions{}
  \GlobalEffects{Updates the dinamic information of the HOLTER variable.}
  \Preconditions{The holter should be initialized.}
  \Performance{  }

  \begin{code}
  --*/
int e2100_refresh( HOLTER this );

/*--
  \end{code}

  \funparagraph{e2100\_set\_address}
  \PurposeReturn{0 if OK, -1 if some communication error occured.}
  \Exceptions{}
  \GlobalEffects{Sets the transfer address (in bytes) from where the
  next block  will be read. Usually it is a multiple of 
  \constanttag{E2100\_BLOCK\_SIZE}. The holter and the package keep
  syncronized copies of that variable.}
  \Preconditions{The holter should be initialized.}
  \Performance{}

  \begin{code}
  --*/

int e2100_set_address( HOLTER this, long address);


/*--
  \end{code}

  \funparagraph{e2100\_read\_block}
  \PurposeReturn{0 if ok, -1 if some communication error occured.
  The read block will be placed in memory at the given location.}
  \Exceptions{}
  \GlobalEffects{The package internal transfer address will be updated.}
  \Preconditions{The memory block should be at least
  \constanttag{E2100\_BLOCK\_SIZE} long. The holter should be initialized.}
  \Performance{}

  \begin{code}
  --*/

int e2100_read_block( HOLTER this, void * block );


/*--
  \end{code}

  \funparagraph{e2100\_sampling\_period}
  \PurposeReturn{The sampling period of the holter expressed in microseconds.}
  \Exceptions{}
  \GlobalEffects{}
  \Preconditions{The holter must be initialized}
  \Performance{}

  \begin{code}
  --*/

short e2100_sampling_period( HOLTER this );


/*--
\end{code}

\funparagraph{e2100\_resolution}
\PurposeReturn{The resolution in bits of the acquired signal}
\Exceptions{
}
\GlobalEffects{ }
\Preconditions{ }
\Performance{ }

\begin{code}
--*/

short e2100_resolution( HOLTER this );


/*--
\end{code}

\funparagraph{e2100\_read\_fill}
\PurposeReturn{The number of bytes occupied by the signal in the holter's 
record ram. This value is updated by the open and the refresh functions.}
\Exceptions{ }
\GlobalEffects{ }
\Preconditions{ }
\Performance{ }

\begin{code}
--*/

long e2100_read_fill( HOLTER this );


/*--
\end{code}

\funparagraph{e2100\_read\_size}
\PurposeReturn{ The total number of samples stored in holter's record ram.
Warning: two samples are generated each sampling period. This value is
updated by the open and refresh  functions.}
\Exceptions{ }
\GlobalEffects{ }
\Preconditions{ }
\Performance{ }

\begin{code}
--*/

long e2100_read_size( HOLTER this );


/*--
\end{code}

\funparagraph{e2100\_nb\_seconds}
\PurposeReturn{The number of seconds elapsed since the start of the record
(the holter starts to record as soon as it is turned on).}
\Exceptions{ }
\GlobalEffects{ }
\Preconditions{ }
\Performance{ }

\begin{code}
--*/

long e2100_nb_seconds( HOLTER this );
/*--
\end{code}

\funparagraph{e2100\_transfer_file}
\PurposeReturn{Nothing}
\Exceptions{ }
\GlobalEffects{Transfers the contents of the holter in the given file.}
\Preconditions{ }
\Performance{ }

\begin{code}
--*/

void e2100_transfer_file( struct holter *, FILE *);

/*--
  \end{code}

  % MOVE THIS COMMENT TO THE END OF THE FILE
  \endoffile

  --*/
