
/*--
  \end{code}

  \section{Holter user interface}
  \PackagePurpose{Provides a full screen graphical user interface
  to the holter communication program.}
  \FunctionInterraction{
  \begin{itemize}
  \item call
  \function_tag{hui\_main} and go home.
  \end{itemize}
  A data structure has to be defined in order to transmit relevant information
  to the caller.
  }
  \Exceptions{}
  \GlobalEffects{The user might request the creation of a file.}
  \Preconditions{The \packagetag{wiport2} package should be initialized.}
  \Performance{}

  \begin{code}
  --*/



/*--
  \end{code}

  \section{Holter user interface}
  \ImplementationNotes{}
  \Portability{This package relies upon the \packagetag{wiport2} package
  for the graphical capabilities and upon the \packagetag{e2100} package for
  communication capabilities.
  }
  \Changes{}

  \begin{code}
  --*/

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#ifdef WIPORT_X11
#include "wiport2x.h"
#else
#include "wiport2.h"
#endif

#include "e2100.h"




/*--
  \end{code}

  \dataparagraph{Global variables}  % a name for the data structure

  \Definition{Data associated with the status info.}
  \SemanticStructure{}
  \ConsistencyRules{}

  \begin{code}
  --*/
static GWIN * hui_status_gwin;
static short hui_status_dwin;


void hui_status_print( char * text ) {
  if( hui_status_gwin!=NULL ) {
    cleardwin( hui_status_gwin, hui_status_dwin, LOWERED_DWIN );
    gtext( hui_status_gwin, hui_status_dwin, 3, 3, text );
  }
  else
    puts( text );
}


void hui_show_progress( char * text, int ratio, int width )  {
   const prg_y0 =  7;		       /* y start of the progress ind. */
   const prg_y1 = 14;		       /* y end   of the progress ind. */
   
   if( hui_status_gwin != NULL )  {
      int x0;
      cleardwin( hui_status_gwin, hui_status_dwin, LOWERED_DWIN );
      gtext( hui_status_gwin, hui_status_dwin, 3, 3, text );
      x0 = 20 + strlen(text) * font_x_size;
      wi_line( hui_status_dwin, x0-2, prg_y0 - 2, x0 + width + 4, prg_y0 - 2);
      wi_line( hui_status_dwin, x0-2, prg_y1 + 2, x0 + width + 4, prg_y1 + 2);
      wi_line( hui_status_dwin, x0-2, prg_y0 - 2, x0 - 2, prg_y1 + 2);
      wi_line( hui_status_dwin, x0+width+4, prg_y0 - 2, x0 + width + 4, prg_y1 + 2);
      wi_rect( hui_status_dwin, x0, prg_y0, (ratio * width)/100, prg_y1 - prg_y0);
   }
}

void hui_show_holter_info( GWIN * gg, short ldwin, HOLTER the_holter) {
  char buffer[1024];
  
  sprintf( buffer, "Sampling period:             %6d microsec"
	  , e2100_sampling_period( the_holter ));
  if( gg!= NULL )
    gtext( gg, ldwin, 10, 10, buffer );
  else
    puts( buffer );
  sprintf( buffer, "Vertical resolution:         %6d bits"
	  , e2100_resolution( the_holter ));
  if( gg!=NULL )
    gtext( gg, ldwin, 10, 25, buffer );
  else
    puts( buffer );
}

void hui_show_holter_status( GWIN * gg, short ldwin, HOLTER the_holter) {
  char buffer[1024];
  char act_y;			/* keep track of the vertical position */

  act_y = 40;
#if 0
  sprintf( buffer, "Dimension of the record: %10ld samples."
	  , e2100_read_size( the_holter ) );
  if( gg!= NULL )
    gtext( gg, ldwin, 10, act_y, buffer );
  else
    puts( buffer );
  act_y += 15;
#endif

  sprintf( buffer, "Modulul de pacient functioneaza de %f4.2 ore."
	  , e2100_nb_seconds( the_holter ) / 3600);
  if( gg!=NULL )
    gtext( gg, ldwin, 10, act_y, buffer );
  else 
    puts( buffer );
  act_y += 15;


  sprintf( buffer, "Memoria interna este umpluta la %f3.1\% din capacitate." 
	  , e2100_read_fill( the_holter ) / (8 * 1048576L));
  if( gg!= NULL )
    gtext( gg, ldwin, 10, act_y, buffer );
  else
    puts( buffer );
  act_y += 15;
}

HOLTER hui_get_the_holter( int serial_port ) {
  HOLTER the_holter;

  the_holter = e2100_open( serial_port ); /* internal high speed interface */
  if( the_holter == NULL ) {
    hui_status_print( "Cannot talk to the holter, check cable and `Update info'." );
  } else {
    hui_status_print( "Holter detected and initialized." );
  }
  return the_holter;
}

void transfer_file_with_progress( struct holter * the_holter
				 , FILE * the_file 
				 , int nb_blocks ) {
  long nb, nbt;			/* number of blocks */
  int terr;			/* number of temporary errors */
  char * buffer;
  long address;
  int status;			

  buffer = (char*)malloc( nb_blocks * E2100_BLOCK_SIZE );
  if( buffer == NULL ) {
    hui_status_print( "Internal memory error!" );
    sleep( 2 );
    return;
  }
  nb = ((e2100_read_fill(the_holter) - 1) / E2100_BLOCK_SIZE) + 1;
  nbt = nb;
  terr = 0;
  address = 0L;
  e2100_set_address( the_holter, address );
  for( ; nb;  ) {
    int chunk;			/* how many blocks to read at once */
    if( (nb % 100 < nb_blocks ) ) {
      char buffer[32];
      sprintf( buffer, "Am transferat %6dk din %6dk:"
	      , nb * 32 / 1024, nbt * 32 / 1024 );
      hui_show_progress( buffer, ((nbt - nb) * 100 ) / nbt, 300 );
    }
    if( nb >= nb_blocks ) {
      chunk = nb_blocks; 
    } else {
      chunk = nb;
    }
    status = e2100_read_blocks( the_holter, (void*)buffer, chunk );
    if( status == 0 ) {
      nb -= chunk;
      fwrite( buffer, chunk * E2100_BLOCK_SIZE, 1, the_file);
    } else {
      while( e2100_refresh_address( the_holter ) != 0);
      terr ++;
    }
  }
  hui_status_print( "Transfer done.");
  free( buffer );
}



#define BUTTON_X_DIM 120
#define BUTTON_Y_DIM 20




/*--
\end{code}

\funparagraph{hui\_main}
% The name of the function is expected to be prezent
% in the sectioning element above

\PurposeReturn{No usefull return value, as a side effect handle
interaction with the used. The display will be in graphical mode
if parameter gg is a valid gwin or in ascii mode is gg is NULL.}
\Exceptions{}
\GlobalEffects{ }
\Preconditions{The graphical library should be initialized.}
\Performance{}

\begin{code}
--*/


void hui_main( GWIN * gg, int serial_port, char * in_filename ) {
  short hwin, pcwin, statusdwin, fulld;
  HOLTER the_holter;
  char filename[255]; FILE * the_file;

  if( in_filename != NULL )
    strncpy( filename, in_filename, 254);
  else
    filename[ 0 ] = 0;
  the_file = NULL;
    
  if( gg!=NULL ) {
    /* initialize the full dwin */
    fulld = newdwin( gg, 1, 1, 640, 480 );
    cleardwin( gg, fulld, PLANE_DWIN );
    
    /* initialize status bar */
    statusdwin = newdwin( gg, 10, 430, 550, 20);
    cleardwin( gg, statusdwin, LOWERED_DWIN );
    wi_set_color( statusdwin, 1 );
    hui_status_gwin = gg; hui_status_dwin = statusdwin;
    wi_def_help_dwin( statusdwin );

    /* initialize the holter window */
    hwin = newdwin( gg, 10, 10, 460, 300 );
    cleardwin( gg, hwin, LOWERED_DWIN );
    gsetcolor( gg, hwin, 1 );
    wi_set_help( wi_button( fulld, 500, 30, BUTTON_X_DIM, BUTTON_Y_DIM
	      , "Update info", 'c' )
	, "Update the window above with fresher information." );
    wi_set_help( wi_button( fulld, 500, 30 + 1 * ( 10 + BUTTON_Y_DIM )
	      , BUTTON_X_DIM, BUTTON_Y_DIM, "Transfer now", 't')
		, "Start the transfer now.");
#ifndef MSDOS
    wi_set_help( wi_button( fulld, 500, 30 + 2 * ( 10 + BUTTON_Y_DIM )
	      , BUTTON_X_DIM, BUTTON_Y_DIM, "Transfer bg", 'e')
		, "Start the transfer in the background." );
#endif
    wi_set_help( wi_button( fulld, 500, 30 + 3 * ( 10 + BUTTON_Y_DIM )
	      , BUTTON_X_DIM, BUTTON_Y_DIM, "Exit", 'x')
		, "End the program.");

    gtext( gg, hwin, 10, 100, "File name:" );
    wi_set_help( wi_textpad( hwin, 160, 100, 14, 1, 20, filename, 'f') 
	  , "Enter the name (without extension) of the destination file.");
  }
  
  /* initialize the communication package */
  the_holter = hui_get_the_holter( 2 );
  if( the_holter != NULL ) {
    if( gg!=NULL )
      cleardwin( gg, hwin, LOWERED_DWIN );
#if 0
    /* to much technical information was generated by the next
     * function so it is not used at the moment */
    hui_show_holter_info( gg, hwin, the_holter);
#endif
    hui_show_holter_status( gg, hwin, the_holter);
  }

  if( gg==NULL ) {
    if( the_holter == NULL ) goto proper_exit;
    the_file = fopen( filename, "w" );
    if( the_file == NULL ) goto proper_exit;
    e2100_transfer_file( the_holter, the_file );
    fclose( the_file );
    goto proper_exit;
  }

  while( 1 ) {
    switch( gwinteract( gg ) ) {
    case 'x': goto proper_exit;
    case 'c':			/* connect or reconnect to the holter */
      if( the_holter != NULL ) {
	e2100_close( the_holter );
      }
      the_holter = hui_get_the_holter( 2 );
      if( the_holter == NULL ) break;
      cleardwin( gg, hwin, LOWERED_DWIN );
      hui_show_holter_info( gg, hwin, the_holter);
      hui_show_holter_status( gg, hwin, the_holter);
      break;
    case 'f':
      gwaccept( gg );
      if( the_holter == NULL ) break;
      the_file = fopen( filename, "w" );
      if( the_file == NULL ) {
	hui_status_print( (char*)strerror(errno) );
      } else {
	char message[64];
	sprintf( message, "File %s opened.", filename );
	hui_status_print( message );
      }
      break;
    case 'e':
      if( the_file == NULL ) 
	hui_status_print( "Please enter a valid file name!" );
      else {
	fclose( the_file );
	wiportclose();
	switch( fork() ) {
	case 0:
/*	  hui_status_print( "Transfering in background, you may exit." ); */
	  goto proper_exit;
	default:
	  {
	    char * argv[] = {"gwo6", filename, NULL };

	    printf( "Running the transfer program.\n" );
#ifndef MSDOS
	    execve( "/home/dev/holter/gwo6", argv, NULL );
#endif
	    printf( "Oops, could not execute the program.\n" );
	  }
	}
      }
      break;
    case 't':
      if( the_file == NULL ) 
	hui_status_print( "Please enter a valid file name!" );
      else {
	transfer_file_with_progress( the_holter, the_file, 128  );
	fclose( the_file );
      }
      break;
    }
    
  }

 proper_exit:
  if( the_holter != NULL ) {
    e2100_close( the_holter );
  }
}

#define HUI_MAIN
#ifdef HUI_MAIN

void test_raw( void ) {
  int transfers;
  HOLTER the_holter;
  char buffer[32], test_buffer[32];

  /* initialize the communication package */
  the_holter = hui_get_the_holter( 2 );
  if( the_holter == NULL ) {
    printf( "Could not initialize holter.\n" );
    return;
  }
  for( transfers=0; transfers<1000; transfers++ ) {
    e2100_set_address( the_holter, 111 * 32 );
    if( e2100_read_block( the_holter, (void*)buffer ) == 0 ) {
      e2100_set_address( the_holter, 111 * 32 );
      e2100_read_block( the_holter, (void*)test_buffer );
      if( memcmp( buffer, test_buffer, E2100_BLOCK_SIZE ) == 1) {
	printf( "Transfer error at transfer %d.\n", transfers );
	fflush( stdout );
      }
    } else {
      printf( "Temporary error at transfer %d.\n", transfers );
    }
  }
}


main(int argc, char * argv[]) {
  GWIN *gg;

  /* parse command line */
  if( argc == 1) {		/* no arguments, graphical interface */
  /* initialize the graphical package */
  wiportinit(NULL, 0, 0, 0, NULL);
  
  gg = newgwin(640, 480, "Holter", "Holter");
  use_color_s_tech(gg);

  hui_main( gg, 2, NULL );
  wiportclose();
  }else {
    hui_main( NULL, 2, argv[1] );
  }

  exit( 0 );
}
#endif
