
/*--

  \begin{verbatim}
  Holter user interface
  Copyright (c) 1995 Marius Seritan

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  \end{verbatim}

  \section{Holter user interface}
  \PackagePurpose{Provides a full screen graphical user interface
  to the holter communication program.}
  \FunctionInterraction{
  \begin{itemize}
  \item call
  \function_tag{hui\_main} and go home.
  \end{itemize}
  }
  \Exceptions{}
  \GlobalEffects{The user might request the creation of a file.}
  \Preconditions{The \packagetag{wiport2} package should be initialized.}
  \Performance{}
  
  \begin{code}
  --*/



/*--
  \end{code}
  
  \section{Holter user interface}
  \ImplementationNotes{}
  \Portability{This package relies upon the \packagetag{wiport2} package
  for the graphical capabilities and upon the \packagetag{e2100} package for
  communication capabilities.
  }
  \Changes{}

  \begin{code}
  --*/

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#ifdef WIPORT_X11
#include "wiport2x.h"
#else
#include "wiport2.h"
#endif

#include "e2100.h"




/*--
  \end{code}

  \dataparagraph{Global variables}  % a name for the data structure

  \Definition{Data associated with the status info.}
  \SemanticStructure{}
  \ConsistencyRules{}

  \begin{code}
  --*/
static GWIN * hui_status_gwin;
static short hui_status_dwin;
static short hui_status_y = 10;


void hui_status_print( char * text ) {
  if( hui_status_gwin!=NULL ) {
    gtext( hui_status_gwin, hui_status_dwin, 3, hui_status_y, text );
    hui_status_y += font_y_size;
  }
  else
    puts( text );
}


void hui_show_progress( char * text, int ratio, int width, int dw )  {
   const prg_y0 =  7;		       /* y start of the progress ind. */
   const prg_y1 = 14;		       /* y end   of the progress ind. */
   
   if( hui_status_gwin != NULL )  {
      int x0;
      cleardwin( hui_status_gwin, dw, LOWERED_DWIN );
      gtext( hui_status_gwin, dw, 3, 3, text );
      x0 = 20 + strlen(text) * font_x_size;
      wi_line( dw, x0-2, prg_y0 - 2, x0 + width + 4, prg_y0 - 2);
      wi_line( dw, x0-2, prg_y1 + 2, x0 + width + 4, prg_y1 + 2);
      wi_line( dw, x0-2, prg_y0 - 2, x0 - 2, prg_y1 + 2);
      wi_line( dw, x0+width+4, prg_y0 - 2, x0 + width + 4, prg_y1 + 2);
      wi_rect( dw, x0, prg_y0, (ratio * width)/100, prg_y1 - prg_y0);
   }
}

void hui_show_holter_info( GWIN * gg, short ldwin, HOLTER the_holter) {
  char buffer[1024];
  
  sprintf( buffer, "Sampling period:             %6d microsec"
	  , e2100_sampling_period( the_holter ));
  if( gg!= NULL )
    gtext( gg, ldwin, 10, 10, buffer );
  else
    puts( buffer );
  sprintf( buffer, "Vertical resolution:         %6d bits"
	  , e2100_resolution( the_holter ));
  if( gg!=NULL )
    gtext( gg, ldwin, 10, 25, buffer );
  else
    puts( buffer );
}

void hui_show_holter_status( HOLTER the_holter) {
  char buffer[1024];
  long seconds;

  seconds = e2100_nb_seconds( the_holter );
  sprintf( buffer, "Modulul de pacient functioneaza de %2d:%2d ore."
	  ,  seconds / 3600 , (seconds % 3600 ) / 60  );
  hui_status_print( buffer );

  sprintf( buffer, "Memoria interna este umpluta la %4.2f%% din capacitate." 
	  , (double)e2100_read_fill( the_holter ) * 100.0 / (8.0 * 1048576L));
  hui_status_print( buffer );
}

HOLTER hui_get_the_holter( int serial_port ) {
  HOLTER the_holter;

  the_holter = e2100_open( serial_port ); /* internal high speed interface */
  if( the_holter == NULL ) {
    hui_status_print( "Eroare: Modul de pacient nu este accesibil." );
    hui_status_print( "Cauze posibile:  60% cablul nu este conectat." );
    hui_status_print( "                 30% modulul de pacient nu este pornit." );
    hui_status_print( "                 10% placa de interfata nu este instalata." ); 			    
  } else {
    hui_status_print( "Modulul de pacient este conectat." );
  }
  return the_holter;
}

/* gwo_check
 * check if the same block of the given `dimension' is located
 * at `address' both in the file and in the patient module
 *
 * returns 1 if successfull, 0 if not or if some error occured
 */

int gwo_check( FILE * f, struct holter * the_holter
	      , long address, int dimension ) {
  char * f_buffer, * h_buffer;
  
  /* make dimension to be a multiple of 32 */
  dimension = (dimension/32)*32;
  /* allocate memory */
  f_buffer = alloca(dimension);
  h_buffer = alloca( dimension );
  if( ( f_buffer == NULL ) || (h_buffer == NULL ) ) 
    return 0;
  /* read information from the file */
  if( fseek( f, address, 0) != 0 ) return 0;
  if( fread( f_buffer, dimension, 1, f ) != 1 )  return 0;
  /* read information from the holter */
  e2100_set_address( the_holter, address );
  if( e2100_read_blocks( the_holter, h_buffer, dimension/32 ) != 0 ) {
    return 0;
  }
  if( memcmp( f_buffer, h_buffer, dimension ) == 0 )
    return 1;
  return 0;
}

/* fopen_gwo
 * if necessary complete the name of the file with the gwo extension
 * check if a file with the given name already exists
 *   and if so prompt the user to decide if the transfer should continue
 *   or be restarted
 * the real file name will be returned in `name'
 * the holter transfer address will be returned in `address'
 */
FILE  * fopen_gwo( char * name, long * address, struct holter * the_holter ) {
  char * op;			/* operation to be performed by fopen */
  struct stat st;
  FILE * f;

  if( strchr( name,'.' ) == NULL ) {
    strcat( name, ".gwo" );
  }
  if( stat( name, &st ) == -1 ) {
    /* the file does not exist, open it for writting */
    *address = 0L;
    return fopen( name, "wb" );
  }
  /* oops, there is already a file with the given name */
  /* set address to the start of the last complete block */
  f = fopen( name, "r+b");

  if( f==NULL ) {
    *address = 0L; return f;
  }
  *address = st.st_size & ~4095;
  hui_status_print( "Atentiune: exista un fisier cu acelasi nume." );
  /* check to see if the file and the holter hold the same information */
  if( ( gwo_check( f, the_holter, 0L, 4096 ) == 0 ) 
     || ( ( *address > 4096 ) 
          && ( gwo_check(f, the_holter, *address-4096, 4096) == 0 ))){
    hui_status_print( "Atentiune: informatia din holter este diferita de cea din fisier" );
    hui_status_print( "     Alegeti alt nume de fisier daca exista" );
    hui_status_print( "     riscul de a sterge un fisier anterior." );
    *address = 0L;
  } else {
    hui_status_print( "Fisierul contine partial informatia din holter." );
    {
      char buffer[128];
      sprintf( buffer, "Transferul va incepe de la adresa %ldk", *address/1024 ); 
      hui_status_print( buffer );
    }
    hui_status_print( "Daca doriti sa transferati din nou tot fisierul," );
    hui_status_print( "selectati butonul --Sterge fisier-- inainte de a incepe transferul." );
  }

  fseek( f, *address, 0);
  return( f );
}

void transfer_file_with_progress( struct holter * the_holter
				 , FILE * the_file 
				 , int nb_blocks 
				 , long address 
				 , int dw ) {
  long nb, nbt;			/* number of blocks */
  int terr;			/* number of temporary errors */
  char * buffer;
  int status;			

  buffer = (char*)malloc( nb_blocks * E2100_BLOCK_SIZE );
  if( buffer == NULL ) {
    hui_status_print( "Eroare: memorie interna insuficienta, nu pot sa continui." );
    sleep( 2 );
    return;
  }
  /* seek into the file, just to be sure */
  fseek( the_file, address, 0 );
  nb = ((e2100_read_fill(the_holter) - 1 - address) / E2100_BLOCK_SIZE) + 1;
  nbt = nb;
  terr = 0;
  e2100_set_address( the_holter, address );
  for( ; nb;  ) {
    int chunk;			/* how many blocks to read at once */
    if( (nb % 100 < nb_blocks ) ) {
      char buffer[32];
      sprintf( buffer, "Transferat %4dk din %4dk:"
	      , (int)(((nbt-nb) * 32)/1024), (int)((nbt * 32)/1024) );
      hui_show_progress( buffer, ((nbt - nb) * 100 ) / nbt, 150, dw );
    }
    if( nb >= nb_blocks ) {
      chunk = nb_blocks; 
    } else {
      chunk = nb;
    }
    status = e2100_read_blocks( the_holter, (void*)buffer, chunk );
    if( status == 0 ) {
      nb -= chunk;
      if( fwrite( buffer, chunk * E2100_BLOCK_SIZE, 1, the_file) == 0) {
	hui_status_print( "Eroare: am probleme la scrisul pe disc." );
        switch( errno ) {
        case EPERM:
	case EACCESS:
          hui_status_print( "  Nu aveti drept de scriere in fisierul." );
          break;
        case EIO:
          hui_status_print( "  Eroare de scriere/citire, verificati discul." );
	  break;
	case EBUSY:
	case ETXTBUSY:
	  hui_status_print( "  Fisierul este folosit de un alt program." );
	  break;
	case EISDIR:
	  hui_status_print( "  Exista un director cu acest nume." );
	  break;
	}
	free( buffer );
	return;
      }
    } else {
      while( e2100_refresh_address( the_holter ) != 0);
      terr ++;
    }
  }
  hui_show_progress( "Transfer terminat: "
		    , ((nbt - nb) * 100 ) / nbt, 150, dw );
  hui_status_print( "Transfer terminat.");
  free( buffer );
}



#define BUTTON_X_DIM 120
#define BUTTON_Y_DIM 20




/*--
\end{code}

\funparagraph{hui\_main}
% The name of the function is expected to be prezent
% in the sectioning element above

\PurposeReturn{No usefull return value, as a side effect handle
interaction with the used. The display will be in graphical mode
if parameter gg is a valid gwin or in ascii mode is gg is NULL.}
\Exceptions{}
\GlobalEffects{ }
\Preconditions{The graphical library should be initialized.}
\Performance{}

\begin{code}
--*/


void hui_main( GWIN * gg, int serial_port, char * in_filename ) {
  short hwin, statusdwin, fulld, helpdwin;
  HOLTER the_holter;
  char filename[255]; FILE * the_file;

  if( in_filename != NULL )
    strncpy( filename, in_filename, 254);
  else
    filename[ 0 ] = 0;
  the_file = NULL;
    
  if( gg!=NULL ) {
    /* initialize the full dwin */
    fulld = newdwin( gg, 1, 1, 640, 480 );
    cleardwin( gg, fulld, PLANE_DWIN );
    
    /* initialize the upper window */
    hwin = newdwin( gg, 10, 10, 450, 60 );
    cleardwin( gg, hwin, LOWERED_DWIN );
    gsetcolor( gg, hwin, 1 );
    /* initialize status win */
    statusdwin = newdwin( gg, 10, 80, 450, 300);
    cleardwin( gg, statusdwin, LOWERED_DWIN );
    wi_set_color( statusdwin, 1 );
    hui_status_gwin = gg; hui_status_dwin = statusdwin;
    /* initialize on-line help dwin */
    helpdwin = newdwin( gg, 10, 430, 450, 20);
    cleardwin( gg, helpdwin, LOWERED_DWIN );
    wi_set_color( helpdwin, 1 );
    wi_def_help_dwin( helpdwin );

    wi_set_help( wi_button( fulld, 500, 30, BUTTON_X_DIM, BUTTON_Y_DIM
	      , "Modul pacient", 'c' )
	, "Incearca sa se reconecteze la modulul de pacient." );
    wi_set_help( wi_button( fulld, 500, 30 + 1 * ( 10 + BUTTON_Y_DIM )
	      , BUTTON_X_DIM, BUTTON_Y_DIM, "Transfer", 't')
		, "Incepe/continua transferul.");
    wi_set_help( wi_button(fulld, 500, 30 + 2 * ( 10 + BUTTON_Y_DIM )
	      , BUTTON_X_DIM, BUTTON_Y_DIM, "Sterge fisier", 'd' ) 
		, "Sterge fisierul curent.");
    wi_set_help( wi_button(fulld, 500, 30 + 3 * ( 10 + BUTTON_Y_DIM )
	      , BUTTON_X_DIM, BUTTON_Y_DIM, "Verifica fisier", 'v' ) 
		, "Compara informatia din fisierul curent cu cea din modul.");
    
#ifndef MSDOS
    wi_set_help( wi_button( fulld, 500, 30 + 4 * ( 10 + BUTTON_Y_DIM )
	      , BUTTON_X_DIM, BUTTON_Y_DIM, "Transfer bg", 'e')
		, "Start the transfer in the background." );
#endif
    wi_set_help( wi_button( fulld, 500, 30 + 4 * ( 10 + BUTTON_Y_DIM )
	      , BUTTON_X_DIM, BUTTON_Y_DIM, "Terminat", 'x')
		, "Incheie executia programului.");

    gtext( gg, hwin, 10, 10, "Numele fisierului:" );
    wi_set_help( wi_textpad( hwin, 160, 10, 14, 1, 20, filename, 'f') 
	  , "Introduceti numele fisierului destinatie.");
  }
  
  /* initialize the communication package */
  the_holter = hui_get_the_holter( 2 );
  if( the_holter != NULL ) {
    if( gg!=NULL )
/*
      cleardwin( gg, hwin, LOWERED_DWIN );
*/
#if 0
    /* to much technical information was generated by the next
     * function so it is not used at the moment */
    hui_show_holter_info( gg, hwin, the_holter);
#endif
    hui_show_holter_status( the_holter);
  }

  if( gg==NULL ) {
    long address;		/* address in the holter where the transfer */
				/* should begin */
    if( the_holter == NULL ) goto proper_exit;
    the_file = fopen_gwo( filename,  &address, the_holter );
    if( the_file == NULL ) goto proper_exit;
    fseek( the_file, 0, 0 );
    e2100_transfer_file( the_holter, the_file );
    fclose( the_file );
    the_file = NULL;
    goto proper_exit;
  }

  while( 1 ) {
    long address;		/* address in the holter where the transfer 
				 * should start */
    switch( gwinteract( gg ) ) {
    case 'x': goto proper_exit;
    case 'c':			/* connect or reconnect to the holter */
      if( the_holter != NULL ) {
	e2100_close( the_holter );
      }
      the_holter = hui_get_the_holter( 2 );
      if( the_holter == NULL ) break;
/*      hui_show_holter_info( gg, hwin, the_holter);*/
      hui_show_holter_status( the_holter);
      break;
    case 'f':
      gwaccept( gg );
      if( the_holter == NULL ) break;
      the_file = fopen_gwo( filename, &address, the_holter );
      if( the_file == NULL ) {
	hui_status_print( (char*)strerror(errno) );
      } else {
	char message[64];
	sprintf( message, "Fisierul %s a fost deschis.", filename );
	hui_status_print( message );
      }
      break;
    case 'e':
      if( the_file == NULL ) 
	hui_status_print( "Nu ati introdus un nume de fisier valid!" );
      else {
	fclose( the_file );
	wiportclose();
	switch( fork() ) {
	case 0:
/*	  hui_status_print( "Transfering in background, you may exit." ); */
	  goto proper_exit;
	default:
	  {
	    char * argv[] = {"gwo6", filename, NULL };

	    printf( "Running the transfer program.\n" );
#ifndef MSDOS
	    execve( "/home/dev/holter/gwo6", argv, NULL );
#endif
	    printf( "Oops, could not execute the program.\n" );
	  }
	}
      }
      break;
    case 't':
      if( the_file == NULL ) 
	hui_status_print( "Nu ati introdus un nume de fisier corect!" );
      else {
	transfer_file_with_progress( the_holter, the_file, 128
				    , address, helpdwin  );
	fclose( the_file );
	the_file = NULL;
      }
      break;
    case 'd':
      if( the_file != NULL ) {
	fclose( the_file );
	the_file = fopen( filename, "wb" );
	address = 0L;
	hui_status_print( "Urmatorul transfer va incepe de la adresa 0." );
      }
     break;
    case 'v':
      if( the_file != NULL ) {
	struct stat st;
	long address;

	fstat( fileno(the_file), &st );
	for( address=0; address<st.st_size-4096; address+=4096 ) {
	  if( gwo_check( the_file, the_holter, address, 4096 ) == 0) {
	    char buffer[128];
	    sprintf( buffer, "Eroare la adresa %ld.", address );
	    hui_status_print( buffer );
	  }
	}
	hui_status_print( "Verificare terminata." );
      }
      break;
    }
    
  }

 proper_exit:
  if( the_holter != NULL ) {
    e2100_close( the_holter );
  }
}

#define HUI_MAIN
#ifdef HUI_MAIN

void test_raw( void ) {
  int transfers;
  HOLTER the_holter;
  char buffer[32], test_buffer[32];

  /* initialize the communication package */
  the_holter = hui_get_the_holter( 2 );
  if( the_holter == NULL ) {
    printf( "Could not initialize holter.\n" );
    return;
  }
  for( transfers=0; transfers<1000; transfers++ ) {
    e2100_set_address( the_holter, 111 * 32 );
    if( e2100_read_block( the_holter, (void*)buffer ) == 0 ) {
      e2100_set_address( the_holter, 111 * 32 );
      e2100_read_block( the_holter, (void*)test_buffer );
      if( memcmp( buffer, test_buffer, E2100_BLOCK_SIZE ) == 1) {
	printf( "Transfer error at transfer %d.\n", transfers );
	fflush( stdout );
      }
    } else {
      printf( "Temporary error at transfer %d.\n", transfers );
    }
  }
}


main(int argc, char * argv[]) {
  GWIN *gg;

  /* parse command line */
  if( argc == 1) {		/* no arguments, graphical interface */
  /* initialize the graphical package */
  wiportinit(NULL, 0, 0, 0, NULL);
  
  gg = newgwin(640, 480, "Holter", "Holter");
  use_color_s_tech(gg);

  hui_main( gg, 2, NULL );
  wiportclose();
  }else {
    hui_main( NULL, 2, argv[1] );
  }

  exit( 0 );
}
#endif

