; This file is the data acquisition and processing program
; stored in the ECGNOSIS 2100 24h ambulatory ECG recorder
; 
; Copyright (c) 1993--1994 Radu Corlan
; Copyright (c) 1993--1994 Marius Seritan
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by 
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.

; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
; This program uses the verision 1.0 of the compression algorithm and the
; version 1.0 of the structure of the output file (those two aspects are
; presented somewhere in this file).

; do not forget to change those when modifying the program
; 1.1  working program, used during standardization
; 1.2  mseritan dec '94
;      implemented new communication protocols in order to 2 able to
;      read all ram contents and the info from EEPROM
;      tried transversal parity (XOR of all bytes in a packet)
;	    --- not implemented at the moment, useless
;      echo bytes for multibyte commands
;	
; 
VERSION EQU 1		; OF THIS PROGRAM
PATCH	EQU 2		; PATCH LEVEL OF THIS PROGRAM
FFORMAT	EQU 1		; FILE FORMAT
CFORMAT	EQU 1		; COMPRESSION FORMAT

;*************************************************
; Presentation of the data structures 
;*************************************************

; The program uses two FIFOs. Those data structures are used to transmit data
; between three asynchronous processes:
; 
; \item{} the acquisition FIFO is used to connect the data acquisition process
; with the filter module;
; 
; \item{} the processing FIFO is used to connect the filter module with the
; processing module.
; 
; The FIFOs are defined by a data area and by functions operating on that area.
; For efficiency reasons the FIFOs are not accessed only through their
; functional interface.


;*************************************************
; Constants
;*************************************************

; The size of the acquisition FIFO is given by @{FIFOMK@} and the code is
; relying on the fact that the FIFO is located at address 0 in the RAM
; of the microcontroller. As this FIFO is also used as a data area for the
; filter module, the size of the FIFO should be at least the
; double of the number of coefficients of  the filter.

FIFOMK  EQU $3F         ;LENGTH OF FIFO - 1

; The second FIFO is used to permit the partitioning of the memory in
; blocks. This task is complicated by the fact that the compression
; mechanism generates a continous flow of bits that has to be stoped and
; restarted. Also there is a small complication due to the fact that the
; two channels are compressed toghether. Also this FIFO is used as a
; data area for the QRS detection mechanism.
; 
; The dimension of the second FIFO is given by the constant @{DDATAMK@}.
; The second FIFO must be large enough to accomodate at least 9 values
; as there can be 8 values in the buffer of the compacting module. At
; the start of a block there will be a delay between the filtering
; module and the processing module, the second FIFO must be big enough
; to permit the proper functioning of the program in the presence of
; this delay.

DDATAMK EQU $1F         ; length of the DELTA data area - 1

STACK   EQU $FF         ; TOP OF STACK ADDRESS

BLKSIZE EQU $1000       ; dimension of a block of data
BLKTTT  EQU BLKSIZE/256 ; temporary variable
BLKMSK  EQU BLKTTT-1
BTNPER  EQU 10          ; integration period for the button
BTNTHR  EQU 5           ; debouncing threshold
INTVL   EQU 6144        ;SAMPLING INTERVAL IN USEC
;***
; configuration of the LED port
;***
PWLED   EQU 1
CH1LED  EQU 2
CH2LED  EQU 4
MEMLED  EQU 8

; number of valid states for the compression automata
CNBST   EQU 5

; number of filter's coefficients
FNBCOEF EQU 12

        ORG     $0000

FIFO    DS FIFOMK+1    ; START OF FIFO for unfiltered data
DDATA   DS DDATAMK+1   ; data area for the DELTA module

FSP     DS 1           ; FIFO START POINTER
FEP     DS 1           ; FIFO END POINTER
DFSP    DS 1           ; pointers in the data area for the DELTA module
DFEP    DS 1
DSAVED  DS 1           ; pointer to indicate values saved in the memory
; 32 BIT MEMORY ADDRESSES (IN BIG ENDIAN FORMAT)
ADDR1   DS 4           ; THE FIRST FREE DRAM LOCATION
ADDR2   DS 4           ; THE FIRST BYTE TO BE TRANSMITTED ON THE SERIAL LINE
CTIMER  DS 4           ; counter, number of stored values (compressed & raw)
; the next address is also used as mark of end of buffer
BTNAD   DS 4           ; address of the last button event
STATUS  DS 1           ; STATUS BYTE
;       BIT 0           FIFO OVERFLOW
; variables used in order to divide the data in buffers
SAVADDR DS 4           ; place to save true size information
BLKST   DS 1           ; block status, a new block is starting
; data associated with the button
BTNCNT  DS 1           ; integration variable
BTNTIM  DS 1           ; integration timer
BTNST   DS 1           ; button status, 0 released FF pressed
BTNPR   DS 1           ; state of the button automata, see BUTTON:
TDBYTES DS 1           ; NUMBER OF DATA BYTES TO BE TRANSMITTED
TSBYTES DS 1           ; NUMBER OF STATUS BYTES TO BE TRANSMITTED
TABYTES DS 1           ; NUMBER OF ADDRESS BYTES TO BE RECEIVED
SBPTR   DS 2           ; POINTER TO DATA BYTE TO BE TRANSMITTED
ADPTR   DS 1           ; POINTER TO ADDRESS BYTE TO BE RECEIVED

; data associated with LED handling
LCNTCH  DS 2           ; counter for channels' leds
LEDST   DS 1           ; status of the LEDs, communication with PERINT
; data for the QRS detection module
QRSDATA DS 8           ; 4 bytes by channel, 0 threshold 1 counter 2 inhibit
; variables for the compression module
CSTATE  DS 1           ; state of the compression automata
CMETA   DS 1           ; flag (counter) implementing metastable behaviour
CASVAL  DS 1           ; storage to assemble a full value from bits
CASNBB  DS 1           ; number of free bits in the current assembled value
; variables of the filter module
FVAL    DS 2           ; the actual value on which the filter works
; error on the serial line
SRERROR	DS 1

EORAM   EQU     $100

;*************************************************
; CODE section
;*************************************************
        ORG     $F800

;*** Initialization
        LDS     #STACK          ; SET THE STACK

        LDX     #$1000          ; REGISTER BASE

;*** Port initialization
        BSET    $39,X $80

        LDAA    #$2             ; PROGRAM PARALLEL PORTS
        STAA    $02,X           ; HANDSHAKE MODE
        LDAA    #$88
        STAA    $26,X           ; PA7=OUT
        LDAA    #$3E
        STAA    $09,X           ; PD1-5=OUT, PD0=IN

        LDAA    #$C0            ; PUT RAM IN REFRESH
        STAA    $00,X
        LDAA    #$FF
        STAA    $08,X

;*** Variable initialization
        LDAA    #EORAM
        LDY     #0
VINIT01:
        CLR     0,Y
        INY
        DECA
        BNE     VINIT01
        BSET    LEDST PWLED
        LDAA    #DDATA
        STAA    DFEP
        STAA    DFSP
        STAA    DSAVED

        JSR     COMPINIT        ; initialization of the compression module

;*** ADC initialization
        LDAA    #$10            ; ADC START COMMAND
        STAA    $30,X


        BSET    $22,X $80       ; ENABLE INTERRUPT ON OUTP COMPARE 1

;*** Serial port initialization
        CLR     $2C,X           ; SERIAL SCCR1

	LDAA    #$00            ; BAUD (76800)
;       LDAA    #$01            ; BAUD (38400)
;	LDAA    #$20            ; BAUD (19200 @ 4.9152MHZ OSC)
;	LDAA    #$21            ; BAUD (9600 @ 4.9152MHZ OSC)
;       LDAA    #$22            ; BAUD (4800 @ 4.9152MHZ OSC)
        STAA    $2B,X

        LDAA    #$2C            ; PROGRAM SERIAL INTERFACE
        STAA    $2D,X           ; SCCR2 (TCIE, RIE, TXEN, RXEN)

        CLI                     ; ENABLE INTERRUPTS

        LDAA    $2E,X
        LDAA    #$AA            ; WRITE AA TO SERIAL
        STAA    $2F,X

        BSET    BLKST 1         ; make sure BLOCK will be activated

MAIN:
; doing all the necessary processing
; each routine check to see if there is any data available when called
; if no data is available the routines return at once
        JSR     BLOCK
        JSR     PROCESS
        JSR     FILTER
        JSR     BUTTON
        BRA     MAIN


;*************************************************
; FILTER module
;*************************************************

FILTER:
        LDAA    FSP             ; is there any data available?
        CMPA    FEP
        BNE     FILTDO          ; yes filter it
        RTS                     ; no, abandon processing
FILTDO:
; initialize
        CLR     FVAL
        CLR     FVAL+1
        CLRB                    ; in B pointer in the FIFO and in the filter
F_ONE:
; implements one step of the filter
        PSHB
        JSR     RDFIFO          ; read data in A
        PULB                    ; refresh filter pointer
        PSHB
        LDY     #FILTCOEF
        ABY                     ; compute index in the filter
        LDAB    0,Y             ; in B coef of the filter


;*** Multiply unsigned A with signed B
        TSTB
        BGE     MULPOZB         ; is B pozitive?
        NEGB                    ; no, change its sign
        MUL                     ; do the multiplication
        COMA                    ; 2's complement of D
        COMB
        ADDD    #1
        BRA     MULOK
MULPOZB:
        MUL                     ; do the multiplication
MULOK:

        ADDD    FVAL            ; add it to the current value
        STD     FVAL
        PULB                    ; goto next coefficient
        INCB
        CMPB    #FNBCOEF        ; was it the last one?
        BNE     F_ONE           ; no, loop

;*** Save filtered value in the processing FIFO
        CLRA                    ; compute the address
        LDAB    DFEP
        XGDY
        LDAA    FVAL            ; get the filtered value
        STAA    0,Y

        LDAA    DFEP            ; advance data FIFO end pointer
        INCA
        ANDA    #DDATAMK
        ORAA    #DDATA
        STAA    DFEP
        CMPA    DFSP
        BNE     DFIFOK
        BSET    LEDST   CH2LED
DFIFOK:

        LDAA    FSP
        INCA
        ANDA    #FIFOMK
        STAA    FSP

        RTS


PROCESS:
        LDAA    DFSP                    ; any data available?
        CMPA    DFEP
        BNE     DELTADO
        RTS
DELTADO:
        CLRB                            ; retrieve the last value
        JSR     RDDFIF
        PSHA
        LDAB    #1                      ; retrieve the previous value
        JSR     RDDFIF
        TAB
        PULA
        SBA                             ; the processing statement
        JSR     DEFLATE                 ; compress the processed value
        JSR     QRS_DET                 ; blink the LEDS when QRS
; advance the FIFO start pointer
        LDAA    DFSP
        INCA
        ANDA    #DDATAMK                ; normalize the processed value
        ORAA    #DDATA
        STAA    DFSP
        RTS

;*************************************************
; COMPRESSION module
;*************************************************

; @C@<Algorithm@>
; 
; The algorithm used in order to compress the signal is based on a  simple
; automata with @{CNBST@} states (this is a parameter that must be set by
; the user when changing the data defining the compression automata).
; 
; The automata accepts at its input a 8 bit value representing a signed
; char. However, in order to ease the explanation of the automata, this
; value is considered an unsigned quantity.
; 
; In each state there is a minimal and a maximal value that can be
; transmitted. The code for this value is obtained by taking the lower
; bytes of this value, the number of those bytes is dependent on the
; actual state the automata. In the actual value cannot be transmitted
; in this state, the state of the automata is incremented and a special
; value is sent on the output bit stream. When reaching
; the maximum number of states, the value is codified using all the 8
; bits.
; 
; In order to ensure the return of the automata to state zero, all the
; states but the initial state are metastables. When the automata has
; succeded to codify the actual value, the state is kept for one more
; cycle and then it is decremented. The state of the automata is kept in
; the variable @{CSTATE@}. The automata is completely defined by the
; data stored in the array @{CSTRUC@}. In fact only the field @{CNBB@}
; of this array is necessary as the other fields can be automatically
; calculated.
; 
; The compression automata takes its input from the register @{A@} and
; place its output in memory. The bits are assembled using two variables
; @{CASVAL@} to assemble bytes from bits and @{CASNBB@} indicationg the
; number of free bits in the previous  variable.

; initialization routine
COMPINIT:
        CLR     CSTATE
        CLR     CMETA
        CLR     CASNBB
        BSET    CASNBB 8
        RTS

;***
; main entry point to the compression module
; input: A  = value to be compressed
; modifies: A, B, IY
;***

DEFLATE:
; start of the compression algorithm

;*** Implement metastable behaviour
        LDAB    CSTATE
        TSTB
        BEQ     CGTEST01        ; already in state 0, no fun
        TST     CMETA           ; is this state still stable?
        BNE     DECMETA
        DECB                    ; no, falling down
        STAB    CSTATE
        BRA     CGTEST01
DECMETA:
        DEC     CMETA           ; yes, make the state unstable
CGTEST01:

;*** Determinate proper state of the automata
        LDAB    CSTATE
CGTEST02:
;* Compute address for state in B
        LSLB                    ; compute index of the data for actual state
        LSLB
        LDY     #CSTRUC         ; start of the compression table
        ABY                     ; IY=address of the data for the actual state

        CMPA    CMAX,Y          ; is the value within the actual limits?
        BLO     CSENDOK         ; yes
        CMPA    CMIN,Y
        BHI     CSENDOK         ; yes again
        PSHA                    ; no, send code for the next state
        LDAA    CCODEN,Y        ; ie, this one
        LDAB    CNBB,Y          ; number of valid bits
        JSR     CSEND
        PULA
        LDAB    CSTATE
        INCB                    ; no, increment the number of the state
        STAB    CSTATE
        CMPB    #CNBST          ; have we reached the top state ?
        BEQ     CSEND0          ; yes, send the byte as is
        BSET    CMETA 1         ; no, make this state stable
        BRA     CGTEST02        ; we do not have the right state yet
CSEND0:
        CLR     CMETA           ; state too high, not stable
        LSLB                    ; compute index of the data for actual state
        LSLB
        LDY     #CSTRUC         ; start of the compression table
        ABY                     ; IY=address of the data for the actual state

CSENDOK:

; in A we still have the value that must be compressed
        LDAB    CNBB,Y          ; number of valid bits
        JSR     CSEND

;*** Keep track of the values saved in memory
        BRSET   BLKST 1 DEFL010 ; the value was not properly sent
        BRCLR   BLKST 2 DEFL009
        BCLR    BLKST 2
        LDAA    DFSP            ; remember how much values were saved
        STAA    DSAVED          ;   in memory from delta's FIFO
DEFL009:
        LDY     #CTIMER         ; increment the compression timer
        JSR     INCADDR
DEFL010:

        RTS

;*** Bit assemble function
CSEND:
; send the actual value
; in A the value
        PSHB

;*** Flush left the correct bits
        SUBB    #8              ; how many useless bits?
        NEGB
CSEND1:
        TSTB                    ; end of the loop ?
        BEQ     CSEND2          ; yes
        LSLA                    ; no
        DECB
        BRA     CSEND1
CSEND2:

        PULB
CSENDLP:
        ROLA                    ; MSB bit of A in Cy flag
        ROL     CASVAL          ; and now in LSB of CASVAL
        DEC     CASNBB          ; so one bit less available
        BNE     NOCVFULL        ; is CASVAL full?

;*** Write the assembled value
        PSHA
        PSHB
; check to see if we are at the end of this block
        LDD     BTNAD+2         ; taking into account the lower bytes
        SUBD    ADDR1+2
        TSTA
        BNE     NOWR02
        CMPB    #3
        BHI     NOWR02
        BSET    BLKST 1         ; signal start of a new block
        BRA     NOWR03
NOWR02:
        LDAA    CASVAL
        JSR     WRBYTE
        BSET    BLKST  2        ; signal a memory write
        LDAA    #8
        STAA    CASNBB          ; 8 free bits in the assembled value
NOWR03:
        PULB
        PULA

        BRCLR   BLKST $01 NOCVFULL
        RTS
NOCVFULL:
        DECB                    ; any bit left?
        BNE     CSENDLP         ; yes, push another bit
        RTS



;***
; QRS_DET
; QRS detection module
; no data in, no data out, all regs affected
; QRSDATA affected
; BFSP, DDATA used (mainly through RDDFIF, but not only)
; eventually LED affected
;***

QRS_DET:
        LDAB    DFSP
        ANDB    #1
        LSLB                            ; 4 bytes per channel
        LSLB
        LDY     #QRSDATA                ; compute the start of the data
        ABY

        TST     2,Y
        BEQ     QQ1

        DEC     2,Y
        RTS
QQ1:

; compute DAN's difference x[ 0 ] - x[ -3 ]
        PSHY
        CLRB                            ; get the actual value
        JSR     RDDFIF
        PSHA
        LDAB    #3                      ; get the fourth previous value
        JSR     RDDFIF
        PULB
        SBA                             ; in A the delta[4], keep it a while
        PULY
; compare it with DAN's threshold
        CMPA    #8
        BLO     QRS001                  ; below the limit, no QRS
        CMPA    #-8
        BHI     QRS001                  ; above the limit, no QRS
; within the limits, we might have a QRS complex
; compare the signs of the thresholds
        LDAB    1,Y                     ; get the previous counter
        BEQ     TRIGST                  ; no previous trigger, store this one
        PSHA                            ; A still holds the difference
        EORA    0,Y                     ; are the signs different ?
        ROLA                            ;; get sign in carry
        PULA                            ;; does not affect carry
        BCC     TRIGST                  ; wrong trigger, store this one
; haha,  blink the led for the apropriate channel
        CLR     1,Y                     ; clear the counter
        LDAB    #80
        STAB    2,Y
; and now blink the LED (in fact the work is done by PERINT)
        LDAB    DFSP                    ; get the channel
        ANDB    #1
        LDY     #LCNTCH
        ABY
        LDAB    #10
        STAB    0,Y
        RTS
TRIGST:
        STAA    0,Y                     ; save the threshold value
        LDAA    #28                     ; another magic number
        STAA    1,Y                     ; restore the counter
QRS001:
        TST     1,Y                     ; do we have to decrement counter ?
        BEQ     QRS002                  ; no
        DEC     1,Y                     ; yes, do it
QRS002:
        RTS

; Each block of memory has a dimension of @{BLKSIZE@} and is a
;standalone unit of data that can be completely decodified. A block is
;made out of:
;
;\item{} a header that defines:
;\itemitem{} The compression timer at the start of the block (the
;variable @{CTIMER@}. This variable is
;represented on 4 bytes and represent the index of the first sample
;stored in the block (the very first sample has index 0).
;\itemitem{} The last two bytes of the compression timer at the end of
;the block. By substracting from this value the last two bytes of the
;@{CTIMER@} at the start of the block, using 16 bits arithmetic, the
;correct number of samples stored in the block can be computed. This
;number is necessary in order to permit the correct decodification of
;the block.
;\itemitem{} The first address of the block that contains button
;information. See below for an explanation of this limit. This variable
;is represented on two bytes and it is stored as the last two bytes of
;the given memory address.
;
;\item{} The value of the signal on the two channels. This value is
;used to ensure the proper initialization of the @/undelta@/ module in
;the decodification phase.
;
;\item{} The compressed values.
;
;\item{} An array containing the moments when the button was pressed
;during the actual block. One such momemt is defined by the last two
;bytes of the compression timer when the button was pressed. Two bytes
;quantities are stored in the reverse order of their occurence (more
;recent events are stored at lower addresses).

; making the preamble of the block
BLOCK:
        TST     BLKST           ; are we requested to initialize a new block?
        BNE     BLOCK01
        RTS
BLOCK01:
;** Substract from  CTIMER nb of values compressed but not sent
        BCLR    DSAVED 1        ; always start with the first channel
        LDAB    DFSP            ; compute how many values were not saved
        SUBB    DSAVED
        ANDB    #DDATAMK        ;   nice arithmetic
        LDAA    CTIMER+3; substract this value from CTIMER
        SBA
        STAA    CTIMER+3
        LDAA    CTIMER+2        ; eventually propagate Carry
        SBCA    #0
        STAA    CTIMER+2
        LDAA    CTIMER+1
        SBCA    #0
        STAA    CTIMER+1
        LDAA    CTIMER
        SBCA    #0
        STAA    CTIMER
        BCLR    CTIMER+3 1
        LDAA    CTIMER+2        ; save the last two bytes of the timer

;** Save last two bytes of CTIMER at SAVADDR
        LDAA    CTIMER+2
        LDY     #SAVADDR
        JSR     MWRITE
        JSR     INCADDR
        LDAA    CTIMER+3
        JSR     MWRITE
        JSR     INCADDR

;*** Save last two bytes of BTNAD at SAVADDR
        LDAA    BTNAD+2
        LDY     #SAVADDR
        JSR     MWRITE
        JSR     INCADDR
        LDAA    BTNAD+3
        JSR     MWRITE
        JSR     INCADDR

;*** Set ADDR1 to start of the new block
        LDD     #BLKSIZE-1
        LDY     #ADDR1
        JSR     ADDADDR
        CLR     ADDR1+3         ; AND the address with ~(BLKSIZE-1)
        LDAA    ADDR1+2
        ANDA    #$FF^BLKMSK
        STAA    ADDR1+2

        LDAA    ADDR1+1         ; have we reached end of memory ?
        CMPA    #$80
        BNE     MEMOK
        JMP     BLED            ; yes, no more fun
MEMOK:

        LDD     ADDR1           ; first copy ADDR1 into BTNAD
        STD     BTNAD
        LDD     ADDR1+2
        STD     BTNAD+2
        LDD     #BLKSIZE        ; then make it point to (end of block)+1
        LDY     #BTNAD
        JSR     ADDADDR

;*** Save value of the compression timer
        LDAA    CTIMER          ; save the compression timer in memory
        JSR     WRBYTE
        LDAA    CTIMER+1
        JSR     WRBYTE
        LDAA    CTIMER+2
        JSR     WRBYTE
        LDAA    CTIMER+3
        JSR     WRBYTE

;*** Leave room for more status information
        LDD     ADDR1           ; save the actual memory location
        STD     SAVADDR         ; to come back here and add size information
        LDD     ADDR1+2
        STD     SAVADDR+2
        LDD     #4              ; leave four bytes free
        LDY     #ADDR1
        JSR     ADDADDR

;*** Save values of the channels
        LDAB    DSAVED
        STAB    DFSP
        DECB
        DECB
        ANDB    #DDATAMK
        ORAB    #DDATA
        CLRA
        XGDY                    ; in Y pointer in the FIFO
        LDAA    1,Y             ; the second channel
        PSHA                    ; save it
        LDAA    0,Y             ; the first channel
        JSR     WRBYTE          ;   in memory
        PULA                    ; retrieve second channel
        JSR     WRBYTE          ;   in memory

; initialize the compression module
        JSR     COMPINIT        ; initialize the compression module
        CLR     BLKST           ; not at the beginning of a block any more
        RTS                     ; going back to main

;***
; BUTTON
; routine to handle the button
; It is implemented as a simple automata with two states.
; The transition between the states is made only when the button is
; changing its position (pressed/released). Action is executed when the
; button is pressed.
;***
BUTTON:
        TST     BTNPR
        BEQ     BUT000
; in state 1, the button was down at the previous iteration
        TST     BTNST
        BNE     BUT004          ; the button is still pressed, return
        CLR     BTNPR           ; the button was just released, goto state 0
        BRA     BUT004
; in state 0, the button was up at the previous iteration
BUT000:
        TST     BTNST
        BNE     BUT001          ; the button was just pressed, action
BUT004:
        RTS                     ; no action, back to main
BUT001:
        LDD     BTNAD+2         ; check to see if there is still
        SUBD    ADDR1+2         ;    enough place in the current buffer
        TSTA
        BNE     BUT002
        CMPB    #3
        BHS     BUT002
        RTS                     ; no, keep state, store event in next block
BUT002:
        BSET    BTNPR 1         ; goto state 1 of the automata
        LDY     #BTNAD          ; write the lower bytes of CTIMER in mem
        JSR     DECADDR
        LDAA    CTIMER+3
        JSR     MWRITE
        JSR     DECADDR
        LDAA    CTIMER+2
        JSR     MWRITE
        RTS

;*************************************************
; MEMORY functions
;*************************************************

;***
; WRBYTE
; entry point in the routine that writes a byte to the RAM
;***

WRBYTE:
        LDY     #ADDR1
        JSR     MWRITE          ;WRITE TO RAM
        JSR     INCADDR         ; mwrite does not modify iy

        RTS

;**
; MWRITE WRITES THE VALUE IN A TO THE RAM ADDRESS in reg IY
; ASSUME 1M MEMORY AND X=$1000
;**

MWRITE:
        SEI
        PSHA
        BSET    $0,X $20        ; DISABLE REFRESH DURING WRITE
        PSHB
        LDAB    #$FF
        STAB    $07,X           ;PORT C=OUT

        STAA    $03,X           ;OUTPUT DATA
        LDAA    3,Y             ;LS BYTE OF ADDR
        STAA    $04,X
        LDAA    1,Y             ;MS BYTE OF ADDR
        ANDA    #$07            ;KEEP 3 BITS (ADDR16,17,18)
        LSLA
        LSLA
        LSLA
;        LSLA
        LDAB    $8,X
        ANDB    #$C7
        STAB    $8,X
        ORAA    $8,X
        STAA    $8,X            ;OUTPUT ROW A8-A10

        LDAA    #$40            ;THE A22 BIT
        BITA    1,Y
        BEQ     MWBANK2

        BCLR    $0,X $40        ;RAS ACTIVE (BANK 1)
        BRA     MW2

MWBANK2:
        BCLR    $0,X $80        ;RAS ACTIVE (BANK 2)
MW2:
        LDAA    1,Y
        ANDA    #$38            ;KEEP BITS 19, 20, 21
;        LSLA
;        LSLA
        ANDB    #$C3
        STAB    $8,X
        ORAA    $8,X
        STAA    $8,X            ;OUTPUT COLUMN A8-A10 AND WE
        LDAA    2,Y             ;A8-A15
        STAA    $4,X            ;WRITE COLUMN A0-7
        BSET    $8,X $04        ;WE INACTIVE
        BSET    $0,X $C0        ;RAS INACTIVE

        PULB
        BCLR    $0,X $20        ; RESUME REFRESHING
        PULA
        CLI
        RTS

;**
; MREAD READS IN A THE VALUE AT THE RAM ADDRESS ADDR2
; ASSUME 1M MEMORY AND X=$1000
;**

MREAD:
        PSHB
        BSET    0,X $20         ;DISABLE REFRESH DURING READ
        LDAB    #$00
        STAB    $07,X           ;PORT C=IN

        LDAA    ADDR2+3         ;LS BYTE OF ADDR
        STAA    $04,X

        LDAA    ADDR2+1         ;MS BYTE OF ADDR
        ANDA    #$07            ;KEEP THREE BITS (ADDR16,17,18)
        LSLA
        LSLA
        LSLA
;        LSLA
        LDAB    $8,X
        ANDB    #$C7
        STAB    $8,X
        ORAA    $8,X
        STAA    $8,X            ;OUTPUT ROW A8-A10

        LDAA    #$40            ;THE A22 BIT
        BITA    ADDR2+1
        BEQ     MRBANK2

        BCLR    $0,X $40        ;RAS ACTIVE (BANK 1)
        BRA     MR2

MRBANK2:
        BCLR    $0,X $80        ;RAS ACTIVE (BANK 2)
MR2:
        LDAA    ADDR2+1
        ANDA    #$38            ;KEEP BITS 19, 20, 21
;        LSLA
;        LSLA
        STAB    $8,X
        ORAA    $8,X
        STAA    $8,X            ;OUTPUT COLUMN A8-A10
        LDAA    ADDR2+2         ;A8-A15
        STAA    $4,X            ;WRITE COLUMN A0-7
        BSET    $0,X $C0        ;RAS INACTIVE

        PULB

        LDAA    $05,X           ;READ DATA

        BCLR    $0,X $20        ;RESUME REFRESH

        RTS

BLED:
        LDAA    #MEMLED
        STAA    LEDST
        LDY     #$1FFF
RE1:
;        STAA    LEDST
        DEY
        BNE     RE1
        CLR     LEDST
        LDY     #$FFFF
RE2:
;        CLR     LEDST
        DEY
        BNE     RE2
        BRA     BLED


;***
; entry point for the routine that stores a value in the FIFO
; input: A
; Modifies: A,B,IY
;***
PSHFIFO:
        PSHA
        CLRA
        LDAB    FEP
        XGDY                    ;Y<-ADDRESS OF THE NEXT FREE FIFO BTYE
        PULA
        STAA    0,Y
        XGDY
        INCB
        ANDB    #FIFOMK         ;WRAP-AROUND OF FIFO POINTER
        STAB    FEP
        CMPB    FSP
        BNE     ADIOK
        BSET    STATUS $01      ;SET FIFO OVERFLOW FLAG
ADIOK:
        RTS

;***
; FIFOADR
; compute the address coresponding to the B-th value from the FIFO start
;
; The first value stored into the FIFO has index 0.
; This routine takes into account the number
; of channels stored in the FIFO --- 2 channels.
;***
FIFOADR:
        LSLB            ; yes, compute true index
        LDAA    FSP     ; start of the FIFO
        SBA
        ANDA    #FIFOMK ; in A the one byte address
        TAB
        CLRA
        XGDY
        RTS

RDFIFO:
        JSR     FIFOADR
        LDAA    0,Y     ; loading the data
        RTS



DFIFADR:
        LSLB            ; yes, compute true index
        LDAA    DFSP    ; start of the FIFO
        SBA
        ANDA    #DDATAMK; in A the one byte address
        ORAA    #DDATA
        TAB
        CLRA
        XGDY
        RTS

RDDFIF:
        JSR     DFIFADR
        LDAA    0,Y     ; loading the data
        RTS

;*************************************************
; BIG POINTERS handling functions
;*************************************************

;***
; CLRADDR ZEROES ADDR (Y - POINTER TO 4 BYTE MEMORY ADDRESS)
;***
CLRADDR:
        CLR     3,Y
        CLR     2,Y
        CLR     1,Y
        CLR     0,Y
        RTS

;***
; INCADDR INCREMENTS ADDR (Y - POINTER TO 4 BYTE MEMORY ADDRESS)
;***

INCADDR:
        INC     3,Y
        BNE     IN0
        INC     2,Y
        BNE     IN0
        INC     1,Y
        BNE     IN0
        INC     0,Y
IN0:
        RTS

;***
; DECADDR INCREMENTS ADDR (Y - POINTER TO 4 BYTE MEMORY ADDRESS)
;***

DECADDR:
        LDD     2,Y
        SUBD    #1
        STD     2,Y
        BCC     DEC0
        LDD     0,Y
        SUBD    #1
        STD     0,Y
DEC0:
        RTS

;***
; ADDADDR adds the value in D to the value pointed by Y (a 4  byte integer)
; modifies A, B
;***
ADDADDR:
        ADDD    2,Y
        STD     2,Y
        BCC     ADDAD1
        INC     1,Y
        BNE     ADDAD1                  ; Cy not affected by INC
        INC     0,Y
ADDAD1:
        RTS

;*************************************************
; Serial interface handler
;*************************************************

;*** Serial interrupt dispatcher
SERHDLR:
        LDAA    $2E,X           ;TEST RDRF BIT
        ANDA    #$20
        BEQ     SHTXI

        JMP     SRINT

SHTXI:
        JMP     STINT

;****
; SRINT IS THE SERIAL RECEIVE INTERRUPT SERVICE ROUTINE
;****

SRINT:
        LDAA    $2E,X           ; READ SCSR (SERIAL STATUS REG)
        ANDA    #$0A            ; TEST ERROR FLAGS
        BEQ     NOSRERR         ; NO ERR, CONTINUE
	STAA	SRERROR		; save it in srerror

SREXIT:
        LDAA    $2F,X           ; READ SCDR TO CLEAR ERRORS
        RTI                     ; IGNORE THE BYTE

NOSRERR:
        LDAA    TABYTES         ; CHECK IF THERE ARE MORE BYTES TO RECEIVE
        BEQ     NOTB            ; NO, THIS IS A COMMAND

        DEC     TABYTES

        CLRA                    ; YES, STORE THEM
        LDAB    ADPTR
        XGDY
        LDAA    $2F,X           ; READ THE RECEIVED BYTE
        STAA    0,Y
        STAA    $2F,X           ; ECHO BYTE (ACK RECEPTION)
        INC     ADPTR
	RTI

NOTB:
        LDAA    TDBYTES         ; CHECK IF THERE IS ANOTHER OPERATION
        ORAA    TSBYTES
        BNE     SREXIT          ; EXIT IF THERE ARE OTHER OPS ACTIVE

        LDAA    $2F,X           ; READ COMMAND BYTE
        STAA    $2F,X           ; ECHO BYTE (ACK RECEPTION)
        CMPA    #$A0            ; CHECK FOR STATUS READ
        BEQ     RSSR

        CMPA    #$A1            ; CHECK FOR DATA READ
        BEQ     SRDR
	
	CMPA	#$A2		; CHECK FOR INFO READ (added dec94 mseritan)
	BEQ	SRIR

	CMPA	#$a3		; CHECK FOR RAM READ (added dec94 mseritan)
	BEQ	SRRR

        CMPA    #$A8		; CHECK FOR ADDRESS WRITE
	BNE	SREXIT		; EXIT IF NONE OF THESE IS TRUE

; ADDRESS WRITE SETUP

	CLR	TABYTES
        BSET    TABYTES $04     ; SET LENGTH OF ADDRESS
        LDD     #ADDR2          ; SET BEGINNING OF ADDRESS
        STAB    ADPTR
        RTI


; READ STATUS REQUEST

RSSR:
	CLR	TSBYTES
        BSET    TSBYTES $20     ; SET NUMBER OF STATUS BYTES ($20)
        LDD     #ADDR1          ; SET BEGINNING OF STATUS INFO (ADDR1)
        STD	SBPTR
        BSET    $2D,X $80       ; ENABLE TX INTERRUPT
        RTI

; READ DATA REQUEST

SRDR:
        BSET    TDBYTES $20     ; SET NUMBER OF DATA BYTES TO BE TRANSMITTED
        BSET    $2D,X $80       ; ENABLE TX INTERRUPT
        RTI

; READ INFO REQUEST
SRIR:
        BSET    TSBYTES $20	; SET NUMBER OF INFO BYTES ($20)
        LDD     #INFO		; SET BEGINNING OF INFO DATA 
        STD	SBPTR
        BSET    $2D,X $80       ; ENABLE TX INTERRUPT
        RTI

; READ RAM REQUEST
SRRR:
        BSET    TSBYTES $20	; SET NUMBER OF RAM BYTES 
        LDD     ADDR2+2		; GET START ADDRESS REQUESTED
        STD	SBPTR
	CLR	SBPTR		; PROGRAM WRITEONLY PROTECTION !!!
        BSET    $2D,X $80       ; ENABLE TX INTERRUPT
        RTI


;**** END OF RECEIVE INT SERVICE ROUTINE

;****
; STINT IS THE TRANSMIT INTERRUPT SERVICE ROUTINE
;***

STINT:
        LDAA    TSBYTES         ; CHECK IF THERE ARE BYTES TO XMIT
        BNE     STSB

        LDAA    TDBYTES         ; CHECH FOR DATA BYTES TO XMIT
        BNE     STDB
	

        BCLR    $2D,X $80       ; DISABLE FURTHER TX INTS

        RTI                     ; NO ACTIVE OP, EXIT

; TRANSMIT SOME BYTES
STSB:
        LDD	SBPTR
        XGDY
        LDAA    0,Y
        STAA    $2F,X           ; XMIT BYTE
        INC     SBPTR+1		; UPDATE POINTERS AND EXIT
	BNE	STSB1		; A 2 BYTE POINTER
	INC	SBPTR
STSB1:
        DEC     TSBYTES
        RTI

STDB:
        JSR     MREAD           ; READ FROM DRAM
        STAA    $2F,X           ; XMIT BYTE
        LDY     #ADDR2
        JSR     INCADDR         ; UPDATE POINTERS AND EXIT
        DEC     TDBYTES
        RTI

;**** END OF XMIT INT SERVICE ROUTINE

;*************************************************
; Periodic interrupt handler
;*************************************************

PERINT:
;*** Prepare next interrupt
        LDD     $16,X           ;ADJUST TIMER OUTPUT COMPARE REG
        ADDD    #INTVL
        STD     $16,X

        LDAA    #$80            ;CLEAR COUNTER FLAG
        STAA    $23,X

;*** Sample data
        LDAA    $31,X           ; get AD's first channel
        PSHA
        LDAA    $32,X
        JSR     PSHFIFO ; store the second value
        PULA
        JSR     PSHFIFO         ; store the first value
        LDAA    #$10            ;THE ADC COMMAND (MULT ENABLED, SCAN DISABLED)
        STAA    $30,X           ;FOR THE NEXT INTERRUPT

;**** Light LEDs
; send the  apropriate value to the LEDs
        LDAB    LEDST
        LDY     #LCNTCH
; test led for channel 1
        TST     0,Y
        BEQ     LED001
        ORAB    #CH1LED
        DEC     0,Y
LED001:
        TST     1,Y
        BEQ     LED002
        ORAB    #CH2LED
        DEC     1,Y
LED002:
        LDAA    FEP
        ANDA    #6
        BEQ     LED003
        CLRB
LED003:
        LDAA    #$FF
        STAA    $07,X           ;PORT C = OUT
        STAB    $03,X           ;WRITE LED DATA
        BSET    $0,X $08        ;STROBE ON
        BCLR    $0,X $08        ;STROBE OFF

;*** Debounce button
; handling the button
        TST     BTNTIM          ; decrement button timer
        BEQ     BTN001
        DEC     BTNTIM
        BRSET   0,X 2 BTN010    ; is the button pressed
        INC     BTNCNT          ; yes, increment button counter
        BRA     BTN010
BTN001:
; reinitialize the debouncing mechanism
        LDAA    #BTNPER         ; initializing the timer for the button
        STAA    BTNTIM
        CLR     BTNST           ; clear button  status
        LDAA    BTNCNT          ; compare the counter
        PSHA
        CLR     BTNCNT          ;   and reset it at once
        PULA
        CMPA    #BTNTHR         ; with the threshold associated with button
        BLO     BTN010          ; if < threshold leave status zero
        COM     BTNST           ; make the status > 0
BTN010:

        RTI

;*************************************************
; ROM data
;*************************************************


; data for the automata
; description of structure for each state
CMAX    EQU     0               ; max positive value accepted in the state
CMIN    EQU     1               ; min positive value accepted in the state
CNBB    EQU     2               ; number of bits
CCODEN  EQU     3               ; code to go to the next state
CSTRUC:
        FCB  1, -1, 1,  1
        FCB  2, -2, 2,  2
        FCB  4, -4, 3,  4
        FCB 16,-16, 5, 16
        FCB 64,-64, 7, 64
        FCB 128, -127, 8, 128   ; only the 8 used from here
; coefficients of the filter
FILTCOEF:
        FCB -4, -8, -2, 19, 50, 73
        FCB 73, 50, 19, -2, -8, -4

INFO:
	FDB	5000		; SAMPLING PERIOD IN 10^-6 sec
	FCB	8		; RESOLUTION IN BITS
	FCB	FFORMAT, CFORMAT; FILE FORMAT
	FCB	VERSION, PATCH	; OF THE HOLTER PROGRAM
	FCB	77		; MAGIC

;********
;* INT VECTORS
;* written by Radu Corlan
;*********

        ORG     $FFD6
        FDB     SERHDLR         ;SERIAL INT HANDLER

        ORG     $FFE8
        FDB     PERINT          ;TIMER OC1 INT HANDLER

        END
