/* Linux kernel module for holter transmission program interface. 
 * writen by Marius SERITAN on the 5-th of March 1995
 * (C) copyright 1995 Marius SERITAN
 * 
 * History: 9 April 1995 mseritan:
 *   porting the modular device driver to kernel 1.1.62. 
 *   Oops, it was dead easy, a matter of adding the __KERNEL__ define :-)
 *
 *   23 May 1995 mseritan:
 *     speeding up the transfer under Linux     
 */

#define __KERNEL__

#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/major.h>

#include <asm/system.h>
#include <asm/segment.h>

#include <linux/module.h>
#include <linux/version.h>

char kernel_version[] = UTS_RELEASE;

#include "serlio.h"

struct holter_status {
  int 		present;
  int		ready;
  int		active;
  struct wait_queue *wait;
} hs;

/* Functions taken from e2100.c, they will get executed in kernel
 * space */

static short gwo_send_byte(int c, int serial_port) {
  int r;
  r = transmit_byte(serial_port, c);
  if( r<0 ) return r;
  r = receive_byte( serial_port );
  if( (r<0) || (r!=c) ) return -1;
}

static int gwo_get_byte( int serial_port ) {
  return( receive_byte( serial_port ));
}

#define GWO_WRITE_ADDRESS   0x00a8

int gwo_set_address(long adr, int serial_port) {
  /* adr --- the address to be set 
   * serial_port --- number of the serial port used
   */
  short r, c, i;

  r = gwo_send_byte(GWO_WRITE_ADDRESS, serial_port );
#ifdef DEBUG_MODULE
  printk( "Sending gwo command " );
#endif
  if( r < 0 ) {
    return -1;
  }

  for(i=0; i<4; i++) {
    c = 0x00ff & (adr >> 24);
    r = gwo_send_byte(c, serial_port );
#ifdef DEBUG_MODULE
    printk( " . " );
#endif
    if( r<0 ) {
      printk( "Error\n" );
      return r;
    }
    adr = adr << 8;
    }

#ifdef DEBUG_MODULE
  printk( "Okay\n" );
#endif
  return 0;  /* OK */
  }


#define GWO_BYTE_MAX_JUNK 100


static int gwo_init( int serial_port, long baud_rate ) {
  int received;

  /* initialization of the serial port */
  clear_serial_error(serial_port);
  set_timeout(serial_port, 1000L );
  serial_port_init( serial_port, baud_rate, 0x03);
  modem_control(serial_port, 1);

  /* empty serial line, maybe holter is replying to some command
   * if the response is too verbose something must be wrong  
   */

  for( received=0;  receive_byte(serial_port) > 0; received++ )
    if( received >= GWO_BYTE_MAX_JUNK ) return -1;

  clear_serial_error(serial_port); /* of course there is one serial error */
  
  return 0;
}


static long baud_rates[] = { 19200L, 19200L, 76800L };


/* File operations implementation. */

/* Data storage to implement the caching mechanism */
static char holter_buffer[128];
static long holter_address;
static long holter_fill;

static int open_holter(struct inode * inode, struct file * file) {
  if( gwo_init( MINOR(inode->i_rdev), baud_rates[MINOR(inode->i_rdev)]) )
    return  -EINVAL;
  holter_address = 1<<30;
  holter_fill = -1;
  return 0;
}

static void close_holter(struct inode * inode, struct file * file) {
}

/* The write function is used to send the command to the holter.
 * In fact the command is stored for future use.
 */

static int holter_request_code;

static int write_holter(struct inode * inode, struct file * file
		       , char * buffer, int count) {
  if( count != 1 ) return -EINVAL;
  holter_request_code = get_fs_byte( buffer );
  return count;
}



static int seek_holter( struct inode * inode, struct file * file
		       , off_t offset, int orig) {
  unsigned long flags;
  int result;

  file->f_pos = offset;
  save_flags( flags );
  cli();
  result = gwo_set_address( offset, MINOR(inode->i_rdev));
  restore_flags( flags );
  return result;
}


static int read_holter(struct inode * inode, struct file * file
		       , char * buffer, int count) {
  unsigned long flags;		/* processor state */
  int result;
  char local_buffer[33];
  int serial_port;
  int block;


  /* if( count > 32 ) count = 32;*/
  if( count < 32 ) return -EINVAL;
  serial_port = MINOR(inode->i_rdev);
  for(   block = count / 32; block > 0; block -- ) {
    int i;			/* index in the local buffer */

    save_flags( flags );
    cli();
    result = gwo_send_byte( holter_request_code, serial_port );
    if( result < -1 ) {
      restore_flags(flags);
      return -EINVAL;
    }
  
    for(i=0; i<32 && result>=0; i++) {
      result  = gwo_get_byte(serial_port);
      local_buffer[i] = result;
    }
    restore_flags( flags );
    if( result < 0 ) return -EINVAL;
    memcpy_tofs( buffer, local_buffer, 32 );
    buffer += 32;
  }
  return ( count / 32 ) * 32;
}


static struct file_operations holter_fops = {
  seek_holter,
  read_holter,
  write_holter,
  NULL,		/* readdir */
  NULL,		/* select */
  NULL,		/* ioctl */
  NULL,		/* mmap */
  open_holter,
  close_holter
};






/* functions for module management */

int init_module( void ) {
  /* register an unused device, major number should be checked
   * with /usr/include/linux/major.h (no it is not done at the moment)
   */
  hs.present = 0;
  if( register_chrdev(LINUX_DEVICE_MAJOR, "holter", &holter_fops) ) {
    printk( "Oops, major number %d already taken, giving up!\n"
	   , LINUX_DEVICE_MAJOR );
    return -1;
  }
  hs.present = 1;
  holter_request_code = 0xa1;
  return 0;
}

void cleanup_module( void ) {
  if( hs.present )
    unregister_chrdev( 19, "holter" );
};
