/*****************************************************
 * LOW-LEVEL SERIAL LINE I/O ROUTINES 
 * Copyright (c) 1993 Radu Corlan
 * Copyright (c) 1994 Marius Seritan
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * REV A
 *   9 dec '94 mseritan:
 *      speeding up `receive_byte'
 *  10 dec '94 mseritan:
 *      in function `receive_byte'
 *        [] making status info available for interpretation 
 *        [] accesing  8250 status register only when needed
 *      wrote `interpret_error'
 *             
 ****************************************************/
#include "serlio.h"
#include <stdio.h>


int serial_error[3]={0, 0, 0};  /* serial line error flag, incremented whether an 
                                error occured */

long timeout_count[3]={10000L, 10000L, 10000L}; /* the number of count until timeout is signalled
                                 timeout_count = 0 sets infinite retry */

unsigned serial_base[3] = {0x3f8, 0x2f8, 0x360}; /* table of serial port 
					     base addresses */

long baud_base[3] = {BAUDBASE1, BAUDBASE1, BAUDBASE2}; /* Baud bases */


/* serial_port_init initialises the serial port
   args:
       port: 0 for COM1, 1 for COM2
       baud: the desired baud rate (long!)
       line: the line parameters, as follows:
          bits 1-0: character length    00 -- 5 bits
                                        01 -- 6 bits
                                        10 -- 7 bits
                                        11 -- 8 bits

                 2: stop bits            0 -- 1 stop bit
                                         1 -- 2 or 1 1/2 stop bits

                 3: parity enable
                 4: parity odd/even      0 -- odd parity
                                         1 -- even parity

                 5: stick parity
                 6: set break
                 7: always 0
 */

void serial_port_init(int port, long baud, short line)

  {
  unsigned bdiv;

  bdiv = (unsigned) ( ( 10 * baud_base[port] / baud + 5 ) / 10 );

  
  sys_outp (serial_base[port] + INTENR , 0);
  sys_outp (serial_base[port] + LINEC ,  line | 0x80);  
  sys_outp (serial_base[port] + BAUDL ,  bdiv & 0x00ff);
  sys_outp (serial_base[port] + BAUDH ,  (bdiv >> 8) & 0x00ff);
  sys_outp (serial_base[port] + LINEC ,  line & 0x7f);
  }

/* modem_control sets the modem control bits, as follows:
   port:    0 for COM1, 1 for COM2
   mcontrol bit 0 -- DTR active
            bit 1 -- RTS active
            bit 2 -- AUX1 active
            bit 3 -- AUX2 active
            bit 4 -- enable loopback
            bits 5--7 -- always 0
 */

void modem_control(int port, unsigned mcontrol)

  {
  sys_outp (serial_base[port] + MODC , mcontrol & 0x001f);
  }

/* modem_status reads the modem status inputs on a port
   port:    0 for COM1, 1 for COM2
   the return value is encoded as follows:
            bit 0 -- DCD active
            bit 1 -- RI  active
            bit 2 -- DSR active
            bit 3 -- CTS active
            bit 4 -- DCD changed since last read
            bit 5 -- RI  changed since last read
            bit 6 -- DSR changed since last read
            bit 7 -- CTS changed since last read
 */

unsigned modem_status(int port)

  {
  return sys_inp (serial_base[port] + MODS);
  }

/* transmit_byte sends a byte of data through the port.
   Returns the data sent
   On error (timeout) it returns -2 and increments serial_error.
   On serial error returns -1 and increments serial_error.
   If timeout_count == 0 it cycles infinitely until it is able to transmit.
 */

int transmit_byte (int port, unsigned data) {
  long i;
  int status;

  if (timeout_count[port] == 0L)
       {
       while (! tx_ready(port));
       sys_outp (serial_base[port] + TXHR, data & 0x00ff);
       return (data & 0x00ff);
       }
  for (i=timeout_count[port]; i>0L; i--) {
    status = line_status( port );
    if status_tx_ready(status) {
      sys_outp (serial_base[port] + TXHR, data & 0x00ff);
      return (data & 0x00ff);
    }
  }
  serial_error[port] ++;
  return (-2);
  }    

/* Interpret the error codes if a transmission error occured
 * by mseritan on 10 dec '94
 */
#ifndef LINUX
void interpret_error( int error_code, FILE * outfile ) {
  fprintf( outfile, "Error interpretation for %x: " );
  if( error_code & UART_LSR_BI) {
    fprintf( outfile, "Break interrupt " );
  }
  if( error_code & UART_LSR_FE) {
    fprintf( outfile, "Frame error " );
  }
  if( error_code & UART_LSR_PE) {
    fprintf( outfile, "Parity error " );
  }
  if( error_code & UART_LSR_OE) {
    fprintf( outfile, "Overrun error " );
  }
  fprintf( outfile, "\n." );
}
#endif

/* receive_byte receives a byte of data through the port.
   Returns the data received
   On error it returns a -1 for serial error -2 for timeout
     and increments serial_error.
   If timeout_count == 0 it cycles infinitely until it receives
       a character or detects an error condition.
 */

int receive_byte (int port) {
  long i;
  int data, status;
  
  status = line_status(port);
  if( status_rx_ready(status) ) {
    data = sys_inp (serial_base[port] + RXDR);
    return (data & 0x00ff);
  }
  
  if (timeout_count[port] == 0L) {
    /* status already read above */
    while (! ( status_rx_ready(status) || status_rx_error(status) ) )
      status = line_status(port);
    if (status_rx_error(status)) {
/*      interpret_error( status, stderr );*/
      serial_error[port] ++;
      return (-1);
    }
    data = sys_inp (serial_base[port] + RXDR);
    return (data & 0x00ff);
  }
  /* variable `status' already initialized for the first iteration */
  for (i=timeout_count[port]; i>0L; i--) {
    if( status_rx_ready(status) ) {
      data = sys_inp (serial_base[port] + RXDR);           
      return (data & 0x00ff);
    }
    if( status_rx_error(status) ) {
      serial_error[port] ++;
      return (-1);
    }         
    status = line_status( port );
  }
  serial_error[port] ++;
  return (-2);
}    

/*  End of serlio serial line interface code */
