/*****************************************************
 * LOW-LEVEL SERIAL LINE I/O ROUTINES 
 * Copyright (c) 1993 Radu Corlan
 * Copyright (c) 1994-1995 Marius Seritan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * modified june 1994 by Marius Seritan, djgpp port
 * modified jan 1995 by Marius Seritan, LINUX port 
 * modified february 1995 by Marius SERITAN, enabling/disabling interrupts
 *     under the LINUX operating system
 *     Bad luck, no way to play around with interrupts while in user mode,
 *     trying to write a device in another file.
 *
 * REV A
 ****************************************************/


/* include system dependent headers and 
 * translate system dependent functions 
 */
#ifndef unix /* for MSC6.0 */
#include <conio.h>
#define sys_outp(port,value) outp(port,value)
#define sys_inp(port)        inp(port)
#endif

#if defined(unix) && !defined(LINUX) /* for DJGPP */
#define sys_inp(port)        inportb(port)
#define sys_outp(port,value) outportb(port, value)
#define _enable()         asm( "sti" )
#define _disable()        asm( "cli" )
#endif

#if defined(LINUX)
#include <unistd.h>
#include <fcntl.h>
#include <asm/system.h>
#include <asm/io.h>
#define sys_inp(port)        inb(port)
#define sys_outp(port,value) outb(value,port)
/* I have found no way to enable/disable the interrupts under LINUX,
 * it seems like cli, sti are catched by the processor in user mode
 * and generate a SIGSEGV signal (segmentation fault).
 */
#define _enable()
#define _disable()
#endif

extern int serial_error[];  /* serial line error flag, incremented whether an 
                                error occured */

extern long timeout_count[]; /* the number of count until timeout is signalled
                                 timeout_count = 0 sets infinite retry */

extern unsigned serial_base[]; /* table of serial port 
					     base addresses */

/* ************************************************* */

/* these are the offsets to the 8250 registers */

#define TXHR 0	  /* transmitter holding register offset */
#define RXDR 0    /* receiver data register offset */
#define INTENR 1  /* interrupt enable register offset */
#define INTIDR 2  /* interrupt identification register offset */
#define LINEC 3	  /* line control register offset */
#define MODC  4   /* modem control register offset */
#define LINES 5	  /* line status register offset */
#define MODS  6   /* modem status register offset */
#define BAUDL 0	  /* baud rate divisor low offset */
#define BAUDH 1   /* baud rate divisor high offset */

#define BAUDBASE1 119000L  /* the base frequency for baud rates */
#define BAUDBASE2 156250L

/* definitions for Line Status Register
 * taken from /usr/include/linux/serial.h
 * by mseritan on the 10 dec '94
 */
#define UART_LSR_TEMT	0x40	/* Transmitter empty */
#define UART_LSR_THRE	0x20	/* Transmit-hold-register empty */
#define UART_LSR_BI	0x10	/* Break interrupt indicator */
#define UART_LSR_FE	0x08	/* Frame error indicator */
#define UART_LSR_PE	0x04	/* Parity error indicator */
#define UART_LSR_OE	0x02	/* Overrun error indicator */
#define UART_LSR_DR	0x01	/* Receiver data ready */

/* ************************************************* */

/* some macros to ease things */

/* clear the error condition */
#define clear_serial_error(port) serial_error[port]=0

/* get the error condition */
#define get_serial_error(port) serial_error[port]

/* set the timeout count */
#define set_timeout(port, val) timeout_count[port]=val

/* get and interpret line status */
#define line_status(port) (sys_inp(serial_base[port]+LINES))
#define status_tx_ready(status) ((status) & 0x20 ? 1 : 0)
#define status_rx_ready(status) ((status) & 0x01 ? 1 : 0)
#define status_rx_error(status) ((status) & 0x1e ? 1 : 0)

/* test for tx ready condition */
#define tx_ready(port) (sys_inp(serial_base[port]+LINES) & 0x20 ? 1 : 0)

/* test for rx ready condition */
#define rx_ready(port) (sys_inp(serial_base[port]+LINES) & 0x01 ? 1: 0)

/* test for rx error condition */
#define rx_error(port) (sys_inp(serial_base[port]+LINES) & 0x1e ? 1: 0)

/* ************************************************ */

/* serial_port_init initialises the serial port
   args:
       port: 0 for COM1, 1 for COM2
       baud: the desired baud rate (long!)
       line: the line parameters, as follows:
          bits 1-0: character length    00 -- 5 bits
                                        01 -- 6 bits
                                        10 -- 7 bits
                                        11 -- 8 bits

                 2: stop bits            0 -- 1 stop bit
                                         1 -- 2 or 1 1/2 stop bits

                 3: parity enable
                 4: parity odd/even      0 -- odd parity
                                         1 -- even parity

                 5: stick parity
                 6: set break
                 7: always 0
 */

void serial_port_init(int port, long baud, short line);

/* modem_control sets the modem control bits, as follows:
   port:    0 for COM1, 1 for COM2
   mcontrol bit 0 -- DTR active
            bit 1 -- RTS active
            bit 2 -- AUX1 active
            bit 3 -- AUX2 active
            bit 4 -- enable loopback
            bits 5--7 -- always 0
 */

void modem_control(int port, unsigned mcontrol);

/* modem_status reads the modem status inputs on a port
   port:    0 for COM1, 1 for COM2
   the return value is encoded as follows:
            bit 0 -- DCD active
            bit 1 -- RI  active
            bit 2 -- DSR active
            bit 3 -- CTS active
            bit 4 -- DCD changed since last read
            bit 5 -- RI  changed since last read
            bit 6 -- DSR changed since last read
            bit 7 -- CTS changed since last read
 */

unsigned modem_status(int port);

/* transmit_byte sends a byte of data through the port.
   Returns the data sent
   On error (timeout) it returns -1 and increments serial_error.
   If timeout_count == 0 it cycles infinitely until it is able to transmit.
 */

int transmit_byte (int port, unsigned data);

/* receive_byte receives a byte of data through the port.
   Returns the data received
   On error it returns -1 and increments serial_error.
   If timeout_count == 0 it cycles infinitely until it receives
       a character or detects an error condition.
 */

int receive_byte (int port);


